/* Calculator program for incremental (reversible) script
 * execution under a SEED-like session editor.
 */
   
/* Version allowing multi-character variable names and using
 * malloced storage for variable table.
 * The purpose is primarily to test compatibility of C malloc
 * and Pascal 'new' as used in the session editor.
 *
 * N. Holsti 1993.4.1
 */

#include "cseed.h"
#include "seed_lexline.h"

#include <string.h>
#include <malloc.h>

typedef enum {EndSymbol}	SymbolType;
typedef lex_realtype		numeric;

/* The calculator memory (variable table) is a linked list
 * of items, with the latest on top.
 */

typedef struct {char             *name;
		numeric           value;
		struct Mem_Item  *prev;
		} Mem_Item;

typedef Mem_Item       *Mem_Index;
#define Mem_Null	NULL
typedef Mem_Index	Mem_PositionType;

typedef struct {
	lex_positiontype LexPos;
	Mem_PositionType MemPos;
	}
	PositionType;


/* Mem package */

struct  {
	Mem_Index Top;   /* current top of list */
	} Mem;

void Mem_Define (Name,Value)
lex_wordtype Name;
numeric      Value;
/* Store 'Name=Value' in the memory. */
{
    Mem_Item *item;
    item = (Mem_Item *) malloc(sizeof(Mem_Item));
    if (item != NULL) {
      item->name = malloc(Name.length+1);
      if (item->name != NULL) {      	
        strncpy(item->name,Name.text,Name.length);
        *(item->name+Name.length)=0;
        item->value = Value;
        item->prev  = (struct Mem_Item *)Mem.Top;
        Mem.Top = item;
      } else {
      	printf("Memory could not be allocated for variable name.\n");
      	/* if (free(item) != 1) perror("Mem_Define.free"); */
      	free(item);  /* in SysV, free() is void. */
      	lex_error();
      }
    } else {
      printf("The calculator storage is full.\n");
      lex_error();
    }
}

Mem_Index Mem_Defined (Name)
lex_wordtype Name;
/* The index of the Name in memory. Mem_Null if undefined. */
{
  Mem_Index i;
  for (i = Mem.Top; i != NULL;  i = (Mem_Index) i->prev) {
    if ( strlen (i->name) == Name.length &&
         strncmp(i->name,    Name.text, Name.length) == 0) break;
  }
  return i;
}


void Mem_RevPrint (mi)
Mem_Index mi;
/* Print names of memory items in reverse order, i.e. in the
 * order they were defined.
 */
{
  if (mi->prev != NULL) Mem_RevPrint((Mem_Index)mi->prev);
  printf(" %s",mi->name);
}
   
void Mem_Value (Name,Value)
lex_wordtype  Name;
numeric      *Value;
/* Value defined for the Name. Error if Name is undefined. */
{
  Mem_Index i;
  i=Mem_Defined(Name);
  if (i != Mem_Null) { *Value = i->value; }
  else {
    printf("The variable '%1.*s' is not defined.\n",Name.length,Name.text);
    if (Mem.Top != Mem_Null) {
      printf("Variables defined:");
      Mem_RevPrint(Mem.Top);
      printf("\n");
    }
    lex_error();
  }
}

void Mem_Store (P)
Mem_PositionType *P;
{
  *P=Mem.Top;
}

void Mem_Restore (P)
Mem_PositionType P;
{
  Mem_Index i;
  while (Mem.Top != P) {
    i       = (Mem_Index) Mem.Top;
    Mem.Top = (Mem_Index) i->prev;
    /*
    if (free(i->name) != 1) perror("Mem_Restore.free->name");
    if (free(i      ) != 1) perror("Mem_Restore.free");
    */
    free(i->name);   /* In SysV, free() is void. */
    free(i      );
  }
}




/* Combined position-type Store and Restore */

void Pos_Store (P,tag)
PositionType *P;
char *tag;
{
  lex_store(&(P->LexPos));
  Mem_Store(&(P->MemPos));
}

void Pos_Restore (P)
PositionType P;
{
  lex_restore(P.LexPos);
  Mem_Restore(P.MemPos);
}



/* Grammar */

void SkipBlankLines()
{
  lex_get();
  while (lex_token.kind == eoltoken) { lex_next(); }
}


void Err (mess)
char *mess;
{
  printf("%s\n",mess);
  lex_error();
}


/* 'forward' decls. */
void Expression();
void MoreTerms();
void Term ();
void MoreFactors();
void Factor();

typedef int boolean;

boolean FollowMoreTerms();
boolean FollowMoreFactors();


void Calculation()
{
  PositionType here;
  lex_wordtype name;
  numeric      value;
  Pos_Store(&here);
  
  do {
    SkipBlankLines();                       if (lex_cancel()) goto cancel;
    if (lex_issymbol(EndSymbol)) {
      lex_exit(); break;
    }
    if (lex_token.kind != wordtoken) {
      Err("Each expression must be named with a variable name.");
      goto cancel;
    }
    name=lex_token.value.word;
/*
    if (name.length > 1) {
      Err("Please use only 1-letter variable names.");
      goto cancel;
    }
*/
    if (Mem_Defined(name) != Mem_Null) {
      Err("This variable name was defined earlier.");
      goto cancel;
    }
    lex_next();            if (lex_cancel()) goto cancel;
    if (!lex_ismark('=')) {
      Err("Please type an equal sign (=) after the variable name.");
      goto cancel;
    }
    lex_next();           if (lex_cancel()) goto cancel;
    Expression(&value);   if (lex_cancel()) goto cancel;
    if (lex_token.kind == eoltoken) {
      /* ok */
    } else
    if (lex_ismark(';')) {
      /* ok */
    } else {
      Err("Use ';' between formulae within a line.");
      goto cancel;
    }
    Mem_Define(name,value);
    printf("      %-10.*s   =   %10.6f\n",name.length,name.text,value);
/*
    mallocmap();
*/
    lex_null();       if (lex_cancel()) goto cancel;
    Calculation();    if (lex_cancel()) goto cancel;
    lex_exit();
  cancel: if (lex_cancel()) Pos_Restore(here);
  }
  while (!lex_done());
}


void Expression (value)
numeric *value;
{
  PositionType here;
  numeric      s,t;
  Pos_Store(&here);
  do {
    if (lex_ismark('+')) { s= 1.0; lex_next(); }
    else
    if (lex_ismark('-')) { s=-1.0; lex_next(); }
    else                   s= 1.0;
                             if (lex_cancel()) goto cancel;
    Term(&t);                if (lex_cancel()) goto cancel;
    MoreTerms(s*t,value);    if (lex_cancel()) goto cancel;
    lex_exit();
  cancel: if (lex_cancel()) Pos_Restore(here);
  }
  while (!lex_done());
}


void MoreTerms (before,value)
numeric  before;
numeric *value;
{
  PositionType here;
  numeric      t;
  Pos_Store(&here);
  do {
    if (lex_ismark('+')) {
      lex_next();                     if (lex_cancel()) goto cancel;
      Term(&t);                       if (lex_cancel()) goto cancel;
      MoreTerms(before+t,value);      if (lex_cancel()) goto cancel;
    } else {
    if (lex_ismark('-')) {
      lex_next();                     if (lex_cancel()) goto cancel;
      Term(&t);                       if (lex_cancel()) goto cancel;
      MoreTerms(before-t,value);      if (lex_cancel()) goto cancel;
    } else {
    if (FollowMoreTerms()) { *value=before; }
    else {
      Err("Use '+' or '-' to add or subtract more terms.");
      goto cancel;
    }}}
    lex_exit;
  cancel: if (lex_cancel()) Pos_Restore(here);
  }
  while (!lex_done());
}


void Term (value)
numeric *value;
{
  PositionType here;
  numeric      f;
  Pos_Store(&here);
  do {
    Factor(&f);               if (lex_cancel()) goto cancel;
    MoreFactors(f,value);     if (lex_cancel()) goto cancel;
    lex_exit();
  cancel: if (lex_cancel()) Pos_Restore(here);
  }
  while (!lex_done());
}


void MoreFactors (before,value)
numeric  before;
numeric *value;
{
  PositionType here;
  numeric      f;
  Pos_Store(&here);
  do {
    if (lex_ismark('*')) {
      lex_next();                   if (lex_cancel()) goto cancel;
      Factor(&f);                   if (lex_cancel()) goto cancel;
      MoreFactors(before*f,value);  if (lex_cancel()) goto cancel;
    } else {
    if (lex_ismark('/')) {
      lex_next();                   if (lex_cancel()) goto cancel;
      Factor(&f);                   if (lex_cancel()) goto cancel;
      if (f == 0.0) {
        Err("Divisor should not be zero.");
        goto cancel;
      }
      MoreFactors(before/f,value);  if (lex_cancel()) goto cancel;
    } else {
    if (FollowMoreFactors()) { *value=before; }
    else {
      Err("Use '*' or '/' to multiply or divide with more factors.");
      goto cancel;
    }}}
    lex_exit();
  cancel: if (lex_cancel()) Pos_Restore(here);
  }
  while (!lex_done());
}


void Factor (value)
numeric *value;
{
  PositionType here;
  lex_wordtype name;

  Pos_Store(&here);
  do {
    if (lex_token.kind == natnumtoken) {
      *value=lex_token.value.natnumber;
      lex_next();                        if (lex_cancel()) goto cancel;
    } else {
    if (lex_token.kind == realnumtoken) {
      *value=lex_token.value.realnumber;
      lex_next();                        if (lex_cancel()) goto cancel;
    } else {
    if (lex_token.kind == wordtoken) {
      name=lex_token.value.word;
/*
      if (name.length > 1) {
        Err("Variable names can be only one letter long.");
        goto cancel;
      }
*/
      Mem_Value(name,value);            if (lex_cancel()) goto cancel;
      lex_next();                       if (lex_cancel()) goto cancel;
    } else {
    if (lex_ismark('(')) {
      lex_next();                       if (lex_cancel()) goto cancel;
      Expression(value);                if (lex_cancel()) goto cancel;
      if (!lex_ismark(')')) {
        Err("A closing parenthesis seems to be missing.");
        goto cancel;
      }
      lex_next();                       if (lex_cancel()) goto cancel;
    } else {
      printf("A factor should be a number, a variable, or");
      Err   (" a parenthesized expression.");
      goto cancel;
    }}}}
    lex_exit();
  cancel: if (lex_cancel()) Pos_Restore(here);
  }
  while (!lex_done());
}


boolean FollowMoreTerms ()
{
  if (lex_token.kind == eoltoken) {
	return 1;
  } else{
	return lex_ismark(')') || lex_ismark(';');
  }
}

boolean FollowMoreFactors ()
{
  if (lex_ismark('+') || lex_ismark('-')) {
	return 1;
  } else {
	return FollowMoreTerms();
  }
}


void Calculations()
/* A "starting symbol" to store the initial state. */
{
  PositionType here;
  Pos_Store(&here);
  do {
    lex_next();      if (lex_cancel()) goto cancel;
    Calculation();   if (lex_cancel()) goto cancel;
    lex_exit();
  cancel: if (lex_cancel()) Pos_Restore(here);
  }
  while (!lex_done());
}


void CalcInit()
{
  lex_define(*lex_word_c("end"),EndSymbol);
  lex_define(*lex_word_c("bye"),EndSymbol);
  lex_ignore(eolcommenttoken);
  Mem.Top=NULL;
}


int main (int argc, char *argv[])     /* Calcul */
{
  seed_init(argc, argv);
  lex_init ();

  if (seed_stdout() == -1) exit(1);

  printf("Calcul(C) 1994.10.24.A\n");
  printf("Type in a formula, such as x = 2+2. Press Enter. Edit at will.\n");
  printf("For help, type a question mark and press Enter.\n");
  printf("To quit, type 'end' or ctrl-X ctrl-C.\n\n");

  CalcInit();
  Calculations();
  printf("End calcul.\n");

  seed_fini(0);
  return 9;     /* should never happen */

}
