/* keyboa.c

  KeyBoa, raw keyboard setup and keystroke input.

Configures the standard input device to ignore all interrupt and break
keys and not to echo input.  Provides a routine to read a single byte.
This routine does not block simultaneous output to the terminal.
A function to report the number of bytes queued (entered but not read)
is also included. 

  Sun Unix port, 1991.12.23 by Niklas Holsti.
  Output post-processing can be controlled (on, off).
  1993.10.08  Port to Solaris 2.2. VSTATUS is no longer defined.
  1994.03.52  Translate to C, to use C headers (too complex to
              translate easily into Pascal.)
             
*/


#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <stropts.h>
#include <sys/conf.h>
#include <errno.h>
#include <termios.h>

/* private var */

static struct termios wasattr;   /* terminal attributes as they were   */
static struct termios attr;      /*    "         "      "  we set them */

#define standin STDIN_FILENO


void rawkeys ()
  /* Configure keyboard to ignore interrupt and break keys and
    not to echo.
    If standard input is not a tty device, does nothing.
  */
  {
  int status;
  char anul = _POSIX_VDISABLE;

  if (isatty(standin)) {
    status = tcgetattr(standin,&wasattr);
    if (status == -1) perror("tcgetattr");
    attr = wasattr;
    attr.c_iflag = (attr.c_iflag & ~(ISTRIP | IXON | ICRNL)) | IMAXBEL;
    attr.c_oflag = (attr.c_oflag & ~ OPOST                 )          ;
    attr.c_lflag = (attr.c_lflag & ~(ISIG | ICANON | ECHO ))          ;

    /* Many of the following special characters are probably
      disabled already by ISIG | ICANON being off, but it does not
      hurt to make sure. */
    attr.c_cc[VINTR   ] = anul;
    attr.c_cc[VQUIT   ] = anul;
    attr.c_cc[VERASE  ] = anul;
    attr.c_cc[VKILL   ] = anul;
    attr.c_cc[VEOF    ] = anul;
    attr.c_cc[VEOL    ] = anul;
    attr.c_cc[VEOL2   ] = anul;
    attr.c_cc[VSWTCH  ] = anul;
    attr.c_cc[VSTART  ] = anul;
    attr.c_cc[VSTOP   ] = anul;
    attr.c_cc[VSUSP   ] = anul;
    attr.c_cc[VDSUSP  ] = anul;
    attr.c_cc[VREPRINT] = anul;
    attr.c_cc[VDISCARD] = anul;
    attr.c_cc[VWERASE ] = anul;
    attr.c_cc[VLNEXT  ] = anul;
    /*   c_cc[VSTATUS ] = anul; */

    /* Set parameters for reading single bytes. Note, this
      must be done after the special keys are set (since
      the same locations are used). */
    attr.c_cc[VMIN    ] = 1;  /* Must get at least one character */
    attr.c_cc[VTIME   ] = 0;  /* Can wait indefinitely. */

    status = tcsetattr(standin,TCSANOW,&attr);
    if (status == -1) perror("tcsetattr");
   }
  }  /* rawkeys */



void waskeys ()
  /* Return keyboard to state before last RawKeys. */
  {
  int status;

  if (isatty(standin)) {
    status = tcsetattr(standin,TCSANOW,&wasattr);
    if (status == -1) perror("tcsetattr");
    }
  }  /* waskeys */



int keyqueue ()
  /* Report number of bytes in input queue. */
  {
  int qs,stat;

  qs = 0;  /* dummy */
  stat = ioctl(standin,I_NREAD,&qs);
  if (stat == -1) perror("ioctl");
  return qs;
  }   /* keyqueue */


 
void keytime (int delay)
  /* When reading the keyboard, wait at most delay/10 seconds, or
    indefinitely if delay = 0.
  */
  {
  int status;


  attr.c_cc[VTIME] = (char)delay;
  if (delay == 0) attr.c_cc[VMIN] = 1;   /* indef wait for one char */
  if (delay != 0) attr.c_cc[VMIN] = 0;   /* time-out possible */

  status = tcsetattr(standin,TCSANOW,&attr);
  if (status == -1) perror("tcsetattr");
  }  /* keytime */


void keyopost (int yes)
  /* Control output post-processing option. */
  {
  int status;


  if (yes) attr.c_oflag |=  OPOST;
  else     attr.c_oflag &= ~OPOST;

  status = tcsetattr(standin,TCSANOW,&attr);
  if (status == -1) perror("tcsetattr");
  }   /* keyopost */



int  keybyte (char *ch)
  /* Read one byte from the keyboard. Returns zero on success, else
    error code. */
  {
  char k;
  int  n;

  k= (char)0;
  n = read(standin,&k,1);
  *ch = k;
  if (n == 1) return  0;
         else return -1;
  }  /* keybyte */


char *keywhyc (int error)
  /* Explanation of "error" code.
    OS dependent. */
  {
  switch (errno) {
    case EAGAIN     : return "No data ready";
    case EBADF      : return "Invalid file descriptor";
    case ENOTTY     : return "Not a tty";
    case EINVAL     : return "Invalid attribute/options";
    case EFAULT     : return "Invalid buffer address";
    case EINTR      : return "Read interrupted by signal";
    case EIO        : return "I/O error";
    default         : return "??";
    }
  }  /* keywhyc */
