{ Keystroke-to-editor function mapping.

  Maps keystroke sequences to editor functions.
  Is called from Editor. Reads keystrokes until an editor function is
  formed. Any unrecognised sequences are defaulted to Ef_text i.e. text
  characters.

  Performs correctly (but inefficiently) even if the keymap syntax needs
  look-ahead. For example, it is permitted for ESC [ to start an editing
  function and ESC A to mean two (!) insertable characters, ESC and A.

  The matching can be made case-insensitive after specified prefixes,
  including the null prefix.

  No function keystroke may be a proper prefix of another. For
  example, if ESC A maps to a function, ESC A B cannot map to another
  function; it will me mapped to the ESC A function followed by text B
  (or whatever B maps to).

  Any number (zero, one, several) of keystroke sequences can map to a
  given editor function. Keystroke sequence length is limited by
  the default length of the 'string' type.

  Keystroke definitions can be read from a text file.

  Author:  Niklas Holsti
  Changes:
    91.06.17 NH  Added file parameter to Explain routine.
    92.04.08 NH  Port from Mac Turbo Pascal to SunOS.
                 Includes routines ReadChar and KeyPressed, to
                 simulate the Turbo Pascal console module
                 routines. 
    92.04.11 NH  Key symbol 'del' added (ascii 127).
    92.04.17 NH  Local vars declared 'private' for less risk of sym conflicts.
}

module Keymap;

#include "editfuns.p";
#include "keymap_d.p"
#include "strings_d.p"
#include "strings_f.p"
#include "keyboa_f.p"


procedure perror (s: univ_ptr);     { Unix }
  external c;


{ Strings to be matched are stored in an n-ary tree structure. Strings
  with common prefixes share storage. There are two trees, one for
  keystroke sequences and one for editing function names.

  The buffered keystrokes are held in a string (prefix). The partial
  mapping of this keystroke sequence is held as a pointer into the
  mapping tree (prefn).

  As a speed-up, the map keys of the tree root are also stored in a
  bit map (set of char).
}


type
  NodeKind = (prefixNode,  { stands for a prefix of keystroke seqs }
              finalNode,   { complete and unambiguous keystroke seq }
              ambigNode);  { complete but ambiguous keystroke seq }

  Link = ^Node;
  Node = record
         key : char;
         next: Link;
         case kind: NodeKind of
           prefixNode: (suffixes: Link;      { unordered list }
                        upcase  : boolean);
           finalNode : (fun     : EditFun);
           ambignode : ();                   { or set of EditFun }
         end {Node};


private var
  NameMap: Link;    { Function name -> function; pointer to header node. }
  KeysMap: Link;    { Keystroke seq -> function; ditto }

  topkeys: set of char;  { Keystrokes that may start a mapped sequence. }
  prefix : vstring; { Keystrokes read but not used in editing function }
  prefn  : Link;    { Node in KeysMap corresponding to prefix. If the
                      prefix is empty or does not map to anything,
                      prefn = KeysMap. }
  preupc : boolean; { Prefix specified uppercase conversion. }


private
procedure ShowKey (key: char; width: integer);
  { Display a key left-adjusted in a field of specified width. }
  var w: integer;
  begin
  w:=width;
  if ord(key) >= 32 then begin write(key); w:=w-1; end
                    else begin write('^',chr(ord(key)+64)); w:=w-2; end;
  while w >  0 do begin write(' '); w:=w-1; end;
  end {ShowKey};

private
procedure ShowMap (theMap: Link; indent: integer);
  { Display map structure. }
  begin
  if theMap <> nil then 
    with theMap^ do begin
      ShowKey(key,3);
      case kind of
        ambigNode : writeln('ambiguous');
        finalNode : writeln('fun = ',ord(fun));
        prefixNode: begin
                    if upcase then write('<> ');
                    ShowMap(suffixes, indent+3);
                    end;
        end;
      if next <> nil then begin
        write(' ': indent);
        ShowMap(next,indent);
        end;
      end;
  end {ShowMap};

procedure Km_ShowNames;
  begin
  writeln('Editor function names:');
  write(' ':3);
  if NameMap = nil then writeln('None defined.')
                   else ShowMap(NameMap^.suffixes,3);
  end {Km_ShowNames};

procedure Km_ShowKeys;
  begin
  writeln('Keystroke strings:');
  write(' ':3);
  if KeysMap = nil then writeln('None defined.')
                   else ShowMap(KeysMap^.suffixes,3);
  end {Km_ShowKeys};
  
procedure Km_ShowPrefix;
  var i: integer;
  begin
  write('Prefix: "');
  for i:=1 to length(prefix) do ShowKey(prefix[i],1);
  write('".');
  if preupc then write(' Uppercase conversion.');
  if (prefn = KeysMap) or (KeysMap = nil) then writeln(' No mapping.')
  else begin
    writeln(' Possible suffixes:');
    write(' ':3);
    case prefn^.kind of
      prefixNode: ShowMap(prefn^.suffixes,3);
      finalNode : writeln('Final node.');
      ambigNode : writeln('Ambiguous node.');
      end;
    end;
  end {Km_ShowPrefix};


procedure Km_ShowFunName (theFun: EditFun);

  var name: array [1..80] of char;

  procedure ShoFuNa (theMap: Link; level: integer);
    var i: integer;
    begin
    if theMap <> nil then
      with theMap^ do begin
        name[level]:=key;
        case kind of
          ambigNode : ;
          finalNode : if fun = theFun then
                         for i:=1 to level do write(name[i]);
          prefixNode: ShoFuNa(suffixes, level+1);
          end;
        ShoFuNa(next,level);
        end;
    end {ShoFuNa};

  begin
  if NameMap <> nil then ShoFuNa(NameMap^.suffixes,1);
  end {Km_ShowFunName};



procedure Km_Explain (var f: text; result: Km_result);
  begin
  case result of
    Km_Ok       : write(f,'Ok');
    Km_memfull  : write(f,'Memory full, cannot extend map');
    Km_isprefix : write(f,'Ambiguous (new key string is prefix of old)');
    Km_wasprefix: write(f,'Ambiguous (old key string is prefix of new)');
    Km_notOk    : write(f,'Ambiguous and/or undefined');
    Km_notprefix: write(f,'Not a prefix key string');
    end;
  end {Km_Explain};

private
function Raise (k: char): char;
  { Convert lower-case letter to upper-case. }
  begin
  if ('a'<=k) and (k<='z') then Raise:=chr(ord(k)+ord('A')-ord('a'))
                           else Raise:=k;
  end {Raise};

private
function FindKey (key: char; L: Link; upcase: boolean): Link;
  { Look for the key (possibly converted to uppercase) in L
    and successors.
    Return link to node with key, or nil. }
  var done: boolean;
      K   : link;
      kup : char;
  begin
  K:=L;
  if upcase then kup:=Raise(key)
            else kup:=key;
  repeat
    if K=nil then done:=true
             else done:=K^.key = kup;
    if not done then K:=K^.next;
  until done;
  FindKey:=K;
  end {FindKey};


private
procedure MapKeys (    keys  : vstring;
                       theMap: Link;
                   var match : integer;
                   var upc   : boolean;
                   var value : Link);
  { Map string 'keys' using theMap. The map tree is followed as far
    as matching nodes can be found. theMap must be non-nil.
    The 'match' parameter is returned as the length of the prefix of 'keys'
    that could be matched with the map, and 'value' is the last node matched.
    Thus, if match = length(keys), the mapping succeeded and 'value' is
    the result. Otherwise, match < length(keys) and 'value' is the node
    whose suffix list needs an addition, if 'keys' is to be inserted
    in the map. For example, if match=0, then value=theMap.
    If 'value' is not an internal node (kind prefixNode) but
    match < length(keys), the prefix keys[1..match] is mapped to 'value'.
    In this case, 'keys' cannot be inserted in the map.
    The 'upc' parameter reports whether the uppercase conversion was
    activated on some matching level. }
  var V: Link;
  begin
  match:=0;
  value:=theMap;
  V    :=value;
  upc  :=value^.upcase;
  while (match < length(keys))
    and (V <> nil)
  do begin
    upc:= upc or value^.upcase;
    V  := FindKey(keys[match+1],value^.suffixes,upc);
    if V <> nil then begin
      match:=match+1;
      value:=V;
      if value^.kind = prefixNode
      then upc:= upc or value^.upcase
      else V  := nil;
      end;
    end;
  end {MapKeys};


private
procedure InitMap (var theMap: Link; var result: Km_result);
  { Initialise theMap or check that it is initialised. }
  begin
  result:=Km_Ok;  { so far }
  if theMap = nil then begin
    { writeln('Initialising map.'); readln; }
    new(theMap);
    if theMap = nil then result:=Km_memfull
    else begin
      { writeln('Map header allocated.'); readln; }
      theMap^.kind    := prefixNode;
      theMap^.suffixes:= nil;
      theMap^.upcase  := false;
      theMap^.next    := nil;
      { writeln('Map header initialised.'); readln; }
      end;
    end;
  end {InitMap};


private
procedure ExtendMap (    theMap: Link;
                         keys  : vstring;
                         upcase: boolean;
                         kind  : NodeKind;
                     var value : Link;
                     var result: Km_result);
  { Extend theMap to include the argument keys. We assume that it has
    already been determined that theMap^.suffixes does not match keys[1].
    The keys are optionally converted to upper case.
    The kind of the node to which keys is mapped is specified.
    If it is a prefixNode, the suffix list is set to nil and the upper-
    case conversion is turned off. Other node fields are not set.
    The node is returned in the 'value' parameter. }
  var k: integer;
      M: Link;  { the new map branch, built bottom up }
      N: Link;  { a new map node }
  begin
  { writeln('Extend map "',keys,'"'); readln; }
  { The new branch is constructed bottom-up so that the nodes are easy
    to delete if we run into trouble on the way. }
  k:=length(keys);
  result:=Km_Ok;  { so far }
  M:=nil;
  while (k >= 1)
  and   (result = Km_Ok)
  do begin
    new(N);
    if N = nil then result:=Km_memfull
    else begin
      N^.key :=keys[k]; if upcase then N^.key:=Raise(N^.key);
      N^.next:=nil;

      if k = length(keys) then begin
        N^.kind:=kind;
        value  :=N;
        end
      else
        N^.kind:=prefixNode;

      if N^.kind = prefixNode then begin
        N^.suffixes:=M;
        N^.upcase  :=false;
        end;

      M:=N;
      k:=k-1;
      end;
    end;
  if result=Km_Ok then begin
    { Insert the new branch at the head of the existing map. }
    M^.next:=theMap^.suffixes;
             theMap^.suffixes:=M;
    end
  else begin
    { Some error. Delete branch nodes, if any created. }
    while M<>nil do begin
      N:=M^.suffixes;
      dispose(M);
      M:=N;
      end;
    value:=nil;
    end;
  end {ExtendMap};


private
procedure MakeMap (var theMap: Link;
                       theFun: EditFun; keys: vstring;
                   var result: Km_result);
  { Make a map or extend an existing map. }
  var match: integer;
      mapv : Link;
      value: Link;
      upc  : boolean;
  begin
  InitMap(theMap,result);
  if result = Km_Ok then begin
    MapKeys(keys,theMap,match,upc,mapv);
    if match >= length(keys) then result:=Km_isprefix
    else
    if mapv^.kind <> prefixNode then result:=Km_wasprefix
    else begin
         { writeln('Going to extend map.'); readln; }
         ExtendMap(mapv,
                   copy(keys,match+1,length(keys)-match),
                   upc,finalNode,value,result);
         if result = Km_ok then
           value^.fun:=theFun;
         end;
    end;
  end {MakeMap};


procedure Km_NameFun (theFun: EditFun; name: vstring;
                      var result: Km_result);
  begin
  MakeMap(NameMap,theFun,name,result);
  end {Km_NameFun};


private  
procedure NoPrefix;
  { Erase buffered prefix. }
  begin
  prefix:='';
  prefn :=KeysMap;
  if prefn <> nil then preupc:=prefn^.upcase
                  else preupc:=false;
  end {NoPrefix};


procedure Km_Map (    theFun: EditFun;
                      keys  : vstring;
                  var result: Km_result);
  begin
  MakeMap(KeysMap,theFun,keys,result);
  if result = Km_ok then topkeys:=topkeys + [keys[1]];
  NoPrefix;
  end {Km_Map};


procedure Km_Upcase (    prefix: vstring;
                     var result: Km_result);
  var match: integer;
      mapv : Link;
      value: Link;
      upc  : boolean;
  begin
  InitMap(KeysMap,result);
  if result = Km_Ok then begin
    MapKeys(prefix,KeysMap,match,upc,mapv);
    if match >= length(prefix) then
      value:=mapv
    else begin
      { writeln('Going to extend map.'); readln; }
      ExtendMap(mapv,
                copy(prefix,match+1,length(prefix)-match),
                upc,prefixNode,value,result);
      end;
    if result = Km_Ok then
      if value^.kind <> prefixNode then result:= Km_notprefix
      else value^.upcase:=true;
    end;
  end {Km_Upcase};
    

procedure Km_ReadMap (var f: text;
                      var undef,ambig: boolean;
                      var result: Km_result);
  var
    endf : boolean;
    c    : char;
    name : vstring;
    prevname: vstring;
    keys : vstring;
    key  : vstring;
    kord : integer;
    match: integer;
    upc  : boolean;
    mapv : Link;
    resu : Km_result;

  procedure skipsp;
    { Skip past spaces and comments until non-blank or eoln.
      Nonblank character in c. }
    begin
    c:=' ';
    while (c=' ') and (not eoln(f)) do read(f,c);
    if c='#' then begin
      while not eoln(f) do read(f,c);
      c:=' ';
      end;
    end {skipsp};

  procedure getword (var word: vstring);
    { Skip past spaces, read seq of non-blanks into word. }
    begin
    word:='';
    skipsp;
    while c <> ' ' do begin
      word:=concat(word,c);
      if eoln(f) then c:=' '
                 else read(f,c);
      end;
    end {getword};

  procedure problem; 
    { Display data causing a problem. }
    var i: integer;
    begin
    write('Name: "',name,'"');
    write('   Keys: "');
    for i:=1 to length(keys) do ShowKey(keys[i],1);
    write('"   ');
    end {problem};

  procedure undefined;
    begin
    problem; writeln('Undefined name.');
    undef:=true;
    end {undefined};

  procedure checkresu;
    begin
    if resu <> Km_ok then begin
      problem;
      Km_Explain(output,resu);
      writeln;
      if resu in [Km_isprefix,Km_wasprefix] then ambig:=true;
      if result = Km_ok then result:=resu;
      end;
    end {checkresu};

  begin
  result:=Km_Ok;
  undef:=false;
  ambig:=false;
  endf :=eof(f);
  prevname:='';
  while not endf do begin
    keys:='';
    getword(name);
    if name = '"' then name:=prevname;
    prevname:=name;
    if name='.' then endf:=true
    else
    if (name<>'') and (not eoln(f)) then begin
      repeat
        getword(key);
        { writeln('Key: "',key,'"'); }
        if key<>'' then begin
          if (key[1]='^') and (length(key)=2) then begin
            kord:=ord(key[2]);
            if      kord >= 96 then key:=chr(kord-96)
            else if kord >= 64 then key:=chr(kord-64)
                               else key:='';
            end
          else if length(key) > 1 then begin 
            if key='sp'       then key:=' '       else
            if key='hash'     then key:='#'       else
            if key='esc'      then key:=chr(27)   else
            if key='csi'      then key:=chr(155)  else
            if key='del'      then key:=chr(127)  else
                                   key:='';
            end;
          keys:=concat(keys,key);
          end;
      until (key='') or eoln(f);
      end;

    if name='<>' then begin
      Km_Upcase(keys,resu);
      checkresu;
      end

    else
    if  (name<>'')
    and (keys<>'')
    then begin
      MapKeys(name,NameMap,match,upc,mapv);
      if      match < length(name)    then undefined
      else if mapv^.kind <> finalNode then undefined
      else begin
        Km_Map(mapv^.fun,keys,resu);
        checkresu;
        end;
      end;

    readln(f);
    if eof(f) then endf:=true;
    end;
  NoPrefix;
  end {Km_ReadMap};


procedure Km_GetFun (var theFun  : EditFun;
                     var textChar: char;
                         delay   : integer;
                     var timeout : boolean);
  var
    match    : integer;
    funformed: boolean;
    key      : char;

  procedure GetKey;
    { Wait for a key to be pressed. }
    begin
    KeyTime(delay);
    timeout:= KeyByte(key) = -1;
    end {GetKey};

  procedure AppKey;
    { Append key to prefix and map it, or, if the prefix
      is empty and the key cannot start a keystroke sequence,
      return it as text, bypassing the prefix. }
    var keyn: Link;
    begin
    if length(prefix) = 0 then
      if not (key in topkeys) then begin
        funformed:=true;
        theFun   :=Ef_text;
        textChar :=key;
        end;
    if not funformed then begin
      prefix:=concat(prefix,key);
      keyn  :=FindKey(key,prefn^.suffixes,preupc);
      if keyn <> nil then prefn:=keyn
                     else prefn:=KeysMap;   { no map }
      end;
    end {AppKey};

  begin
  timeout :=false;
  theFun  :=Ef_undefined;   { unless changed below }
  textChar:=' ';            { ditto }

  if KeysMap = nil then begin
    { Keystroke map uninitialised. Keystrokes returned unmapped. }
    GetKey;
    if not timeout then begin
      theFun  :=Ef_text;
      textChar:=key;
      end;
    end
  else begin
    { Accumulate keypresses until a function is formed, or timeout. }
    funformed:=false;
    repeat
      { See if the existing prefix forms a function of some kind. }
      if prefn = KeysMap then begin
        { Existing prefix does not map to anything. }
        { If prefix is not empty, its head can be returned as text. }
        funformed := length(prefix) > 0;
        if funformed then begin
          { Take first keystroke of prefix. }
          theFun  :=Ef_text;
          textChar:=prefix[1];
          delete(prefix,1,1);
          { See if the remainder maps to something. }
          MapKeys(prefix,KeysMap,match,preupc,prefn);
          end;
        end
      else begin
        { Existing prefix maps to a node, but a function is not }
        { yet formed if the node is internal (prefixNode). }
        funformed:= prefn^.kind <> prefixNode;
        if funformed then begin
          case prefn^.kind of
            finalNode: theFun:= prefn^.fun;
            ambigNode: theFun:= Ef_ambiguous;
            end;
          NoPrefix;
          end {funformed}
        else
          preupc:=preupc or prefn^.upcase;
        end;

      if not funformed then begin
        { Need more keystrokes. }
        GetKey;
        if not timeout then AppKey;
        end;

    until funformed or timeout;

    end;
  end {Km_GetFun};


function KeyPressed: boolean;
  begin
  KeyPressed:= KeyQueue > 0;
  end {KeyPressed};


function ReadChar: char;
  var c: char;
  begin
  KeyTime(0);
  if KeyByte(c) <> 0 then begin
    perror(addr('ReadChar'));
    c:=chr(0);
    end;
  ReadChar:=c;
  end {ReadChar};


procedure Km_Init;
  begin
  { writeln('Keymap for script editor. 1991.04.14.A'); }
  NameMap:=nil;
  KeysMap:=nil;
  topkeys:=[];
  NoPrefix;
  RawKeys;
  end {Km_Init};

procedure Km_Fini;
  begin
  WasKeys;
  end {Km_Fini};
