{
  Lexical routines in Pascal to interface with incremental (reversible)
  script execution under a SEED-like session editor.
  Line-buffered version.

  Author :  Niklas Holsti
  Changes:
     91.06.17  NH  Power routine corrected.
                   Type Lex_RealType added.
                   CancelSignal token-kind added.
     91.06.21  NH  Log module used before Windows (which needs it).
     91.09.03  NH  Lex_Define reports errors, does not have 'Ok' param.
                   Lex_Scan added.
     91.09.05  NH  Symbol matching ignores letter case.
                   Overlong symbols cause lexical error.
                   Lex_Line and Lex_Scan set the 'used' position.
     91.09.25  NH  Length of current token kept, given to Ed_Tell.
     92.04.11  NH  Port to Sun Unix.
                   Editor no longer exports "nochar" or "eolchar", so
                   we have local vars for them.
     92.04.17  NH  Module caused any program using it to dump core
                   when terminating. Reason: local variable exit: boolean
                   conflicted with exit() routine. Defined variables
                   as private.
     92.12.01  NH  To support 'manual' execution mode, believes what
                   Ed_Getln tells it about the position of the
                   line to be analysed.
                   Lex_CharSets added (reason mistaken - no comment).
     93.04.02  NH  Ed_written is called at start of Lex_Get, Lex_Next,
                   and Lex_Null, to help handle output from applications
                   that use Ed_WriteFrom.
     93.04.02  NH  position.char = nochar checked also after each ignored
                   token. Also, each ignored token is skipped by advancing
                   position. (Problem was: if EolToken was ignored, the
                   nochar produced a MarkToken.)
     93.04.19  NH  Adding quoted string tokens, quote chars settable.
                   Word head and body char set selectable.
                   Comment style selectable (Ada-style, or any
                   single char as rest-of-line comment introducer).
                   Max word length increase to 80.
     93.04.21  NH  OpOnSet string parameter is value (not var). For some
                   reason, having it a var parameter caused a bus error
                   (data alignment) when calling from C.
     94.07.28  NH  Editor can now send "abort" messages (ctrl-G keystroke).
                   These are ignored for now.
     94.09.19  NH  Keep "cancel" index at "infinite" when not in cancelling phase.
                   This lets one restore a freshly stored position, which
                   is nice in some programming styles.
     94.10.24  NH  Omit "tag" string from Lex positions. This has not been
                   used for some time and is potentially expensive in
                   execution time (string copying).
 }

module Lex;

#include "strings_d.p"
#include "log_f.p"
#include "edit_d.p"
#include "edit_f.p"

#include "lexline_d.p"

const
  lexlog  =  1;   { log principal operations. }
  MaxSyms = 40;   { at most this many symbols can be defined }

type
  Lex_CharSet   = set of char;
  TokenCharSet  = array [TokenKind] of Lex_CharSet;
  TokenKinds    = set of TokenKind;

var

  Lex_token: TokenType;        { the current token }

private var

  ignored  : array [TokenKind] of boolean;
  symbols  : array [0..MaxSyms] of record { a very trivial implementation }
                 word  : Lex_wordType;    { also, [0] is dummy }
                 symnum: integer;
                 end;
  topsym   : integer;

  firstc, bodyc: TokenCharSet;
  quotepair    : array [char] of char;    { quotepair[opener] = closer }

  numerals : Lex_CharSet;
  blanks   : Lex_CharSet;
  letters  : Lex_CharSet;
  alphas   : Lex_CharSet;
  
  position : Lex_PositionType; { current analyser position }

  linebuf  : Lex_LineType;     { one script line (or end of it) }
  linepos  : ScriptIndex;      { index of linebuf[1], <0 if linebuf empty }
  linelen  : integer;          { length of linebuf, 0 if linebuf empty }
  nochar   : char;             { signifies "character not present" }
  eolchar  : char;             { signifies "newline" }

  error    : ScriptIndex;
  cancelled: boolean;          { exception }
  exit     : boolean;          { part of exception handling }
  cancel   : ScriptIndex;      { the position given in a cancel signal, or
                                 infinite if not in cancelling phase }

  infiniteindex: ScriptIndex;  { Ed_InfiniteIndex (constant) }


private
procedure Logpos (pos: Lex_PositionType);
  { Describe Lex position on logfile. }
  begin
  with pos do begin
    write(logfile, { '[', tag, '] ', } 
                   'last=',last:1,' start=',start:1,' char=');
    if character = nochar then write(logfile,'none')
    else write(logfile,'"',character,'"');
    end;
  end {Logpos};

procedure Lex_Store (var safe: Lex_PositionType { ; tag: vstring } );
  begin
{ position.tag:= tag;   }
  safe := position;
  if Log_includes(lexlog) then begin
    write(logfile,'Lex_Store');
    Logpos(safe);
    writeln(logfile);
    end;
  end {Lex_Store};

procedure Lex_Null;
  begin
  Ed_Written;
  Lex_token.kind := NullToken;
  position.start := position.last + 1;
  position.length:= 0;
  end {Lex_Null};

procedure Lex_exit;
  begin
  exit:=true;
  end {Lex_exit};

function Lex_done: boolean;
  begin
  Lex_done := exit;
  end {Lex_done};

function Lex_cancel: boolean;
  begin
  Lex_cancel := cancelled;
  end {Lex_cancel};


private
procedure NextChar;
  { Gets script[position.start] into position.character.
    Assumed to be within current line, or the newline
    after it, or at the start of the next line.
    Fills line buffer if necessary.
    If an error is present, the rest of the current line is abandoned.
    May raise cancelled exception, and is certain to do so if an
    error is present. Sets token.kind to CancelSignal in this case. }
  var
    fst: integer;   { first char of ((erroneous) token }
    i  : integer;   { index in line }
    msg: Ed_MsgType;
  begin
  i:= position.start - linepos + 1;
  if (i     > linelen+1    )   { in next line (past newline) }
  or (error < infiniteindex)   { an error is present }
  then begin
    repeat
      fst:=position.start;
      if error < infiniteindex then
        position.start:= position.last + 1;  { otherwise, editor thinks we are
                                               re-reading some input. }
      Ed_Tell (position.last,error,position.start,
               fst,position.length);
      Ed_Getln(linebuf,linelen,msg);
      case msg.kind of
        ProcessMsg: begin
            linepos:= msg.processfrom;
            i      := 1;
            position.start:= linepos;
            end;
        CancelMsg: begin
            cancel   := msg.cancelto;
            cancelled:= true;
            Lex_token.kind:= CancelSignal;
            if Log_includes(lexlog) then
              writeln(logfile,'Lex.NextChar cancel=',cancel:1);
            end;
        AbortMsg: begin
            if Log_includes(lexlog) then
              writeln(logfile,'Lex.NextChar received and ignored abort');
            end;
        end;
    until msg.kind <> AbortMsg;
    end {get next line};

  if not cancelled then begin
    with position do begin
      if i <= linelen then character:= linebuf[i]
                      else character:= eolchar;  { or eofchar? }
      last := start;
      start:= last+1;
      end;
    end;
  end {NextChar};      

private
procedure TellUsed;
  { Report last-used position to editor. }
  begin
  Ed_Used(position.last);
  end {TellUsed};


procedure Lex_Error;
  begin
  error := position.last;
  if Log_includes(lexlog) then
    writeln(logfile,'Lex_Error Err=',error:1,
           ' Start=',position.start:1,
           ' Length=',position.length:1);
  NextChar;   { should raise cancelled exception }
  if not cancelled then
    writeln(logfile,'Lex_Error (',error:1,') did not raise cancelled !');
  end {Lex_Error};


procedure Lex_Scan (var eol: boolean);
  begin
  Lex_Null;

  if (position.character = nochar)    { last character was consumed }
  or (linepos < 0)                    { or line buffer is empty }
  then NextChar;

  while (position.character = ' ') and (not cancelled) do NextChar;

  if not cancelled then begin
    TellUsed;
    eol := position.character = eolchar;
    if eol then position.character:= nochar;
    end;
  end {Lex_Scan};


procedure Lex_Line (var line  : Lex_LineType;
                    var length: integer    );
  { Get rest of current script line, without lexical analysis.
    Returns text of line and its length. }
  begin
  Lex_Null;
  if (position.character = nochar)    { last character was consumed }
  or (linepos < 0)                    { or line buffer is empty }
  then NextChar;
  length:=0;
  while (position.character <> eolchar) and (not cancelled) do begin
    length:=length+1;
    line[length]:=position.character;
    NextChar;
    end;
  if not cancelled then begin
    TellUsed;
    position.character:= nochar;
    end;
  end {Lex_Line};


function Power (x: Lex_RealType; n: integer): Lex_RealType;
  { Return x to the power n. }
  var k: integer;
      p,z: Lex_RealType;
  begin
  if n >= 0 then k:= n
            else k:=-n;
  p:=1.0; z:=x;
  while k > 0 do begin
    if odd(k) then p:=p*z;
    k:=k div 2;
    z:=sqr(z);
    end;
  if n < 0 then p:=1.0/p;
  Power:=p;
  end {Power};
  

private
procedure Upcase (mixed: vstring; var upper: Lex_WordType);
  { upper := mixed with lowercase letters converted to uppercase. }
  var i: integer;
      c: char;
  begin
  upper:=mixed;
  for i:=1 to length(upper) do begin
    c:=upper[i];
    if (c >= 'a') and (c <= 'z') then
      upper[i]:= chr(ord(c) - ord('a') + ord('A'));
    end;
  end {Upcase};


procedure Lex_Analyse;
  { Skip blanks, try to recognise a token. }

  const
    MaxDigits =   30;  { at most this many digits (whole + decimals) }
    MaxExp    = 4930;  { number range is 10^(-MaxExp)..10^(MaxExp).  }

  type
    DigitSeq  = record
                digit: array [1..MaxDigits] of integer;
                num  : integer;
                end;
  var
    tokenstart: ScriptIndex;
    complete  : boolean;

  procedure GetDigits (var Ds: DigitSeq; var overflow: boolean);
    { Add more digits. Report whether there are too  many. }
    begin
    while (position.character in numerals)
      and (not cancelled)
    do begin
      if position.character in ['0'..'9'] then begin
        Ds.num:=Ds.num+1;
        if Ds.num <= MaxDigits then
          Ds.digit[Ds.num]:=ord(position.character)-ord('0');
        end;
      NextChar;
      end;
    overflow:= Ds.num > MaxDigits;
    if overflow then begin
      TellUsed;
      Ed_write('Too many digits...');
      end;
    end {GetDigits};

  procedure IntDig (var {i} Ds: DigitSeq;
                    var {o} N : integer;
                    var {o} overflow: boolean);
    { Return integer value of decimal representation.
      Report overflows. }
    var i,d: integer;
    begin
    N:=0; overflow:=false; i:=1;
    while (i <= Ds.num) and (not overflow) do begin
      d:=Ds.digit[i];
      overflow:= N > (maxint - d) div 10;
      if not overflow then begin
        N:=N*10 + d;
        i:=i+1;
        end;
      end;
    if overflow then begin
      TellUsed;
      Ed_write('Integer overflow...');
      end;
    end {IntDig};

  function  RealDig (var {i} Ds: DigitSeq): Lex_RealType;
    { Return real value of decimal representation, with an implied
      decimal point in front of the first digit.
      Thus, the value is >=0 and < 1. }
    var R: Lex_RealType;
        i: integer;
    begin
    R:=0.0;
    i:=Ds.num;
    while i >= 1 do begin
      R:=(R + Ds.digit[i])*0.1;  { Note, should use high precision const }
      i:=i-1;
      end;
    RealDig:=R;
    end {RealDig};

  procedure isNumber;
    label 66;
    var
      digits,expdigits: DigitSeq;
      natnum          : integer;
      expdot,expval   : integer;
      expsign         : integer;
      overflow        : boolean;
    begin
    { Digits before decimal point. }
    digits.num:=0;
    GetDigits(digits,overflow);   if cancelled or overflow then goto 66;
    { Decimal point or exponent part? }
    if not (position.character in ['.','E','e','D','d']) then begin
      { Natural number without decimals or exponent. }
      IntDig(digits,natnum,overflow);  if overflow then goto 66;
      Lex_token.kind     := NatNumToken;
      Lex_token.natnumber:= natnum;
      { position.character is not included in token. }
      end {nat num}

    else begin
      { Decimal and/or exponent parts. }
      expdot:=digits.num;
      if position.character = '.' then begin
        NextChar;
        GetDigits(digits,overflow); if cancelled or overflow then goto 66;
        end;
      if position.character in ['E','e','D','d'] then begin
        NextChar;                   if cancelled then goto 66;
        if position.character ='-' then expsign:=-1
                                   else expsign:=+1;
        if position.character in ['+','-'] then NextChar;
        expdigits.num:=0;
        GetDigits(expdigits,overflow); if cancelled or overflow then goto 66;
        IntDig(expdigits,expval,overflow); if overflow then goto 66;
        if expval < maxint div 2 then expval:=expdot + expsign*expval;
        end {exp}
      else
        expval:=expdot;
      if abs(expval) > MaxExp then begin
        TellUsed;
        Ed_write('Real number overflow...');
        goto 66;
        end;
      Lex_token.kind      := RealNumToken;
      Lex_token.realnumber:= RealDig(digits) * Power(10.0,expval);
      { position.character is not included in token. }
      end {decimal};
 66 : ;
    end {isNumber};

  procedure isWordOrSym;
    label 66;
    var word: Lex_WordType;
        upwo: Lex_WordType;
        i   : integer;
    begin
    word:='';
    repeat
      if length(word) >= Lex_maxWord then begin
        TellUsed;
        Ed_write('This word is too long.');
        goto 66;
        end;
      word:=word + position.character;
      NextChar;
    until (not (position.character in alphas)) or cancelled;
    if not cancelled then begin
      i:=topsym;
      Upcase(word,upwo);
      while (symbols[i].word <> upwo) and (i > 0) do i:=i-1;
      if i > 0 then begin
        Lex_token.kind  := SymbolToken;
        Lex_token.symnum:= symbols[i].symnum;
        end
      else begin
        Lex_token.kind := WordToken;
        Lex_token.word := word;
        end;
      { token does not include position.character }
      end;
  66: ;
    end {isWordOrSym};

  procedure isEol;
    begin
    Lex_token.kind:= EolToken;
    position.character:=nochar;  { position.character is included in token. }
    end {isEol};
      
  procedure isString;
    label 66;
    var body: vstring;
        opener,closer: char;
    begin
    opener:= position.character;
    closer:= quotepair[opener];
    body:='';
    repeat
      NextChar;     if cancelled then goto 66;
      if position.character = eolchar then begin
        TellUsed;
        Ed_write('String closing quote is missing.');
        goto 66;
        end;
      if position.character <> closer then begin
        if length(body) >= Lex_maxString then begin
          TellUsed;
          Ed_write('String too long.');
          goto 66;
          end;
        body:= body + position.character;
        end;
    until position.character = closer;

    Lex_token.kind  := StringToken;
    Lex_token.quoted:= body;
    Lex_token.opener:= opener;
    Lex_token.closer:= closer;
    position.character:= nochar; { token includes position.character }

  66: ;
    end {isString};
      
  procedure isComOrMark;
    { Perhaps an end-of-line comment, or a special mark. }
    var mark: char;
    begin
    mark:=position.character;
    NextChar;
    if not cancelled then begin
      if (position.character in bodyc[EolCommentToken])
      or ( ' '               in bodyc[EolCommentToken])
      then begin
        repeat NextChar
        until (position.character = eolchar) or cancelled;
        if not cancelled then begin
          Lex_token.kind:=EolCommentToken;
          { position.character (=eol) is not included in token }
          end
        end {comment}
      else begin
        Lex_token.kind     :=MarkToken;
        Lex_token.character:=mark;
        { position.character is not included in token }
        end {mark};
      end;
    end {isComOrMark};

  procedure isMark;
    begin
    Lex_token.kind     :=MarkToken;
    Lex_token.character:=position.character;
    position.character :=nochar;  { position.character is included in token }
    end {isMark};

  procedure Atoken;
    var c: char;
    begin
    { Skip blanks }
    while (position.character in blanks) and not cancelled
    do NextChar;

    if not cancelled then begin
      Lex_token.kind:=NullToken;      { unless analysis succeeds }
      tokenstart    :=position.last;  { first character of token }

      { Decide what kind of token it seems to be. }
      c:= position.character; { just for abbreviation }
      if      c in firstc[NatNumToken    ] then isNumber
      else if c in firstc[WordToken      ] then isWordOrSym
      else if c in firstc[StringToken    ] then isString
      else if c in firstc[EolCommentToken] then isComOrMark
      else if c  = eolchar                 then isEol
      else                                      isMark;

      if Lex_token.kind = NullToken then begin
        { A lexical error message has already been emitted. }
        Ed_Writeln;
        position.length:=0;
        Lex_Error;
        end
      else begin
        TellUsed;
        position.start :=tokenstart;
        position.length:=position.last - tokenstart;
        if position.character = nochar then
          position.length:=position.length  + 1;
        end;
      end;
    end {Atoken};

  begin {Lex_Analyse}
  repeat
    if position.character = nochar then NextChar;
    Atoken;
    if cancelled then complete:= true
    else begin
      complete:= not ignored[Lex_token.kind];
      if not complete then begin
        { Skip the ignored token, like Lex_Null. }
        position.start := position.last + 1;
        position.length:= 0;
        end;
      end;
  until complete; 
  end {Lex_Analyse};


procedure Lex_Restore (safe: Lex_PositionType);
  begin
  cancelled := safe.last >= cancel;

  if Log_includes(lexlog) then begin
    write(logfile,'Lex_Restore');
    Logpos(safe);
    write(logfile,' cancel=',cancel:1);
    if cancelled then writeln(logfile,' cancelled')
                 else writeln(logfile,' restored');
    end;

  if not cancelled then begin
    position:=safe;
    error   :=infiniteindex;
    cancel  :=infiniteindex;
    linepos :=-1;  { The line buffer is empty. }
    linelen := 0;
    if position.start <= position.last then begin
      { The safe position had a valid token, which must be re-read. }
      position.character:= nochar;  { to force filling of line buffer }
      Lex_Analyse;
      end
    else
      Lex_token.kind := NullToken;
    end;
  exit:=cancelled;
  end {Lex_Restore};


procedure Lex_Next;
  var safe: Lex_PositionType;
  begin
  Ed_Written;
  position.start:= position.last + 1;
  Lex_Store(safe {,'Lex_Next'});
  repeat
    Lex_Analyse;
    exit:=true;
    if cancelled then Lex_Restore(safe);
  until exit;
  end {Lex_Next};

procedure Lex_Get;
  begin
  Ed_Written;
  if Lex_token.kind = NullToken then Lex_Next;
  end {Lex_Get};


function Lex_isMark (mark: char): boolean;
  begin
  Lex_Get;
  if cancelled then Lex_isMark:=false
  else
  if Lex_token.kind <> MarkToken then Lex_isMark:=false
  else
  Lex_isMark:= Lex_token.character = mark;
  end {Lex_isMark};


function Lex_isSymbol (num: integer): boolean;
  begin
  Lex_Get;
  if cancelled then Lex_isSymbol:=false
  else
  if Lex_token.kind <> SymbolToken then Lex_isSymbol:=false
  else
  Lex_isSymbol:= Lex_token.symnum = num;
  end {Lex_isSymbol};

procedure Lex_Int (var N: integer);
  label 66;
  var safe: Lex_PositionType;
  begin
  Lex_Store(safe{ ,'Lex_Int'});
  repeat
    Lex_Get;                                if cancelled then goto 66;
    if Lex_token.kind = NatNumToken then begin
      N := Lex_token.natnumber;
      Lex_Null;
      end
    else begin
      Ed_Write('Integer number expected...');
      Ed_Writeln;
      Lex_Error;
      end;
    exit:=true;
  66: if cancelled then Lex_Restore(safe);
  until exit;
  end {Lex_Int};

procedure Lex_Define (word  : Lex_wordType;
                      symnum: integer);
  var ok: boolean;
  begin
  ok:=topsym < MaxSyms;
  if ok then begin
    topsym:=topsym+1;
    Upcase(word,symbols[topsym].word);
                symbols[topsym].symnum:=symnum;
    end
  else begin
    Ed_write('Lex_Define: symbol "');
    Ed_write(word);
    Ed_write('" not defined (table full)"');
    Ed_writeln;
    end;
  end {Lex_Define};

procedure Lex_Ignore (kind: TokenKind);
  begin
  ignored[kind]:=true;
  end {Lex_Ignore};


private
procedure Equivalent (kind: TokenKind; equiv: TokenKinds;
                      var tokc: TokenCharSet);
  { If kind is in the equivalence set equiv, propagate its
    character set to all equivalent token classes. }
  var k: TokenKind;
  begin
  if kind in equiv then
    for k:=NullToken to EofToken do
      if k in equiv then tokc[k] := tokc[kind];
  end {Equivalent};


private
procedure Derived;
  { Derive some character sets from the firstc/bodyc sets. }
  begin
  numerals:= firstc[NatNumToken] + bodyc[NatNumToken]; { or RealNum...}
  alphas  :=                       bodyc[WordToken  ]; { or Symbol... }
  end {Derived};


private
procedure OpOnSet (var      charset: Lex_CharSet;
                            op   : Lex_CsOpType;
                   {var in} chars: vstring);
  { Set/extend/reduce the charset as specified by op and chars. }
  var
    i: integer;
    cs: Lex_CharSet;
  begin
  { Collect the set of characters from the string. }
  cs:=[];
  for i:=1 to length(chars) do cs:= cs + [chars[i]];
  { Modify the char set for the token. }
  case op of
    Lex_Set    : charset :=           cs;
    Lex_Include: charset := charset + cs;
    Lex_Exclude: charset := charset - cs;
    otherwise
      writeln('Lex: char-set operator "',op,'" not understood.');
    end {case};
  end {OpOnSet};


procedure Lex_FirstChar (kind: TokenKind; op: Lex_CsOpType; chars: vstring);
  { Define the characters that can start a token of the given kind.
    The 'op' is one of '=', '+', '-' and the character set is
    correspondingly set equal to the given 'chars' or the given
    chars are added to or removed from the set. (The type string
    is used, rather than 'set of char', to ease calls from C. }
  begin
  OpOnSet(firstc[kind],op,chars);
  Equivalent(kind, [NatNumToken,RealNumToken], firstc);
  Equivalent(kind, [SymbolToken,WordToken   ], firstc);
  Derived;
  end {Lex_FirstChar};


procedure Lex_BodyChar (kind: TokenKind; op: Lex_CsOpType; chars: vstring);
  { Define the characters that can appear in the body of a token of
    the given kind. The parameters correspond to those of Lex_FirstChar.
    For EolCommentToken, the 'body' means just the second character; if
    a blank space can appear in the body, a comment can be introduced by
    any one character from the allowed first character set.
    Thus, Ada-style comments are specified by first = '-', body  = '-',
    and C-style comments by first = '#', body = ' '. }
  begin
  OpOnSet(bodyc[kind],op,chars);
  Equivalent(kind, [NatNumToken,RealNumToken], bodyc);
  Equivalent(kind, [SymbolToken,WordToken   ], bodyc);
  Derived;
  end {Lex_BodyChar};


procedure Lex_QuotePair (left,right: char);
  begin
  quotepair[left] := right;
  end {Lex_QuotePair};


procedure Lex_BlankChar (op: Lex_CsOpType; chars: vstring);
  begin
  OpOnSet(blanks,op,chars);
  Derived;
  end {Lex_BlankChar};


procedure Lex_Init;
  { Initialize local variables. }
  var k: TokenKind;
      c: char;
      any: Lex_CharSet;
  begin
  { writeln('Lex for incremental script processing. 1991.05.13.A'); }

  for k:=NullToken to EofToken do begin
    ignored[k]:=false;
    firstc [k]:= [];    { default, changed for some tokens below }
    bodyc  [k]:= [];    { ditto }
    end;

  topsym:= 0;
  symbols[0].word:='';

  any := [];
  for c:=chr(0) to chr(255) do begin
    any := any + [c];
    quotepair[c] := c;
    end;

  firstc[NatNumToken ] := ['0'..'9'];
  bodyc [NatNumToken ] := firstc[NatNumToken];
  firstc[RealNumToken] := ['0'..'9'];
  bodyc [RealNumToken] := firstc[RealNumToken];
  firstc[SymbolToken ] := ['A'..'Z','a'..'z'];
  bodyc [SymbolToken ] := firstc[SymbolToken] + ['0'..'9','_'];
  firstc[WordToken   ] := firstc[SymbolToken];
  bodyc [WordToken   ] := bodyc [SymbolToken];
  firstc[StringToken ] := ['"',''''];
  bodyc [StringToken ] := any;
  firstc[EolCommentToken] := ['-'];
  bodyc [EolCommentToken] := ['-'];

  blanks   :=[' ',chr(9)];  { 9 = HT }
  Derived;

  position.last  :=0;  { before start of script, position 0 }
  position.start :=1;  { start of script, position 1 }
  position.length:=0;

  infiniteindex      := Ed_InfiniteIndex;
  nochar             := chr(0);
  eolchar            := chr(10);
  position.character := nochar;
{ position.tag       := '';              }
  error              := infiniteindex;
  cancelled          := false;
  cancel             := infiniteindex;
  Lex_token.kind     := NullToken;
  linepos            :=-1;  { The line buffer is empty }
  linelen            := 0;
  end {Lex_Init};
