{ Screen output module for text display.

  Provides buffered output of raw data (control codes and such)
  and "rendered" output of text (with embedded control codes presented
  in some nice way).

  Provides its own channel (file descriptor) to the output, as captured
  in the initialization routine from fd 1. Thus, the application program
  being run under the session editor can redirect standard output to e.g.
  a pipe to the editor.

  Rendered output is not yet implemented (it just does raw output).

  Author: Niklas Holsti    around 1993.03.29
  Changes:
    NH  93.03.29

}

module Render;

#include "unistd_p.h"
#include "strings_d.p"

const
  wrbufmax = 500;  { max bytes buffered for one sys write }

private var
  wrfd : integer;  { own file descriptor for output (dupped from fd 1) }
  wrown: boolean;  { wrfd is owned by this module }
  wrbuf: packed array [0..wrbufmax] of char;
  wrlen: 0..wrbufmax;

procedure perror (s: univ_ptr);
  external c;

procedure Rn_Flush;
  { Make sure all data intended for screen get there, ie.
    dump buffer. }
  var w,wrpos: integer;
  begin
  if wrlen > 0 then begin
    wrpos:=0;
    repeat
      w:=sys_write(wrfd,addr(wrbuf[wrpos]),wrlen);
      if w > 0 then begin
        wrpos:= wrpos + w;
        wrlen:= wrlen - w;
        end;
    until (wrlen = 0) or (w < 0);
    if w < 0 then perror(addr('Render-write'));
    wrlen:=0;
    end;
{$w-}        { Yes, I know that ADDR is applied to an internal var, wrbuf }
  end {Rn_Flush};
{$w+}


function Rn_WriteC (c: char; filep: univ_ptr): integer;
  { Writes (buffers) one byte for the screen.
    The strange calling interface is designed to
    match the "terminfo" routine "tputs". }
  begin
  wrbuf[wrlen]:=c; wrlen:=wrlen+1;
  if wrlen >= wrbufmax then Rn_Flush;
  Rn_WriteC:=1;
{$w-}         { Yes, I know that 'filep' is not used, thank you very much. }
  end {Rn_WriteC};
{$w+}


procedure Rn_Raw  (data: vstring);
  { Output the data as such. No 'rendering'. }
  var i: 1..maxstring; dum: integer;
  begin
  for i:=1 to length(data) do
    dum:=Rn_WriteC(data[i],nil);
  end {Rn_Raw};


procedure Rn_Text (    txt   : vstring;
                       maxcol: integer;
                   var actcol: integer);
  { Output the text, rendering normally unprintable characters in
    some visible way -- once this is implemented.
    Up to 'maxcol' screen columns are used; the number actually
    used is returned in actcol. For printable text, actcol = length(txt),
    or maxcol, whichever is smaller. }
  var
    i: 0..maxstring;
    m: 0..maxstring;
    dum: integer;
  begin
  m:=length(txt);
  if m > maxcol then m:=maxcol;
  for i:=1 to m do
    dum:=Rn_WriteC(txt[i],nil);  { crudely done }
  actcol:=m;
  end {Rn_Text};


procedure Rn_TextVar (var txt   : vstring;
                          first,len: integer;
                          maxcol: integer;
                      var actcol: integer);
  { Same as Rn_Text, but renders only txt[first .. first+len-1].
    Also avoids parameter copying, since the text is a 'var'. }
  var
    i  : 0..maxstring;
    m  : 0..maxstring;
    dum: integer;
  begin
  m:=length(txt)-first+1;
  if m > len    then m:=len   ;
  if m > maxcol then m:=maxcol;
  for i:=first to m+first-1 do
    dum:=Rn_WriteC(txt[i],nil);  { crudely done }
  actcol:=m;
  end {Rn_TextVar};


procedure Rn_Init;
  { Initialize module. }
  begin
  { Set up own file descriptor and buffer for screen output. }
  wrfd := dup(1);
  wrown:= wrfd <> -1;
  if not wrown then begin
    perror(addr('Render-dup'));
    wrfd:=1;
    end;
  wrlen:=0;
  end {Rn_Init};


procedure Rn_Fini;
  { Finalize module. }
  begin
  Rn_Flush;
  if wrown then
    if sys_close(wrfd) = -1 then
      perror(addr('Render-close'));
  end {Rn_Fini};
