{ scanline.p

  Line scanner routines in Pascal to interface with incremental (reversible)
  script execution under a SEED-like session editor.

  Author :  Niklas Holsti
  Changes:
     94.10.01  Started, using lexline package as base.
     94.01.05  First published version.
 }

module ScanLine;

#include "strings_d.p"
#include "log_f.p"
#include "edit_d.p"
#include "edit_f.p"

#include "scanline_d.p"

const
  scanlog = 1;   { log principal operations. }

var
  Scan_eol : char;             { the newline input symbol }
  Scan_can : char;             { cancel (reverse) symbol from Scan_char }

private var

  position : Scan_PositionType;
  linebuf  : Ed_LineType;      { the (tail of) the current input line }
  linepos  : ScriptIndex;      { position of linebuf[1]. <0 if buf not filled }
  linelen  : integer;          { length of linebuf, if linebuf filled }

  error    : ScriptIndex;      { where error noted, if < inf }
  cancelled: boolean;          { exception }
  exit     : boolean;          { part of exception handling }
  cancel   : ScriptIndex;      { the position given in a cancel signal, or
                                 infinite if not in cancelling phase }

  infiniteindex: ScriptIndex;  { Ed_InfiniteIndex (constant) }


private
procedure Logpos (pos: Scan_PositionType);
  { Describe Scan position on logfile. }
  begin
  with pos do begin
    writeln(logfile,'[',tag,'] last=',last:1,' mark=',mark:1);
    end;
  end {Logpos};

procedure Scan_Store (var safe: Scan_PositionType;
                          tag : vstring);
  begin
  position.tag:= tag;
  safe:= position;
  if Log_includes(scanlog) then begin
    write(logfile,'Scan_Store');
    Logpos(safe);
    writeln(logfile);
    end;
  end {Scan_Store};

function Scan_Next: ScriptIndex;
  begin
  Scan_Next:= position.last+1;
  end {Scan_Next};

procedure Scan_exit;
  begin
  exit:=true;
  end {Scan_exit};

function Scan_done: boolean;
  begin
  Scan_done := exit;
  end {Scan_done};

function Scan_cancel: boolean;
  begin
  Scan_cancel := cancelled;
  end {Scan_cancel};

private
procedure GetLine (start: ScriptIndex);
  { Reads script starting at "start" into linebuf[1..]
    up to end of line.
    If an error is present, the rest of the current line is abandoned.
    May raise cancelled exception, and is certain to do so if an
    error is present. }
  var
    msg  : Ed_MsgType;
  begin
  repeat
    Ed_Tell (position.last, error, start,
             position.mark, position.last-position.mark+1);
    Ed_Getln(linebuf,linelen,msg);
    case msg.kind of
      ProcessMsg: begin
          linepos:= start;
          end;
      CancelMsg: begin
          cancel   := msg.cancelto;
          cancelled:= true;
          if Log_includes(scanlog) then
            writeln(logfile,'Scan.GetLine cancel=',cancel:1);
          end;
       AbortMsg: begin
          if Log_includes(scanlog) then
            writeln(logfile,'Scan.GetLine received and ignored abort');
          end;
       end;
  until msg.kind <> AbortMsg;
  end {GetLine};


function Scan_Char (i: Scriptindex): char;
  var off: integer;
  begin
  { Ensure linebuf has the data. }
  if linepos < 0 then
    GetLine(i)
  else
    if (i < linepos)
    or (i > linepos+linelen)
    then GetLine(i);

  if cancelled then
    Scan_Char:= Scan_can
  else begin
    { Note we have looked at position i. }
    if i > position.last then position.last:= i;
    { Pick  up the char at pos i }
    off:= i - linepos;
    if off < linelen then
      Scan_Char:= linebuf[off+1]
    else
      Scan_Char:= Scan_eol;
    end;
  end {Scan_Char};


procedure Scan_Used;
  begin
  Ed_Used(position.last);
  end {TellUsed};


procedure Scan_Mark;
  begin
  position.mark:= position.last;
  end {Scan_Mark};


procedure Scan_Error;
  begin
  error:= position.last;
  Scan_Used;
  if Log_includes(scanlog) then
    writeln(logfile,'Scan_Error error=',error:1,
           ' mark=',position.mark:1,
           ' length=',error-position.mark+1:1);
  GetLine(error+1);  { should raise cancelled exception }
  if not cancelled then
    writeln(logfile,'Scan_Error (',error:1,') did not raise cancelled !');
  end {Scan_Error};


procedure Scan_Restore (safe: Scan_PositionType);
  begin
  cancelled := safe.last >= cancel;

  if Log_includes(scanlog) then begin
    write(logfile,'Scan_Restore');
    Logpos(safe);
    write(logfile,' cancel=',cancel:1);
    if cancelled then writeln(logfile,' cancelled')
                 else writeln(logfile,' restored');
    end;

  if not cancelled then begin
    position:=safe;
    error   :=infiniteindex;
    cancel  :=infiniteindex;
    linepos := -1;
    end;
  exit:=cancelled;
  end {Scan_Restore};


procedure Scan_Init;
  { Initialize local variables. }
  begin
  { writeln('Scan for incremental script processing. 1994.10.01.A'); }

  Scan_eol      := chr(10);
  Scan_can      := chr( 3);

  infiniteindex := Ed_InfiniteIndex;

  position.last := 0;  { before start of script, position 0 }
  position.mark := 1;
  position.tag  := '';

  linepos       := -1; { line buffer is empty }
  error         := infiniteindex;
  cancelled     := false;
  cancel        := infiniteindex;
  end {Scan_Init};
