{ Screen handling module for text display. Provides cursor addressing;
  char insert, overstrike, and delete; settable screen port and
  scrolling region (line interval); line insert and delete; clear to
  end of line and clear display. 

  Author: Niklas Holsti    around 1991.02.25
  Changes:
    NH  91.06.17  Added file parameter to Explain routine.
    NH  92.04.08  Port to Sun OS; ANSI terminal commands.
                  Some declarations in screen_d.p; routine
                  declarations in screen_f.p
    NH  92.04.11  Caching of current scrolling region, insert/overstrike
                  mode, and cursor position.
                  Sc_Wrote, Sc_Limbo, Sc_Fini added.
    NH  92.04.17  Local vars 'private' for less risk of sym conflicts.
    NH  92.05.23  Uses Termcap routines.
    NH  92.05.28  Uses Terminfo routines.
    NH  92.11.29  Missing terminal capabilites are displayed in error msg.
    NH  92.11.30  Missing insert-line and delete-line capabilites simulated
                  with scrolling region and index, if available.
    NH  92.11.30  Screen max size increase to 80 * 150.
    NH  93.03.29  Render module used for all output.
}

module Screen;

#include "terminfo_p.h"
#include "strings_d.p"    { for render }
#include "screen_d.p"
#include "render_f.p"

{
 The present version operates with Sys V Unix Terminfo terminal database.
 Insert, overstrike, and delete-char are used, if the terminal
 description includes such capabilities.
}

type
  TextMode  = (modeUnknown, modeInsert, modeOverstrike);
  CursorPos = record row,col: integer; end;

  tcap      = univ_ptr;     { a terminal capability (string) }

private var
  screenArea  : record
                rows: Sc_rowRange;
                cols: integer;
                end;
  cursor      : CursorPos;
  mode        : TextMode;
  scrollRows  : Sc_rowRange;       { 0,0 for "unknown" }

  dummy       : tcap;

  { Absolutely required capabilities: }
  clr_eol             : tcap;
  clear_screen        : tcap;
  cursor_address      : tcap;
  insert_line         : tcap;  { or sim with scroll region and reverse index }
  delete_line         : tcap;  { or sim with scroll region and index }

  { Optional but nice: }
  enter_ca_mode       : tcap;
  exit_ca_mode        : tcap;
  change_scroll_region: tcap;  { req'd if insert-line or delete-line absent }
  scroll_forward      : tcap;  { = index, ditto }
  scroll_reverse      : tcap;  { = reverse index, ditto }
  enter_insert_mode   : tcap;
  exit_insert_mode    : tcap;
  delete_character    : tcap;

  { Flags for capabilities }
  can_set_scroll      : boolean;   { Can set scrolling region }
  sim_insert_line     : boolean;   { Can/must simulate insert_line }
  sim_delete_line     : boolean;   { Ditto             delete_line }


#define tgoto(cm,c,r) tparm(cm,r,c)

procedure Sc_Explain (var f: text; result: Sc_result);
  { Write an explanation of the result. }
  begin
  case result of
    Sc_ok         : write(f,'Operation succeeded');
    Sc_undefterm  : write(f,'No TERM environment variable');
    Sc_badtermcap : write(f,'Could not get terminal data');
    Sc_notermcap  : write(f,'Required terminal capability missing');
    Sc_areawrong  : write(f,'Screen area out of bounds');
    Sc_poswrong   : write(f,'Cursor position out of bounds');
    Sc_scrollwrong: write(f,'Scroll region out of bounds');
    Sc_delwrong   : write(f,'Delete-line outside scroll region');
    Sc_inswrong   : write(f,'Insert-line outside scroll region');
    end;
  end {Sc_Explain};


procedure Sc_Flush;
  { Make sure all data intended for screen get there, ie.
    dump buffer. }
  begin Rn_Flush; end;


procedure Sc_Limbo;
  { We no longer know where the cursor is ... }
  begin
  cursor.row:=0;
  cursor.col:=0;
  end {Sc_Limbo};

procedure Scrolling (top,bot: integer);
  { Change scrolling area. Internal routine.
    Assumes that there is such a capability.
    Does NOT modify 'cursor' - although we no
    longer know where the cursor is, we may know
    where it should be. }
  begin
  dummy:=tputs(tparm(change_scroll_region,top-1,bot-1),
               1,Rn_WriteC);
  end {Scrolling};

procedure Sc_SetScroll (rows: Sc_rowRange; var result: Sc_result);
  begin
  if rows <> scrollRows then begin
    if  (screenArea.rows.top <= rows.top)
    and (           rows.top <= rows.bot)
    and (           rows.bot <= screenArea.rows.bot)
    then begin
      scrollRows:=rows;
      if can_set_scroll then begin
        Scrolling(rows.top,rows.bot);
        Sc_Limbo;
        end;
      result:=Sc_ok;
      end
    else
      result:=Sc_scrollwrong;
    end;
  end {Sc_SetScroll};

procedure Sc_SetArea (    rows  : Sc_rowRange;
                          cols  : integer;
                      var result: Sc_result);
  begin
  if  (   1     <= rows.top  )
  and (rows.top <= rows.bot  )
  and (rows.bot <= Sc_MaxRows)
  and (cols >= 1) and (cols <= Sc_MaxCols)
  then begin
    screenArea.rows:=rows;
    screenArea.cols:=cols;
    Sc_SetScroll(rows,result);
    end
  else
    result:=Sc_areawrong;
  end {Sc_SetArea};

procedure Sc_GetArea (var rows: Sc_rowRange;
                      var cols: integer);
  begin
  rows:=screenArea.rows;
  cols:=screenArea.cols;
  end {Sc_GetArea};

procedure GotoRC (row,col: integer);
  { Move cursor to .. }
  begin
  dummy:=tputs(tgoto(cursor_address,col-1,row-1),1,Rn_WriteC);
  end {GotoRC};

procedure Sc_Goto (row,col: integer; var result: Sc_result);
  var newpos: CursorPos;
  begin
  newpos.row:=row;
  newpos.col:=col;
  if newpos <> cursor then begin
    if  (row >= screenArea.rows.top) and (row <= screenArea.rows.bot)
    and (col >=     1              ) and (col <= screenArea.cols    )
    then begin
      GotoRC(row,col);
      cursor := newpos;
      result:=Sc_ok;
      end
    else
      result:=Sc_poswrong;
    end;
  end {Sc_Goto};


procedure Sc_Wrote (cols: integer);
  { Informs us that a number of columns have been written
    at the cursor, which thus has moved along. }
  begin
  if cursor.col > 0 then begin
    cursor.col:= cursor.col + cols;
    if cursor.col > screenArea.cols then Sc_Limbo;
    end;
  end {Sc_Written};

procedure Sc_Clear;
  begin
  dummy:=tputs(clear_screen,1,Rn_WriteC);
  cursor.row:=1;
  cursor.col:=1;
  end {Sc_Clear};

procedure Sc_ClearEOL;
  begin
  dummy:=tputs(clr_eol,1,Rn_WriteC);
  end {Sc_ClearEOL};

function  Sc_CanInsert: boolean;
  begin
  Sc_CanInsert:= enter_insert_mode <> nil;
  end {Sc_CanInsert};

procedure Sc_InsertMode;
  begin
  if mode <> modeInsert then begin
    dummy:=tputs(enter_insert_mode,1,Rn_WriteC);
    mode:=modeInsert;
    end;
  end {Sc_InsertMode};

procedure Sc_OverstrikeMode;
  begin
  if mode <> modeOverstrike then begin
    dummy:=tputs(exit_insert_mode,1,Rn_WriteC);
    mode:=modeOverstrike;
    end;
  end {Sc_OverstrikeMode};

function  Sc_CanDeleteChar: boolean;
  begin
  Sc_CanDeleteChar:= delete_character <> nil;
  end {Sc_CanDeleteChar};

procedure Sc_DeleteChar (n: integer);
  var i: integer;
  begin
  { Should use parm_dch capability if present. }
  for i:=1 to n do dummy:=tputs(delete_character,1,Rn_WriteC);
  end {Sc_DeleteChar};

procedure Sc_InsertLine (var result: Sc_result);
  begin
  if  (cursor.row >= scrollRows.top)
  and (cursor.row <= scrollRows.bot)
  then begin
    if not can_set_scroll then
      { Simulate scrolling region by deleting bottom line. }
      if scrollRows.bot < screenArea.rows.bot then begin
        with scrollRows do GotoRC(bot,1);
        dummy:=tputs(delete_line,1,Rn_WriteC);
        with cursor do GotoRC(row,col);
        end;
    if sim_insert_line then begin
      { Simulate insert_line by scroll_reverse (reverse index)
        with the proper scrolling region. }
      Scrolling(cursor.row,scrollRows.bot);
      GotoRC(cursor.row,1); 
      dummy:=tputs(scroll_reverse,1,Rn_WriteC);
      with scrollRows do Scrolling(top,bot);
      with cursor do GotoRC(row,col);
      end
    else
      dummy:=tputs(insert_line,1,Rn_WriteC);
    result:=Sc_ok;
    end
  else
    result:=Sc_inswrong;
  end {Sc_InsertLine};

procedure Sc_DeleteLine (var result: Sc_result);
  begin
  if  (cursor.row >= scrollRows.top)
  and (cursor.row <= scrollRows.bot)
  then begin
    if sim_delete_line then begin
      Scrolling(cursor.row,scrollRows.bot);
      GotoRC(scrollRows.bot,screenArea.cols);
      dummy:=tputs(scroll_forward,1,Rn_WriteC);
      with scrollRows do Scrolling(top,bot);
      with cursor do GotoRC(row,col);
      end
    else begin
      dummy:=tputs(delete_line,1,Rn_WriteC);
      if not can_set_scroll then
        { Simulate scrolling region by inserting a line at bottom. }
        if scrollRows.bot < screenArea.rows.bot then begin
          with scrollRows do GotoRC(bot,1);
          dummy:=tputs(insert_line,1,Rn_WriteC);
          with cursor do GotoRC(row,col);
          end;
      end;
    result:=Sc_ok;
    end
  else
    result:=Sc_delwrong;
  end {Sc_DeleteLine};


function getenv (name: univ_ptr): univ_ptr;
  external c;

procedure Sc_Init;
  type
    cid   = varying[10] of char;
  var
    name  : ^string;
    gotit : integer;
    errret: integer;
    rows  : Sc_rowRange;
    cols  : integer;
    result: Sc_result;

  procedure GetCap (var cap: tcap; id: cid; required: boolean);
    { Get a string-valued terminal capability. }
    var idz: cid;
    begin
    idz:=id+chr(0);
    cap:=tigetstr(addr(idz[1]));
    if cap = univ_ptr(-1) then begin
      writeln('Sc_Init malfunction: "',id,
              '" is not a string capability');
      cap:=nil;
      end;
    if required then
      if cap = nil then begin
        writeln('screen/Sc_Init: missing capability "',id,'"');
        result:=Sc_notermcap;
        end;
{$w-}                 { Yes, ADDR is applied to idz, a dynamic var. }
    end {GetCap};
{$w+}

  begin
  result:=Sc_ok;  { unless ...}

  name:=getenv(addr('TERM'));
  if name = nil then result:= Sc_undefterm
  else begin
    gotit:=setupterm(name,1,errret);
    if errret <> 1 then result:= Sc_badtermcap
    else begin

      { Optional capabilities }
      GetCap(enter_ca_mode        ,'smcup' ,false);
      GetCap(exit_ca_mode         ,'rmcup' ,enter_ca_mode <> nil);
      GetCap(change_scroll_region ,'csr'   ,false);
      GetCap(enter_insert_mode    ,'smir'  ,false);
      GetCap(exit_insert_mode     ,'rmir'  ,enter_insert_mode <> nil);
      GetCap(delete_character     ,'dch1'  ,false);
      GetCap(scroll_forward       ,'ind'   ,false);
      GetCap(scroll_reverse       ,'ri'    ,false);

      can_set_scroll:= change_scroll_region <> nil;

      { Can we simulate insert/delete-line if needed? }
      sim_insert_line:= can_set_scroll and (scroll_reverse <> nil);
      sim_delete_line:= can_set_scroll and (scroll_forward <> nil);

      if enter_ca_mode <> nil then
        dummy:=tputs(enter_ca_mode,1,Rn_WriteC);

      { Required capabilities }
      with rows do begin
        top:=1;
        bot:=tigetnum(addr('lines'));
        if bot = -2 then result:=Sc_notermcap;
        end;

      cols:=tigetnum(addr('cols'));
      if cols = -2 then result:=Sc_notermcap;

      GetCap(clr_eol       ,'el'   ,true);
      GetCap(clear_screen  ,'clear',true);
      GetCap(cursor_address,'cup'  ,true);
      GetCap(insert_line   ,'il1'  ,not sim_insert_line);
      GetCap(delete_line   ,'dl1'  ,not sim_delete_line);

      { Should we simulate insert/delete-line? }
      if insert_line <> nil then sim_insert_line:=false;
      if delete_line <> nil then sim_delete_line:=false;

      Sc_Limbo;
      if enter_insert_mode = nil then mode:=modeOverstrike
                                 else mode:=modeUnknown;

      with scrollRows do begin top:=0; bot:=0; end;   { unset }

      { Init screen. }
      if result = Sc_ok then Sc_SetArea(rows,cols,result);
      if result = Sc_ok then Sc_Clear;
      end;
    end;

  if result <> Sc_ok then begin
    write('Screen module init failed: ');
    Sc_Explain(output,result);
    writeln;
    end;

{$w-}   { Yes, I know that argument to ADDR is an internal variable. }
  end {Sc_Init};
{$w+}


procedure Sc_Fini;
  var rows  : Sc_rowRange;
      result: Sc_result;
  begin
  with rows do begin top:=1; bot:=tigetnum(addr('lines')); end;
  with scrollRows do begin top:=0; bot:=0; end;   { unset }
  Sc_SetArea(rows,tigetnum(addr('cols')),result);
  if exit_insert_mode <> nil then begin
    mode:=modeUnknown;
    Sc_OverstrikeMode;
    end;
  Sc_Limbo;
  Sc_Goto(rows.bot,1,result);
  if exit_ca_mode <> nil then
    dummy:=tputs(exit_ca_mode,1,Rn_WriteC);
  end {Sc_Fini};
