/* Output from p2c, the Pascal-to-C translator */
/* From input file "transcripts.p" */


/* Transcript handling for SEED-like session editors supporting incremental
   (reversible) script processing.

   A transcript is a stored dialog between a user and an application
   program. It contains both the user's inputs and the program's outputs.
   Its purpose is to support interaction by being editable, with an
   automatic undo-redo mechanism that updates the output when the input
   is changed. Thus, instead of interacting directly with the application
   program, the user manipulates and views the transcript with an editor,
   and the editor sends input text from the transcript to the application
   program, and receives in reply output text which it inserts into the
   transcript.

   The simplest model of a transcript T is a sequence of positions, numbered
   1,2,...infinity. Each position i can hold an input symbol T.input(i) and
   and output symbol T.output(i), which is the output of the program (script
   processor) after reading T.input(i) (having already read T.input(1..i-1)).
   If the user then changes T.input(j), with j <= i, a "cancel" signal is
   sent to the program, and the program is expected to reverse its
   execution (return to a previous state) to a state in which the last
   symbol read is T.input(k) with k < j. The program chooses k.
   The changed input string T.input(k+1..) is then sent to the program,
   and the corresponding old outputs are replaced by new ones.

   Many interactive programs read input a line at a time, and format
   their output as lines. The transcript must therefore be divided into
   lines. Output can either be inserted within an input line, following
   the input that caused it, or be on placed on new output lines and tagged
   to show which input position actually generated it (this caters to
   programs that read input line by line, but may still use grains smaller
   than a line.

   Output elements are marked (time stamped) with the number of the last
   input symbol interpreted by the script processor when the output was
   produced. Note the distinction between "reading" and "interpreting"
   input symbols: the processor may read input symbols into a buffer
   (e.g. a line), and then interpret them. If the processor claims that
   symbol T.input(i) is the last one interpreted for some output, although
   it has read symbols T.input(i+1..k), the processor guarantees that
   the output is the same whatever the symbols >i are (as long as one
   of these symbols makes the processor start interpreting).

   This module implements the transcript as a data structure. It does not
   concern itself with the editor user interface - the commands, display,
   scrolling etc. (except for the Ts_Show routine), nor with the
   communication between the editor and the application program.

   The module defines two abstract data types, "transcript" and "spot".
   A transcript holds an interactive dialog, a sequence of input and output
   elements. A spot is a location in the transcript where new input or
   output can be inserted, or some other editing done. A spot is thus
   an insertion point and lies between two transcript items (usually
   characters) or at either end of the transcript.

   Author: Niklas Holsti    around 1991.02.17
   Notes :
     91.02.17  NH  Status: mostly functional, but pre-dated output
                   not allowed (Ts_Used does not work) and region
                   cut/copy/paste is not implemented.
                   Null text elements are sometimes used, and may cause
                   problems in eoln & eot tests (should check).
                   The use of type string in transcript elements
                   fragments storage, especially since we do not join
                   all possible elements. The module handles only one
                   transcript - it should be made into a proper exported
                   data type.
     91.04.12  NH  No null elements will occur (text may be empty if the
                   element has a newline). Deletion of input automatically
                   deletes (unwrites) immediately following output.
                   Ts_Unwrite added and Ts_Read no longer deletes any
                   output. Ts_Skip added.
     91.04.13  NH  Handles numerous transcripts. Copy and Cut implemented.
     91.05.15  NH  Ts_Erase added.
     91.04.19  NH  hiSpot field and Ts_MoreSpots added.
     91.04.20  NH  Separate bot and eot elements.
     91.04.26  NH  UpSpotPos uses Order, rewritten.
                   Ts_Delete corrected (join line to empty line).
     91.05.05  NH  Ts_Newline handles tagged output.
                   Ts_Joinline for deleting newlines, with proper handling
                   of tagged and embedded output. Ts_Delete deletes only
                   characters.
                   SkipForw/Back on Ts_Insert/Delete/Joinline, so they can
                   be used with the cursor placed more freely.
     91.06.17  NH  Added file parameter to Explain routine.
                   Ts_Insert gets preRow/Col params (since SkipForw may
                   move point before inserting char).
     91.09.03  NH  Ts_Explain fixed (always wrote 'Ts_Result' to standard
                   output rather than specified file).
     91.09.05  NH  Ts_Show no longer clears to end of line (whether to do so
                   must be decided based on whether the right margin has been
                   reached or not - if done after writing at right margin,
                   the next screen line may be erased). To this end, Ts_Show
                   returns numCol, the number of columns written.
     91.09.14  NH  Change handling of output in Ts_Load - now discards
          -15      leading output, and marks other output invalid.
     91.09.23  NH  Ts_Newline yields preRow,preCol values.
                   Ts_AtOutput, Ts_CharAfter added.
     91.09.24  NH  Ts_GotoEnd added.
     92.04.10  NH  Port to Sun Unix.
                   Ts_Init renamed to Ts_Create.
                   Ts_Init and Ts_Fini added (package initing and finiting).
                   Ts_LineLength added.
     92.04.17  NH  Local vars 'private' for less risk of sym conflicts.
     93.03.29  NH  Render module used for transcript output.
                   NOTE -- the reckoning of columns to display / columns
                   displayed is not yet implemented properly (since the
                   render module does not yet render anything as other than
                   single characters).
     93.05.14  NH  Dewrite can now be applied also when the write point
                   is at the end of an output element (reversal to a
                   position corresponding to the middle of an output
                   line did not delete later output properly).
                   Ts_Writeln fixed to initialize the result flag.
     94.04.04  NH  Ts_Reads, read more than one char at a time (faster).
                   (From Mac version, 93.02.13).
     94.10.07  NH  Flag for transcript modified since last load.
*/


#include <p2c/p2c.h>


#define TRANSCRIPTS_G
#include "transcripts.h"
/* p2c: strings_f.p, line 5: 
 * Warning: Expected IMPLEMENT, found 'external' [227] */
/* p2c: strings_f.p, line 5: 
 * Warning: Expected END, found 'external' [227] */
/* p2c: transcripts.p, line 125: 
 * Warning: Could not open include file transcripts_d.p [230] */


void _Transcripts_init()
{

  /* for test routines */
  /* render_f.p
     External declarations of render.p routines.
  */
  /* Make sure all data intended for screen get there, ie.
     dump buffer. */

  /* Writes (buffers) one byte for the screen.
     The strange calling interface is designed to
     match the "terminfo" routine "tputs". */

  /* Output the data as such. No 'rendering'. */

  /* Output the text, rendering normally unprintable characters in
     some visible way -- once this is implemented.
     Up to 'maxcol' screen columns are used; the number actually
     used is returned in actcol. For printable text, actcol = length(txt),
     or maxcol, whichever is smaller. */

  /* Same as Rn_Text, but renders only txt[first .. first+len-1].
     Also avoids parameter copying, since the text is a 'var'. */

  /* Initialize module. */

  /* Finalize module. */
  /* */

  /*
    The transcript is represented in a blocked manner as a sequence of
    input and output elements in a doubly linked list. An element is
    usually a string of characters (possibly empty). Other kinds of
    elements (e.g. graphical output) may be implemented later.
    There is a special beginning-of-transcript marker element, botElem,
    and a special end-of-transript element, eotElem. No spot should
    ever be sited (other than transiently) on the botElem.

    The transcript is divided into lines. A line consists of one or
    more elements. Elements are never split across lines. Each element has
    a boolean flag (newline) which is true iff the element is the last on a
    line. The last (real) element in the transcript may or may not have
    newline=true, but eot/botElem.newline is always true.

    Each output element is marked with the number "safe" of the last
    input symbol interpreted by the script processor when the output was
    produced. Most output elements are also linked into a stack, with links
    pointing from an output element to the preceding (earlier) one.
    The very first output element points to omElem.
    The output stack contains only elements that are "valid" in the
    sense that the processing that produced them has not been
    cancelled (reversed). Output elements after the topmost (latest)
    stacked element are "invalid" (out of date), and will be deleted
    when the corresponding input is (re)processed. Invalid output elements
    are further distinguished by safe=maxint.

    To mark output elements, a transcript records the number of the last
    input symbol read by the processor, and the number of the last input
    symbol the processor claims it has interpreted.

    Spots are implemented as pairs (element pointer, offset). To support
    editing, row and column numbers are also provided for each spot,
    although they could also be computed by scanning the transcript.
    To support script processing and reversal, each spot also caches
    the number of input symbols that precede it. Caching this redundant
    data means that any change in the transcript forces us to update many
    later spots. The editor should not use very many spots, and should
    delete any spot it no longer needs.

    Note that when a spot is between elements e and f, where e^.newline
    is false, the spot can be represented in two equivalent ways, as

           ( element=>e, offset=>length(e) )
    or
           ( element=>f, offset=>0 )

    If e^.newline is true, the first form is at the end of e's line, before
    the line separator, and the second form is at the start of f's line.

    The column number of a spot may temporarily be unknown. This happens
    only when the spot is moved backwards past a line boundary. The true
    column numbers for all spots must be known before any changes are made
    in the transcript, as the spot-updating routine (UpSpotPos) depends
    on this. Essentially, this means that any routine whose name begins with
    Ts_ can assume that all columns are known on entry, and must ensure that
    all columns are known on exit.
  */

  /* "safe" mark on invalid output elements. */


  /* Testing */
  /* Number of elements that can be allocated before
                         memory is full (for testing memory overflow
                         situation) */
/* p2c: transcripts.p, line 198: Warning: Symbol 'F' is not defined [221] */
  static int _was_initialized = 0;
  if (_was_initialized++)
    return;
  /*Ts_explain*/
  printf("%12ldTs_result: ", f);
/* p2c: transcripts.p, line 199:
 * Warning: Symbol 'RESULT' is not defined [221] */
  switch (result) {

  case Ts_Ok:
    printf("%12ldOk", f);
    break;

/* p2c: transcripts.p, line 200:
 * Warning: Symbol 'TS_OK' is not defined [221] */
/* p2c: transcripts.p, line 200: Warning: Symbol 'F' is not defined [221] */
  case Ts_spotsfull:
    printf("%12ldSpot table full", f);
    break;

/* p2c: transcripts.p, line 201:
 * Warning: Symbol 'TS_SPOTSFULL' is not defined [221] */
/* p2c: transcripts.p, line 201: Warning: Symbol 'F' is not defined [221] */
  case Ts_memfull:
    printf("%12ldMemory full", f);
    break;

/* p2c: transcripts.p, line 202:
 * Warning: Symbol 'TS_MEMFULL' is not defined [221] */
/* p2c: transcripts.p, line 202: Warning: Symbol 'F' is not defined [221] */
  case Ts_elemfull:
    printf("%12ldTranscript element full", f);
    break;

/* p2c: transcripts.p, line 203:
 * Warning: Symbol 'TS_ELEMFULL' is not defined [221] */
/* p2c: transcripts.p, line 203: Warning: Symbol 'F' is not defined [221] */
  case Ts_immutable:
    printf("%12ldTranscript element immutable", f);
    break;

/* p2c: transcripts.p, line 204:
 * Warning: Symbol 'TS_IMMUTABLE' is not defined [221] */
/* p2c: transcripts.p, line 204: Warning: Symbol 'F' is not defined [221] */
  case Ts_nontext:
    printf("%12ldNot a text element", f);
    break;

/* p2c: transcripts.p, line 205:
 * Warning: Symbol 'TS_NONTEXT' is not defined [221] */
/* p2c: transcripts.p, line 205: Warning: Symbol 'F' is not defined [221] */
  case Ts_noelem:
    printf("%12ldNo element (nil)", f);
    break;

/* p2c: transcripts.p, line 206:
 * Warning: Symbol 'TS_NOELEM' is not defined [221] */
/* p2c: transcripts.p, line 206: Warning: Symbol 'F' is not defined [221] */
  case Ts_nullelem:
    printf("%12ldNull element (zero length)", f);
    break;

/* p2c: transcripts.p, line 207:
 * Warning: Symbol 'TS_NULLELEM' is not defined [221] */
/* p2c: transcripts.p, line 207: Warning: Symbol 'F' is not defined [221] */
  case Ts_wrongorder:
    printf("%12ldTwo spots in wrong order", f);
    break;

/* p2c: transcripts.p, line 208:
 * Warning: Symbol 'TS_WRONGORDER' is not defined [221] */
/* p2c: transcripts.p, line 208: Warning: Symbol 'F' is not defined [221] */
  case Ts_nonewline:
    printf("%12ldNewline missing where expected", f);
    break;

/* p2c: transcripts.p, line 209:
 * Warning: Symbol 'TS_NONEWLINE' is not defined [221] */
/* p2c: transcripts.p, line 209: Warning: Symbol 'F' is not defined [221] */
  case Ts_wasnewline:
    printf("%12ldNewline present, should not be", f);
    break;

/* p2c: transcripts.p, line 210:
 * Warning: Symbol 'TS_WASNEWLINE' is not defined [221] */
/* p2c: transcripts.p, line 210: Warning: Symbol 'F' is not defined [221] */
  case Ts_noteot:
    printf("%12ldSpot should be at EOT", f);
    break;

/* p2c: transcripts.p, line 211:
 * Warning: Symbol 'TS_NOTEOT' is not defined [221] */
/* p2c: transcripts.p, line 211: Warning: Symbol 'F' is not defined [221] */
  case Ts_atomic:
    printf("%12ldElement cannot be split", f);
    break;

/* p2c: transcripts.p, line 212:
 * Warning: Symbol 'TS_ATOMIC' is not defined [221] */
/* p2c: transcripts.p, line 212: Warning: Symbol 'F' is not defined [221] */
  case Ts_notinput:
    printf("%12ldNot at input item", f);
    break;

/* p2c: transcripts.p, line 213:
 * Warning: Symbol 'TS_NOTINPUT' is not defined [221] */
/* p2c: transcripts.p, line 213: Warning: Symbol 'F' is not defined [221] */
  case Ts_endinput:
    printf("%12ldEnd of input (at EOT)", f);
    break;

/* p2c: transcripts.p, line 214:
 * Warning: Symbol 'TS_ENDINPUT' is not defined [221] */
/* p2c: transcripts.p, line 214: Warning: Symbol 'F' is not defined [221] */
  case Ts_validoutput:
    printf("%12ldValid output, not dewritten", f);
    break;

/* p2c: transcripts.p, line 215:
 * Warning: Symbol 'TS_VALIDOUTPUT' is not defined [221] */
/* p2c: transcripts.p, line 215: Warning: Symbol 'F' is not defined [221] */
  case Ts_fixedoutput:
    printf("%12ldCannot insert text before output", f);
    break;

/* p2c: transcripts.p, line 216:
 * Warning: Symbol 'TS_FIXEDOUTPUT' is not defined [221] */
/* p2c: transcripts.p, line 216: Warning: Symbol 'F' is not defined [221] */
  case Ts_wrongform:
    printf("%12ldInput file format is wrong", f);
    break;

/* p2c: transcripts.p, line 217:
 * Warning: Symbol 'TS_WRONGFORM' is not defined [221] */
/* p2c: transcripts.p, line 217: Warning: Symbol 'F' is not defined [221] */
  case Ts_invalid:
    printf("%12ldInvalid or unimplemented operation", f);
    break;
/* p2c: transcripts.p, line 218:
 * Warning: Symbol 'TS_INVALID' is not defined [221] */
/* p2c: transcripts.p, line 218: Warning: Symbol 'F' is not defined [221] */
  }
}
/* p2c: transcripts.p, line 222: 
 * Warning: Expected a '.', found 'private' [227] */
/* p2c: transcripts.p, line 222: 
 * Warning: Junk at end of input file ignored [277] */
/* p2c: Note: Remember to call _Transcripts_init() in main program [215] */




/* End. */
