{ Transcript handling for SEED-like session editors supporting incremental
  (reversible) script processing.

  A transcript is a stored dialog between a user and an application
  program. It contains both the user's inputs and the program's outputs.
  Its purpose is to support interaction by being editable, with an
  automatic undo-redo mechanism that updates the output when the input
  is changed. Thus, instead of interacting directly with the application
  program, the user manipulates and views the transcript with an editor,
  and the editor sends input text from the transcript to the application
  program, and receives in reply output text which it inserts into the
  transcript.

  The simplest model of a transcript T is a sequence of positions, numbered
  1,2,...infinity. Each position i can hold an input symbol T.input(i) and
  and output symbol T.output(i), which is the output of the program (script
  processor) after reading T.input(i) (having already read T.input(1..i-1)).
  If the user then changes T.input(j), with j <= i, a "cancel" signal is
  sent to the program, and the program is expected to reverse its
  execution (return to a previous state) to a state in which the last
  symbol read is T.input(k) with k < j. The program chooses k.
  The changed input string T.input(k+1..) is then sent to the program,
  and the corresponding old outputs are replaced by new ones.

  Many interactive programs read input a line at a time, and format
  their output as lines. The transcript must therefore be divided into
  lines. Output can either be inserted within an input line, following
  the input that caused it, or be on placed on new output lines and tagged
  to show which input position actually generated it (this caters to
  programs that read input line by line, but may still use grains smaller
  than a line.

  Output elements are marked (time stamped) with the number of the last
  input symbol interpreted by the script processor when the output was
  produced. Note the distinction between "reading" and "interpreting"
  input symbols: the processor may read input symbols into a buffer
  (e.g. a line), and then interpret them. If the processor claims that
  symbol T.input(i) is the last one interpreted for some output, although
  it has read symbols T.input(i+1..k), the processor guarantees that
  the output is the same whatever the symbols >i are (as long as one
  of these symbols makes the processor start interpreting).

  This module implements the transcript as a data structure. It does not
  concern itself with the editor user interface - the commands, display, 
  scrolling etc. (except for the Ts_Show routine), nor with the
  communication between the editor and the application program.

  The module defines two abstract data types, "transcript" and "spot".
  A transcript holds an interactive dialog, a sequence of input and output
  elements. A spot is a location in the transcript where new input or
  output can be inserted, or some other editing done. A spot is thus
  an insertion point and lies between two transcript items (usually
  characters) or at either end of the transcript.

  Author: Niklas Holsti    around 1991.02.17
  Notes :
    91.02.17  NH  Status: mostly functional, but pre-dated output
                  not allowed (Ts_Used does not work) and region
                  cut/copy/paste is not implemented.
                  Null text elements are sometimes used, and may cause
                  problems in eoln & eot tests (should check).
                  The use of type string in transcript elements
                  fragments storage, especially since we do not join
                  all possible elements. The module handles only one
                  transcript - it should be made into a proper exported
                  data type.
    91.04.12  NH  No null elements will occur (text may be empty if the
                  element has a newline). Deletion of input automatically
                  deletes (unwrites) immediately following output.
                  Ts_Unwrite added and Ts_Read no longer deletes any
                  output. Ts_Skip added.
    91.04.13  NH  Handles numerous transcripts. Copy and Cut implemented.
    91.05.15  NH  Ts_Erase added.
    91.04.19  NH  hiSpot field and Ts_MoreSpots added.
    91.04.20  NH  Separate bot and eot elements.
    91.04.26  NH  UpSpotPos uses Order, rewritten.
                  Ts_Delete corrected (join line to empty line).
    91.05.05  NH  Ts_Newline handles tagged output.
                  Ts_Joinline for deleting newlines, with proper handling
                  of tagged and embedded output. Ts_Delete deletes only
                  characters.
                  SkipForw/Back on Ts_Insert/Delete/Joinline, so they can
                  be used with the cursor placed more freely.
    91.06.17  NH  Added file parameter to Explain routine.
                  Ts_Insert gets preRow/Col params (since SkipForw may
                  move point before inserting char).
    91.09.03  NH  Ts_Explain fixed (always wrote 'Ts_Result' to standard
                  output rather than specified file).
    91.09.05  NH  Ts_Show no longer clears to end of line (whether to do so
                  must be decided based on whether the right margin has been
                  reached or not - if done after writing at right margin,
                  the next screen line may be erased). To this end, Ts_Show
                  returns numCol, the number of columns written.
    91.09.14  NH  Change handling of output in Ts_Load - now discards
         -15      leading output, and marks other output invalid.
    91.09.23  NH  Ts_Newline yields preRow,preCol values.
                  Ts_AtOutput, Ts_CharAfter added.
    91.09.24  NH  Ts_GotoEnd added.
    92.04.10  NH  Port to Sun Unix.
                  Ts_Init renamed to Ts_Create.
                  Ts_Init and Ts_Fini added (package initing and finiting).
                  Ts_LineLength added.
    92.04.17  NH  Local vars 'private' for less risk of sym conflicts.
    93.03.29  NH  Render module used for transcript output.
                  NOTE -- the reckoning of columns to display / columns
                  displayed is not yet implemented properly (since the
                  render module does not yet render anything as other than
                  single characters).
    93.05.14  NH  Dewrite can now be applied also when the write point
                  is at the end of an output element (reversal to a
                  position corresponding to the middle of an output
                  line did not delete later output properly).
                  Ts_Writeln fixed to initialize the result flag.
    94.04.04  NH  Ts_Reads, read more than one char at a time (faster).
                  (From Mac version, 93.02.13).
    94.10.07  NH  Flag for transcript modified since last load.
}

module Transcripts;

#include "strings_d.p"
#include "strings_f.p"
#include "console_f.p"     { for test routines }
#include "render_f.p"
#include "transcripts_d.p"

{
  The transcript is represented in a blocked manner as a sequence of
  input and output elements in a doubly linked list. An element is
  usually a string of characters (possibly empty). Other kinds of
  elements (e.g. graphical output) may be implemented later.
  There is a special beginning-of-transcript marker element, botElem,
  and a special end-of-transript element, eotElem. No spot should
  ever be sited (other than transiently) on the botElem.
  
  The transcript is divided into lines. A line consists of one or
  more elements. Elements are never split across lines. Each element has
  a boolean flag (newline) which is true iff the element is the last on a
  line. The last (real) element in the transcript may or may not have
  newline=true, but eot/botElem.newline is always true.

  Each output element is marked with the number "safe" of the last
  input symbol interpreted by the script processor when the output was
  produced. Most output elements are also linked into a stack, with links
  pointing from an output element to the preceding (earlier) one.
  The very first output element points to omElem.
  The output stack contains only elements that are "valid" in the
  sense that the processing that produced them has not been
  cancelled (reversed). Output elements after the topmost (latest)
  stacked element are "invalid" (out of date), and will be deleted
  when the corresponding input is (re)processed. Invalid output elements
  are further distinguished by safe=maxint.

  To mark output elements, a transcript records the number of the last
  input symbol read by the processor, and the number of the last input
  symbol the processor claims it has interpreted.

  Spots are implemented as pairs (element pointer, offset). To support
  editing, row and column numbers are also provided for each spot, 
  although they could also be computed by scanning the transcript.
  To support script processing and reversal, each spot also caches
  the number of input symbols that precede it. Caching this redundant
  data means that any change in the transcript forces us to update many
  later spots. The editor should not use very many spots, and should
  delete any spot it no longer needs.

  Note that when a spot is between elements e and f, where e^.newline
  is false, the spot can be represented in two equivalent ways, as

         ( element=>e, offset=>length(e) )
  or
         ( element=>f, offset=>0 )

  If e^.newline is true, the first form is at the end of e's line, before
  the line separator, and the second form is at the start of f's line.

  The column number of a spot may temporarily be unknown. This happens
  only when the spot is moved backwards past a line boundary. The true
  column numbers for all spots must be known before any changes are made
  in the transcript, as the spot-updating routine (UpSpotPos) depends
  on this. Essentially, this means that any routine whose name begins with
  Ts_ can assume that all columns are known on entry, and must ensure that
  all columns are known on exit.
}

const
  invalid = maxint;    { "safe" mark on invalid output elements. }

private var

  { Testing }
  memElems : integer;  { Number of elements that can be allocated before
                         memory is full (for testing memory overflow
                         situation) }

procedure Ts_Explain (var f: text; result: Ts_result);
  begin
  write(f,'Ts_result: ');
  case result of
    Ts_Ok         : write(f,'Ok');
    Ts_spotsfull  : write(f,'Spot table full');
    Ts_memfull    : write(f,'Memory full');
    Ts_elemfull   : write(f,'Transcript element full');
    Ts_immutable  : write(f,'Transcript element immutable');
    Ts_nontext    : write(f,'Not a text element');
    Ts_noelem     : write(f,'No element (nil)');
    Ts_nullelem   : write(f,'Null element (zero length)');
    Ts_wrongorder : write(f,'Two spots in wrong order');
    Ts_nonewline  : write(f,'Newline missing where expected');
    Ts_wasnewline : write(f,'Newline present, should not be');
    Ts_noteot     : write(f,'Spot should be at EOT');
    Ts_atomic     : write(f,'Element cannot be split');
    Ts_notinput   : write(f,'Not at input item');
    Ts_endinput   : write(f,'End of input (at EOT)');
    Ts_validoutput: write(f,'Valid output, not dewritten');
    Ts_fixedoutput: write(f,'Cannot insert text before output');
    Ts_wrongform  : write(f,'Input file format is wrong');
    Ts_invalid    : write(f,'Invalid or unimplemented operation');
    end;
  end {Ts_explain};

private
function LengthElem (elem: ElementRef): integer;
  { Length (number of characters/columns) "shown" for the element. }
  begin
  with elem^ do
    case kind of
      BotElement  : LengthElem := 0; { invisible }
      TextElement : LengthElem := length(theText);
      OtherElement: LengthElem := 7; { <other>   }
      EotElement  : LengthElem := 0; { invisible }
      end;
  end {LengthElem};

private
procedure ShowElement (var f: text; elem: ElementRef; offset,len: integer);
  { Display (part of) the element, starting from the given
    offset and proceeding for the given length (number of
    characters). }
  var txt: vstring;
  begin
  with elem^ do begin
    case kind of
      BotElement  : txt:='';
      TextElement : txt:=theText;
      OtherElement: txt:='<other>';
      EotElement  : txt:='';
      end;
    end;
  write(f,Copy(txt,offset+1,len));
  end {ShowElement};

private
procedure RenderElement (elem: ElementRef; offset,len: integer);
  { Render (part of) the element, starting from the given
    offset and proceeding for the given length (number of
    characters). }
  var actcol: integer;  { dummy for now }
  begin
  with elem^ do begin
    case kind of
      BotElement  : ;
      TextElement : Rn_TextVar(theText  ,offset+1,len,len,actcol);
      OtherElement: Rn_Text   ('<other>'             ,len,actcol);
      EotElement  : ;
      end;
    end;
  end {RenderElement};

private
procedure FlipPrev (var elem: ElementRef; var offset: integer);
  { If at the start of an element, flip to an equivalent position
    at the end of the preceding element. Of course, this is not
    possible if the preceding element has a newline. }
  begin
  if  (offset = 0)
  and (not elem^.prev^.newline) then begin
    elem  :=elem^.prev;
    offset:=LengthElem(elem);
    end;
  end {FlipPrev};

private
procedure FlipNext (var elem: ElementRef; var offset: integer);
  { If at the end of an element, flip to an equivalent position
    at the start of the next element. Of course, this is not
    possible if the element has a newline. }
  begin
  if  (not elem^.newline)
  and (offset >= LengthElem(elem))
  then begin
    elem  :=elem^.next;
    offset:=0;
    end;
  end {FlipNext};

private
procedure Flip (var elem: ElementRef; var offset: integer);
  { If at the start (or end) of an element, flip to an equivalent
    position at the end of the previous (or at the start of the next)
    element. Of course, this is not possible if a newline is in
    the way. }
  begin
  if offset = 0 then begin
    if not elem^.prev^.newline then begin
      elem  :=elem^.prev;
      offset:=LengthElem(elem);
      end
    end
  else
  if offset = LengthElem(elem) then
    if not elem^.newline then begin
      elem  :=elem^.next;
      offset:=0;
      end;
  end {Flip};

private
function FirstOnLine (elem: ElementRef): boolean;
  { Whether the element is the first on its line. }
  begin
  FirstOnLine := elem^.prev^.newline;
  end {FirstOnLine};

private
function LastOnLine (elem: ElementRef): boolean;
  { Whether the element is the last on its line. }
  begin
  LastOnLine := elem^.newline;
  end {LastOnLine};

private
function BolElement (elem: ElementRef): ElementRef;
  { Returns a reference to the element that starts the line of which
    elem is a part. }
  var b: ElementRef;
  begin
  b:=elem;
  while not FirstOnLine(b) do b:=b^.prev;
  BolElement:=b;
  end {BolElement};

private
function ColElem (elem: ElementRef): integer;
  { Column of start of element. }
  var
    e: ElementRef;
    c: integer;
  begin
  e:=elem; c:=1;
  while not FirstOnLine(e) do begin
    e:=e^.prev;
    c:=c + LengthElem(e);
    end;
  ColElem:=c;
  end {ColElem};

private
procedure BotSpot (var pt : SpotOb;
                   var tip: Transcript);
  { Move the spot to the beginning of the transcript. }
  begin
  with pt do begin
    elem   :=tip.botElem^.next;
    offset :=0;
    row    :=1;
    col    :=1;
    knowcol:=true;
    inpos  :=0;
    end;
  end {BotSpot};


function  Ts_MoreSpots (var tip: Transcript): integer;
  var i,more: Spot;
  begin
  with tip do begin
    more:=MaxSpots-hiSpot;
    for i:=1 to hiSpot-1 do
      if not (i in spotlist) then more:=more+1;
    end;
  Ts_MoreSpots:=more;
  end {Ts_MoreSpots};

procedure Ts_NewSpot (var theSpot: Spot;
                      var tip    : Transcript;
                      var result : Ts_result);
  var i: Spot;
  begin
  with tip do begin
    theSpot:=0;
    for i:=hiSpot-1 downto 1 do
      if not (i in spotlist) then theSpot:=i;
    if theSpot = 0 then
      if hiSpot < MaxSpots then begin
        theSpot:=hiSpot+1;
        hiSpot :=theSpot;
        end;
    if theSpot = 0 then
      result:=Ts_spotsfull
    else begin
      spotlist:=spotlist + [theSpot];
      BotSpot(spots[theSpot],tip);
      result:=Ts_ok;
      end;
    end;
  end {Ts_NewSpot};

procedure Ts_DelSpot (var theSpot: Spot;
                      var tip    : Transcript);
  var i,hi: Spot;
  begin
  with tip do begin
    if theSpot in spotlist then begin
      spotlist:= spotlist - [theSpot];
      theSpot := 0;
      hi:=0;
      for i:=1 to hiSpot do if i in spotlist then hi:=i;
      hiSpot:=hi;
      end
    else
      writeln(errout,'Ts_DelSpot: theSpot is not in spotlist.');
    end;
  end {TsDelSpot};


private
procedure ColKnow (theSpot: Spot; var tip: Transcript);
  { Make sure that the column of the spot is known. }
  begin
  with tip do begin
    with spots[theSpot] do begin
      if not knowcol then begin
        col    :=ColElem(elem) + offset;
        knowcol:=true;
        end;
      end;
    end;
  end {ColKnow};


procedure Ts_Where (    theSpot: Spot;
                    var tip    : Transcript;
                    var row,col: integer);
  begin
  ColKnow(theSpot,tip);
  row:=tip.spots[theSpot].row;
  col:=tip.spots[theSpot].col;
  end {Ts_Where};


function  Ts_Pos (theSpot: Spot; var tip: Transcript): integer;
  begin
  Ts_Pos:=tip.spots[theSpot].inpos;
  end {Ts_Pos};


private
function Order (var {i} p1,p2: SpotOb): integer;
  { Row-column order of two spots = sign(p1-p2).
    Columns assumed known. }
  var diff: integer;
  begin
  if not (p1.knowcol and p2.knowcol) then           {assert}
    writeln(errout,'Order: some column unknown.');
  diff:= p1.row - p2.row;
  if diff = 0 then diff:= p1.col - p2.col;
  if      diff < 0 then Order:= -1
  else if diff > 0 then Order:= +1
                   else Order:=  0;
  end {Order};


function Ts_Order (var spot1,spot2: Spot;
                   var tip        : Transcript): integer;
  begin
  { ColKnow(spot1,tip); }  { It is assumed that columns are known. }
  { ColKnow(spot2,tip); }
  with tip do Ts_Order:= Order(spots[spot1],spots[spot2]);
  end {Ts_Order};
      
function Ts_AtEol (theSpot: Spot; var tip: Transcript): boolean;
  begin
  with tip,spots[theSpot] do
    if      offset <> LengthElem(elem) then Ts_AtEol:=false
    else if elem^.newline              then Ts_AtEol:=true  { also XotElem }
    else                                    Ts_AtEol:=elem^.next = eotElem;
  end {Ts_AtEol};

function Ts_AtEot (theSpot: Spot; var tip: Transcript): boolean;
  begin
  with tip,spots[theSpot] do begin
    FlipNext(elem,offset);
    Ts_AtEot := elem = eotElem;
    end;
  end {Ts_AtEot};


private
procedure UpSpotOff (var theSpot  : Spot;
                     var tip      : Transcript;
                         newoffset: integer);
  { Update all spot offsets because of changes in the transcript.
    The transcript has been changed immediately before the given
    spot by inserting or deleting data so that the spot now lies
    on the given new offset in its element. Here we set the new offset
    for the given spot and also update the offset of other spots,
    to keep them in their original positions, if possible.
    The spots are updated as follows:
      a. Spots on the same element, at or after the given spot,
         are given new offsets
      b. Spots on the same element that lie within a deleted
         offset range (before the given spot) are moved to coincide
         with the given spot
    The new values for the given spot are set by the same rules. We
    assume that the given spot already refers to the correct element.
    The routine UpSpotPos is usually called after this one, to update
    row and column numbers and input symbol counts.
  }
  var
    s: Spot;
    p: SpotOb;
    doff: integer;
  begin
  with tip do begin
    p:=spots[theSpot];  { the old values. }
    doff:=newoffset - p.offset;
    for s:=1 to hiSpot do if s in spotlist then
      with spots[s] do
        if elem = p.elem then begin
          if      offset  >= p.offset  then { a } offset:=offset + doff
          else if offset  >  newoffset then { b } offset:=newoffset;
          end;
    end;
  end {UpSpotOff};

private
procedure UpSpotPos (var theSpot: Spot;
                     var tip    : Transcript;
                         newrow,newcol: integer;
                         newinpos     : integer);
  { Update all spot row/col/inpos because of changes in the transcript.
    The transcript has been changed immediately before the given
    spot by inserting or deleting data so that the spot now lies
    on the given new row and column in the transcript and has the
    given new number of input symbols (inpos) to its left.

    Here we set the new values for this spot and also update the
    row/col/inpos of all other spots to keep them in the same positions
    as before, if this is possible. The routine UpSpotOff is usually
    called before this, to update spot offset values.

    The true column numbers of all spots must have been known before
    the changes in the transcript, since these values (which now are
    false for some spots) are used here to select which spots need
    updating and how.

    In general, the routine leaves unchanged all spots that lie before the
    changes. Spots that lie after the position (old and new) of the
    given spot are assigned new row, column, and/or input symbol numbers.
    Spots that lie on a deleted region - in the range (new pos)..(old pos)
    of the given spot - are moved to coincide with the new position of
    the given spot, i.e. they are "left behind" in the place where the
    text was deleted.

    The new values for the given spot are set by the same rules. We
    assume that the given spot already refers to the correct element.

    As a special case, spots on the same line and same column as the
    given spot are moved to the same element and offset (and same row and
    column) as the given spot. This handles spots that were equivalent
    to the given spot before the transcript was changed, but would no
    longer be equivalent because of inserted newlines
  }
  var
    s: Spot;
    p,newp: SpotOb;
    drow,dcol,dinp: integer;
    dp: integer;

  begin
  with tip do begin
    p   :=spots[theSpot];  { the old values of row, col, inpos }
    newp:=p;               { the new values, all }
    newp.row  :=newrow;
    newp.col  :=newcol;
    newp.inpos:=newinpos;
    drow:=newrow    - p.row;
    dcol:=newcol    - p.col;
    dinp:=newinpos  - p.inpos;

    for s:=1 to hiSpot do if s in spotlist then begin
      dp:=Order(spots[s],p);
      if dp > 0 then
        { Spot s lies after changes, gets new row, col, inpos. }
        with spots[s] do begin
          if row = p.row then col:=col + dcol;
          row  :=row   + drow;
          inpos:=inpos + dinp;
          end
      else
      if dp = 0 then
        { Spot s is coincident with the given spot, and is set to the
          same position (see note re special case above. }
        spots[s]:=newp
      else
      if Order(spots[s],newp) >= 0 then
        { Spot s lies within a deleted region, and is set to the
          same position as the given spot. }
        spots[s]:=newp;
    { else }
        { Spot s lies before changes, and is not affected. }
      end {for s};
    end {with tip};
  end {UpSpotPos};

private
procedure SpotNewline (var theSpot: Spot;
                       var tip    : Transcript;
                       var result : Ts_result);
  { Given a spot at the start of an element (offset zero), make a
    line break by setting the newline flag on the preceding element (the
    flag is assumed to be off). }
  var newinpos: integer;
  begin
  with tip.spots[theSpot] do begin
    with elem^.prev^ do begin
      if newline then result:=Ts_wasnewline
                 else result:=Ts_ok;
      newline:=true;
      if source = InputSource then newinpos:=inpos+1
                              else newinpos:=inpos;
      end;
    UpSpotPos(theSpot, tip,  {row   } row+1,
                             {col   } offset+1,
                             {inpos } newinpos)
    end;
  end {SpotNewline};

private
procedure SplitSpots (    tail    : ElementRef;
                          splitoff: integer;
                      var tip     : Transcript);
  { Move spots after an element has been split into two at an offset.
    The "tail" parameter is the split element, which now contains the
    tail part of the text, starting from the offset "splitoff" in the
    original element. The text before splitoff has been moved to a
    new element which is tail^.prev.
    This routine updates spots that referred to the split element.
    Spots in the first part are referred to the new element,
    and spots in the second part are reduced in offset.
    Row and column numbers and input symbol counts are not changed. }
  var s: Spot;
  begin
  with tip do
    for s:=1 to hiSpot do if s in spotlist then
      with spots[s] do
        if elem = tail then
          if offset < splitoff then elem  :=tail^.prev
                               else offset:=offset - splitoff;
  end {SplitSpots};

private
procedure JoinSpots (    tail   : ElementRef;
                         joinoff: integer;
                     var tip    : Transcript);
  { Move spots after joining two adjacent elements.
    The "tail" parameter is the second element, from which all text
    has been appended to the first element, tail^.prev. The length
    of the text in the first element was originally "joinoff", so this
    is the offset at which the new text starts.
    This routine updates spots that referred to the tail element.
    They are changed to refer to the joined element (tail^.prev) and
    their offsets are suitably increased.
    Row and column numbers and input symbol counts are not changed.
  }
  var s: Spot;
  begin
  with tip do
    for s:=1 to hiSpot do if s in spotlist then
      with spots[s] do
        if elem = tail then begin
          elem  :=tail^.prev;
          offset:=offset + joinoff;
          end;
  end {JoinSpots};

private
procedure DelNullSpots (    nullelem,newelem: ElementRef; 
                            newoffset: integer;
                        var tip      : Transcript);
  { Move spots off a null element that is to be deleted.
    The spots are moved to the given new element and offset.
    Row and column numbers and input symbol counts are not changed. }
  var s: Spot;
  begin
  with tip do
    for s:=1 to hiSpot do if s in spotlist then
      with spots[s] do
        if elem = nullelem then begin
          if offset > 0 then
            writeln(errout,'DelNullSpots: nonzero offset = ',offset:1);
          elem  :=newelem;
          offset:=newoffset;
          end;
  end {DelNullSpots};


private
procedure Link (a,b: ElementRef);
  { Connect  a -> b  and the corresponding reverse link. }
  begin
  a^.next:=b;
  b^.prev:=a;
  end {Link};

private
procedure InsertBefore (beforeThis,insertThis: ElementRef);
  { Twiddle links to insert an element before another element. }
  var p: ElementRef;
  begin
  p:=beforeThis^.prev;
  Link(p,insertThis           );
  Link(  insertThis,beforeThis);
  end {InsertBefore};

private
procedure InsertAfter (afterThis,insertThis: ElementRef);
  { Twiddle links to insert and element after another element. }
  var n: ElementRef;
  begin
  n:=afterThis^.next;
  Link(afterThis,insertThis  );
  Link(          insertThis,n);
  end {InsertAfter};

private
procedure DisposeElement (var elem: ElementRef);
  { Dispose (deallocate) an element. No links are changed. }
  begin
  Dispose(elem);
  elem:=nil;
  memElems:=memElems+1;
  end {DisposeElement};

private
procedure DelElement (var elem: ElementRef;
                      var tip : Transcript);
  { Twiddle links to delete this element from the transcript.
    Also dispose of it. Returns elem=nil.
    Note, spots and output stack are NOT updated. }
  begin
  if elem=tip.botElem then
    writeln(errout,'Transcript: tried to delete botElem, naughty!')
  else
  if elem=tip.eotElem then
    writeln(errout,'Transcript: tried to delete eotElem, naughty!')
  else begin
    Link(elem^.prev,elem^.next);
    DisposeElement(elem);
    end;
  end {DelElement};

private
procedure NewElement (var e     : ElementRef;
                          kind  : ElementKind;
                      var result: Ts_result);
  { Allocate a new element, if possible. Also sets e^.kind. }
  begin
  if memElems > 0 then new(e {,kind})
                  else e:=nil;
  if e<>nil then begin
    e^.kind :=kind;
    result  :=Ts_ok;
    memElems:=memElems-1;
    end
  else
    result:=Ts_memfull;
  end {NewElement};


procedure Ts_SetMemSize (numberElems: integer);
  begin
  memElems:=numberElems;
  end {Ts_SetMemSize};


private
procedure SetOff (var p        : SpotOb;
                      newoffset: integer);
  { Change the offset value of a spot.
    We assume that the new offset given is possible, i.e. >= 0
    and <= length of element.
    Row,col, and inpos counts are updated, too. }
  var dx: integer;
  begin
  with p do begin
    if newoffset < 0 then                             { assert }
      writeln(errout,'Setoff <0: ',newoffset:1);
    if newoffset > LengthElem(elem) then              { assert }
      writeln(errout,'Setoff >len: ',newoffset:1,'>',LengthElem(elem):1);
    dx    :=newoffset-offset;
    offset:=newoffset;
    if knowcol then col:=col + dx;
    if elem^.source = InputSource then inpos:=inpos + dx;
    end;
  end {SetOff};

private
procedure IncPos (var spot: SpotOb);
  { Update row,col,knowcol,inpos when a spot moves from elem:offset to
    the start of the next element. Note, element and offset are NOT
    changed. The spot is assumed not to be at eot. }
  var dx: integer;
  begin
  with spot do begin
    dx:=LengthElem(elem)-offset;
    if elem^.source = InputSource then begin
      inpos:=inpos + dx;
      if elem^.newline then inpos:=inpos+1;
      end;
    if elem^.newline then begin
      row:=row+1;
      col:=1;
      knowcol:=true;
      end
    else
      if knowcol then col:=col + dx;
    end;
  end {IncPos};

private
procedure IncElem (var spot: SpotOb);
  { Moves to start of next element. The spot is assumed not to be at eot. }
  begin
  IncPos(spot);
  with spot do begin
    elem  :=elem^.next;
    offset:=0;
    end;
  end {IncElem};

private
procedure DecElem (var spot: SpotOb);
  { Move to end of preceding element, which is assumed not to be the
    bot element. }
  var prev: ElementRef;
      len : integer;
  begin
  with spot do begin
    prev:=elem^.prev;
    if prev = nil then writeln(errout,'DecElem = nil!')  { assert }
    else begin
      len :=LengthElem(prev);
      if elem^.source = InputSource then inpos:=inpos - offset;
      if prev^.newline then begin
        row    :=row-1;
        knowcol:=false;
        if prev^.source = InputSource then inpos:=inpos - 1;
        end
      else
        if knowcol then col:=col - offset;
      elem  :=prev;
      offset:=len ;
      end {not nil};
    end;
  end {DecElem};
  
private
procedure NextElem (var theSpot: Spot; var tip: Transcript);
  { Move spot to start of next element.
    The spot is assumed not to be at eotElem.
    The new position may be at eotElem. }
  begin
  IncElem(tip.spots[theSpot]);
  end {NextElem};

private
procedure PrevElem (var theSpot: Spot; var tip: Transcript);
  { Move spot to start of preceding element.
    We assume that the spot is not on the first element
    in the transcript. }
  begin
  DecElem(tip.spots[theSpot]  );
  SetOff (tip.spots[theSpot],0);
  end {PrevElem};

private
procedure SkipForw (var spot: SpotOb);
  { Move forward to just before next input symbol, or eot. }
  begin
  with spot do FlipNext(elem,offset);
  while spot.elem^.source = OutputSource do IncElem(spot);
  end {SkipForw};

private
procedure SkipBack (var spot: SpotOb);
  { Move backward to just after preceding input symbol, or bot. }
  var done: boolean;
  begin
  with spot,elem^ do done:= (source = InputSource) and (offset > 0);
  while not done do begin
    if spot.offset > 0 then SetOff(spot,0);
    done:= spot.elem^.prev^.source <> OutputSource;
    if not done then DecElem(spot);
    end;
  end {SkipBack};
          

procedure Ts_Move (    fromSpot : Spot;
                       rows,cols: integer;
                   var movedSpot: Spot;
                   var tip      : Transcript);
  var newrow   : integer;
      newcol   : integer;
      newoffset: integer;
      len      : integer;

  begin
  tip.spots[movedSpot] := tip.spots[fromSpot];
  with tip,spots[movedSpot] do begin
    ColKnow(movedSpot,tip);
    newrow:=row+rows;
    newcol:=col+cols;
    if newcol<1 then newcol:=1;      { accept nothing else }

    { Increase row? }
    while row < newrow do
      if elem = eotElem then newrow:=row  { at end of transcript }
                        else NextElem(movedSpot,tip);

    { Decrease row? }
    while row > newrow do
      if elem^.prev = botElem then newrow:=row  { at start of transcript }
                              else PrevElem(movedSpot,tip);

    { Decrease col? }
    ColKnow(movedSpot,tip);

    while col > newcol do begin
      newoffset:=offset+newcol-col;
      if newoffset >= 0 then SetOff(spots[movedSpot],newoffset)
      else begin
        PrevElem(movedSpot,tip);  { this may go too far, but the loop }
        ColKnow (movedSpot,tip);  { to increase col (below) will correct }
        end;
      end;

    { Increase col? }
    while col < newcol do begin
      newoffset:=offset + newcol-col;
      len:=LengthElem(elem);
      if newoffset <= len then SetOff(spots[movedSpot],newoffset)
      else
      if LastOnLine(elem) then begin
        SetOff(spots[movedSpot],len);
        newcol:=col;
        end
      else NextElem(movedSpot,tip);
      end;

    end {with movedSpot};
  end {Ts_Move};


procedure Ts_Goto (var theSpot: Spot;
                   var tip    : Transcript;
                       row,col: integer);
  var oldrow,oldcol: integer;
  begin
  Ts_Where(theSpot,tip,oldrow,oldcol);
  Ts_Move (theSpot,row-oldrow,col-oldcol,theSpot,tip);
  end {Ts_Goto};


procedure Ts_Skipln (var theSpot: Spot;
                     var tip    : Transcript;
                     var result : Ts_result);
  var wasnl,eot: boolean;
  begin
  repeat
    with tip.spots[theSpot].elem^ do begin
      wasnl:=(source = InputSource) and newline;
      eot  := source = OmSource;
      end;
    if not eot then NextElem(theSpot,tip);
  until wasnl or eot;
  if eot then result:=Ts_nonewline
         else result:=Ts_ok;
  end {Ts_Skipln};


procedure Ts_GotoEnd (var theSpot: Spot;
                      var tip    : Transcript);
  begin
  Ts_Goto(theSpot,tip,maxint,MaxCol);
  end {Ts_GotoEnd};


function Ts_LineLength (theSpot: Spot; var tip: Transcript): integer;
  var len: integer;
      e  : ElementRef;
      eol: boolean;
  begin
  with tip, spots[theSpot] do begin
    len:=col-offset-1;         { Length up to start of this element }
    e  :=elem;
    repeat
      len:=len+LengthElem(e);    { Add in this element }
      eol:=e^.newline;
      e  :=e^.next;
    until eol or (e = eotElem);
    end;
  Ts_LineLength:=len;
  end {Ts_LineLength};


procedure Ts_Show (    theSpot      : Spot; var tip: Transcript;
                       fromCol,toCol: integer;
                   var numCol       : integer);
  var elem: ElementRef;  { an element on the line }
      col : integer;     { starting column of elem }
      offs: integer;     { starting offset in elem }
      offq: integer;     { ending offset (+1) in elem   }
      len : integer;     { length of elem }
      cols: integer;     { columns written from elem }
      done: boolean;
  begin
  elem  :=BolElement(tip.spots[theSpot].elem);
  col   :=1;
  numCol:=0;
  repeat
    len :=LengthElem(elem);
    offs:=fromCol - col; if offs <  0  then offs:= 0 ;
    offq:=toCol+1 - col; if offq > len then offq:=len;
    cols:=offq - offs;
    if cols > 0 then begin
      RenderElement(elem,offs,cols);
      numCol:=numCol + cols;
      end;
    done:=LastOnLine(elem) or (col > toCol);
    if not done then begin
      elem:=elem^.next;
      col :=col + len;
      end;
  until done;
  end {Ts_Show};


private
procedure SplitElem (var elem  : ElementRef;
                         offset: integer;
                     var tip   : Transcript;
                     var result: Ts_result);
  { Split the given element into two elements by  moving the text
    before the given offset to a new element. The new element is
    linked into the transcript before the given element, so the content
    of the transcript is not changed, just the element structure.
    Spots on the element are also updated. 
    Can also split non-text elements, if the offset is zero or length(elem).
    The new, empty element is always a text element in this case; the
    caller should make it non-null at once. }
  var x: ElementRef;
  begin
  if elem^.kind = TextElement then begin
    NewElement(x,TextElement,result);
    if result=Ts_ok then begin
      x^.source :=elem^.source;  { probably Input }
      x^.newline:=false;
      InsertBefore(elem,x);
      x^.theText:=Copy(elem^.theText,1,offset);
      Delete(elem^.theText,1,offset);
      SplitSpots(elem,offset,tip);
      end;
    end {text}
  else {nontext}
  if (offset > 0) and (offset < LengthElem(elem))
  then result:=Ts_atomic
  else begin
    NewElement(x,TextElement,result);
    if result=Ts_ok then begin
      x^.source :=elem^.source;
      x^.theText:='';
      if offset=0 then begin
        InsertBefore(elem,x);
        x^.newline:=false;
        end
      else begin
        InsertAfter(elem,x);
        x^.newline:=elem^.newline;
        elem^.newline:=false;
        elem:=x;
        end;
      end;
    end;
  end {SplitElem};

private
procedure JoinElem (var elem: ElementRef; var tip: Transcript);
  { If possible, coalesce the element with its successor. }
  var next: ElementRef;
      len : integer;
  begin
  next:=elem^.next;
  if  (elem^.source = next^.source)
  and (elem^.kind   = next^.kind  )
  and (elem^.kind   = TextElement)
  and (not elem^.newline)
  then begin
    len:=LengthElem(elem);
    if len + LengthElem(next) <= MaxElemLength then begin
      elem^.theText:=concat(elem^.theText,next^.theText);
      elem^.newline:=next^.newline;
      JoinSpots(next,len,tip);
      DelElement(next,tip);
      end;
    end;
  end {JoinElem};
 

procedure Ts_Unwrite (var theSpot: Spot;
                      var tip    : Transcript;
                          safePos: integer);
  var here: boolean;
  begin
  with tip, spots[theSpot] do begin
    repeat
      case elem^.source of
        OutputSource: begin
                      here:= elem^.safe <= safePos;
                      if not here then begin
                        elem^.safe:= invalid;
                        modified:= true;
                        end;
                      end;
        InputSource : here:= false;
        OmSource    : here:= elem = botElem;
        end;
      if not here then PrevElem(theSpot,tip);
    until here;
    if elem^.newline then NextElem(theSpot,tip)  { also covers botElem }
                     else SetOff(spots[theSpot],LengthElem(elem));
    ColKnow(theSpot,tip);
    end;
  end {Ts_Unwrite};


private
procedure Dewrite (var theSpot: Spot;
                   var tip    : Transcript;
                       force  : boolean;
                   var some   : boolean;
                   var result : Ts_result);
  { Deletes all output that immediately follows the given spot.
    If the given spot is within an input element nothing is done.
    If the given spot is at the end of an element (other than
    the eot), it is moved to the start of the next element.
    If that element is an output element, it and its successors
    are candidates for deletion. All output to be deleted must be
    invalid (safe = invalid), unless the 'force' parameter is true.
    Returns with spot before the next input symbol or at eot.
    Spots are NOT updated; the row and column counters refer to
    the new position of the spot (e.g. the input symbol after the
    output) before the output was deleted.
    The parameter 'some' returns true if some output was deleted. }

  var x: ElementRef;
  begin
  result:=Ts_ok;
  some  :=false;
  with tip.spots[theSpot] do begin
    { If at end of element, flip to next. }
    if elem <> tip.eotElem then FlipNext(elem,offset);
    if (elem^.source = OutputSource) and (offset = 0) then
      repeat
        if (elem^.safe = invalid) or force then begin
          x:=elem;
          NextElem(theSpot,tip);
          DelElement(x,tip);
          some:=true;
          end
        else
          result:=Ts_validoutput;
      until (elem^.source <> OutputSource)  { input or eot }
         or (result <> Ts_ok);
    end {with};
  end {Dewrite};


procedure Ts_Dewrite (var theSpot: Spot;
                      var tip    : Transcript;
                      var preRow,preCol: integer;
                      var result : Ts_result);
  var
    row,col: integer;  { row and col the spot will have }
    some   : boolean;  { some output was actually deleted }
  begin
  Ts_Where (theSpot,tip,row,col);
  Dewrite  (theSpot,tip,{force} false,some,result);
  Ts_Where (theSpot,tip,preRow,preCol);
  if some then begin
    tip.modified:= true;
    UpSpotPos(theSpot,tip,row,col,Ts_Pos(theSpot,tip));
    end;
  end {Ts_Dewrite};


private
function MutableText (elem: ElementRef): Ts_result;
  { Check that elem is mutable (input) text. }
  begin
  if      elem = nil                  then MutableText:=Ts_noelem
  else if elem^.source <> InputSource then MutableText:=Ts_immutable
  else if elem^.kind <> TextElement   then MutableText:=Ts_nontext
                                      else MutableText:=Ts_ok;
  end {MutableText};

private
procedure GrowEnd (var elem  : ElementRef;
                   var offset: integer;
                   var result: Ts_result);
  { If elem:offset is at end of transcript, insert a null
    input text element (must be filled in at once by the caller)
    and update elem:offset to point to it.
    If not at eot, return Ts_immutable (elem:offset may be flipped
    to next element). }
  var x: ElementRef;
  begin
  FlipNext(elem,offset);
  if elem^.kind = EotElement then begin
    NewElement(x,TextElement,result);
    if result=Ts_ok then begin
      x^.source :=InputSource;
      x^.theText:='';
      x^.newline:=false;
      Insertbefore(elem,x);
      elem  :=x;
      offset:=0;
      end;
    end {om}
  else
    result:=Ts_immutable;
  end {GrowEnd};

private
procedure PrepInsert (var elem  : ElementRef;
                      var offset: integer;
                      var result: Ts_result);
  { Prepare to insert input text at elem:offset.
    May flip to next or previous element, and may insert a null
    input text element before eot, if required. }
  begin
  result:=MutableText(elem);
  if result <> Ts_ok then begin
    Flip(elem,offset);
    result:=MutableText(elem);
    end;
  if result = Ts_ok then begin
    if  (offset = LengthElem(elem))
    and (not elem^.newline)
    and (elem^.next^.source = OutputSource)
    then result:=Ts_fixedoutput;
    end
  else
    GrowEnd(elem,offset,result);
  end {PrepInsert};

private
procedure PrepNewline (var elem  : ElementRef;
                       var offset: integer;
                       var result: Ts_result);
  { Prepare to insert an input newline at elem:offset.
    May flip to next or previous element, and may insert a null
    input text element before eot, if required. }
  begin
  if elem^.source <> InputSource then Flip(elem,offset);
  if elem^.source = InputSource then begin
    if  (offset = LengthElem(elem))
    and (not elem^.newline)
    and (elem^.next^.source = OutputSource)
    then result:=Ts_fixedoutput
    else result:=Ts_ok
    end
  else
    GrowEnd(elem,offset,result);
  end {PrepNewline};

private
procedure PrepDest (var theSpot: Spot;
                    var tip    : Transcript;
                    var result : Ts_result);
  { Check and prepare a destination for cut or copy.
    The destination must be within or adjacent to input data.
    If within an input text element, the element is split.
    If successful, offset is returned as zero and the cut or copied
    text can be handled with InsertBefore(elem). }
  begin
  with tip.spots[theSpot] do begin
    FlipNext(elem,offset);
    if elem^.kind = EotElement then result:=Ts_ok
    else begin
      if elem^.source = InputSource then begin
        if offset = 0 then result:=Ts_ok
                      else SplitElem(elem,offset,tip,result);
        end
      else begin
        FlipPrev(elem,offset);
        if elem^.source = InputSource then result:=Ts_fixedoutput
                                      else result:=Ts_immutable;
        end;
      end;
    end {with};
  end {PrepDest};


procedure Ts_Insert (var theSpot      : Spot;
                     var tip          : Transcript;
                         ch           : char;
                     var preRow,preCol: integer;
                     var result       : Ts_result);
  var mid: integer;
  begin
  SkipForw(tip.spots[theSpot]);
  with tip.spots[theSpot] do begin
    PrepInsert(elem,offset,result);
    if result=Ts_ok then begin
      if LengthElem(elem) >= MaxElemLength then begin
        mid:=LengthElem(elem) div 2;
        SplitElem(elem,mid,tip,result);
        end;
      if result=Ts_ok then begin
        Ts_Where(theSpot,tip,preRow,preCol);
        Insert(ch,elem^.theText,offset+1);
        UpSpotOff(theSpot,tip,offset+1);
        UpSpotPos(theSpot,tip,row,col+1,inpos+1);
        tip.modified:= true;
        end;
      end;
    end {with spot};
  end {Ts_Insert};


private
function SavedOut (e: ElementRef; saved: integer): boolean;
  { Whether e is output that is due to input <= saved. }
  begin
  with e^ do
    if      source <> OutputSource then SavedOut:=false
    else if safe = invalid         then SavedOut:=false
    else if safe > saved           then SavedOut:=false
                                   else SavedOut:=true;
  end {SavedOut};


procedure Ts_Newline (var theSpot: Spot;
                      var tip    : Transcript;
                      var preRow,preCol: integer;
                      var result : Ts_result);
  var saved : integer;  { number of input symbol before new newline }
      newrow,newcol,newinp: integer;
      pt: SpotOb;
      newl,outp: boolean;
      first,last: ElementRef;  { the tail of the line is first..last }

  begin
  SkipForw(tip.spots[theSpot]);
  with tip.spots[theSpot] do begin
    PrepNewline(elem,offset,result);
    preRow:=row;
    preCol:=col;
    saved :=inpos;
    if result=Ts_ok then begin
      if  (    offset = 0)                          { a special case }
      and (    elem^.prev^.source = InputSource)    { that might be }
      and (not elem^.prev^.newline)                 { common. }
      then SpotNewline(theSpot,tip,result)
      else
      if  (    offset = LengthElem(elem))           { avoid a null input }
      and (not elem^.newline)
      then begin
           FlipNext(elem,offset);
           SpotNewline(theSpot,tip,result);
           end
      else begin
        SplitElem(elem,offset,tip,result);
        if result = Ts_ok then SpotNewline(theSpot,tip,result);
        end {split};
      end {allowed};
    end {with spot};

  if result = Ts_ok then begin
    tip.modified:= true;
    { May have to move tail of line to proper position in output.
      Find the end of the line. If there is any embedded output, then
      the line is already in correct position. }
    pt:=tip.spots[theSpot];
    first :=pt.elem;   { first element on tail }
    newrow:=pt.row;
    repeat
      newl:=pt.elem^.newline;                  { includes eotElem }
      outp:=pt.elem^.source = OutputSource;
      if not (newl or outp) then IncElem(pt);
    until newl or outp;
    if outp or (pt.elem = tip.eotElem) then
      { tail of line is already in place. }
    else begin
      { End of line found, no embedded output. See if there are
        output elements that are due to head of line. }
      last:=pt.elem;  { last element on tail }
      if SavedOut(last^.next,saved) then begin
        { Must move tail of line. Set theSpot to after the tail. }
        IncElem(pt);
        newinp:=pt.inpos;
        tip.spots[theSpot]:=pt;
        { Cut out the tail. }
        Link(first^.prev,last^.next);
        UpSpotPos(theSpot,tip,newrow,1,saved+1);
        { Find the place to reinsert the tail. }
        while SavedOut(tip.spots[theSpot].elem,saved)
           do NextElem(theSpot,tip);
        { Reinsert the tail. }
        with tip.spots[theSpot] do begin
          newrow:=row;
          newcol:=col;
          Link(elem^.prev,first);
          Link(last,elem);
          end;
        UpSpotPos(theSpot,tip,newrow+1,1,newinp);
        { Move the spot back to the start of the reinserted tail. }
        Ts_Goto(theSpot,tip,newrow,newcol);
        end {moving tail};
      end {no embedded output nor eot};
    end {splitting ok};
  end {Ts_Newline};


private
procedure DelNull (var theSpot: Spot; var tip: Transcript);
  { Given a spot on an empty input text element, see if the element
    can be deleted.
    An empty newline element cannot be deleted, unless the newline can
    be moved to the preceding element.
    If the element could not be deleted, no changes in the transcript
    are made and the spot is not moved.
    If the element was deleted, the spot is moved to another element that
    lies  at the same row,col position. }
  var
    prev,x: ElementRef;
  begin
  with tip.spots[theSpot] do begin
    x:=elem;   { because elem will be changed in DelNullSpots. }
    prev:=x^.prev;
    if x^.newline then begin
      { Try to move newline to prev element. }
      if  (prev^.source = InputSource)
      and (not prev^.newline)
      then begin
        DelNullSpots(x,prev,LengthElem(prev),tip);
        prev^.newline:=true;
        x^.newline:=false;
        DelElement(x,tip);
        end {deleting a newline and saving it}
      end {newline}
    else begin
      DelNullSpots(x,x^.next,0,tip);
      DelElement(x,tip);
      end;
    end {with spot};
  end {DelNull};


procedure Ts_Delete (var theSpot      : Spot;
                     var tip          : Transcript;
                     var preRow,preCol: integer;
                     var result       : Ts_result);
  var
    pt: SpotOb;
    newrow,newcol,newinpos: integer;
  begin
  pt:=tip.spots[theSpot];
  if pt.inpos < 1 then result:=Ts_notinput
  else begin
    SkipBack(pt);
    if (pt.offset = 0) and FirstOnLine(pt.elem)
    then result:=Ts_wasnewline
    else begin
      tip.spots[theSpot]:=pt;
      ColKnow(theSpot,tip);
      with tip.spots[theSpot] do begin
        FlipPrev(elem,offset);       { must not be at start of element }
        result:=MutableText(elem);
        if result = Ts_ok then begin
          newrow  := row;
          newcol  := col-1;
          newinpos:= inpos-1;
          Delete(elem^.theText,offset,1);
          UpSpotOff(theSpot,tip,offset-1);
          if LengthElem(elem) = 0 then DelNull(theSpot,tip);
          preRow:= row;
          preCol:= col;
          UpSpotPos (theSpot,tip,newrow,newcol,newinpos);
          tip.modified:= true;
          end {was input text};
        end {with};
      end {not newline};
    end {some input precedes};
  end {Ts_Delete};


procedure Ts_Joinline (var theSpot      : Spot;
                       var tip          : Transcript;
                       var delRow,delCol: integer;
                       var preRow,preCol: integer;
                       var result       : Ts_result);
  var
    pt     : SpotOb;
    nextrow: integer;       { number of row after joined row }
    nlpos  : integer;       { input symbol numer of deleted newline }
    saved  : integer;       { number of input symbol before deleted newline }
    first,last: ElementRef; { delimit the line to be moved (appended) }
    inlen  : integer;       { number of input symbols in moved line }
    nlelem : ElementRef;    { element with the deleted newline }
    newrow,newcol,newinpos: integer;
    arrived: boolean;       { finding the preceding newline }
    some   : boolean;       { dummy for Dewrite }
  begin
  pt:=tip.spots[theSpot];

  { Find the preceding input newline, if any. }
  repeat
    with pt.elem^.prev^ do arrived:= (source <> OutputSource) and newline;
    if not arrived then DecElem(pt);
  until arrived {at element after input newline or bot};

  if pt.elem^.prev = tip.botElem
  then result:=Ts_nonewline  { there was no preceding input newline }

  else begin
    result:=Ts_ok;           { so far }
    SetOff(pt,0);            { immediately after the newline }
    nextrow:=pt.row;         { number of row after joined row }
    nlpos  :=pt.inpos;       { input symbol number of newline to delete }
    saved  :=nlpos-1;        { number of saved input symbol }

    { Pass over saved output, if any. }
    while SavedOut(pt.elem,saved) do IncElem(pt);

    { Any other output, up to the end of the trailing input line
      (or eot), is dewritten. The input line is delimited in
      preparation for cutting it out. If there is no input to cut,
      first will be nil. }

    tip.spots[theSpot]:=pt;
    ColKnow(theSpot,tip);
    with tip.spots[theSpot] do begin
      delRow:=row;
      delCol:=col;
      first :=nil;             { will be start of right line, if any }
      last  :=nil;             { will be end ditto }
      while (last=nil)
        and (elem<>tip.eotElem)
        and (result = Ts_ok)
      do
        case elem^.source of
          OutputSource: Dewrite(theSpot,tip,{force} true,some,result);
          InputSource : begin
             if first = nil   then first:=elem;
             if elem^.newline then last :=elem;
             NextElem(theSpot,tip);
             end;
          end;
      end {with};

    if result <> Ts_ok then first:=nil;  { some problem, move no input }
    { The spot is now after the region to be removed (the output
      already has been removed). Its row and column numbers are virtual,
      referring to the transcript as it was before anything was removed. }
    with tip.spots[theSpot] do begin
      preRow:=row;
      preCol:=col;
      inlen :=inpos - nlpos;
      end;

    { Cut out the input to be moved, if any. }
    if first<>nil then begin
      if last=nil then begin   { There was no input newline before eot. }
        last:=tip.spots[theSpot].elem^.prev;   { = eotElem^.prev }
        end;
      Link(first^.prev,last^.next);
      end;

    UpSpotPos(theSpot,tip,delRow,delCol,nlpos);

    { Join lines, if any text to join. }
    Ts_Goto(theSpot,tip,nextrow,{col} 1);
    with tip.spots[theSpot] do begin
      nlelem:=elem^.prev;
      nlelem^.newline:=false;
      if first<>nil then begin
        Link(nlelem,first); Link(last,elem);
        end
      else
        last:=nlelem;               { to merge row,col,inpos computation }
      newinpos:=nlpos - 1 + inlen;  { -1 for the deleted newline }
      if (not last^.newline)        { no newline after spot }
      and (elem <> tip.eotElem)     { some saved output }
      then begin
        last^.newline:=true;
        newinpos:=newinpos+1;
        end;
      if last^.newline then begin
        newrow:=row;
        newcol:=col;
        end
      else begin
        newrow:=row-1;
        newcol:=ColElem(last) + LengthElem(last);
        end;
      UpSpotPos(theSpot,tip,newrow,newcol,newinpos);
      end;

    { Go back to nlelem, see if it should be deleted or joined. }
    repeat DecElem(tip.spots[theSpot])
    until tip.spots[theSpot].elem = nlelem;
    with tip.spots[theSpot] do
      if LengthElem(elem) = 0 then DelNull(theSpot,tip)
                              else JoinElem(elem,tip);
    ColKnow(theSpot,tip);
    tip.modified:= true;

    end {was preceding newline};
  end {Ts_Joinline};



procedure Ts_Cut  (var spot1,spot2   : Spot;
                   var fromTranscript: Transcript;
                   var toSpot        : Spot;
                   var toTranscript  : Transcript;
                   var preRow,preCol : integer;
                   var result        : Ts_result);
  var
    order  : integer;
    fie,lae: ElementRef;    { first and last elements in cut }
    rows   : integer;       { number of rows (newlines) in cut }
    insyms : integer;       { number of input symbols in cut }

  { Note re row,col,inpos counts: spot1 counts will be updated as it
    scans toward spot2, as if no text were deleted. Although spot2
    may be moved toward the start of the transcript, its counts will
    not be updated until the cut is done. }

  procedure Prep1;
    { Flip spot1 to first element to be cut, check that it is input,
      and split it if necessary. }
    begin
    with fromTranscript.spots[spot1] do begin
      FlipNext(elem,offset);
      if elem^.source <> InputSource then result:=Ts_notinput
      else
      if offset > 0 then SplitElem (elem,offset,fromTranscript,result);
      end;
    end {Prep1};

  procedure Prep2;
    { Flip spot2 to element after last element to be cut, perhaps
      splitting an element. Check that the last element before spot2
      is input. }
    begin
    with fromTranscript.spots[spot2] do begin
      FlipNext(elem,offset);
      if offset = 0 then begin
        if elem^.prev^.source <> InputSource then result:=Ts_notinput
        end
      else
      { offset > 0 }
      if elem^.source <> InputSource then result:=Ts_notinput
      else begin
        SplitElem(elem,offset,fromTranscript,result);
        FlipNext (elem,offset);  { make sure }
        end;
      end;
    end {Prep2};


  begin {Ts_Cut}
  result:=Ts_ok;                                 { to start with  }
  Ts_Where(spot2,fromTranscript,preRow,preCol);  { unless changed }

  order:=Ts_Order(spot1,spot2,fromTranscript);
  if order > 0 then
    result:=Ts_wrongorder

  else if order = 0 then
    { Nothing to cut. This case is treated specially since otherwise
      the checks below could raise useless alarms. }
    Ts_Where(spot2,fromTranscript,preRow,preCol)

  else begin { spot1 < spot2, something to cut }
    Prep1;
    if result = Ts_ok then Prep2;
    if result = Ts_ok then PrepDest(toSpot,toTranscript,result);
    if result = Ts_ok then begin
      { spot1 is now at start of first element to be cut, and
        spot2 is at the start of the element following the last
        element to be cut. All elements are to be cut whole. }
      with fromTranscript do begin
        fie   := spots[spot1].elem;
        lae   := spots[spot2].elem^.prev;
        rows  := spots[spot2].row   - spots[spot1].row  ;
        insyms:= spots[spot2].inpos - spots[spot1].inpos;
        end;
      Link(fie^.prev,lae^.next);
      Ts_Where(spot2,fromTranscript,preRow,preCol);
      with fromTranscript do begin
        with spots[spot1] do
          UpSpotPos(spot2,fromTranscript,row,col,inpos);
        spots[spot1]:=spots[spot2];  { elem needs to be set }
        end;

      { Insert cut data before toSpot in toTranscript. }
      with toTranscript.spots[toSpot] do begin
        Link(elem^.prev,fie); Link(lae,elem);
        UpSpotPos(toSpot,toTranscript,
                  row   + rows  , ColElem(elem),
                  inpos + insyms);
        end;

      fromTranscript.modified:= true;
        toTranscript.modified:= true;

      end {prep ok};
    end {something to cut};
  end {Ts_Cut};


procedure Ts_Copy (var spot1,spot2   : Spot;
                   var fromTranscript: Transcript;
                   var toSpot        : Spot;
                   var toTranscript  : Transcript;
                       asInput       : boolean;
                   var result : Ts_result);
  var
    newOb: SpotOb;    { Counts row,col etc. for new pos of toSpot. }

  procedure Copier (var p1,p2,tos: SpotOb);
    { Copy (part of) element at p1. }
    var
      x      : ElementRef;
      fio,lao: integer;   { Copy theText between offsets fio and lao }
    begin
    NewElement(x,p1.elem^.kind,result);
    if result = Ts_ok then begin
      if asInput then x^.source := InputSource
                 else x^.source := p1.elem^.source;
      x^.newline:= p1.elem^.newline and (p1.elem <> p2.elem);
      newOb.elem  := x;
      newOb.offset:= 0;
      if x^.kind = TextElement then begin
        { See which part of text to copy }
        if p1.offset > 0     then fio:=p1.offset
                             else fio:=0;
        if p1.elem = p2.elem then lao:=p2.offset
                             else lao:=LengthElem(p1.elem);
        x^.theText:= copy(p1.elem^.theText,fio+1,lao-fio);
        end {text element};
      IncPos(newOb);
      InsertBefore(tos.elem,x);
      end;
    end {Copier};

  begin {Ts_Copy}
  if Ts_Order(spot1,spot2,fromTranscript) > 0
  then result:=Ts_wrongorder
  else
  if Ts_Order(spot1,spot2,fromTranscript) < 0 then begin
    { Something to copy. }
    PrepDest(toSpot,toTranscript,result);
    if result = Ts_ok then begin
      with fromTranscript do begin
        with spots[spot1] do FlipNext(elem,offset);
        with spots[spot2] do FlipPrev(elem,offset);
        end;
      ColKnow(toSpot,toTranscript);
      newOb := toTranscript.spots[toSpot];
      result:= Ts_ok;

      while (Ts_Order(spot1,spot2,fromTranscript) < 0)  { something to copy }
      and   (result = Ts_ok)                            { and memory left }
      do begin
        Copier(fromTranscript.spots[spot1  ],
               fromTranscript.spots[spot2  ],
                 toTranscript.spots[toSpot]);
        NextElem(spot1,fromTranscript);                 { may go > spot2 }
        end {when all done or error};

      with fromTranscript do
        spots[spot1]:=spots[spot2];
      with newOb do
        UpSpotPos(toSpot,toTranscript,row,col,inpos);

      fromTranscript.modified:= true;
      
      end {destination ok};
    end {something to copy};
  end {Ts_Copy};


procedure Ts_Erase (var tip: Transcript);
  var e,n: ElementRef;
      i  : Spot;
  begin
  e:=tip.botElem^.next;            { first element. }
  while e <> tip.eotElem do begin
    n:=e^.next;
    DelElement(e,tip);
    e:=n;
    end;
  for i:=1 to tip.hiSpot do
    if i in tip.spotlist then BotSpot(tip.spots[i],tip);
  tip.modified:= false;
  end {Ts_Erase};


procedure Ts_Skip (var theSpot: Spot;
                   var tip    : Transcript;
                   var result : Ts_result);
  begin
  SkipForw(tip.spots[theSpot]);
  if tip.spots[theSpot].elem^.source = InputSource
  then result:=Ts_ok
  else result:=Ts_endinput;
  end {Ts_Skip};


function  Ts_AtOutput (var theSpot: Spot;
                       var tip    : Transcript): boolean;
  begin
  with tip.spots[theSpot] do begin
    FlipNext(elem,offset);
    if elem^.source <> OutputSource then Ts_AtOutput:=false
    else if offset > 0 then Ts_AtOutput:=true
    else Ts_AtOutput:= elem^.prev^.source = OutputSource;
    end {with};
  end {Ts_AtOutput};


procedure Ts_Seek (     fromSpot : Spot;
                        beforePos: integer;
                   var  movedSpot: Spot;
                   var  tip      : Transcript;
                   var  result   : Ts_result);
  var newinpos  : integer;
      newoffset : integer;  { Offset newinpos would have in this element. }
  begin
  tip.spots[movedSpot]:=tip.spots[fromSpot];
  newinpos:=beforePos-1;
  if newinpos < 0 then result:=Ts_invalid
  else
  with tip,spots[movedSpot] do begin

    { Move backward? }
    while inpos > newinpos do
      if elem^.source <> InputSource
      then PrevElem(movedSpot,tip)
      else begin
        newoffset:=offset + newinpos - inpos;
        if newoffset >= 0
        then SetOff(spots[movedSpot],newoffset)  { is within this element }
        else PrevElem(movedSpot,tip);  { may go too far, but the loop }
        end;                           { to move forward will correct. }

    ColKnow(movedSpot,tip);

    { Move forward? }
    while (inpos < newinpos) and (elem <> eotElem) do
      if elem^.source <> InputSource
      then NextElem(movedSpot,tip)
      else begin
        newoffset:=offset + newinpos - inpos;
        if newoffset <= LengthElem(elem)
        then SetOff(spots[movedSpot],newoffset)  { is within this element }
        else NextElem(movedSpot,tip);
        end;

    if      inpos = newinpos then Ts_Skip(movedSpot,tip,result)
    else if inpos < newinpos then result:=Ts_endinput
                             else result:=Ts_invalid;
    end {with};
  end {Ts_Seek};


procedure Ts_Read (var theSpot: Spot;
                   var tip    : Transcript;
                   var ch     : char;
                   var pos    : integer;
                   var result : Ts_result);
  var len: integer;
  begin
  { Find out if any input is left. }
  Ts_Skip(theSpot,tip,result);
  if result = Ts_ok then
  with tip.spots[theSpot] do begin
    case elem^.source of
      OutputSource: result:=Ts_notinput;  { should never occur }
      OmSource    : result:=Ts_endinput;  { should neven occur }
      InputSource : begin
        len:=LengthElem(elem);
        if offset < len then begin
          if elem^.kind = TextElement then begin
            { Read the input symbol }
            ch  :=elem^.theText[offset+1];
            pos :=inpos+1;
            result:=Ts_ok;
            { Advance to next position }
            offset:=offset+1;     {assert( offset <= len )}
            if knowcol then col:=col+1;
            inpos :=inpos +1;
            end {text}
          else
            result:=Ts_nontext;
          end {within element}

        else begin
          { At end of element - must be a newline }
          if not elem^.newline then
            writeln(errout,'Ts_Read: end of element without newline!?');
          ch  :=' ';
          pos :=inpos+1;
          result:=Ts_ok;
          { Advance to next element. }
          NextElem(theSpot,tip);
          end {newline};
        end {input};
      end {case source};
    end {with spot};
  end {Ts_Read};


procedure Ts_Reads (var theSpot: Spot;
                    var tip    : Transcript;
                    var data   : vstring;
                    var pos    : integer;
                    var result : Ts_result);

  var len : integer;  { length of input element }
      left: integer;  { number of symbols left in input element }
  begin
  { Find out if any input is left. }
  Ts_Skip(theSpot,tip,result);
  if result = Ts_ok then
  with tip.spots[theSpot] do begin
    case elem^.source of
      OutputSource: result:=Ts_notinput;  { should never occur }
      OmSource    : result:=Ts_endinput;  { should neven occur }
      InputSource : begin
        len := LengthElem(elem);
        left:= len - offset;
        if left > 0 then begin  { some real input left in element }
          if elem^.kind = TextElement then begin
            { Read the remaining input symbols }
            data  := copy(elem^.theText,offset+1,left);
            pos   := inpos+1;
            result:= Ts_ok;
            { Advance past data read, i.e. to end of element }
            SetOff(tip.spots[theSpot],len);

            { We could loop here, reading several input elements
              into 'data'. Loop stoppers would be newlines,
              non-text input, output, and end of transcript.
              Perhaps in the future... }

            end {text}
          else
            result:=Ts_nontext;
          end {within element}

        else begin
          { At end of element - must be a newline }
          if not elem^.newline then
            writeln('Ts_Read: end of element without newline!?');
          data  := ' ';
          pos   := inpos+1;
          result:= Ts_ok;
          { Advance to next element. }
          NextElem(theSpot,tip);
          end {newline};
        end {input};
      end {case source};
    end {with spot};
  end {Ts_Reads};


function Ts_Eoln (theSpot: Spot; var tip: Transcript): boolean;
  begin
  with tip.spots[theSpot] do
    Ts_Eoln := (elem^.newline) and (offset=LengthElem(elem));
  end {Ts_Eoln};

function Ts_Eof (theSpot: Spot; var tip: Transcript): boolean;
  begin
  Ts_Eof := Ts_AtEot(theSpot,tip);  { no embedded eof's implemented yet }
  end {Ts_Eof};


procedure Ts_CharAfter (    theSpot: Spot;
                        var tip    : Transcript;
                        var c      : char;
                        var eol    : boolean;
                        var source : Ts_dataSource;
                        var result : Ts_result);
  begin
  with tip.spots[theSpot] do begin
    FlipNext(elem,offset);
    source:=elem^.source;
    if elem^.kind  <> TextElement then result:=Ts_nontext
    else
    if offset >= LengthElem(elem) then begin
      c  :=' ';
      eol:=elem^.newline;
      if eol then result:=Ts_ok
             else result:=Ts_invalid;   { seems like internal error }
      end
    else begin
      c  :=elem^.theText[offset+1];
      eol:=false;
      result:=Ts_ok;
      end;
    end {with};
  end {Ts_CharAfter};


procedure Ts_Write (    theSpot: Spot;
                    var tip    : Transcript;
                        theText: vstring;
                        duepos : integer;
                    var result : Ts_result);
  var
    ende: boolean;  { spot is at the end of its element, incl. newline }
    txti: integer;  { theText[txti..] remains to write }
    len : integer;  { length of text (remaining)       }

  procedure WriteMore;
    { The spot is at the end of an output text element (with no newline)
      to which (part of) the new text can be written.
      Spots are updated. }
    var k: integer;
    begin
    with tip.spots[theSpot] do begin
      k:=MaxElemLength-length(elem^.theText);
      if k > len then k:=len;
      if k > 0 then begin  { some text fits }
        ColKnow(theSpot,tip);
        elem^.theText:=concat(elem^.theText,Copy(theText,txti,k));
        txti:=txti+k;
        len :=len -k;
        UpSpotOff(theSpot, tip,      offset+k       );
        UpSpotPos(theSpot, tip, row, col   +k, inpos);
        end;
      end;
    end {WriteMore};

  procedure WriteBefore;
    { The spot is as the start of an element.
      Write remaining text to new output elements and insert them
      before the spot. Spots are updated. }
    var tot,w: integer;
        writ : ElementRef;
    begin
    if len <= 0 then result:=Ts_ok
    else
      with tip.spots[theSpot] do begin
        tot:=0;
        repeat
          NewElement(writ,TextElement,result);
          if result=Ts_ok then begin
            writ^.source :=OutputSource;
            writ^.newline:=false;
            writ^.safe   :=duepos;
            w:=len;
            if w > MaxElemLength then w:=MaxElemLength;
            writ^.theText:=Copy(theText,txti,w);
            txti:=txti+w;
            len :=len -w;
            tot :=tot +w;
            InsertBefore(elem,writ);
            end;
        until (len<=0) or (result<>Ts_ok);
        ColKnow  (theSpot,tip);
        UpSpotPos(theSpot,tip,row,col+tot,inpos);
        end {with};
    end {WriteBefore};

  begin
  txti:=1;
  len :=length(theText);
  with tip.spots[theSpot] do begin

    if len = 0 then result:=Ts_ok   { so, do nothing }

    else begin

      ende:= (offset = LengthElem(elem)) and (not elem^.newline);

      { Split input element if necessary and possible }
      if  (offset > 0) and (not ende) then
        case elem^.source of
          InputSource : SplitElem(elem,offset,tip,result);
          OutputSource: result:=Ts_atomic;
          end
      else begin
        { At beginning or end of element. }
        { See if there is already an output element we could add to. }
        if  (offset = 0)
        and not elem^.prev^.newline
        then begin
          FlipPrev(elem,offset);
          ende  :=true;
          end;
        if   ende
        and (elem^.source = OutputSource)
        and (elem^.kind   = TextElement )
        and (elem^.safe   = duepos      )
        then WriteMore;
        result:=Ts_ok;
        end;

      if result=Ts_ok then begin
        { At beginning or end of element. }
        FlipNext(elem,offset);
        WriteBefore;
        end;

      tip.modified:= true;

      end {some output};
    end {with spot};                              
  end {Ts_Write};


procedure Ts_Writeln (    theSpot: Spot;
                      var tip    : Transcript;
                          duepos : integer;
                      var result : Ts_result);
  var x: ElementRef;
  begin
  with tip.spots[theSpot] do begin
    result:=Ts_ok;  { unless... }

    FlipNext(elem,offset);

    { Split input element, if required. }
    if offset > 0 then begin
      if elem^.source = InputSource then
        SplitElem(elem,offset,tip,result)
      else
        result:=Ts_atomic;
      end;

    if result = Ts_ok then begin
      { See if newline can be put on prev element. }
      x:=elem^.prev;
      if (x^.source <> OutputSource)
      or (x^.safe   <> duepos)
      or (x^.newline)
      then begin
        { Cannot put newline on prev element - make a new one. }
        NewElement(x,TextElement,result);
        if result=Ts_ok then begin
          x^.source :=OutputSource;
          x^.theText:='';
          x^.safe   :=duepos;
          x^.newline:=false;
          InsertBefore(elem,x);
          end;
        end;
      if result = Ts_ok then begin
        SpotNewline(theSpot,tip,result);
        tip.modified:= true;
        end;
      end;
    end {with spot};  
  end {Ts_Writeln};


procedure Ts_Store (var tip   : Transcript;
                    var f     : text;
                        opt   : Ts_fileOpt;
                    var lines : integer;
                    var result: Ts_result );
  var
    elem : ElementRef;
    pass : boolean;
    tag  : boolean;
    i    : integer;
    c    : char;
  begin
  with tip do begin
    elem :=botElem^.next;
    lines:=0;
    while elem <> eotElem do begin
      with elem^ do begin

        if source = OutputSource then begin
          pass:= opt <> StoreInput;
          tag := opt <> StoreAsText;
          end
        else begin
          pass:= true;
          tag := false;
          end;

        if pass then begin
          if tag then begin
            write(f,'~');
            if safe <> invalid then write(f,safe:1);
            write(f,'[');
            end;
          case kind of
            BotElement   : write(f,'[BOT]');   { should never occur }
            TextElement  : for i:=1 to length(theText) do begin
                             c:=theText[i];
                             write(f,c);
                             if c = '~' then write(f,c);
                             end;
            OtherElement : write(f,'[nontext]');
            EotElement   : write(f,'[EOT]');   { should never occur }
            end;
          if newline then begin
            writeln(f);
            lines:=lines+1;
            end;
          if tag then write(f,'~]');
          end {pass};
        end {with elem};
      elem:=elem^.next;
      end {at eot};
    result:=Ts_ok;
    end;
  end {Ts_Store};


procedure Ts_Load (var theSpot: Spot;
                   var tip    : Transcript;
                   var f      : text;
                   var preRow,preCol: integer;
                   var lines  : integer;
                   var result : Ts_result);
  var
    inorg: integer;       { Input offset to add to output safe tags }
    here : ElementRef;    { Element after load point }
    elem : ElementRef;    { Element being loaded into }
    src  : ElementSource; { Source of elem. OmSource means input text
                            will be discarded. }
    safe : integer;       { Tag for output element }
    n    : integer;       { Chars in elem }
    sym  : (              { What was read from the input file: }
            aChar,        { - a normal character (in "c") }
            aNewline,     { - a line separator }
            anEof);       { - end of file. }
    c    : char;          { Char read from file, if sym = aChar }
    cOp  : (              { What to do: }
            cLoad    ,    { - load sym into transcript  }
            cBeginOut,    { - begin an output element }
            cEndOut  ,    { - end output element }
            cSkip    ,    { - skip this text }
            cError        { - error in input file format }
           );
    row,col,inpos: integer;   { Cumulative position of insertion point }

    procedure GetSym;
      { Get the next symbol from the file.
        Sorry about making this a procedure. }
      begin
      if eof(f) then begin
        sym:=anEof;
        c  :=' ';
        end
      else if eoln(f) then begin
        sym:=aNewline;
        c  :=' ';
        readln(f);
        lines:=lines+1;
        end
      else begin
        sym:=aChar;
        read(f,c);
        end;
      end {GetSym};

    procedure GetSafe;
      { Read an output tag. }
      begin
      safe:=0;
      repeat
        safe:=10*safe + (ord(c) - ord('0'));
        GetSym;
      until not (c in ['0'..'9']);
      safe:=safe + inorg;
      end {GetSafe};

    procedure GetOp;
      { The current symbol is aChar or aNewline.
        Form it into a load operation, possibly reading more input. }
      var axout: boolean;   { accept output, else skip it }
      begin
      axout:= inpos > inorg;  { some input has been loaded. }
      if      sym = aNewline then cOp:=cLoad
      else if  c <> '~'      then cOp:=cLoad
      else begin
        { Tilde escapes: }
        GetSym;
        if sym <> aChar then cOp:=cError
        else 
        if c = '~' then cOp:=cLoad
        else
        if c = '[' then begin
          safe:= invalid;
          if axout then cOp := cBeginOut
                   else cOp := cSkip;
          end
        else
        if c in ['0'..'9'] then begin
          GetSafe;
          safe:=invalid;    { ignore filed tag, just mark invalid }
          if   sym <> aChar then cOp:= cError
          else if c <> '['  then cOp:= cError
          else if axout     then cOp:= cBeginOut
                            else cOp:= cSkip;
          end
        else
        if c = ']' then begin
          if src = InputSource then cOp:= cError
                               else cOp:= cEndOut;
          end
        else
          cOp:=cError;
        end;
      end {GetOp};

    procedure InsElem;
      { Insert the buffer element in the transcript, and update row,col,
        inpos counters. If the element was empty, it is discarded instead. }
      begin
      if (n > 0) or elem^.newline then begin
        InsertBefore(here,elem);
        if elem^.newline then begin
          row:=row+1;
          col:=1;
          end
        else
          col:=col + n;
        if elem^.source = InputSource then begin
          inpos:=inpos + n;
          if elem^.newline then inpos:=inpos+1;
          end;
        elem:=nil;
        tip.modified:= true;
        end
      else
        DisposeElement(elem);
      end {InsElem};

    procedure Load;
      { Load text, starting with the current symbol (which is not anEof),
        until the current element is full, or a '~' marker, end of line,
        or end of file is found.
        If the current symbol is '~', it is loaded (because it was an
        "escaped" marker) and more text is read as usual.
        If necessary, a new element is created before the load starts.
        On exit, the first unloaded symbol is current.
      }
      var done: boolean;
      begin

      { Has source changed? }
      if elem <> nil then
        if src <> elem^.source then InsElem;

      { Create new element? }
      if elem = nil then begin
        NewElement(elem,TextElement,result);
        if result = Ts_ok then begin
          elem^.source := src;
          elem^.newline:= false;  { just to begin with }
          elem^.theText:= '';
          n:=0;
          if src = OutputSource then elem^.safe:= safe;
          end;
        end;

      if result = Ts_ok then begin

        { Load stuff. }
        done:= sym = aNewline;
        while not done do begin
          { Load the current character. }
          with elem^ do theText:=concat(theText,c);
          n:=n+1;
          GetSym;
          if sym = aChar then done:=  (c = '~')
                                   or (n >= MaxElemLength)
                         else done:= true;
          end;

        { Finish this element? }
        if (sym  = aNewline)
        or (sym  = anEof)
        or (n   >= MaxElemLength)
        then begin
          elem^.newline := sym = aNewline;
          InsElem;
          if sym = aNewline then GetSym;
          end;
        end;
      end {Load};

    procedure Skip;
      { Skip text, including the current symbol, until a '~' marker
        or end of file.
        No effect on transcript or anything else. }
      var done: boolean;
      begin
      done:= sym = anEof;
      while not done do begin
        GetSym;
        if sym = aChar then done:= c = '~'
                       else done:= sym = anEof;
        end;
      end {Skip};

    procedure DoOp;
      { Perform a load operation. }
      begin
      case cOp of
        cLoad    : if src <> OmSource then Load
                                      else Skip;
        cBeginOut: begin
                   src:=OutputSource;
                   GetSym;
                   if sym = anEof then result:=Ts_wrongform;
                   end;
        cEndOut  : begin
                   src:=InputSource;
                   GetSym;
                   end;
        cSkip    : begin
                   src:=OmSource;
                   GetSym;
                   if sym = anEof then result:=Ts_wrongform;
                   end;
        cError   : result:=Ts_wrongform;
        end;
      end {DoOp};

  begin {Ts_Load}
  SkipForw (tip.spots[theSpot]);
  PrepDest (theSpot,tip,result);
  Ts_Where (theSpot,tip,preRow,preCol);
  lines := 0;
  if result = Ts_ok then begin
    here  := tip.spots[theSpot].elem;
    inorg := Ts_Pos(theSpot,tip);
    elem  := nil;
    src   := InputSource;
    row   := preRow;
    col   := preCol;
    inpos := inorg;
    GetSym;
    while (sym <> anEof) and (result = Ts_ok) do begin
      GetOp;
      DoOp;
      end {at eof or error};
    if elem <> nil then InsElem;
    UpSpotPos(theSpot,tip,row,col,inpos);
    end;
  end {Ts_Load};


procedure Ts_ShowElements (var tip   : Transcript;
                               row   : integer;
                               rows  : integer;
                               col   : integer;
                           var result: Ts_result);
  var
    spt : Spot;
    r,re: integer;
    elem: ElementRef;
    nl  : boolean;
  begin
  Ts_NewSpot(spt,tip,result);
  if result=Ts_ok then begin
    Ts_Move(spt,row-1,0,spt,tip);
    elem:=tip.spots[spt].elem;
    r:=row; re:=r+rows; nl:=true;
    while r<re do begin
      if nl then begin
        GotoXY(col,r-row+1); ClearEOL;
        write('#',r:1,': ');
        end;
      with elem^ do begin
        case source of
          InputSource : write('i');
          OutputSource: begin
                        write('o');
                        if safe <> invalid then write(safe:1);
                        end;
          OmSource    : ;
          end;
        case kind of
          BotElement   : write('B');
          TextElement  : write('T');
          OtherElement : write('X');
          EotElement   : write('E');
          end;
        write(LengthElem(elem):1);
        if newline then write('/');
        nl:=LastOnLine(elem);
        if nl then r:=r+1
              else write(' ');
        end {with elem};
      if elem=tip.eotElem then r:=re
                          else elem:=elem^.next;
      end {when number of rows shown, or elements end };
    Ts_DelSpot(spt,tip);
    end;
  end {Ts_ShowElements};


procedure Ts_ShowSpots (var tip: Transcript);
  var s: Spot;
      i: integer;
  begin
  with tip do begin
    writeln('hiSpot = ',hiSpot:1);
    for s:=1 to hiSpot do
      if s in spotlist then
        with spots[s] do begin
          write(s:2,'->');
          ShowElement(output,elem,0,LengthElem(elem));
            ClearEOL; writeln;
          write('    '); for i:=1 to offset do write(' ');
          write('^ [',row:1,'.');
          if knowcol then write(col:1)
                     else write('?');
          write('] +',offset:1,' (',inpos:1,'..)');
          ClearEOL; writeln;
          end;
    end;
  end {Ts_ShowSpots};

        
procedure Ts_Check (var f  : text;
                    var tip: Transcript;
                    var ok : boolean);
  const
    maxElem = 200;  { during testing, max number of elements in tip }
  var
    s: Spot;

  procedure CheckXot (elem: ElementRef; name: vstring; nature: ElementKind);
    { Check bot/eot element. }
    begin
    if elem = nil then begin
      writeln(f,name,'=nil');
      ok:=false;
      end
    else with elem^ do begin
      if not newline then begin
        writeln(f,name,'^.newline = false');
        ok:=false;
        end;
      if kind <> nature then begin
        writeln(f,name,'^.kind is wrong');
        ok:=false;
        end;
      if source <> omSource then begin
        writeln(f,name,'^.source <> omSource');
        ok:=false;
        end;
      end;
    end {CheckXot};

  procedure CheckElems;
    { Check that there are no null elements, and that 'prev' links
      are connected properly. }
    var q: SpotOb;
        p,x: ElementRef;
        n: integer;
        safe: integer;
    procedure Bad;
      begin
      write(f,'* Element ',n:1,' [',q.row:1,'.',q.col:1,'], ');
      ok:=false;
      end {Bad};
    begin
    BotSpot(q,tip);
    p:=tip.botElem;
    n:=1;
    safe:=0;
    repeat
      x:=q.elem;
      if x = nil then begin
        Bad; writeln(f,'elem = nil');
        end
      else begin
        if x^.prev <> p then begin
          Bad; writeln(f,'prev is wrong.');
          end;
        if x^.kind = TextElement then
          if (LengthElem(x) = 0)
          and (not x^.newline) then begin
            Bad; writeln(f,' is NULL');
            end;
        if x^.source = OutputSource then begin
          if x^.safe < safe then begin
            Bad; writeln(f,'safe = ',x^.safe:1,
                     ' < prev max safe = ',safe:1);
            end;
          if x^.safe > safe then safe:=x^.safe;
          end;
        p:=x;
        IncElem(q);
        n:=n+1;
        end;
    until (p = tip.eotElem)
       or (x = nil)
       or (n >= maxElem);
    if p <> tip.eotElem then begin
      Bad; writeln(f,'* Transcript has too many elements.');
      end;
    end {CheckElems};
    
  procedure CheckSpot (var p: SpotOb; s: Spot);
    { Check this spot. }
    var
      q: SpotOb;
      n,len: integer;
    procedure Bad;
      begin write(f,'* Spot ',s:1,': '); end;
    begin
    BotSpot(q,tip);
    n:=0;
    while (q.elem <> p.elem)
      and (q.elem <> tip.eotElem)
      and (n      <  maxElem)
    do begin
      IncElem(q);
      n:=n+1;
      end;
    if p.elem <> q.elem then begin
      Bad; writeln(f,'element not found.');
      if q.elem <> tip.eotElem then
        writeln(f,'* Too many elements in transcript.');
      BotSpot(p,tip);
      ok:=false;
      end
    else begin
      len:=LengthElem(p.elem);
      if p.offset < 0 then begin
        Bad; writeln(f,'offset=',p.offset:1,' < 0');
        p.offset:=0;
        ok:=false;
        end
      else
      if p.offset > len then begin
        Bad; writeln(f,'offset=',p.offset:1,
                    ' > element length=',len:1);
        p.offset:=len;
        ok:=false;
        end;
      SetOff(q,p.offset);
      if p.row <> q.row then begin
        Bad; writeln(f,'row=',p.row:1,' <> ',q.row:1);
        p.row:=q.row;
        ok:=false;
        end;
      if p.col <> q.col then begin
        Bad; writeln(f,'col=',p.col:1,' <> ',q.col:1);
        p.col:=q.col;
        end;
      if p.inpos <> q.inpos then begin
        Bad; writeln(f,'inpos=',p.inpos:1,' <> ',q.inpos:1);
        p.inpos:=q.inpos;
        ok:=false;
        end;
      end {elem found};
    end {CheckSpot};

  begin {Ts_CheckSpots}
  ok:=true;
  with tip do begin
    CheckXot(botElem,'* botElem',BotElement);
    if botElem^.prev <> nil then begin
      writeln(f,'* botElem^.prev <> nil');
      ok:=false;
      end;
    CheckXot(eotElem,'* eotElem',EotElement);
    if eotElem^.next <> nil then begin
      writeln(f,'* eotElem^.next <> nil');
      ok:=false;
      end;
    CheckElems;
    if hiSpot < 0  then begin
      writeln(f,'* hiSpot=',hiSpot:1,' < 0');
      hiSpot:=0;
      ok:=false;
      end
    else
    if hiSpot > MaxSpots then begin
      writeln(f,'* hiSpot=',hiSpot:1,' > Max');
      hiSpot:=MaxSpots;
      ok:=false;
      end
    else
    if not (hiSpot in spotlist) then begin
      writeln(f,'* hiSpot=',hiSpot:1,' not in spotlist');
      ok:=false;
      end;
    for s:=1 to hiSpot do
      if s in spotlist then CheckSpot(spots[s],s);
    end;
  if not ok then writeln(f,'* end Ts_Check');
  end {Ts_Check};


procedure Ts_Dump (var f: text; var tip: Transcript);
  var
    spt   : Spot;
    r,re  : integer;
    elem  : ElementRef;
    nl    : boolean;
    result: Ts_result;
    s     : Spot;
    i     : integer;
  begin
  Ts_NewSpot(spt,tip,result);
  if result<>Ts_ok then
    writeln(f,'Ts_Dump: could not make new spot.')
  else
  with tip do begin
    elem:=spots[spt].elem;
    r:=1; re:=r+1000; nl:=true;
    while r<re do begin
      if nl then begin
        writeln(f);
        write(f,'#',r:1,': ');
        end;
      with elem^ do begin
        case source of
          InputSource : write(f,'i');
          OutputSource: begin
                        write(f,'o');
                        if safe <> invalid then write(f,safe:1);
                        end;
          OmSource    : ;
          end;
        case kind of
          BotElement   : write(f,'B');
          TextElement  : write(f,'T');
          OtherElement : write(f,'X');
          EotElement   : write(f,'E');
          end;
        write(f,LengthElem(elem):1);
        if newline then write(f,'/');
        nl:=LastOnLine(elem);
        if nl then r:=r+1
              else write(f,' ');
        end {with elem};
      if elem=eotElem then r:=re
                      else elem:=elem^.next;
      end {when number of rows shown, or elements end };
    Ts_DelSpot(spt,tip);
    writeln(f);
    end;

  with tip do begin
    writeln(f,'hiSpot = ',hiSpot:1);
    for s:=1 to hiSpot do
      if s in spotlist then
        with spots[s] do begin
          write(f,s:2,'->');
          ShowElement(f,elem,0,LengthElem(elem)); writeln(f);
          write(f,'    '); for i:=1 to offset do write(f,' ');
          write(f,'^ [',row:1,'.');
          if knowcol then write(f,col:1)
                     else write(f,'?');
          write(f,'] +',offset:1,' (',inpos:1,'..)');
          writeln(f);
          end;
    end;
  end {Ts_Dump};
        

procedure Ts_Create (var tip   : Transcript;
                     var result: Ts_result);
  begin
  with tip do begin
    spotlist:=[];
    hiSpot  :=0;
    NewElement(botElem,BotElement,result);
    if result=Ts_ok then begin
      NewElement(eotElem,EotElement,result);
      if result<>Ts_ok then 
        DisposeElement(botElem)
      else begin
        with botElem^ do begin
          source :=OmSource;
          newline:=true;
          prev   :=nil;
          end;
        with eotElem^ do begin
          source :=OmSource;
          newline:=true;
          next   :=nil;
          end;
        Link(botElem,eotElem);
        end;
      end;
    modified:= false;
    end {with};
  end {Ts_Create};


function Ts_Modified (var {in} tip: Transcript): boolean;
  begin
  Ts_Modified:= tip.modified;
  end {Ts_Modified};


procedure Ts_Original (var {io} tip: Transcript);
  begin
  tip.modified:= false;
  end {Ts_Original};


procedure Ts_Init;
  begin
  { writeln('Transcripts 91.04.19.A'); }
  Ts_SetMemSize(maxint-50);
  { writeln('Ts initialised.'); }
  end {Ts_Init};

procedure Ts_Fini;
  begin
  end {Ts_Fini};
