{ transcripts_f.p
  External declarations for routines in transcripts.p
  Requires: strings_d.p
}

procedure Ts_Init;
  { Initialize module. }
  external;

procedure Ts_Fini;
  { Finalize module. }
  external;

procedure Ts_Explain (var f: text; result: Ts_result);
  { Write an explanation of the result. }
  external;

procedure Ts_Create (var tip   : Transcript;
                     var result: Ts_result);
  { Initialise a transcript to be empty. The result can be
    ok or memory full. }
  external;

function  Ts_MoreSpots (var tip: Transcript): integer;
  { At least this many new spots can still be defined in the transcript. }
  external;

procedure Ts_NewSpot (var theSpot: Spot;
                      var tip    : Transcript;
                      var result : Ts_result);
  { Initialise a spot variable to refer to the start of the transcript,
    before the first element. }
  external;

procedure Ts_DelSpot (var theSpot: Spot;
                      var tip    : Transcript);
  { Delete a spot. To be done for each spot initialised, before the
    variable is deleted (e.g. by block exit). }
  external;

procedure Ts_Where (    theSpot: Spot;
                    var tip    : Transcript;
                    var row,col: integer);
  { Returns row and column number of the character following the
    given spot. Both row and col start from 1. }
  external;

function  Ts_Pos (theSpot: Spot; var tip: Transcript): integer;
  { Returns the number of input symbols that precede the given spot.
    Input symbols are input characters and input newlines. }
  external;

function Ts_Order (var spot1,spot2: Spot;
                   var tip        : Transcript): integer;
  { Determines order of two spots in a transcript.
    Returns sign(spot1-spot2), i.e. negative if spot1 precedes
    spot2 in the transcript. }
  external;

function Ts_AtEol (theSpot: Spot; var tip: Transcript): boolean;
  { Whether the spot is at the right end of its line. Note, there is
    not necessarily a newline at the end of the last transcript line,
    so Ts_AtEol can yield true although Ts_Eoln is false. }
  external;

function Ts_AtEot (theSpot: Spot; var tip: Transcript): boolean;
  { Whether the spot is just before the end of the transcript. }
  external;

function Ts_LineLength (theSpot: Spot; var tip: Transcript): integer;
  { The length, in screen columns, of the line on which the spot
    lies. For the last transcript line, works as if there were
    always a newline at its end. }
  external;

procedure Ts_Show (    theSpot      : Spot;  var tip: Transcript;
                       fromCol,toCol: integer;
                   var numCol       : integer);
  { Write the line holding this spot of the transcript. This may be
    an output line, an input line, or a mixed input/output line.
    The output is limited to the column range fromCol..toCol,
    counting the first character as column 1.
    If the line is short, the last column actually written may be
    less than toCol.The number of columns (characters) written is
    returned in numCol. }
  external;

procedure Ts_Move (    fromSpot : Spot;
                       rows,cols: integer;
                   var movedSpot: Spot;
                   var tip      : Transcript);
  { Move from a spot by the specified number of rows (up: -, down: +)
    and columns (left: -, right: +). Motion by rows stops at first line
    and last line. Motion by columns stops at left or right margin.
    Note that the end of the transcript is considered to be on a line
    of its own if and only if the transcript proper ends in a newline.
    The new position is returned in movedSpot, which must have been
    initialised by Ts_NewSpot (movedSpot can be the same variable as
    fromSpot).
    Input and output lines and columns are both counted equally. }
  external;

procedure Ts_Goto (var theSpot: Spot;
                   var tip    : Transcript;
                       row,col: integer);
  { Move a spot to the specified row and column position.
    Motion by rows stops at first line and last line.
    Motion by columns stops at left or right margin.
    The spot must have been initialised.
    Input and output lines and columns are both counted equally. }
  external;

procedure Ts_Skipln (var theSpot: Spot;
                     var tip    : Transcript;
                     var result : Ts_result);
  { Move spot forward over the next input newline. If there is no
    input newline between the spot and eot, leave the spot at eot
    and return Ts_nonewline. }
  external;

procedure Ts_GotoEnd (var theSpot: Spot;
                      var tip    : Transcript);
  { Move the spot to the end of the transcript, just after the
    last input or output symbol.
    The spot must have been initialised. }
  external;

procedure Ts_Insert (var theSpot      : Spot;
                     var tip          : Transcript;
                         ch           : char;
                     var preRow,preCol: integer;
                     var result       : Ts_result);
  { Insert (input) character ch at the spot position, pushing the
    remaining text on the line right. The spot moves one step right, to
    follow the inserted character.
    If the spot is not in an input region, it is moved forward to before
    the next input symbol.
    The parameters preRow,preCol return the row and column the spot
    had before the character was inserted, but after the spot was moved
    forward (these values are useful for screen updating).  }
  external;

procedure Ts_Newline (var theSpot: Spot;
                      var tip    : Transcript;
                      var preRow,preCol: integer;
                      var result : Ts_result);
  { Insert a newline at the spot. The line holding the spot breaks into
    two pieces, with the spot at the start of the later piece.
    The later piece is moved past output that resulted from the
    earlier piece (perhaps splitting an output line).
    If the spot is not in an input region, it is moved forward to before
    the next input symbol.
    The parameters preRow,preCol return the row and column the spot
    had before the newline was inserted, but after the spot was moved
    forward (these values are useful for screen updating). }
  external;

procedure Ts_Delete (var theSpot      : Spot;
                     var tip          : Transcript;
                     var preRow,preCol: integer;
                     var result       : Ts_result);
  { Delete input character preceding the spot. The spot does not have
    to be placed immediately after the input.
    After the character is deleted, the line to the right of the
    spot is moved left to fill the gap, The spot is left in the position
    of the deleted character. No output is deleted (here).
    The parameters preRow,preCol return the row and column the spot
    had before the character was deleted (these values are useful for
    screen updating).
    If the input symbol preceding the spot is a newline, the operation
    is rejected with result = Ts_wasnewline. Ts_Joinline must be used
    to delete newlines. }
  external;

procedure Ts_Joinline (var theSpot      : Spot;
                       var tip          : Transcript;
                       var delRow,delCol: integer;
                       var preRow,preCol: integer;
                       var result       : Ts_result);
  { Delete input newline preceding the spot. The spot does not have to
    be placed immediately after the newline.
    After the newline is deleted, the input line to the right
    of the spot is joined (appended) to the input line to the left of
    the spot.
    Any embedded output in the right line is deleted (it is not necessary
    to "unwrite" it first). Output following the deleted newline, but due
    to input preceding it, is preserved, and will follow the joined line.
    If there is such saved output, but no input newline to the right of
    the spot,  one is added to the joined line anyway (effectively, the
    newline preceding the spot is moved to the end of the transcript input).
    The spot is left at the position of the join.
    The transcript is thus changed in two, possibly disjoint places.
    The first place is the joined line, which is indicated by the spot.
    The second place is the region of the right-hand line, where
    both input and (embedded) output data were removed.
    The parameters delRow,delCol return the start of the removed text
    region, and preRow,preCol return the place after the region. }
  external;

procedure Ts_Cut  (var spot1,spot2   : Spot;
                   var fromTranscript: Transcript;
                   var toSpot        : Spot;
                   var toTranscript  : Transcript;
                   var preRow,preCol : integer;
                   var result : Ts_result);
  { Cuts the part between spot1 and spot2 from fromTranscript.
    All of spot1, spot2, and toSpot must lie within or adjacent to
    input data, and spot1 must be <= spot2.
    All input and output between spot1 and spot2 is cut. 
    The data cut from fromTranscript are inserted before toSpot
    in toTranscript. Input remains input and output remains output.
    If the cut is successful, spot1 and spot2 are both placed in
    the space where the text was cut.
    The parameters preRow,preCol return the row and column spot2
    had after it was moved past output to be unwritten, but before
    any text (input or output) was deleted (these values are useful
    for screen updating). }
  external;

procedure Ts_Copy (var spot1,spot2   : Spot;
                   var fromTranscript: Transcript;
                   var toSpot        : Spot;
                   var toTranscript  : Transcript;
                       asInput       : boolean;
                   var result        : Ts_result);
  { Copies the part between spot1 and spot2 from fromTranscript,
    both input and output. spot1 must precede spot2.
    fromTranscript is not changed, but spot1 is moved forward to
    coincide with spot2.
    The data copied from fromTranscript are inserted before
    toSpot in toTranscript. This position must be within or adjacent
    to input data. If 'asinput' is true, output data are
    converted to input data when copied. }
  external;

procedure Ts_Erase (var tip: Transcript);
  { Erase all of transcript. Any spots in the transcript are moved
    to the start (= the end) of the transcript. }
  external;

procedure Ts_Skip (var theSpot: Spot;
                   var tip    : Transcript;
                   var result : Ts_result);
  { Skip output until an input symbol  (or end of
    transcript) is found. The new position is immediately
    before the first input symbol after theSpot (or at end of
    transcript, in which case result is Ts_endinput). }
  external;

function  Ts_AtOutput (var theSpot: Spot;
                       var tip    : Transcript): boolean;
   { True iff the spot is embedded within output, i.e. has output
     both to the left and to the right.
     This function is useful for the context-dependent choice of
     what a key (e.g. the Enter key) does. }
  external;

procedure Ts_Seek (    fromSpot : Spot;
                       beforePos: integer;
                   var movedSpot: Spot;
                   var tip      : Transcript;
                   var result   : Ts_result);
  { Move from a spot to a new position immediately before the input
    symbol number beforePos, counting the first input symbol as 1. Note
    that beforePos is absolute, not relative to fromSpot; fromSpot
    is expected to lie close to the sought position and be a good place
    to start seeking.
    The new position is returned in movedSpot, which must have been
    initialised by Ts_NewSpot (movedSpot can be the same variable as
    fromSpot). After a successful move, Ts_Pos(movedSpot) = beforePos-1.
    Input symbols are input characters and input newlines.
    Output elements are not counted.
    The seek may fail if beforePos is nonpositive or too large. 
    If beforePos is nonpositive, the movedSpot is placed at fromSpot and
    result is Ts_invalid. If beforePos was too large, the movedSpot is
    defined at end of transcript and result is Ts_endinput. }
  external;

procedure Ts_Read (var theSpot: Spot;
                   var tip    : Transcript;
                   var ch     : char;
                   var pos    : integer;
                   var result : Ts_result);
  { Reads the character to the right of the spot and places the spot
    just after the symbol read. The parameter pos returns the number
    of the input symbol read, counting the first symbol as 1.
    Reading a line separator (where Ts_Eoln is true) returns a blank,
    and the spot is moved to the start of the next (input) line.
    If there is no input between the given spot and the end of the
    transcript, a result of Ts_endinput is given, ch and pos are
    undefined, and the spot is set at end of transcript. }
  external;

procedure Ts_Reads (var theSpot: Spot;
                    var tip    : Transcript;
                    var data   : vstring;
                    var pos    : integer;
                    var result : Ts_result);
  { Reads input symbols to the right of the spot and places the spot
    just after the last symbol read. The symbols are returned in the 
    parameter data. The parameter pos returns the number of the first
    input symbol read, counting the first symbol in the transcript as 1.
    The number of symbols read and returned depends on the structure
    of the transcript, but reading stops at the next line separator,
    if not before.
    If the first symbol read is a line separator (where Ts_Eoln is true)
    then a blank character is returned and the spot is moved to the
    start of the next (input) line.
    If there is no input between the given spot and the end of the
    transcript, a result of Ts_endinput is given, data and pos are
    undefined, and the spot is set at end of transcript. }
  external;

procedure Ts_CharAfter (    theSpot: Spot;
                        var tip    : Transcript;
                        var c      : char;
                        var eol    : boolean;
                        var source : Ts_dataSource;
                        var result : Ts_result);
  { Returns character immediately after (to the right of) the spot.
    The character is in c, unless it was a line separator in which
    case c is blank and eol is true.
    The source of the character (or line separator) is also returned.
    An error results if the spot is at the end of the transcript. }
  external;

function Ts_Eoln (theSpot: Spot; var tip: Transcript): boolean;
  { Whether the spot is just before the end of its line. }
  external;

function Ts_Eof  (theSpot: Spot; var tip: Transcript): boolean;
  { Whether the spot is just before the end of an input file
    within the transcript. }
  external;

procedure Ts_Write (    theSpot: Spot;
                    var tip    : Transcript;
                        theText: vstring;
                        duepos : integer;
                    var result : Ts_result);
  { Insert output text in the transcript at the given spot.
    The rest of the text to the right of the spot, is moved right
    to make room. The spot follows it.
    The output is marked as "due to" or "caused by" the input symbol
    number duepos.
    Writing a null string has no effect on the transcript or the spot
    position, but may be rejected (result other than Ts_ok) if it would
    be illegal to write a non-null string at the given spot. }
  external;

procedure Ts_Writeln (    theSpot: Spot;
                      var tip    : Transcript;
                          duepos : integer;
                      var result : Ts_result);
  { Insert an output-newline in the transcript at the given spot.
    The text to the right of the spot, and the spot itself, are thus
    moved to the start of a new line.
    The output is marked as "due to" or "caused by" the input symbol
    number duepos. }
  external;

procedure Ts_Unwrite (var theSpot: Spot;
                      var tip    : Transcript;
                          safePos: integer);
  { Moves the spot back to just after the last output resulting from the
    input symbols up to and including symbol number safePos (output with
    safe <= safePos).
    On the way, all output resulting from input symbols > safePos is
    marked as invalid, which means it will be deleted when that part of the
    input is reprocessed. Output after the given spot is not considered.
    If no output with safe <= safePos exists, the spot moves to beginning of
    the transcript. }
  external;

procedure Ts_Dewrite (var theSpot: Spot;
                      var tip    : Transcript;
                      var preRow,preCol: integer;
                      var result : Ts_result);
  { Deletes all output that immediately follows the given spot.
    If the given spot is within input text, the nothing is done.
    done. Otherwise, any output that immediately follows the
    spot is deleted, up to the next input symbol or eot.
    The output must have been marked invalid, otherwise the result
    code is returned as Ts_validoutput.
    Returns with spot before the next input symbol or at eot.
    Returns, in preRow and preCol, the row and column of the new
    position of the spot (e.g. the input symbol after the output)
    before the output was deleted (these values are useful for
    screen updating).
  }
  external;

procedure Ts_Store (var tip   : Transcript;
                    var f     : text;
                        opt   : Ts_fileOpt;
                    var lines : integer;
                    var result: Ts_result );
  { Store transcript in a text file, with optional handling of output.
    The file is assumed to be open. It is only written to, and not closed.
    The positions of the spots are not stored in the file.
    The character sequence '~[' marks the start and '~]' the end of
    an output element. If the output element is tagged with a 'safe'
    position, the position is stored as a decimal integer between the
    '~' and the '['.
    If the character '~' occurs in the text, it is stored as '~~'.
    The total number of lines written is also returned. }
  external;

procedure Ts_Load (var theSpot: Spot;
                   var tip    : Transcript;
                   var f      : text;
                   var preRow,preCol: integer;
                   var lines  : integer;
                   var result : Ts_result);
  { Load transcript from text file, inserting its text at the given spot.
    If the spot is not in an input region, it is moved forward to before
    the next input symbol.
    The file is assumed to be open, and is read until end of file (or a
    format error). The file is not closed.
    If the file does not contain tags that separate input from output, all
    the text is considered input. If there are tags in the file, input and
    output are loaded as input and output respectively. However, leading
    output (output preceding the first input) is discarded, and other
    output is marked invalid, so it will be removed when the input
    is processed by the application.
    The parameters preRow,preCol return the row and column the spot
    had before the character was inserted, but after the spot was moved
    forward (these values are useful for screen updating).
    The total number of lines read from the file is returned, as well
    as a result code. }
  external;

function Ts_Modified (var {in} tip: Transcript): boolean;
  { Reports whether the content of the transcript has been modified
    after the transcript was created or since the last defined "original"
    state (see Ts_Original). Both input and output changes are counted,
    but not changes in the set of spots or the positions of spots. }
  external;

procedure Ts_Original (var {io} tip: Transcript);
  { Declares that the content of the transcript is now in an "original"
    state. Ts_Modified will return false until there are further
    changes in the content of the transcript. }
  external;


{* Testing routines *}

procedure Ts_ShowElements (var tip   : Transcript;
                               row   : integer;
                               rows  : integer;
                               col   : integer;
                           var result: Ts_result);
  { Displays element structure, starting from the given row and
    continuing for the given number of rows.
    Display is on screen, one line per row, starting from the given
    column. }
  external;

procedure Ts_ShowSpots (var tip: Transcript);
  { Displays the positions of the currently defined spots,
    including the spot number, text of element, offset within
    element, and row and column numbers. }
  external;

procedure Ts_Check (var f  : text;
                    var tip: Transcript;
                    var ok : boolean);
  { Check that the transcript and the defined spots are consistent.
    Report problems on file f. }
  external;

procedure Ts_Dump (var f  : text;
                   var tip: Transcript);
  { List element structure and defined spots on the output file f. }
  external;

procedure Ts_SetMemSize (numberElems: integer);
  { For testing memory overflow situations - memory will be filled
    when numberElems elements have been allocated. }
  external;

