{ views.p

  A view is a combination (screen window, transcript, cursor pos).

  Author: Niklas Holsti around 1994.8.2

Changes:
  94.10.01 Titlebar shows (View) for unmodifiable files.
  94.10.07 Titlebar shows '*' for modified files.
  94.10.23 Add origname handling to transcript creation and Vw_Original.
  94.10.24 Include Seed version info from version.p.
}

module Views;

#include "version.p"
#include "strings_d.p"
#include "log_f.p"
#include "screen_d.p"
#include "windows_d.p"
#include "transcripts_d.p"
#include "keymap_d.p"
#include "editfuns.p"

#include "render_f.p"
#include "windows_f.p"
#include "screen_f.p"
#include "transcripts_f.p"
#include "keymap_f.p"

#include "edit_d.p"
#include "views_d.p"

public var  { NOTE -- bad to have public vars, change! }

  edwin    : array [WinIndex] of EditWindow;


private var

  { Screen space }
  TotalCols: integer;      { Screen width  }

  allrows   : Sc_rowRange;  { Total screen area. }
  datarows  : Sc_rowRange;  { The screen area for data windows, excluding }
                            { command line and message line.         }
  queryrows : Sc_rowRange;  { The screen area for user queries (file names
                              etc.) - consists of the command line and
                              the message line. }
  statusrows: Sc_rowRange;  { Rows used for status displays. May overlap
                              other areas. }

  { Window set }
  { edwin    : array [WinIndex] of EditWindow; } { is public, for now }
  numwin   : integer;    { edwin[1..numwin] are the existing windows, in
                           order from screen top to screen bottom. }
  

  { Transcript set }
  etipList : EtipRef;    { Linked list of all transcripts held in editor. }


  { Query transcript }
  queryEtip: EtipRef;      { If not nil, the query area is in use. }


  { Message line }
  messPlace: ScreenPlace;  { It's place on the screen }
  messed   : boolean;      { There is something on the message line }


  { Scratch "result" flags }
  tsresult : Ts_result;
  scresult : Sc_result;
  wdresult : Wd_result;

  tracing  : boolean;      { log various operations }

{ Module option control }

procedure Vw_Trace (sw: boolean);
  { Turn tracing on or off. }
  begin tracing:= sw; end;


procedure Vw_Fini;
  begin
  Ts_Fini;
  Wd_Fini;
  Sc_Fini;
  Rn_Fini;
  Km_Fini;
  Log_Fini;
  end {Vw_Fini};


{ Status line and similar output routines. }

procedure Vw_WriteChr (c: char);
  { Display (render) a character for status line or other such. }
  begin
  if Rn_WriteC(c,nil) <> 1 then;
  end {WriteChr};

procedure Vw_WriteStr (s: vstring);
  { Display (render) a string for status line or other such. }
  var ac: integer;
  begin
  Rn_TextVar(s,1,length(s),length(s),ac);
  end {WriteStr};

procedure Vw_DecimalString (n: integer; w: integer; var s: vstring);
  { Format n as decimal integer. Fill with leading blanks to at least
    w characters. Result in s.
    Note: implementation is not efficient. }
  var a: integer;
  begin
  s:='';
  a:=abs(n);
  repeat
    s:= chr(ord('0')+(a mod 10)) + s;
    a:=               a div 10;
  until a = 0;
  if n < 0 then s:= '-' + s;
  while length(s) < w do s:=' ' + s;
  end {DecimalString};
  
procedure Vw_ScProblem (result: Sc_result);
  { Show error result from Screen module. }
  begin
  if result <> Sc_ok then begin
    Sc_OverstrikeMode; Rn_Flush;
    Sc_Explain(errout,result);
    Log_Open;
    write(logfile,'Screen: ');
      Sc_Explain(logfile,result);
      writeln(logfile);
    end;
  end {ScProblem};

private
procedure GoRC (r,c: integer);
  { Go to given row,column. }
  begin
  Sc_Goto(r,c,scresult); Vw_ScProblem(scresult);
  end {GoRC};

private
procedure GoMess;
  { Move screen cursor to message position, and note
    that a message is (will be) displayed. }
  begin
  with messPlace do GoRC(row,col);
  messed:=true;
  Sc_Limbo;
  Sc_OverstrikeMode;
  end {GoMess};

private
procedure Beep;
  begin
  Vw_WriteChr(chr(7));
  end {Beep};


procedure Vw_StatusPlace (num, col: integer; var place: ScreenPlace);
  { Define a place on a screen status line.
    Status line number num is = 0, 1, 2, .. relative to the first
    line reserved for status.
    Column position is relative to left margin, col = 1, 2, 3.
  }
  begin
  place.row:= statusrows.top + num;
  place.col:= col;
  end {Vw_StatusPlace};


procedure Vw_TsProblem (result: Ts_result);
  { Show error result from Transcript module. }
  begin
  if result <> Ts_ok then begin
    GoMess; Beep; Rn_Flush;
    Ts_Explain(errout,result);
    Log_Open;
    write(logfile,'Transcripts: ');
      Ts_Explain(logfile,result);
      writeln(logfile);
    end;
  end {TsProblem};

procedure Vw_WdProblem (result: Wd_result);
  { Show error result from Windows module. }
  begin
  if result <> Wd_ok then begin
    GoMess; Beep; Rn_Flush;
    Wd_Explain(errout,result);
    Log_Open;
    write(logfile,'Windows: ');
      Wd_Explain(logfile,result);
      writeln(logfile);
    end;
  end {WdProblem};

procedure Vw_KmProblem (result: Km_result);
  { Show error result from Keymap module. }
  begin
  if result <> Km_ok then begin
    GoMess; Beep; Rn_Flush;
    Km_Explain(errout,result);
    Log_Open;
    write(logfile,'Keymap: ');
      Km_Explain(logfile,result);
      writeln(logfile);
    end;
  end {KmProblem};

procedure Vw_Message (txt: vstring);
  { Display message. }
  begin
  GoMess;
  Vw_WriteStr(txt);
  Rn_Flush;
  end {Message};

procedure Vw_ClearMess;
  { Clear message line and note absence of message.
    Nothing done if there was no message to start with. }
  begin
  if messed then begin
    GoMess;
    Sc_ClearEOL;
    messed:=false;
    end;
  end {ClearMess};


function Vw_MessPause (txt: vstring; w: WinIndex): boolean;
  { Message and pause.
    User can dump the indicated window, increase/decrease log levels,
    and ask program to stop.
    Returns "true" if execution is to continue.
  }
  var ch: char;
  begin
  Log_Open;
  if Log_includes(pauselog) then writeln(logfile,'Views Pause ',txt);
  Vw_Message(txt);
  repeat
    flush(errout);
    Rn_Flush;
    flush(logfile);
    ch:=ReadChar;
    if      ch = '+' then Log_level(+1)
    else if ch = '-' then Log_level(-1)
    else if ch = 'w' then Wd_ShowWin(logfile,edwin[w].theWindow)
    else if ch = 't' then Ts_Dump(logfile,
                                  edwin[w].refEtip^.theTranscript)
    else if ch = '!' then { program will stop }
    else if ch = ' ' then { program will continue }
    else Beep;
  until (ch = ' ') or (ch = '!');
  Vw_ClearMess;
  Vw_MessPause:= ch = ' ';
  end {Vw_MessPause};

procedure exit(c: integer);
  external c;

private
procedure HaltProgram;
  { Cruel abortion. }
  begin
  Log_close('Module Views halts program.');
  Vw_Fini;
  flush(errout);
  exit(1);
  end {HaltProgram};

private
procedure MessPause (txt: vstring; w: WinIndex);
  { Message and pause, handle stop request, too. }
  begin
  if not Vw_MessPause(txt, w) then HaltProgram;
  end {MessPause};


{ View list handling }


private
function FindView (etr: EtipRef; views: ViewRef): ViewRef;
  { Look for a view of etr in a list of views.
    Return ref to view if found, else nil.
    List is not changed.
  }
  var
     v, vs: ViewRef;
  begin
  v := nil;
  vs:= views;
  while (v = nil) and (vs <> nil) do
    if vs^.refEtip = etr then v := vs
                         else vs:= vs^.nextView;
  FindView:= v;
  end {FindView};


private
procedure AppendView (v: ViewRef; var views: ViewRef);
  { Append a view (or list of views) to a list of views. }
  var
    vs: ViewRef;
  begin
  if views = nil then
    views:= v
  else begin
    vs:= views;
    while vs^.nextView <> nil do vs:= vs^.nextView;
    vs^.nextView:= v;
    end;
  end {AppendView};


private
procedure DeleteView (v: ViewRef; var views: ViewRef);
  { Delete a view from a list of views.
    The view is not disposed, it is just unlinked from the
    list. To discard it, do dispose(v) afterwards.
    The view must not be nil; the list may be nil.
    An unspecific message is given if the view is not found.
  }
  var
    vs   : ViewRef;
    found: boolean;
  begin
  if v = views then begin
    views      := views^.nextView;
    found      := true;
    end
  else
    begin
    vs:= views; found:= false;
    while (vs <> nil) and (not found) do
      if vs^.nextView = v then begin
        vs^.nextView:= v^.nextView;  { = vs^.nextView^.nextView }
        found       := true;
        end
      else
        vs:= vs^.nextView; 
    end;

  if found then
    v^.nextView:= nil
  else
    Message('DeleteView: view not found in list.');

  end {DeleteView};

private
procedure DisposeViews (var views: ViewRef);
  { Discard a list of views.
    Also discards the cursor spot of each view.
    Returns views = nil.
  }
  var rest: ViewRef;
  begin
  while views <> nil do begin
    with views^ do begin
      rest:= nextView;
      Ts_DelSpot(cursor, refEtip^.theTranscript);
      end;
    dispose(views);
    views:= rest;
    end;
  end {DisposeViews};


function Vw_InfiniteIndex: ScriptIndex;
  begin
  Vw_InfiniteIndex:=maxint-1000;
  end {Vw_InfiniteIndex};



procedure Vw_NewEditableTranscript (    aname  : vstring;
                                        amode  : Ed_ExecMode;
                                        modify : boolean;
                                    var theEtip: EtipRef);
  { Creates a new editable transcript. The transcript name (window title)
    and initial execution mode must be specified, also whether the
    transcript is modifiable by the user.
    The transcript is initially empty and has a null filename.
    The new transcript is added to the set of transcripts maintained
    by the editor.
    The transcript is not yet displayed in any windows.
  }
  var
    tsres: Ts_result;
    etl  : EtipRef;
  begin
  new(theEtip);
  if theEtip <> nil then begin
    with theEtip^ do begin
      theEtip^.name:= aname;
      origname   := '';
      filename   := '';
      Ts_Create(theTranscript,tsres);
      if tsres = Ts_ok then
        Ts_NewSpot(theMark,theTranscript,tsres);
      if tsres = Ts_ok then
        Ts_NewSpot(readpt,theTranscript,tsres);
      if tsres = Ts_ok then
        Ts_NewSpot(writept,theTranscript,tsres);
      Vw_TsProblem(tsres);
      modifiable := modify;
      modified   := false;
      writefd    := -1;
      outputfd   := -1;
      marked     := false;
      startPos   :=  1;
      lastPos    :=  0;
      execPos    := -1;
      editPos    := Vw_InfiniteIndex;
      errorPos   := Vw_InfiniteIndex;
      errorFirst := 0;    { but N/A if errorPos = inf }
      errorLength:= 0;    { ditto }
      usedPos    := Vw_InfiniteIndex;
      skipwrite  := false;
      aborting   := false;
      modeTop    := 1;
      with modeStack[modeTop] do begin
        mode    := amode;
        startPos:= 0;
        end;
      nextEtip  := nil;
      end {with};

    { Append the new one to the list of transcripts. }
    if etipList = nil then
      etipList:= theEtip
    else begin
      etl:= etipList;
      while etl^.nextEtip <> nil do etl:= etl^.nextEtip;
      etl^.nextEtip:= theEtip;
      end;

    end {new ok};

  end {NewEditableTranscript};


procedure Vw_KillEditableTranscript (var theEtip: EtipRef);
  { Destroys an editable transcript.
    Removes it from the the list of transcripts first.
    Assumes that no window is actively viewing the
    transcript (otherwise, dangling refs will result.)
    Removes all stored views of this transcript from all windows.
  }
  var
    p,q  : EtipRef;
    w    : WinIndex;
    v    : ViewRef;
  begin

  { First search for and destroy stored views. }
  for w:= 1 to numwin do
    with edwin[w] do begin
      if refEtip = theEtip then
        Message('KillEditableTranscript: the transcript is being viewed!');
        { -- should take some action here, such as closing the window or
          -- not deleting the transcript, after all. }
      v:= FindView (theEtip, views);
      if v <> nil then begin
        DeleteView (v, views);
        DisposeViews (v);
        end;
      end;
    
  { Second, delete it from the list. }
  if etipList = theEtip then begin
    etipList:= etipList^.nextEtip;
    end
  else begin
    p:=etipList;
    repeat
      q:= p^.nextEtip;
      if q = theEtip then begin
        p^.nextEtip:= q^.nextEtip;
        q^.nextEtip:= nil;   { just to be super neat }
        p:= nil;             { break from loop }
        end
      else
        p:= q;
    until p = nil;
    end;

  { Next, destroy it's parts. }
  with theEtip^ do begin
    Ts_Erase(theTranscript);  { SHOULD perhaps delete spots separately. }
                              { Need a Ts_Kill function. }
    end;

  { Finally, delete the thing itself. }
  dispose(theEtip);
  theEtip:= nil;
  end {KillEditableTranscript};


{ Window operations }


function Vw_NumWindows: integer;
  { Returns number of windows open at present.
   The window numbers are 1..NumWindows.
  }
  begin Vw_NumWindows:= numwin; end;


procedure Vw_WriteTitle (w: WinIndex);
  { Write title bar for this window. }

  var
    c, m: char;
    left: integer;     { Room left on line. }

  procedure Wrif (s: vstring);
    { Write as much as there is room for. }
    var used: integer;
    begin
    if left > 0 then begin
      Rn_Text (s, left, used);
      left:= left-used;
      end;
    end {Wrif};

  begin
  if edwin[w].title <> NoTitle then begin
    with edwin[w], refEtip^ do begin
      if title = BoldTitle then c:='=' else c:='-';
      if modified then m:= '*' else m:= c;
      GoRC(titleRow,1);
      Sc_OverstrikeMode;
      left:= TotalCols;
      Wrif (c+m+' '+seed_version+': ');
      Wrif (name);
      Wrif(' (');
      case modeStack[modeTop].mode of
        NoExec : Wrif( 'Doc' );
        ManExec: Wrif( 'Man' );
        JobExec: Wrif( 'Job' );
        IncExec: Wrif( 'Inc' );
        end;
      if not modifiable then Wrif(' View');
      Wrif(') ');
      if filename <> name then Wrif(filename+' ');
      while left > 0 do begin
        Vw_WriteChr(c); left:=left-1;
        end;
      end;
    Sc_Limbo;
    end {titled};
  end {WriteTitle};


procedure Vw_WriteTitles (theEtip: EtipRef);
  { Write title bar for all windows showing this transcript. }
  var w: WinIndex;
  begin
  for w:= 1 to numwin do
    if edwin[w].refEtip = theEtip then
       Vw_WriteTitle(w);
  end {WriteTitles};


procedure Vw_Title (w: WinIndex; title: TitleStyle);
  { Define the title bar style for a window and redraw
    the title bar, if the title style changes. }
  begin
  if edwin[w].title <> title then begin
    edwin[w].title:= title;
    Vw_WriteTitle (w);
    end;
  end {Vw_Title};


function Vw_Split (w: WinIndex; var low: WinIndex): boolean;
  { Divide a window into two halves, if possible.
    The upper half has the same index, the lower half a new one.
  }
  const
    space = 1;  { space between halves for upper window title bar. }
  var
    lower       : Window;   { lower half of window }
    lowcur      : Spot;     { lower half cursor }
    row,col     : integer;  { original position of cursor }
    uppbot      : integer;  { upper half bottom text row }
    lowtop      : integer;  { lower half top    text row }
    i           : integer;
    ok          : boolean;
  begin
  ok:= false;  { unless we succeed totally. }
  if numwin >= MaxWindows then
    Message('[Too many windows]')
  else if Ts_MoreSpots(edwin[w].refEtip^.theTranscript) < 2 then
    Message('[Too many windows for this text]')
  else begin
    wdresult:=Wd_ok;   { so far }
    with edwin[w], refEtip^ do begin
      { Create cursor spot for the lower half. }
      Ts_NewSpot(lowcur,theTranscript,tsresult);
      if tsresult <> Ts_ok then Vw_TsProblem(tsresult)
      else begin
        Wd_Split(theWindow,
                { upperRows } (theWindow.rows-space) div 2,
                { spaceRows } space,
                lower, wdresult);
        if wdresult <> Wd_ok then begin
          Vw_WdProblem(wdresult);
          Ts_DelSpot(lowcur,theTranscript);
          end
        else begin
          { Make sure that the cursors are visible in each half. }
          uppbot:= Wd_TopRow(theWindow) + theWindow.rows - 1;
          lowtop:= Wd_TopRow(lower);
          Ts_Move (cursor, 0,0, lowcur,theTranscript);
          Ts_Where(cursor,theTranscript,row,col);
          if not Wd_VisRow(theWindow,row) then
            Ts_Goto(cursor,theTranscript,
                           {row=>} uppbot,
                           {col=>} Wd_LeftCol(theWindow,uppbot));
          if not Wd_VisRow(lower,row) then
            Ts_Goto(lowcur,theTranscript,
                           {row=>} lowtop,
                           {col=>} Wd_LeftCol(lower,lowtop));
          end;
        end;
      end {with};

    if (tsresult = Ts_ok) and (wdresult = Wd_ok) then begin
      { Make space for the new window (lower) in the window list. }
      { It will be window number w+1.                             }
      low:= w+1;
      for i:=numwin downto low do edwin[i+1]:=edwin[i];
      numwin:=numwin+1;
      { Update upper half. }
      with edwin[w] do begin
        titleRow := theWindow.screenBotRow + 1;
        end;
      Vw_WriteTitle (w);
      { Define lower half. }
      with edwin[low] do begin
        theWindow:= lower;
        refEtip  := edwin[w].refEtip;
        cursor   := lowcur;
        title    := NormalTitle;
        titleRow := theWindow.screenBotRow + 1;
        views    := nil;
        end;
      Vw_WriteTitle (low);
      ok:= true;
      end {ok};
    end {not max windows};

  Vw_Split:= ok;
  end {Vw_Split};

private
procedure ClearRow (r: integer);
  { Clear this screen row. }
  begin
  Sc_Goto(r, 1, scresult);
  if scresult = Sc_ok then Sc_ClearEOL
                      else Vw_ScProblem(scresult);
  end {ClearRow};

private
procedure ClearWindow (w: WinIndex);
  { Erase the part of the screen covered by the window. }
  var r: integer;
  begin
  with edwin[w], theWindow do begin
    for r:= screenTopRow to screenBotRow do ClearRow(r);
    if title <> NoTitle then ClearRow(titlerow);
    end;
  end {ClearWindow};


procedure Vw_CloseWindow (w: integer);
  { Close this window. Leave screen area unused. }
  begin
  with edwin[w], refEtip^ do begin
    Ts_DelSpot   (cursor,theTranscript);
    DisposeViews (views);
    ClearWindow  (w);
    Wd_Close     (theWindow,wdresult);
    Vw_WdProblem (wdresult);
    end;
  end {CloseWindow};

procedure Vw_CloseOthers (var except: WinIndex);
  { Close all except the given window. Make the given window
    grow to use all of the screen (data rows). }
  var w: integer;
  begin
  if numwin <= 1 then Message('[Only one window open]')
  else begin
    for w:=1 to numwin do if w<>except then Vw_CloseWindow(w);
    edwin[1]:= edwin[except];
    except  := 1;
    numwin  := 1;
    Wd_Move  (edwin[except].theWindow,
              datarows.top, datarows.bot,
              wdresult);
    Vw_WdProblem(wdresult);
    edwin[except].titleRow := datarows.bot + 1;
    Vw_WriteTitle(except);
    end;
  end {CloseOthers};


procedure Vw_RefreshScreen;
  { Redraw all windows on screen, including title bars. }
  { Note, SHOULD also reinitialize terminal, but this is not done. }
  var w: integer;
  begin
  for w:=1 to numwin do begin
    with edwin[w] do begin
      ClearWindow(w);
      Wd_FreshAll(theWindow);
      refEtip^.modified:= Ts_Modified(refEtip^.theTranscript);
      end;
    Vw_WriteTitle (w);
    end;
  end {RefreshScreen};


procedure Vw_View (w: WinIndex; etr: EtipRef);
  { Make the window w view the transcript etr. }
  var
    s,v    : ViewRef;
    row,col: integer;
  begin
  with edwin[w] do
    if refEtip <> etr then begin
      { Store the old view. }
      new(s);
      if s = nil then begin
        Message('Vw_View: could not store current view.');
        Beep;
        end
      else begin

        s^.refEtip := refEtip;
        s^.cursor  := cursor;
        s^.nextView:= nil;

        { Look for a stored view of the new transcript. }
        v := FindView (etr, views);
        if v = nil then begin
          { This is a new transcript for this window. }
          Ts_NewSpot (cursor, etr^.theTranscript, tsresult);
          if tsresult <> Ts_ok then
            Vw_TsProblem(tsresult)  { and keep old view }
          else
            refEtip:= etr;
          end
        else begin
          { The transcript has been viewed in this window before. }
          refEtip:= v^.refEtip;   { = etr }
          cursor := v^.cursor;
          { Remove it from the list of stored views. }
          DeleteView (v, views);
          dispose(v);
          tsresult:= Ts_ok;
          end;

        if tsresult = Ts_ok then begin  { Success }
          AppendView (s, views);
          { Fresh up }
          Ts_Where (cursor, refEtip^.theTranscript, row, col);
          ClearWindow(w);
          Wd_FreshAll(theWindow);
          Wd_SeeRow (theWindow, row);
          Wd_SeeCol (theWindow, row, col);
          Vw_WriteTitle(w);
          end {success}
        else
          dispose(s);
        end {could store view};
      end {not same old transcript};    
          
  end {Vw_View};


function Vw_FirstTranscript: EtipRef;
  { Return the first transcript. }
  begin Vw_FirstTranscript:= etipList; end;


function Vw_NextTranscript (etr: EtipRef): EtipRef;
  { Return the next transcript, or nil if none. }
  begin Vw_NextTranscript:= etr^.nextEtip; end;


procedure Vw_CleanWindow (w: WinIndex; var clean: boolean);
  { Clean up (refresh) window number w until it is clean or 
    user presses a key. Parameter 'clean' shows whether the
    window was cleaned.
    If the "modified" status of the transcript has
    changed, this redraws all title bars for the transcript.
  }
  label
      99;
  var ops  : Wd_ScreenOps;
      work : boolean;
      press: boolean;
      clpt : Spot;
      range: Sc_rowRange;
      c    : integer;
      maxc : integer;
      row,col: integer;
      nonover: boolean;
      overuse: boolean;
      numcol : integer;  { number of columns written }
      logon  : boolean;
      nowmod : boolean;  { new modified status }

  begin
  logon:=Log_includes(cleanlog);
  if logon then writeln(logfile,'Cleaning window ',w:1);
  if tracing then MessPause('Cleaning starts.', w);
  with edwin[w], refEtip^ do begin
    maxc:=2*theWindow.rows;  { maximum clean-ops that should occur }
    c:=0; work:=false;
    scresult:=Sc_Ok;
    tsresult:=Ts_Ok;
    press   :=false;
    repeat
      if tracing then MessPause('Wd_Clean..', w);
      if Log_includes(clwinlog) then
        Wd_ShowWin(logfile,edwin[w].theWindow);
      Wd_Clean(theWindow,ops,clean);
      if not clean then with ops do begin
        if not work then begin
          range.top:= theWindow.screenTopRow;
          range.bot:= theWindow.screenBotRow;
          Sc_SetScroll(range,scresult);   if scresult<>Sc_Ok then goto 99;
          Ts_NewSpot(clpt,theTranscript,tsresult);
                                          if tsresult<>Ts_Ok then goto 99;
          work:=true;
          Ts_Move(cursor,0,0,clpt,theTranscript);
          end;

        if deleteRow > 0 then begin
          if logon then writeln(logfile,'CW delete row ',deleteRow:1);
          Sc_Goto(deleteRow,1,scresult); if scresult<>Sc_Ok then goto 99;
          Sc_DeleteLine(scresult);       if scresult<>Sc_Ok then goto 99;
          end;

        if insertRow > 0 then begin
          if logon then writeln(logfile,'CW insert row ',insertRow:1);
          Sc_Goto(insertRow,1,scresult); if scresult<>Sc_Ok then goto 99;
          Sc_InsertLine(scresult);       if scresult<>Sc_Ok then goto 99;
          end;

        with writeFull do
          if screenRow > 0 then begin
            if logon then writeln(logfile,'CW write full row ',screenRow:1,
                                          ' text ',textRow:1); 
            Ts_Where(clpt,theTranscript,row,col);
            Ts_Move (clpt,textRow-row,textCol-col,clpt,theTranscript);
            Ts_Where(clpt,theTranscript,row,col);
            Sc_Goto(screenRow,1,scresult); if scresult<>Sc_Ok then goto 99;
            if row = textRow then begin { row exists }
              Sc_OverstrikeMode;
              Ts_Show (clpt,theTranscript,
                            textCol,textCol+theWindow.cols-1,
                            numcol);
              Sc_Wrote(numcol);
              if numcol < theWindow.cols then Sc_ClearEOL;
              end
            else Sc_ClearEOL;           { row beyond eof }
            end;

        with withinRow do
          if screenRow > 0 then begin
            if logon then writeln(logfile,'CW within row ',screenRow:1,
                                      ' text ',textRow:1);
            Ts_Where(clpt,theTranscript,row,col);
            overuse:=false;
            if insdel.chars <> 0 then
              if logon then writeln(logfile,'...insdel ',insdel.chars:1,
                                      ' screen ',insdel.screenCol:1,
                                      ' text ',insdel.textCol:1);
            if insdel.chars > 0 then begin
              if Sc_CanInsert then begin
                Ts_Move(clpt,textRow-row,insdel.textCol-col,clpt,
                        theTranscript);
                Ts_Where(clpt,theTranscript,row,col);
                if  (row = textRow)
                and (col = insdel.textCol)
                then begin   { within text }
                  Sc_Goto(screenRow,insdel.screenCol,scresult);
                                     if scresult <> Sc_Ok then goto 99;
                  Sc_InsertMode;
                  Ts_Show(clpt,theTranscript,
                          insdel.TextCol,insdel.textCol+insdel.chars-1,
                          numcol);
                  Sc_Wrote(numcol);
                  end {within text};
                end {can insert}
              else
                overuse:=true;
              end {insert requested}
            else

            if insdel.chars < 0 then begin
              if Sc_CanDeleteChar then begin
                Sc_Goto(screenRow,insdel.screenCol,scresult);
                                   if scresult <> Sc_Ok then goto 99;
                Sc_DeleteChar(-insdel.chars);
                end {can delete car}
              else
                overuse:=true;
              end {delete requested};

            if overuse then begin
              { Modify overstrike request to handle insert/delete. }
              if overstrike.chars = 0 then nonover:=true
              else nonover:= overstrike.textCol > insdel.textCol;
              if nonover then begin
                { Start overstrike from insert/delete point. }
                overstrike.screenCol:=insdel.screenCol;
                overstrike.textCol  :=insdel.textCol;
                end;
              { Extend overstrike to right margin. }
              overstrike.chars:=theWindow.cols-overstrike.screenCol+1;
              end {overuse};

            if overstrike.chars > 0 then begin
              if logon then writeln(logfile,
                                      '...overstrike ',overstrike.chars:1,
                                      ' screen ',overstrike.screenCol:1,
                                      ' text ',overstrike.textCol:1);
              Ts_Where(clpt,theTranscript,row,col);
              Ts_Move (clpt,textRow-row,overstrike.textCol-col,clpt,
                            theTranscript);
              Ts_Where(clpt,theTranscript,row,col);
              if  (row = textRow)
              and (col = overstrike.textCol)
              then begin       { text exists }
                Sc_Goto(screenRow,overstrike.screenCol,scresult);
                               if scresult<>Sc_Ok then goto 99;
                Sc_OverstrikeMode;
                Ts_Show (clpt,theTranscript,
                              overstrike.textCol,
                              overstrike.textCol+overstrike.chars-1,
                              numcol);
                Sc_Wrote(numcol);
                if overstrike.screenCol+numcol <= theWindow.cols then
                  Sc_ClearEOL;
                end;
              { else Sc_ClearEOL; }          { row beyond eof }
              end {overstrike};
            end;
        end {ops};
      c:=c+1;
      press:=KeyPressed;
    until clean or (c >= maxc) or press;

    if not press then begin
      { Changes in "modified" status? }
      nowmod:= Ts_Modified(theTranscript);
      if nowmod <> modified then begin
        modified:= nowmod;
        Vw_WriteTitles(refEtip);
        end {mod change};
      end;

    99:
    if not (clean or press) then begin
      MessPause('Window still dirty.', w);
      Vw_TsProblem(tsresult);
      Vw_ScProblem(scresult);
      end {unclean};
    if work then Ts_DelSpot(clpt,theTranscript);

    end {with};
  end {CleanWindow};


procedure Vw_GoCursor (w: WinIndex);
  { Move screen cursor to text cursor position of this window.
    Scroll the window if necessary.
    Includes cleaning of this window, but not others.
  }
  var
    clean       : boolean;
    txrow,txcol : integer;
    scrow,sccol : integer;
  begin
  with edwin[w], refEtip^ do begin
    if tracing then MessPause('GoCursor..', w);
    Ts_Where (cursor,theTranscript,txrow,txcol);
    Wd_SeeRow(theWindow,txrow);
    Vw_CleanWindow(w,clean);   { Must clean window before Wd_SeeCol. }
    if clean then begin
      Wd_SeeCol(theWindow,txrow,txcol);
      Vw_CleanWindow(w,clean);
      if clean then begin
        scrow:=txrow - Wd_TopRow (theWindow)       + theWindow.ScreenTopRow;
        sccol:=txcol - Wd_LeftCol(theWindow,txrow) + 1;
        GoRC(scrow,sccol);
        end;
      end;
    end;
  end {Vw_GoCursor};



{ Query window creation and destruction. }

function Vw_InQueryMode: boolean;
  { Is the query area in use? }
  begin Vw_InQueryMode:= queryEtip <> nil; end;


procedure Vw_CreateQuery (var w: WinIndex; var ok: boolean);
  { Open a new interaction (window + transcript) to ask the
    user for something, e.g. a file name to store the transcript in.
    The new window is returned in w. Success/failure in ok.
  }
  begin
  ok:= false;  { unless we succeed below }
  if Vw_InQueryMode then begin
    Vw_Message('Please complete the current query first.');
    Beep;
    end
  else
  if numwin >= MaxWindows then begin
    Vw_Message('Please close a window first - cannot open more windows.');
    Beep;
    end
  else begin
    Vw_NewEditableTranscript ('[Query]', IncExec, {modify} true, queryEtip);
    if queryEtip = nil then begin
      Vw_Message('Sorry, could not create a transcript for the query.');
      Beep;
      end
    else begin
      numwin  := numwin+1;
      w:= numwin;
      with edwin[w] do begin
        refEtip := queryEtip;
        title   := NoTitle;
        titleRow:= 1;  { just in case }
        Ts_NewSpot(cursor,refEtip^.theTranscript,tsresult);
           Vw_TsProblem(tsresult);
        views   := nil;
        Wd_Open (theWindow,
              {screenTopRow =>}     queryrows.top ,
              {screenBotRow =>}     queryrows.bot ,
              {topRow       =>}     1  ,
              {leftCol      =>}     1  ,
              {cols         =>}     TotalCols  ,
              {band         =>}     1  ,
              wdresult);
        Vw_WdProblem(wdresult);
        end;
      ok:= true;
      end;
    end;
  end {Vw_CreateQuery};


procedure Vw_QuitQuery;
  { Delete the query interaction. }
  begin
  if Vw_InQueryMode then begin

    { Close and clear the query window, which must be the last one.
      Thus, no other window indices are changed. }
    with edwin[numwin] do
      if refEtip = queryEtip then begin
        { This window is going, going, gone. }
        ClearWindow(numwin);
        Wd_Close   (theWindow, wdresult);  Vw_WdProblem(wdresult);
        { Should delete the cursor spot, too, but skip this since
          the whole transcript will be killed later. }
        DisposeViews(views);
        numwin:= numwin-1;  { what remains }
        end
      else
        if not Vw_MessPause('Vw_QuitQuery: last window is not query window!',
                     numwin)
        then HaltProgram;

    { Finalize and get rid of the query transcript. }
    Vw_KillEditableTranscript(queryEtip);
    end;
  end {Vw_QuitQuery};


function Vw_NewTranscript (w: WinIndex; modify: boolean): boolean;
  { Create a new transcript for view in this window, specifying if
    it will be modifiable or for view only.
    Return success/failure.
  }
  var
    rEtip  : EtipRef;
    s      : ViewRef;
    row,col: integer;
  begin
  Vw_NewEditableTranscript('', NoExec, modify, rEtip);
  if rEtip = nil then begin
    Vw_Message('Could not make a new transcript - memory full?');
    Beep;
    end
  else begin
    with edwin[w] do begin
      { Store the old view. }
      new(s);
      if s = nil then begin
        Message('Vw_View: could not store current view.');
        Beep;
        end
      else begin
        s^.refEtip := refEtip;
        s^.cursor  := cursor;
        s^.nextView:= nil;
        AppendView (s, views);
        end;

      refEtip:= rEtip;
      with refEtip^ do begin
        Ts_NewSpot(cursor, theTranscript, tsresult);
        if tsresult <> Ts_ok then begin
          Vw_TsProblem(tsresult);
          row:= 1; col:= 1;
          end
        else
          Ts_Where (cursor, theTranscript, row, col);
        end {with refEtip};

      ClearWindow(w);
      Wd_FreshAll(theWindow);
      Wd_SeeRow  (theWindow, row);
      Wd_SeeCol  (theWindow, row, col);
      { The window title is not written yet, since it is likely that
        the transcript and/or file names will be changed soon. }
      end {with edwin};
    end {new etip ok};

  Vw_NewTranscript:= rEtip <> nil;
  end {Vw_NewTranscript};


private
procedure BaseName (path: vstring; var base: vstring);
  { Extract the "base name" of a file path name.
    This is everything after the last '/' in the path name.
    If the path name has no slashes, the base name is the same
    as the path name.
  }
  var n: integer;
  begin
  n:=length(path);
  if n = 0 then base:= ''
  else begin
    while (n > 1) and (path[n] <> '/') do n:=n-1;
    if path[n] = '/' then base:= substr(path, n+1, length(path)-n)
                     else base:= path;
    end;
  end {BaseName};


{ Distribution of changes to all windows affected. }

procedure Vw_DelCharWs (etr: EtipRef; row,col: integer; num: integer);
  { Delete a number of characters at row,col in all windows
    for transcript etr. }
  var w: integer;
  begin
  for w:=1 to numwin do with edwin[w] do
    if refEtip = etr then Wd_DeleteChar(theWindow,row,col,num);
  end {DelCharWs};

procedure Vw_InsCharWs (etr: EtipRef; row,col: integer; num: integer);
  { Insert a number of characters at row,col in all windows
    for transcript etr. }
  var w: integer;
  begin
  for w:=1 to numwin do with edwin[w] do
    if refEtip = etr then Wd_InsertChar(theWindow,row,col,num);
  end {InsCharWs};

procedure Vw_DelLineWs (etr: EtipRef; row: integer; num: integer);
  { Delete a number of lines starting with 'row' in all windows
    for transcript etr. }
  var w: integer;
  begin
  for w:=1 to numwin do with edwin[w] do 
    if refEtip = etr then Wd_DeleteLine(theWindow,row,num);
  end {DelLineWs};

procedure Vw_InsLineWs (etr: EtipRef; row: integer; num: integer);
  { Insert a number of lines at 'row' in all windows
    for transcript etr. }
  var w: integer;
  begin
  for w:=1 to numwin do with edwin[w] do 
    if refEtip = etr then Wd_InsertLine(theWindow,row,num);
  end {InsLineWs};

procedure Vw_FreshLineWs (etr: EtipRef; row,col: integer);
  { Freshen (redisplay) line 'row' from column 'col' in all windows
    for transcript etr. }
  var w: integer;
  begin
  for w:=1 to numwin do with edwin[w] do 
    if refEtip = etr then Wd_FreshLine(theWindow,row,col);
  end {FreshLineWs};

procedure Vw_NewlineWs (etr: EtipRef; row,col: integer);
  { Make a newline at 'row' column 'col' in all windows for
    transcript etr. }
  var w: integer;
  begin
  for w:=1 to numwin do with edwin[w] do 
    if refEtip = etr then begin
      Wd_BreakLine(theWindow,row,col);
      end;
  end {NewlineWs};


procedure Vw_InsShow (theSpot: Spot; refEtip: EtipRef;
                      preRow,preCol: integer);
  { Update screen after text has been inserted.
    Text (perhaps including newlines) was inserted at the left
    of the given spot. Before the insertion, the spot was at
    the given row and column. The transcript is refEtip. }
  var row,col: integer;
  begin
  Ts_Where(theSpot,refEtip^.theTranscript,row,col);
  if row = preRow then
    Vw_InsCharWs  (refEtip,preRow  ,preCol, col-preCol)
  else begin
    Vw_FreshLineWs(refEtip,preRow  ,preCol            );
    Vw_InsLineWs  (refEtip,preRow+1,        row-preRow);
    end;
  end {InsShow};


procedure Vw_DelShow (refEtip: EtipRef;
                      delRow,delCol,preRow,preCol: integer);
  { Update screen after text has been deleted in a transcript.
    Text (perhaps including newlines) was deleted, starting at
    delRow,delCol and continuing up to but not including preRow,preCol. }
  var nc,nr: integer;
  begin
  nc:= preCol-delCol;
  nr:= preRow-delRow;
  if delRow = preRow then begin
    if nc > 0 then Vw_DelCharWs  (refEtip,delRow, delCol, nc);
    end
  else
  if delCol = 1 then begin
    Vw_DelCharWs  (refEtip,preRow  ,delCol, nc);
    Vw_DelLineWs  (refEtip,delRow  ,        nr);
    end
  else begin
    { delCol > 1 and delRow < preRow - remaining part of preRow
      is appended to delRow. }
    Vw_DelLineWs  (refEtip,delRow+1,        nr);
    Vw_FreshLineWs(refEtip,delRow  ,delCol    );
    end;
  end {Vw_DelShow};


procedure Vw_DelShot (theSpot: Spot; refEtip: EtipRef;
                      preRow,preCol: integer);
  { Update screen after text (perhaps including newlines) was deleted
    at the left of the given spot. Before the deletion, the spot was
    at the given row and column. The transcript is refEtip. }
  var row,col: integer;
  begin
  Ts_Where(theSpot,refEtip^.theTranscript,row,col);
  Vw_DelShow(refEtip,row,col,preRow,preCol);
  end {DelShot};


{ Original / modified status }


procedure Vw_Modifiable (w: WinIndex; modify: boolean);
  { Control whether transcript in the current view in this window is
    to be modifiable (editable, changeable) or not.
  }
  begin
  edwin[w].refEtip^.modifiable:= modify;
  end {Vw_Modifiable};


function Vw_Modified (w: WinIndex): boolean;
  begin
  Vw_Modified:= Ts_Modified(edwin[w].refEtip^.theTranscript);
  end {Vw_Modified};


procedure Vw_Original (w: WinIndex);
  begin
  with edwin[w].refEtip^ do begin
    Ts_Original(theTranscript);
    origname:= filename;
    end;
  end {Vw_Original};


{ Bulk editing routines }


procedure Vw_NameFile (w: WinIndex; fname: vstring; override: boolean);
  { Assign the file name of the current view in this window, if no
    name has yet been assigned.
    Use override = true to change a name assigned earlier.
    The transcript name is set to the base name of the file.
  }
  var fbase: vstring;
  begin
  with edwin[w], refEtip^ do begin
    if (filename = '') or override then begin
      filename:= fname;
      if (name = '') or override then begin
        BaseName (fname, fbase);
        if length(fbase) <= MaxTipName
        then name:= fbase
        else name:= substr(fbase, 1, MaxTipName);
        end;
      Vw_WriteTitles(refEtip);
      end;
    end;
  end {Vw_NameFile};


procedure Vw_ReadFile (w: WinIndex; var ff: text; fname: vstring;
                       var ok: boolean);
  { Read in a text or transcript from a file into the current view.
    The file ff is assumed to be opened already using the name fname and
    can contain pure text or a transcript.
    The file is then closed. Success is reported in ok.
  }
  var
    preRow,preCol: integer;
    lines        : integer;
    linestr      : vstring;
  begin
  with edwin[w], refEtip^ do begin
    Ts_Load(cursor,theTranscript,ff,preRow,preCol,lines,tsresult);
    Vw_InsShow(cursor,refEtip,preRow,preCol);
    if tsresult = Ts_ok then begin
      Vw_DecimalString(lines,1,linestr);
      Vw_Message('(Read ' + linestr + ' lines from ' + fname +')');
      end
    else
      Vw_TsProblem(tsresult);
    end;
  close(ff);
  ok:= tsresult = Ts_ok;
  end {Vw_ReadFile};

{ -- }

procedure Vw_init;
  { Initialize module. }
  var
    numrows  : integer;
    etipr    : EtipRef;

  begin

  Log_Init;
  Km_Init;
  Rn_Init;
  Sc_Init;
  { Get the screen size and cut it up into data and query areas. }
  Sc_GetArea (allrows, TotalCols);
  numrows:= allrows.bot - allrows.top + 1;
  if numrows < 4 then begin
    writeln(errout,'The window is too small - at least 4 lines needed, only ',
      numrows:1,' lines available.');
    Sc_Fini;
    Rn_Fini;
    Km_Fini;
    Log_Fini;
    end;

  Wd_Init;
  Ts_Init;

  queryrows.top:= allrows.bot-1;
  queryrows.bot:= allrows.bot  ;

  datarows .top:= allrows.top  ;
  datarows .bot:= allrows.bot-3;  { less 2 rows query, less 1 row title bar }

  statusrows.top:= queryrows.top;
  statusrows.bot:= queryrows.top;

  with messPlace do begin row:= allrows.bot  ; col:= 1; end;

  messed   := false;
  tracing  := false;

  etipList := nil;
  queryEtip:= nil;

  Sc_SetScroll(datarows, scresult);         Vw_ScProblem(scresult);

  { The first transcript. }
  Vw_NewEditableTranscript('', IncExec, {modify} true, etipr);
  if etipr = nil then begin
    Vw_Message('Could not make the main transcript.');
    HaltProgram;
    end;

  { Window number one. }
  numwin:=1;
  with edwin[1] do begin
    refEtip := etipr;
    title   := NormalTitle;;
    titleRow:= datarows.bot + 1;
    Ts_NewSpot(cursor,refEtip^.theTranscript,tsresult);
      Vw_TsProblem(tsresult);
    views:= nil;
    Wd_Open (theWindow,
            {screenTopRow =>}     datarows.top ,
            {screenBotRow =>}     datarows.bot ,
            {topRow       =>}     1  ,
            {leftCol      =>}     1  ,
            {cols         =>}     TotalCols  ,
            {band         =>}     1  ,
            wdresult);
      Vw_WdProblem(wdresult);
    end {window number one};

  Vw_WriteTitle(1);

  end {Vw_Init};

