{ Window logic module for text display.

  A window is a part of a screen. It covers a contiguous interval of
  screen rows, and the whole width of the screen. The screen space can
  be divided among several tiled (non-overlapping) windows.

  In its screen area, a window displays a rectangular portion of a
  text structured in lines (rows) and columns. A window displays some
  contiguous range of text lines, and from each line a contiguous range
  of text columns (not necessarily the same for all lines).

  The text in a window can be scrolled vertically to cover a different
  line range. Lines can be scrolled horizontally to cover a different
  column range.

  The text can be changed by

     - inserting or deleting text on a line;
     - inserting or deleting lines;
     - breaking a line (inserting a line break).

  Other changes within a line (e.g. inserting and deleting at different
  points on the line) are implemented by redisplaying from a
  given column to end of line. The line-break operation is included
  because of an important special case: insertion of new text at
  the start of an existing line followed by a break at the boundary
  between the new and the old text. This happens frequently in transcript
  editing (new output text inserted before an old input line).

  With regard to the screen position, a window can be moved and resized.
  A window can also be split into two windows.

  This module stores no text, it just selects the screen operations needed
  to update the screen and indicates the point (row,column) in the text
  that contains the text to be displayed.

  To reduce redundant screen operations and support multiple windows into
  one text, the module provides recording and delayed execution of
  line insertion and deletion, scrolling, and some within-line changes.
  There are some restrictions on horizontal scrolling in this regard
  (see comments on Wd_ScrollH and Wd_SeeCol).

  Author: Niklas Holsti    around 1991.02.26
  Changes:
    91.06.17  NH  Added file parameter to Explain routine.
    91.06.18  NH  Uses Log module.
    91.06.19  NH  Show topDirt marker (*) also for deleted line.
                  Wd_Clean/UpRow, changed selection of current screen row
                  of row inserted to compensate for row deleted.
                  Wd_BreakLine added.
    91.06.21  NH  OpComb routine added, used in InsertChar, DeleteChar,
                  and BreakLine.
    91.06.22  NH  Insert & Delete routines get line and char counts.
                  Wd_Clean/DelRow now handles redisplay of compensatory
                  insertion. In case there is no insertion, the number
                  of additional deleted lines is considered when selecting
                  the row to scroll in at the bottom.
                  Cache for last screen row added, to find the current row
                  quicker.
    91.09.26  NH  Within a line, a deletion followed by an insertion never
                  combine (I'm not sure about this, but its a safe choice).
    92.04.09  NH  Port to Sun Unix.
    92.04.17  NH  Local vars 'private' for less risk of sym conflicts.
    93.02.12  NH  Wd_Clean worked badly when the leading lines were going up
                  off the screen, and the following lines were going
                  down off the bottom. Tried to fix it (changed lines
                  marked with date). This was in DelRow; similar change
                  made in DownRow for the opposite-direction case.
                  [ Date refers to Mac change; on Unix, date is 93.04.27 ]
}

module Windows;

#include "strings_d.p"
#include "log_f.p"
#include "windows_d.p"
#include "console_f.p"   { for test routines }

const
  oplog = 1;   { log level to list operations }

type

  { The most important data structure is the "row fate" map. This is an
    array indexed by row number referred to the screen as it was
    before the current batch of changes. It shows which screen row
    this line should occupy after the changes (perhaps off screen).
    It also shows the past and current horizontal position.
    For each line the map can record one insertion or deletion
    of text. More complex within-line edits are implemented by
    redisplaying the line from the leftmost changed column to eol.
    For a deleted row, only the fact of deletion is held. 
  }

  RowFate = record    { What happened to a screen row }
     deleted    : boolean;    { If true, rest of record is undefined. }
     screenRow  : record      { Screen row for this text.. }
                { past            }   { ..when last displayed  }
                  current: integer;   { ..to be displayed next }
                  end;
     textLeftCol: record      { Leftmost text column visible.. }
                  past   : integer;   { ..when last displayed  }
                  current: integer;   { ..to be displayed next }
                  end;
     insdel     : record      { Insert or delete some characters }
                  chars  : integer;   { ..this number of chars }
                  textCol: integer;   { ..starting at this text column }
                  end;                {   in the current text line. }
     fresh      : record      { Redisplay to end of line.. }
                  textCol: integer;   { ..from this text col (if < maxint)}
                  end;
     end {RowFate};

  RowUsage = (           { Whether a screen row is in use }
             rowUsed,    { .. some window covers the row  }
             rowFree     { .. no   window covers the row  }
             );

private var
  usage: array [1..Wd_MaxRows] of RowUsage;
  fate : array [1..Wd_MaxRows] of RowFate;

  { The 'fate' vector is indexed by past screen row.
    Each window W uses only indices from W.screenTopRow to
    W.screenTopRow + W.rows - 1.

    A window display is up to date when -
      for each i:=screenTopRow to screenBotRow
         fate[i] satisfies  deleted = false
                       and  screenRow.current = i
                       and  textLeftCol.current = textLeftCol.past
                       and  insdel.chars = 0
                       and  freshCol = maxint.
    Note that the W.topDirt is only a hint; if > bottom row,
    the window is sure to be clean, but if <= bottom row, the 
    window is not necessarily unclean. The matter is decided by Wd_Clean,
    see below. topDirt can point to a deleted row.
  }

procedure Wd_Explain (var f: text; result: Wd_result);
  begin
  case result of
    Wd_ok        : write(f,'Window operation sucessful.');
    Wd_norows    : write(f,'Window would have no rows.');
    Wd_offscreen : write(f,'Window would extend off screen');
    Wd_norowsleft: write(f,'No rows would be left in lower part.');
    Wd_negspace  : write(f,'Negative space in split.');
    Wd_overlap   : write(f,'Windows would overlap.');
    Wd_unused    : write(f,'Some rows were unused.');
    end;
  end {Wd_Explain};


private
procedure OpComb (var c1,n1: integer;
                      c2,n2: integer;
                  var cf   : integer);
  { Combines two within-line operations, (c1,n1) and (c2,n2).
    First, n1 characters are inserted (n1 > 0) or deleted (n1 < 0)
    at text column c1. Then, n2 characters are inserted or deleted
    at column c2 (counting columns as they are after the first
    operation). This routine tries to combine the two operations
    into one. If it succeeds, the combined operation is returned
    in c1,n1 and cf is updated to account for the change in column
    numbering. If it fails, it sets cf to min(cf,c1,c2); this indicates
    that the line should be redisplayed from column cf.
    Finally, if cf <= min(c1,c2), n1 is set to zero, since it serves
    no purpose to insert/delete in an area that will be redisplayed.
    Special cases: if n1 = 0, the first operation is absent; the combined
                              operation is equal to the second operation.
                   if n2 = 0, there is no second operation; the combined
                              operation is equal to the first operation.
    In both special cases, the final check of cf is still done.
  }
  var
    c    : integer;  { min(c1,c2) }
    e1,e2: integer;  { column after extent of respective operation }
    comb : boolean;  { combination is possible }
  begin
  if n1 = 0 then begin
    { No first operation. }
    c   :=c2;
    comb:=true;
    end
  else
  if n2 = 0 then begin
    { No second operation. }
    c   :=c1;
    comb:=true;
    end
  else begin
    { Two operations to combine. }
    if c1 <= c2 then c:=c1
                else c:=c2;
    e1:=c1 + abs(n1);
    e2:=c2 + abs(n2);

    { Case analysis.
      Two insertions combine if the second one is within or just
      adjacent to the first one.
      An insertion followed by a deletion combine if one is entirely
      covered by the other.
      A deletion followed by an insertion never combine.
      Two deletions combine if the second one contains or is just
      adjacent to the first one.
    }
    if n1 > 0 then begin
      if n2 > 0 then comb:=  (c1 <= c2) and (c2 <= e1)  {ins-ins}
                else comb:= ((c1 <= c2) and (e2 <= e1)) {ins-del}
                         or ((c2 <= c1) and (e1 <= e2))
      end
    else begin
      if n2 > 0 then comb:=  false                      {del-ins}
                else comb:= (c2 <= c1) and (c1 <= e2)   {del-del}
      end;
    end {two ops};

  { Update cf for new column numbers. }
  if cf < maxint then
    if n2 <> 0 then
      if cf >= c2 then begin
        cf:=cf + n2;
        if cf < c2 then cf:=c2;
        end;

  { Combine or redisplay. }
  if comb then n1:=n1+n2
          else if cf > c then cf:=c;

  c1:=c;
  if cf <= c then n1:=0;

  end {OpComb};


private
function AnyRow (use: RowUsage; top,bot: integer): boolean;
  { Whether any screen rows top..bot are in the 'use' state. }
  var i: integer;
      u: boolean;
  begin
  u:=false;
  i:=top;
  while (i<=bot) and (not u) do begin
    if usage[i] = use then u:=true
                      else i:=i+1;
    end;
  AnyRow:=u;
  end {AnyRow};

private
procedure AllRows (use: RowUsage; top,bot: integer);
  { Set usage[top..bot]:=use. }
  var i: integer;
  begin
  for i:=top to bot do usage[i]:=use;
  end {AllRows};

private
procedure OpenRows (textCol: integer; top,bot: integer);
  { Initialize fates of rows top..bot to redisplay the rows
    completely with textCol at the left margin. }
  var i: integer;
  begin
  for i:=top to bot do
    with fate[i] do begin
      deleted            := false;
      screenRow.current  := i;
      textLeftCol.past   := textCol;
      textLeftCol.current:= textCol;
      insdel.chars       := 0;
      fresh .textCol     := 1;     { to redisplay row }
      end;
  end {OpenRows};

private
procedure NinFate (var fate: RowFate);
  { Erase internal edits in the fate. }
  begin
  with fate do begin
    insdel.chars :=0;
    fresh.textCol:=maxint;
    end;
  end {NinFate};

private
function TextRowAt (screenRow: integer;
                    var {in} win: Window): integer;
  { Text row corresponding to screen row. }
  begin
  with win do TextRowAt:=textTopRow + screenRow - screenTopRow;
  end {TextRowAt};

private
function ScreenRowAt (textRow: integer;
                      var {in} win: Window): integer;
  { Screen row corresponding to text row. }
  begin
  with win do ScreenRowAt:=screenTopRow + textRow - textTopRow;
  end {ScreenRowAt};


private
procedure FindTextRow (var {i} win: Window;
                               textRow: integer;
                       var {o} ind    : integer);
  { Find a given text row in the screen of a window, returning the row
    fate map index, or zero if the text row is not now on screen.
    Note that we are concerned with what is on screen now,
    rather than what will be on screen when the window is
    "cleaned" later on. Thus, we do not mind if the row is marked to
    be destined to scroll off the screen, since later scroll operations
    may well make it stay on screen, after all. A deleted line is
    not found, since there is no way to revoke the deletion.
    Binary search in the fate map is used (if there are deleted lines,
    the effective search partition may not be binary).
    The last found line is cached in the window record. }
  var i,j,m,k,c,r: integer;
  begin
  with win do begin
    r:=ScreenRowAt(textRow,win);

    { Check the cache. }
    ind:= cached.screenRow;
    with fate[ind] do
      if deleted then ind:= 0
      else if screenRow.current <> r then ind:= 0;

    if ind = 0 then begin
      { No match in cache, must do some work. }
      i:=screenTopRow;
      j:=screenBotRow;
      while (fate[i].deleted) and (i<=j) do i:=i+1;
      while (fate[j].deleted) and (i<=j) do j:=j-1;
      if i > j
      then { all rows were deleted }
      else
      if  (r < fate[i].screenRow.current)
      or  (r > fate[j].screenRow.current)
      then { row cannot be in i..j }
      else
      if  (r = fate[i].screenRow.current) then ind:=i
      else
      if  (r = fate[j].screenRow.current) then ind:=j
      else
      while (i<j) and (ind=0) do begin
        { Neither row i nor row j were deleted, and
          i.screenRow < r < j.screenRow. }
        m:=(i+j) div 2;
        k:=m;                               { i <= k < j      }
        while fate[k].deleted do k:=k-1;    { stops at k >= i }
        c:=fate[k].screenRow.current;
        if      c = r then ind:=k           { eureka!             }
        else if c > r then j  :=k           { look at i..k, k < j }
        else begin
          k:=m + 1;                         { i < k <= j      }
          while fate[k].deleted do k:=k+1;  { stops at k <= j }
          c:=fate[k].screenRow.current;
          if      c = r then ind:=k         { eureka!             }
          else if c < r then i  :=k         { look at k..j, k > i }
          else               i  :=j+1;      { quit, not found     }
          end;
        end {while};

      if ind <> 0 then cached.screenRow:= ind;

      end {cache miss};
    end {with};
  end {FindTextRow};


private
procedure Stain (i: integer; var win: Window);
  { Note that screen row i is dirty. }
  begin
  with win do if i < topDirt then topDirt:=i;
  end {Stain};


function  Wd_VisRow (var win: Window;
                         row: integer): boolean;
  begin
  Wd_VisRow :=  (row >= win.textTopRow)
            and (row <  win.textTopRow + win.rows);
  end {Wd_VisRow};

function  Wd_VisCol (var win: Window;
                         screenRow: integer;
                         col      : integer): boolean;
  begin
  with fate[screenRow].textLeftCol do
    Wd_VisCol:=  (col >= current)
             and (col <  current + win.cols);
  end {Wd_VisCol};

function Wd_TopRow (var {in} win: Window): integer;
  begin
  Wd_TopRow:=win.textTopRow;
  end {Wd_TopRow};


function Wd_LeftCol (var {in} win: Window;
                              textRow: integer): integer;
  var i: integer;
  begin
  FindTextRow(win,textRow,i);
  if i <= 0 then Wd_LeftCol:=0
  else           Wd_LeftCol:=fate[i].textLeftCol.current;
  end {Wd_LeftCol};


procedure Wd_ScrollV (var win   : Window;
                         topRow: integer);
  var i,dr: integer;
  begin
  with win do
    if topRow <> textTopRow then begin
      dr:=topRow-textTopRow;
      for i:=screenTopRow to screenBotRow do
        with fate[i].screenRow do begin
          current:=current - dr;
          if current <> i then Stain(i,win);
          end;
      textTopRow:=topRow;
      end;
  end {Wd_ScrollV};


procedure Wd_ScrollH (var win    : Window;
                          textRow: integer;
                          leftCol: integer);
  var rlo,rhi,i,r: integer;
  begin
  with win do begin
    rlo:=ScreenRowAt(textRow-band,win);
    rhi:=ScreenRowAt(textRow+band,win);
    for i:=screenTopRow to screenBotRow do begin
      r:=fate[i].screenRow.current;
      if (rlo <= r) and (r <= rhi) then
        with fate[i].textLeftCol do begin
          current:=leftCol;
          if current <> past then Stain(i,win);
          end;
      end;
    end;
  end {Wd_ScrollH};


procedure Wd_SeeRow (var win    : Window;
                         textRow: integer);
  var
    drtop,drbot: integer;

  function Vstep (drmin,drmax: integer): integer;
    { Select amount to scroll. The possible range is
      drmin..drmax (counting both directions as positive). }
    begin
    if      drmin > Wd_Vscroll then Vstep:=drmin
    else if drmax < Wd_Vscroll then Vstep:=drmax
    else                            Vstep:=Wd_Vscroll;
    end {Vstep};

  begin
  if Log_includes(oplog) then
    writeln(logfile,'Wd_SeeRow textRow=',textRow:1);
  drtop:=textRow-win.textTopRow;   { textRow -> top window row }
  drbot:=drtop-win.rows+1;         { textRow -> bottom window row }
  if      drtop<0 then Wd_ScrollV(win,win.textTopRow-Vstep(-drtop,-drbot))
  else if drbot>0 then Wd_ScrollV(win,win.textTopRow+Vstep( drbot, drtop));
  end {Wd_SeeRow};


procedure Wd_SeeCol (var {io} win    : Window;
                              textRow: integer;
                              textCol: integer);
  var
    i                : integer;
    dcleft,dcright,dc: integer;
    leftCol          : integer;

  function Hstep (dcmin,dcmax: integer): integer;
    { Select amount to scroll. The possible range is
      dcmin..dcmax (counting both directions as positive). }
    begin
    if      dcmin > Wd_Hscroll then Hstep:=dcmin
    else if dcmax < Wd_Hscroll then Hstep:=dcmax
    else                            Hstep:=Wd_Hscroll;
    end {Hstep};

  begin
  if Log_includes(oplog) then
    writeln(logfile,'Wd_SeeCol textRow=',textRow:1,
                             ' textCol=',textCol:1);
  FindTextRow(win,textRow,i);
  if i > 0 then begin
    with fate[i] do begin
      dcleft :=textCol-textLeftCol.current;  { textCol -> left  margin }
      dcright:=dcleft-win.cols +1;           { textCol -> right margin }
      if      dcleft <0 then dc:=-Hstep(-dcleft ,-dcright)
      else if dcright>0 then dc:= Hstep( dcright, dcleft )
      else                   dc:=0;
      leftCol:=textLeftCol.current+dc;
      if leftCol < 1 then leftCol:=1;
      end;
    Wd_ScrollH (win,textRow,leftCol);
    end
  else
    if Log_includes(oplog) then writeln(logfile,'...row not found');
  end {Wd_SeeCol};


private
procedure InsLine (var win: Window; textRow: integer; num: integer);
  { Insert a number of lines. Internal routine. }
  var r,i: integer;
  begin
  with win do begin
    if textRow <= textTopRow then
      { Lines inserted before the window. Window unchanged except
        for number of top line. }
      textTopRow:= textTopRow + num
    else begin
      r:= ScreenRowAt(textRow,win);
      if r <= screenBotRow then
        { All rows at or below this one move one step down. }
        for i:=screenTopRow to screenBotRow do
          with fate[i], screenRow do
            if current >= r then begin
              current:=current+num;
              if current <> i then Stain(i,win);
              end;
      end;
    end;
  end {InsLine};


procedure Wd_InsertLine (var win: Window;
                             textRow: integer;
                             num    : integer);
  begin
  if Log_includes(oplog) then
    writeln(logfile,'Wd_InsertLine textRow=',textRow:1,' num=',num:1);
  InsLine(win,textRow,num);
  end {Wd_InsertLine};


procedure Wd_DeleteLine (var win    : Window;
                             textRow: integer;
                             num    : integer);
  var
    n: integer;   { number of rows left to delete }
    d: integer;   { number of rows that are deleted before TopTextRow }
    r: integer;   { First deleted screen row in window }
    b: integer;   { Last deleted screen row in window, +1 }
    i: integer;

  begin
  if Log_includes(oplog) then
    writeln(logfile,'Wd_DeleteLine textRow=',textRow:1,' num=',num:1);
  n:= num;
  with win do begin
    if textRow < textTopRow then begin
      { Deletions before the top window row only affect row numbering. }
      d:= textTopRow - textRow;
      if d > n then d:= n;
      textTopRow:= textTopRow - d;
      n:= n - d;
      end;
    if n > 0 then begin
      r:= ScreenRowAt(textRow,win);         { First deleted screen row }
      if r <= screenBotRow then begin       { Window is affected }
        b:= ScreenRowAt(textRow + n,win);   { Last deleted screen row + 1 }
        for i:=screenTopRow to screenBotRow do
          with fate[i].screenRow do
            if current >= b then begin
              current:= current - n;
              if current <> i then Stain(i,win);
              end
            else
            if current >= r then begin
              fate[i].deleted:= true;
              Stain(i,win);
              end
        end { r <= screen bot };
      end { n > 0 };
    end {with win};
  end {Wd_DeleteLine};


procedure Wd_DeleteChar (var win: Window;
                             textRow,textCol: integer;
                             num            : integer);
  var i: integer;
  begin
  if Log_includes(oplog) then
    writeln(logfile,'Wd_DeleteChar textRow=',textRow:1,
                                '  textCol=',textCol:1,
                                    '  num=',num:1);
  FindTextRow(win,textRow,i);
  if i<>0 then begin
    with fate[i] do
      OpComb(insdel.textCol,insdel.chars, textCol,-num, fresh.textCol);
    Stain(i,win);
    end
  else
    if Log_includes(oplog) then writeln(logfile,'...row not found');
  end {Wd_DeleteChar}; 


procedure Wd_InsertChar (var win: Window; 
                             textRow,textCol: integer;
                             num            : integer);
  var i: integer;
  begin
  if Log_includes(oplog) then
    writeln(logfile,'Wd_InsertChar textRow=',textRow:1,
                                '  textCol=',textCol:1,
                                    '  num=',num:1);
  FindTextRow(win,textRow,i);
  if i<>0 then begin
    with fate[i] do
      OpComb(insdel.textCol,insdel.chars, textCol,+num, fresh.textCol);
    Stain(i,win);
    end
  else
    if Log_includes(oplog) then writeln(logfile,'...row not found');
  end {Wd_InsertChar};


private
procedure SetFresh (var fate   : RowFate;
                        textCol: integer);
  { Set the fresh column. }
  begin
  with fate do
    if not deleted then begin
      if textCol < fresh.textCol then fresh.textCol:=textCol;
      if insdel.chars <> 0 then
        if fresh.textCol <= insdel.textCol then insdel.chars:=0;
      end;
  end {SetFresh};


procedure Wd_FreshLine (var win: Window;
                            textRow: integer;
                            textCol: integer);
  var i: integer;
  begin
  if Log_includes(oplog) then
    writeln(logfile,'Wd_FreshLine textRow=',textRow:1,
                               '  textCol=',textCol:1);
  FindTextRow(win,textRow,i);
  if i > 0 then begin
    SetFresh(fate[i],textCol);
    Stain(i,win);
    end
  else
    if Log_includes(oplog) then writeln(logfile,'...row not found');
  end {Wd_FreshLine};


procedure Wd_BreakLine  (var win: Window;
                             textRow,textCol: integer);
  var i,j: integer;  { row indices }
  begin
  if Log_includes(oplog) then
    writeln(logfile,'Wd_BreakLine textRow=',textRow:1,
                               '  textCol=',textCol:1);
  FindTextRow(win,textRow,i);
  if i = 0 then begin
    if Log_includes(oplog) then writeln(logfile,'...row not found.');
    InsLine(win,textRow+1,1);
    end
  else begin
    { All rows at or below this one move one step down. }
    for j:=i to win.screenBotRow do
      with fate[j],screenRow do current:=current+1;
    { The right-hand part of the broken line is obtained by
      deleting textCol-1 characters from the start of the whole line. }
    with fate[i] do
      OpComb(insdel.textCol,insdel.chars, 1,1-textCol, fresh.textCol);
    Stain(i,win);
    end;
  end {Wd_BreakLine};


procedure Wd_FreshAll (var win: Window);
  var i: integer;
  begin
  if Log_includes(oplog) then
    writeln(logfile,'Wd_FreshAll, screenTopRow=',win.screenTopRow:1);
  for i:=win.screenTopRow to win.screenBotRow do begin
    SetFresh(fate[i],fate[i].textLeftCol.current);
    Stain(i,win);
    end;
  end {Wd_FreshAll};


private
function Proper (top,bot: integer): Wd_result;
  { Check screen area request for window. }
  begin
  if       top > bot                      then Proper:= Wd_norows
  else if (top < 1) or (Wd_MaxRows < bot) then Proper:= Wd_offscreen
                                          else Proper:= Wd_ok;
  end {Proper};


procedure Wd_Open (var win         : Window;
                       screenTopRow: integer;
                       screenBotRow: integer;
                       TopRow      : integer;
                       LeftCol     : integer;
                       cols        : integer;
                       band        : integer;
                   var result      : Wd_result);

  begin
  if Log_includes(oplog) then
    writeln(logfile,'Wd_Open, screen=',screenTopRow:1,'..',screenBotRow:1,
                    ' text=',TopRow:1,'..');
  result:=Proper(screenTopRow,screenBotRow);
  if result = Wd_ok then
    if AnyRow(rowUsed,screenTopRow,screenBotRow) then
      result:= Wd_overlap;

  if result = Wd_ok then begin
    win.screenTopRow    := screenTopRow;
    win.screenBotRow    := screenBotRow;
    win.rows            := screenBotRow-screenTopRow+1;
    win.textTopRow      := TopRow;
    win.cols            := cols;
    win.band            := band;
    AllRows (rowUsed,screenTopRow,screenBotRow);
    OpenRows(LeftCol,screenTopRow,screenBotRow);
    win.topDirt         := screenTopRow;
    win.cached.screenRow:= screenTopRow;
    result:= Wd_ok;
    end;
  end {Wd_Open};


procedure Wd_Move  (var {io} win         : Window;
                             screenTopRow: integer;
                             screenBotRow: integer;
                    var { o} result      : Wd_result);
  { Move and/or resize window on screen. The new topmost screen row and
    new number of rows are given. All screen rows covered by the
    convex union of the old and new window screen areas must be
    free (or used by this window), otherwise Wd_overlap is returned.
    (Window reconfigurations where this condition is not met must be
    implemented by deleting and recreating windows.)
    The text position of the reshaped window is chosen so that the
    existing screen rows are valid, if possible. }
  var ctop,cbot : integer;  { total rows covered by old & new windows }
      newTextTop: integer;  { new textTopRow }
  begin
  if Log_includes(oplog) then writeln(logfile,'Wd_Move');
  result:= Proper(screenTopRow,screenBotRow);
  if result = Wd_ok then begin
    AllRows(rowFree,win.screenTopRow,win.screenBotRow);  { optimistic }
    ctop:=win.screenTopRow;
    if screenTopRow < ctop then ctop:=screenTopRow;
    cbot:=win.screenBotRow;
    if screenBotRow > cbot then cbot:=screenBotRow;
    if AnyRow(rowUsed,ctop,cbot) then begin
      result:=Wd_overlap;
      AllRows(rowUsed,win.screenTopRow,win.screenBotRow);  { we told you }
      end
    else begin
      newTextTop:=win.textTopRow + screenTopRow - win.screenTopRow;
      OpenRows(fate[win.screenTopRow].textLeftCol.current,
               screenTopRow,win.screenTopRow-1);
      OpenRows(fate[win.screenBotRow].textLeftCol.current,
               win.screenBotRow+1,screenBotRow);
      win.screenTopRow    := screenTopRow;
      win.screenBotRow    := screenBotRow;
      win.rows            := screenBotRow - screenTopRow + 1;
      win.textTopRow      := newTextTop;
      win.topDirt         := screenTopRow;  { may be pessimistic }
      win.cached.screenRow:= screenTopRow;  { to keep it in range }
      AllRows(rowUsed,screenTopRow,screenBotRow);
      if newTextTop < 1 then Wd_ScrollV(win,1);
      result:=Wd_ok;
      end;
    end;
  end {Wd_Move};

                    
procedure Wd_Split (var {io} wholeWin : Window;
                             upperRows: integer;
                             spaceRows: integer;
                    var { o} lowerWin : Window;
                    var { o} result   : Wd_result);

  var dist     : integer;  { distance from top of upper to top of lower }
      lowerRows: integer;  { number of rows in lower window }

  begin
  if Log_includes(oplog) then writeln(logfile,'Wd_Split');
  dist:=upperRows + spaceRows;
  lowerRows:=wholeWin.rows - dist;
  if      lowerRows < 1 then result:=Wd_norowsleft
  else if upperRows < 1 then result:=Wd_norows
  else if spaceRows < 0 then result:=Wd_negspace
  else begin
    with lowerWin do begin
      textTopRow      := wholeWin.textTopRow   + dist;
      screenTopRow    := wholeWin.screenTopRow + dist;
      screenBotRow    := wholeWin.screenBotRow;
      rows            := lowerRows;
      cols            := wholeWin.cols;
      band            := wholeWin.band;
      if wholeWin.topDirt < screenTopRow
      then topDirt    := screenTopRow
      else topDirt    := wholeWin.topDirt;
      cached.screenRow:= screenTopRow;
      end {lower};
    with wholeWin do begin
      rows            := upperRows;
      screenBotRow    := screenTopRow + upperRows -1;
      cached.screenRow:= screenTopRow;
      end {upper};
    AllRows(rowFree,wholeWin.screenBotRow+1,lowerWin.screenTopRow-1);
    result:=Wd_ok;
    end;
  end {Wd_Split};


procedure Wd_Close (var {i } win   : Window;
                    var { o} result: Wd_result);
  { Close (discard) the window. The corresponding screen rows
    are marked unused. (If some of them were already so marked,
    Wd_unused is returned.) The window variable cannot be used
    until it is reinitialized with Wd_Open. }
  begin
  if Log_includes(oplog) then
    writeln(logfile,'Wd_Close screenTopRow=',win.screenTopRow:1);
  if AnyRow(rowFree,win.screenTopRow,win.screenBotRow)
  then result:= Wd_unused
  else result:= Wd_ok;
  AllRows  (rowFree,win.screenTopRow,win.screenBotRow);
  end {Wd_Close};


procedure Wd_CloseAll;
  begin
  if Log_includes(oplog) then writeln(logfile,'Wd_CloseAll');
  AllRows(rowFree,1,Wd_MaxRows);
  end {Wd_CloseAll};


procedure Wd_Clean (var {io} win  : Window;
                    var { o} ops  : Wd_ScreenOps;
                    var { o} clean: boolean);
  var
    top     : integer;   { Top screen row of window }        { 93.2.12 }
    bot     : integer;   { Bottom screen row of window }
    topText : integer;   { textTopRow for the subwindow topDirt..bot }
    saved   : boolean;   { Found a non-deleted line }
    topRow  : integer;   { First non-deleted screen row >= topDirt }


  procedure DelRow (r: integer; var j: integer);
    { Delete line at row r.
      If, below the deleted row, there is a line that should not move up,
      a compensatory insertion is done before it, otherwise all lines
      down to and including bot are moved up. The number of the
      screen row inserted in compensation is returned in j. If no
      rows could be moved, j = r.
      Update fates.
      May use all of ops.deleteRow, ops.insertRow, and ops.WriteFull. }
    var
      k    : integer;   { Adjacent screen rows, j=k-1 }
      s,c  : integer;   { New 'current' screen row for inserted line }
      done : boolean;   { Row adjustments done. }
    begin
    with win do begin
      clean:=false;
      ops.deleteRow:=r;
      { Scroll following rows up one step, until the bottom of the
        window, or until we find a row that should not be moved up. }
      j:=r;
      s:=fate[j].screenRow.current;
      done:=false;
      repeat
        k:=j+1;
        if k > bot then done:=true
        else
        if fate[k].deleted then begin
          fate[j]:=fate[k];
          j:=k;
          end
        else begin
          c:=fate[k].screenRow.current;
          if c < k then begin
            { Row k can be moved up. }
            fate[j]:=fate[k];
            j:=k;
            s:=c+1;
            end
          else begin
            { Row k should not be moved up. }
            done:=true;
            if j > r then ops.insertRow:=j  { some could be moved }
                     else ops.deleteRow:=0; { none could be moved }
           end;
          end;
      until done;

      { Rewrite row j. }
      NinFate(fate[j]);
      with fate[j] do begin
        if s >= top then screenRow.current:= s    { still visible } { 93.2.12 }
                    else screenRow.current:= top; { skip dropoffs } { 93.2.12 }
        deleted:=false;
        with textLeftCol do past:=current;
        end;
      with ops.writeFull do begin
        screenRow:=j;
        textRow  :=TextRowAt(fate[j].screenRow  .current,win);
        textCol  :=          fate[j].textLeftCol.current;
        end;
      end;
    end {DelRow};

  procedure InsRow (r: integer);
    { Insert line at row r.
      If, below r, there is a line that should not move down,
      a compensatory deletion is done before it, otherwise all lines
      down to and including bot are moved down, and the line at bot
      falls off the screen.
      Update fates. fate[r] will need redefining.
      May use both ops.deleteRow and ops.InsertRow. }
    var
      j,k        : integer;     { Two adjacent screen rows, j=k-1 }
      fatek,fatej: RowFate;     { Fate copies for j and k }
      done       : boolean;     { Row adjustments done }
    begin
    with win do begin
      { Insert a line to move row j down. }
      clean:=false;
      ops.insertRow:=r;
      { Move following rows one step down, until the bottom of the
        window, or until a row that should not move down is found.
        In the latter case, we delete one row (to compensate for
        the insertion) and stop moving. (The deletion will in fact
        take place before the insertion, thus no text is lost from
        the bottom of the window.) }
      j:=r; fatej:=fate[j];
      NinFate(fate[j]);
      done:=false;
      repeat
        k:=j+1;        
        if k > bot then done:=true
        else begin
          fatek:=fate[k];
          with fatek do
            if not deleted then done:= screenRow.current <= k;
          if done then ops.deleteRow:=j   { fatej.deleted is assured }
          else begin
            fate[k]:=fatej;
            j:=k; fatej:=fatek;
            end;
          end;
      until done;
      end;
    end {InsRow};

  procedure InsRowTop;
    { Rows before topDirt are clean.
      Insert new row at topDirt, to remain there. }
    begin
    with win do begin
      InsRow(topDirt);
      with ops.writeFull do begin
        screenRow:=topDirt;
        textRow  :=topText;
        textCol  :=fate[topDirt].textLeftCol.current;
        end;
      with fate[topDirt] do begin
        screenRow.current:=topDirt;
        with textLeftCol do past:=current;
        end;
      topDirt:=topDirt+1;
      end;
    end {InsRowTop};

  procedure InsRowBot;
    { Rows before topDirt are clean.
      Insert new row at topDirt, to lie just before the final
      position of the current topDirt row (except if that would
      be > bot, in which case the new line will lie at bot.) }
    begin
    with win do begin
      InsRow(topDirt);
      with fate[topDirt] do begin
        with screenRow do begin
          current:=current-1;
          if current>bot then current:=bot;
          end;
        with textLeftCol do past:=current;
        end;
      with ops.writeFull do begin
        screenRow:=topDirt;
        textRow  :=TextRowAt(fate[topDirt].screenRow.current,win);
        textCol  :=fate[topDirt].textLeftCol.current;
        end;
      if fate[topDirt].screenRow.current = topDirt then
        topDirt:=topDirt+1;
      end;
    end {InsRowBot};

  procedure FillRow;
    { Rows before topDirt are clean.
      Row topDirt is deleted, and needs to be filled. }
    begin
    with win do begin
      clean:=false;
      with ops.writeFull do begin
        screenRow:=topDirt;
        textRow  :=topText;
        textCol  :=fate[topDirt].textLeftCol.current;  { or win.leftCol }
        end;
      with fate[topDirt] do begin
        deleted:=false;
        screenRow.current:=topDirt;
        with textLeftCol do past:=current;
        end;
      topDirt:=topDirt+1;  { this row is now clean }
      end
    end {FillRow};

  procedure FillRest;
    { All lines topDirt..bot were deleted from the screen.
      Scroll in a line from next page. }
    begin
    with win do begin
      ops.deleteRow:=topDirt;
      with ops.writeFull do begin
        screenRow:=bot;
        textRow  :=topText;
        textCol  :=fate[topDirt].textLeftCol.current;
        end;
      clean:=false;
      with fate[bot] do begin
        deleted:=false;
        screenRow.current:=topDirt;
        with textLeftCol do past:=current;
        end;
      NinFate(fate[bot]);
      end;
    end {FillRest};

  procedure DownRow;
    { Rows before topDirt are clean.
      Rows from topDirt up to but not including topRow are deleted.
      Row topRow moves down on screen (screenRow.current > topRow).
    }
    begin
    with win do begin
      if topDirt < topRow then FillRow
      else
      { topDirt = topRow }
      { 93.2.12 added: }
      if fate[topDirt].screenRow.current > bot then
        FillRow { The rows from topDirt down will go off the bottom of
                  the screen. We do not bother to scroll them out, just
                  fill in the new rows that replace them. }
      else
      { end additions 93.2.12 }
      if topDirt > screenTopRow then InsRowTop
      else                           InsRowBot;
      end;
    end {DownRow};
       
  procedure UpRow;
    { Rows before topDirt are clean.
      Rows from topDirt up to but not including topRow are deleted, if
      there are any such rows.
      Row topRow moves up on screen (screenRow.current < topRow).
      Note that topRow may be screenTopRow, in which case row topRow
      moves off window.
    }
    var d,j: integer;
    begin
    with win do begin
      if topDirt < topRow then d:=topDirt
                          else d:=topDirt-1;
      if d < screenTopRow then d:=screenTopRow;
      DelRow(d,j);
      if j = topDirt then topDirt:=topDirt+1;  { this row is now clean }
      end;
    end {UpRow};

  procedure WithinRow;
    { Screen row topDirt is at correct vertical position, but
      perhaps not at correct horizontal position.
      Perform within-row edits and refresh, if needed.
      Note, this is done in the old horizontal position. }
    var sc,ec,tc,cs: integer;

    procedure InTop;
      begin
      clean:=false;
      with ops.withinRow do begin
        screenRow:=win.topDirt;
        textRow  :=    topText;
        end;
      end {InTop};

    begin
    with win,fate[topDirt] do begin

      if insdel.chars <> 0 then begin
        sc:=insdel.textCol-textLeftCol.past+1;
        if sc <= cols then begin
          InTop;
          if sc >= 1 then ec:=sc
                     else ec:=1;
          cs:=cols-ec+1;  { max number that can sensibly be ins or del }
          if insdel.chars > 0 then begin
            if cs > insdel.chars then cs:=insdel.chars;
            end
          else begin
            cs:=-cs;
            if cs < insdel.chars then cs:=insdel.chars;
            end;
          tc:=insdel.textCol + (ec-sc);
          with ops.withinRow.insdel do begin
            chars    :=cs;
            screenCol:=ec;
            textCol  :=tc;
            end;
          if cs < 0 then begin
            { Fill end of line. }
            tc:=textLeftCol.past + cols + cs;
            if tc < fresh.textCol then fresh.textCol:=tc;
            end;
          end {within visible part};
        insdel.chars:=0;
        end {ins or del};

      if fresh.textCol < maxint then begin
        sc:=fresh.textCol - textLeftCol.past + 1;
        if sc <= cols then begin
          InTop;
          if sc >= 1 then ec:=sc
                     else ec:=1;
          cs:=cols-ec+1;
          tc:=fresh.textCol + (ec-sc);
          with ops.withinRow.overstrike do begin
            chars    :=cs;
            screenCol:=sc;
            textCol  :=tc;
            end;
          end;
        fresh.textCol:=maxint;
        end {fresh};
      end;
    end {WithinRow};

  procedure HorRow;
    { Screen row topDirt is at correct vertical position.
      Perform horizontal scroll if needed. }
    var dc: integer;
    begin
    with win,fate[topDirt] do begin
      dc:=textLeftCol.past - textLeftCol.current;
      if dc <> 0 then begin
        with ops.withinRow do begin
          screenRow:=topDirt;
          textRow  :=topText;
          end;
        if abs(dc) < cols then { Some columns can be saved on screen. }
          with ops.withinRow.insdel do begin
            chars    :=dc;     { insert or delete, depending on sign }
            screenCol:=1;
            textCol  :=textLeftCol.current;
            end
        else { Redisplay whole line using rightMargin. }
          dc:=-cols;
        if dc < 0 then
          with ops.withinRow.overstrike do begin
            chars    :=-dc;
            screenCol:=cols-chars+1;
            textCol  :=textLeftCol.current+screenCol-1;
            end;
        with textLeftCol do past:=current;
        clean:=false;
        end {diff col};
      topDirt:=topDirt+1;  { all clean now }
      end;
    end {HorRow};

  procedure SavedRow;
    { Rows before topDirt are clean.
      Rows from topDirt up to, but not including topRow were deleted.
      Row topRow is not deleted, but it may have moved to another screen
      row (perhaps even off screen) and it may have scrolled horizontally. }
    begin
    with win,fate[topRow] do begin
      if      screenRow.current > topRow then DownRow
      else if screenRow.current < topRow then UpRow
      else if      topDirt      < topRow then FillRow
      else if (insdel.chars  <> 0)
           or (fresh.textCol <  maxint)  then WithinRow
      else                                    HorRow;
      end;
    end {SavedRow};

  begin { Wd_Clean }
  clean:=true;  { unless some dirt is found below }
  with win do begin
    top:=screenTopRow;                             { 93.2.12 }
    bot:=screenBotRow;
    if topDirt <= bot then begin

      with ops do begin { Set no operation }
        deleteRow          :=0;
        insertRow          :=0;
        writeFull.screenRow:=0;
        withinRow.screenRow:=0;
        withinRow.insdel    .chars:=0;
        withinRow.overstrike.chars:=0;
        end;

      repeat
        { Find some non-deleted line }
        saved  :=false;
        topRow :=topDirt;
        topText:=TextRowAt(topDirt,win);
        repeat if fate[topRow].deleted then topRow:=topRow+1
                                       else saved:=true;
        until saved or (topRow > bot);
        if saved then SavedRow
                 else FillRest;
      until (topDirt > bot) or (not clean);

      end {some dirt};
    end;
  end {Wd_Clean};


procedure Wd_ShowV (var {i } win: Window;
                             srow,scol: integer);
  var
    i: integer;
  begin
  with win do
    for i:=screenTopRow to screenBotRow do begin
      GotoXY(scol,i-screenTopRow+srow+1);
      if fate[i].deleted then write('  /')
                         else write(fate[i].screenRow.current:3);
      if i = topDirt then write('*')
                     else write(' ');
      end;
  end {Wd_ShowV};


private
procedure ShowHorFate (var f: text; var {i} fate: RowFate);
  { Show a row fate as regards horizontal movements. }
  begin
  with fate do begin
    if deleted then write(f,' /')
    else begin
      with textLeftCol do begin
        write(f,current:2);
        if      current < past then write(f,'_"')
        else if current > past then write(f,'"_')
                               else write(f,'  ');
        end;
      if insdel.chars <> 0 then begin
        write(f,'@',insdel.textCol:1);
        if insdel.chars > 0 then write(f,'+',insdel.chars:1)
                            else write(f,    insdel.chars:1);
        end;
      if fresh.textCol < maxint then write(f,'/',fresh.textCol:1);
      end {not deleted};
    end {with};
  end {ShowHorFate};


procedure Wd_ShowH (var {i } win: Window;
                             srow,scol: integer);
  var 
    i: integer;
  begin
  with win do
    for i:=screenTopRow to screenBotRow do begin
      GotoXY(scol,i-screenTopRow+srow+1);
      ShowHorFate(output,fate[i]);
      end {for i};
  end {Wd_ShowH};


procedure Wd_ShowWin (var f: text; var {i } win: Window);
  var 
    i: integer;
  begin
  writeln(f,'Window data:');
  with win do begin
    writeln(f,'  screenTopRow=',screenTopRow:1,
              '  textTopRow=',textTopRow:1,
              '  rows=',rows:1,'  cols=',cols:1,'  band=',band:1);
    
    for i:=screenTopRow to screenBotRow do begin
      write(f,i:6);
      if i = topDirt then write(f,'*  ')
                     else write(f,'   ');
      if fate[i].deleted then write(f,'/')
      else begin
        write(f,fate[i].screenRow.current:4);
        ShowHorFate(f,fate[i]);
        end;
      writeln(f);
      end {i};
    end {with};
  end {Wd_ShowWin};


procedure Wd_Init;
  begin
  { writeln('Windows vs. 91.04.14.A'); }

  Wd_CloseAll;

  end {Wd_Init};


procedure Wd_Fini;
  begin
  end {Wd_Fini};
