{ windows_f.p
  External declarations for the routines in windows.p
}

procedure Wd_Explain (var f: text; result: Wd_result);
  { Write a short explanation of the result code. }
  external;

procedure Wd_Open (var win: Window;
                       screenTopRow: integer;
                       screenBotRow: integer;
                       TopRow      : integer;
                       LeftCol     : integer;
                       cols        : integer;
                       band        : integer;
                   var result      : Wd_result);
  { Open a window with given parameters. TopRow and LeftCol are
    the topmost text row and leftmost text column to be visible.
    The top and bottom screen rows, the number of columns, and the
    'band' for horizontal scrolling are also specified.
    If the window as specified would extend beyond the screen limits,
    Wd_offscreen is returned and the window is not opened.
    If some of the screen rows requested are already in use by other
    windows, Wd_overlap is returned and the window is not opened. }
  external;

procedure Wd_Move  (var {io} win         : Window;
                             screenTopRow: integer;
                             screenBotRow: integer;
                    var { o} result      : Wd_result);
  { Move and/or resize window on screen. The new top and bottom screen
    rows are given. All screen rows covered by the
    convex union of the old and new window screen areas must be
    free (or used by this window), otherwise Wd_overlap is returned.
    (Window reconfigurations where this condition is not met must be
    implemented by deleting and recreating windows.)
    The text position of the reshaped window is chosen so that the
    existing screen rows are not moved, if possible. }
  external;
                    
procedure Wd_Split (var {io} wholeWin : Window;
                             upperRows: integer;
                             spaceRows: integer;
                    var { o} lowerWin : Window;
                    var { o} result   : Wd_result);
  { Split window wholeWin into an upper part (returned in wholeWin)
    and a lower part (returned in lowerWin). The upper part is
    given upperRows screen rows, followed by spaceRows that are left
    unused, and the lower window is given the remaining screen rows
    covered by the original window. If upperRows + spaceRows is so
    large that nothing is left for the lower window, Wd_norowsleft
    is returned and the window is not split.
    Both windows continue to show their respective parts of the
    text. Nothing is done to the spaceRows. }
  external;

procedure Wd_Close (var {i } win   : Window;
                    var { o} result: Wd_result);
  { Close (discard) the window. The corresponding screen rows
    are marked unused. (If some of them were already so marked,
    Wd_unused is returned.) The window variable cannot be used
    until it is reinitialized with Wd_Open. }
  external;

procedure Wd_CloseAll;
  { Close all windows by marking all screen rows as unused.
    No window variable can be used until it is reinitialized
    with Wd_Open. }
  external;

function  Wd_VisRow (var {in} win: Window;
                              row: integer): boolean;
  { Whether the window covers the given text row. }
  external;

function  Wd_VisCol (var {in} win: Window;
                              screenRow: integer;
                              col      : integer): boolean;
  { Whether the indicated row of the window covers the given text column. }
  external;

function Wd_TopRow (var {in} win: Window): integer;
  { Number of topmost visible text row in window. }
  external;

function Wd_LeftCol (var {in} win: Window;
                              textRow: integer): integer;
  { Number of leftmost visible text column of the indicated text row
    in the window. If the row is not visible at all, return 0. }
  external;

procedure Wd_ScrollV (var win   : Window;
                          topRow: integer);
  { Scroll window vertically to reach the given topmost text line. }
  external;

procedure Wd_ScrollH (var win    : Window;
                          textRow: integer;
                          leftCol: integer);
  { Scroll given text row (and surrounding band) horizontally to reach
    the given leftmost text column.
    Note, this routine may have no effect if the given text row is not
    now on the screen, before any cleaning operations. To ensure that
    it has effect, clean the window first. }
  external;

procedure Wd_SeeRow (var {io} win    : Window;
                              textRow: integer);
  { Scroll window to cover the given text row.
    No effect if window already covers it.
    Least scrolling step is Wd_Vscroll lines, but limited
    also by window size, e.g. a 1-line window can always
    be scrolled in 1-line steps. }
  external;

procedure Wd_SeeCol (var {io} win    : Window;
                              textRow: integer;
                              textCol: integer);
  { Scroll given text row (and surrounding band) to cover the given column.
    No effect if row already covers it.
    Least scrolling step is Wd_Hscroll columns, but limited
    also by window width, e.g. a 5-column window can always be
    scrolled in 5-column steps.
    Note, this routine may have no effect if the given text row is
    not on screen now, before any cleaning. To ensure that it has
    effect, clean the window first. }
  external;

procedure Wd_InsertLine (var win: Window;
                             textRow: integer;
                             num    : integer);
  { Insert a number of lines before the given text row. }
  external;

procedure Wd_DeleteLine (var win    : Window;
                             textRow: integer;
                             num    : integer);
  { Delete a number of lines starting with the given text row. }
  external;

procedure Wd_DeleteChar (var win: Window;
                             textRow,textCol: integer;
                             num            : integer);
  { Delete a number of characters starting with the indicated
    row-column position.
    Note, the column is really that of the first character deleted,
    and not the column to the right of it. }
  external;

procedure Wd_InsertChar (var win: Window;
                             textRow,textCol: integer;
                             num            : integer);
  { Insert a number of characters on the given text row, at the
    given column. The rest of the line moves right. }
  external;

procedure Wd_BreakLine  (var win: Window;
                             textRow,textCol: integer);
  { Insert a line break in the given text row, at (just before) the
    given column. The rest of the line is inserted below the given
    row, as a line of its own. }
  external;

procedure Wd_FreshLine (var win: Window;
                            textRow: integer;
                            textCol: integer);
  { Inform the window that the given text row needs to be redisplayed
    from the given column to end of line. }
  external;

procedure Wd_FreshAll (var win: Window);
  { Inform the window that it needs to be completely redisplayed.
    As always, the actual display will be done with the help of
    Wd_Clean. }
  external;

procedure Wd_Clean (var {io} win  : Window;
                    var { o} ops  : Wd_ScreenOps;
                    var { o} clean: boolean);
  { Update display of window.
    Implements line deletion, insertion, and scrolling operations.
    Returns necessary screen operations in the ops parameter, or
    signals that the display is up to date by returning clean = true.
    May have to be called many times, in the form
         repeat
            Wd_Clean (win,ops,clean);
            if not clean then implement ops on screen;
            if (user interrupt) then exit loop;
         until clean (or user interrupt)
    See comments for Wd_ScreenOps for more information.
  }
  external;


{** Testing routines **}

procedure Wd_ShowV (var {i } win: Window;
                             srow,scol: integer);
  { List window "vertical" (i.e. row number) information on screen
    in vertical format starting at the indicated screen row and column. }
  external;

procedure Wd_ShowH (var {i } win: Window;
                             srow,scol: integer);
  { List window "horizontal" (i.e. within-row edits) information on screen
    in vertical format starting at the indicated screen row and column. }
  external;

procedure Wd_ShowWin (var f: text; var {i } win: Window);
  { List window information on text output file. }
  external;

procedure Wd_Init;
  external;

procedure Wd_Fini;
  external;
