{  Calculator program for incremental (reversible) script
   execution under a SEED-like session editor. }

program Calcul (input,output);

#include "seed_p.h"

const
  Mem_Size  = 200;

type

  SymbolType       = (EndSymbol);
  numeric          = Lex_RealType;
  Mem_Index        = 0..Mem_Size;
  Mem_PositionType = Mem_Index;

  PositionType = record
                 LexPos: Lex_PositionType;
                 MemPos: Mem_PositionType;
                 end;

var

{ Mem package }

  Mem: record
       Names : array [Mem_Index] of char;  { loc [0] is dummy }
       Values: array [Mem_Index] of numeric;
       Top   : Mem_PositionType;
       end {Mem};


procedure Calculation;
  { Parses a sequence of lines, each holding one expression,
    preceded by a name (a letter) and an equals sign. }
  forward;

procedure Expression (var value: numeric);
  { Parses an expression and returns its value. }
  forward;

procedure MoreTerms (before: numeric; var value: numeric);
  { Adds or subtracts more terms to an expression of value
    'before', resulting in 'value'. }
  forward;

procedure Term (var value: numeric);
  { Parses a term, result in 'value'. }
  forward;

procedure MoreFactors (before: numeric; var value: numeric);
  { Multiplies or divides more factors into a term of value
    'before', resulting in 'value'. }
  forward;

procedure Factor (var value: numeric);
  { Parses a factor and returns its value. }
  forward;

function FollowMoreTerms: boolean;
  { Whether the current token can follow a MoreTerms structure. }
  forward;

function FollowMoreFactors: boolean;
  { Whether the current token can follow a MoreFactors structure. }
  forward;




{ Mem package }

procedure Mem_Define (Name: char; Value: numeric);
  { Store 'Name=Value' in the memory. }
  begin
  with Mem do
    if Top < Mem_Size then begin
      Top:= Top + 1;
      Names [Top]:=Name;
      Values[Top]:=Value;
      {
      Ed_write('Defined ');
      Ed_write('"' + Name + '"'));
      Ed_writeln;
      }
      end
    else begin
      Ed_write('The calculator storage is full.'); Ed_writeln;
      Lex_Error;
      end;
  end {Mem_Define};

function Mem_Defined (Name: char): Mem_Index;
  { The index of the Name in memory. Zero if undefined. }
  var i: Mem_Index;
  begin
  with Mem do begin
    Names[0]:=Name;  { loop stopper }
    i:=Top;
    while (Names[i] <> Name) do i:=i-1;
    Mem_Defined:=i;
    end;
  end {Mem_Defined};

procedure Mem_Value (Name: char; var Value: numeric);
  { Value defined for the Name. Error if Name is undefined. }
  var i: Mem_Index;
  begin
  with Mem do begin
    i:=Mem_Defined(Name);
    if i > 0 then Value:=Values[i]
    else begin
      Ed_write('The variable "' + Name + '" is not defined.');
           Ed_writeln;
      if Top > 0 then begin
        Ed_write('Variables defined:');
        for i:=1 to Top do Ed_write(' ' + Names[i]);
        Ed_writeln;
        end;
      Lex_Error;
      end;
    end;
  end {Mem_Value};

procedure Mem_Store (var P: Mem_PositionType);
  begin
  P:=Mem.Top;
  end {Mem_Store};

procedure Mem_Restore (P: Mem_PositionType);
  begin
  Mem.Top:=P;
  end {Mem_Restore};




{ Combined position-type Store and Restore }

procedure Pos_Store (var P: PositionType);
  begin
  Lex_Store(P.LexPos);
  Mem_Store(P.MemPos);
  end {Pos_Store};

procedure Pos_Restore (P: PositionType);
  begin
  Lex_Restore(P.LexPos);
  Mem_Restore(P.MemPos);
  end {Pos_Restore};



{ Grammar }

procedure SkipBlankLines;
  begin
  Lex_Get;
  while (Lex_token.kind = EolToken) do Lex_Next;
  end {SkipBlankLines};

procedure Err (mess: vstring);
  begin
  Ed_write(mess);
  Ed_writeln;
  Lex_Error;
  end {Err};

procedure Calculation;
  label 66;
  var
    here : PositionType;
    name : vstring;
    value: numeric;
  begin
  Pos_Store(here);
  repeat
    SkipBlankLines;                         if Lex_cancel then goto 66;
    if Lex_isSymbol(ord(EndSymbol)) then begin
      Lex_exit; goto 66;
      end;
    if Lex_token.kind <> WordToken then begin
      Err('Each expression must be named with a letter (A..Z)');
      goto 66;
      end;
    name:=Lex_token.word;
    if length(name) > 1 then begin
      Err('Please use only 1-letter variable names.');
      goto 66;
      end;
    if Mem_Defined(Name[1]) <> 0 then begin
      Err('This variable name was defined earlier.');
      goto 66;
      end;
    Lex_Next;          if Lex_cancel then goto 66;
    if not Lex_isMark('=') then begin
      Err('Please type an equal sign (=) after the variable name.');
      goto 66;
      end;
    Lex_Next;          if Lex_cancel then goto 66;
    Expression(value); if Lex_cancel then goto 66;
    if Lex_token.kind = EolToken then
      { ok }
    else if Lex_isMark(';') then
      { ok }
    else begin
      Err('Use ";" between formulae within a line.');
      goto 66;
      end;
    Mem_Define(name[1],value);
    Ed_write('    ');
    Ed_write(name); Ed_write(' = '); Ed_wrreal(value,10,6);
    Ed_writeln;
    Lex_Null;         if Lex_cancel then goto 66;
    Calculation;      if Lex_cancel then goto 66;
    Lex_exit;
  66: if Lex_cancel then Pos_Restore(here);
  until Lex_done;
  end {Calculation};

procedure Expression {var value: numeric};
  label 66;
  var
    here: PositionType;
    s,t : numeric;
  begin
  Pos_Store(here);
  repeat
    if Lex_isMark('+') then begin s:=+1.0; Lex_Next; end
    else
    if Lex_isMark('-') then begin s:=-1.0; Lex_Next; end
    else                          s:=+1.0;
                             if Lex_cancel then goto 66;
    Term(t);                 if Lex_cancel then goto 66;
    MoreTerms(s*t,value);    if Lex_cancel then goto 66;
    Lex_exit;
  66: if Lex_cancel then Pos_Restore(here);
  until Lex_done;
  end {Expression};

procedure MoreTerms {before: numeric; var value: numeric};
  label 66;
  var
    here: PositionType;
    t   : numeric;
  begin
  Pos_Store(here);
  repeat
    if Lex_isMark('+') then begin
      Lex_Next;                       if Lex_cancel then goto 66;
      Term(t);                        if Lex_cancel then goto 66;
      MoreTerms(before+t,value);      if Lex_cancel then goto 66;
      end
    else
    if Lex_isMark('-') then begin
      Lex_Next;                       if Lex_cancel then goto 66;
      Term(t);                        if Lex_cancel then goto 66;
      MoreTerms(before-t,value);      if Lex_cancel then goto 66;
      end
    else
    if FollowMoreTerms then value:=before
    else begin
      Err('Use "+" or "-" to add or subtract more terms.');
      goto 66;
      end;
    Lex_exit;
  66: if Lex_cancel then Pos_Restore(here);
  until Lex_done;
  end {MoreTerms};

procedure Term {var value: numeric};
  label 66;
  var
    here: PositionType;
    f   : numeric;
  begin
  Pos_Store(here);
  repeat
    Factor(f);                if Lex_cancel then goto 66;
    MoreFactors(f,value);     if Lex_cancel then goto 66;
    Lex_exit;
  66: if Lex_cancel then Pos_Restore(here);
  until Lex_done;
  end {Term};

procedure MoreFactors {before: numeric; var value: numeric};
  label 66;
  var
    here: PositionType;
    f   : numeric;
  begin
  Pos_Store(here);
  repeat
    if Lex_isMark('*') then begin
      Lex_Next;                     if Lex_cancel then goto 66;
      Factor(f);                    if Lex_cancel then goto 66;
      MoreFactors(before*f,value);  if Lex_cancel then goto 66;
      end
    else
    if Lex_isMark('/') then begin
      Lex_Next;                     if Lex_cancel then goto 66;
      Factor(f);                    if Lex_cancel then goto 66;
      if f = 0.0 then begin
        Err('Divisor should not be zero.');
        goto 66;
        end;
      MoreFactors(before/f,value);  if Lex_cancel then goto 66;
      end
    else
    if FollowMoreFactors then value:=before
    else begin
      Err('Use "*" or "/" to multiply or divide with more factors.');
      goto 66;
      end;
    Lex_exit;
  66: if Lex_cancel then Pos_Restore(here);
  until Lex_done;
  end {MoreFactors};

procedure Factor {var value: numeric};
  label 66;
  var
    here: PositionType;
    name: vstring;
  begin
  Pos_Store(here);
  repeat
    if Lex_token.kind = NatNumToken then begin
      value:=Lex_token.natnumber;
      Lex_Next;                          if Lex_cancel then goto 66;
      end
    else
    if Lex_token.kind = RealNumToken then begin
      value:=Lex_token.realnumber;
      Lex_Next;                          if Lex_cancel then goto 66;
      end
    else
    if Lex_token.kind = WordToken then begin
      name:=Lex_token.word;
      if length(name) > 1 then begin
        Err('Variable names can be only one letter long.');
        goto 66;
        end;
      Mem_Value(name[1],value);         if Lex_cancel then goto 66;
      Lex_Next;                         if Lex_cancel then goto 66;
      end
    else
    if Lex_isMark('(') then begin
      Lex_Next;                         if Lex_cancel then goto 66;
      Expression(value);                if Lex_cancel then goto 66;
      if not Lex_isMark(')') then begin
        Err('A closing parenthesis seems to be missing.');
        goto 66;
        end;
      Lex_Next;                         if Lex_cancel then goto 66;
      end
    else begin
      Ed_write('A factor should be a number, a variable, or');
      Err     (' a parenthesized expression.');
      goto 66;
      end;
    Lex_exit;
  66: if Lex_cancel then Pos_Restore(here);
  until Lex_done;
  end {Factor};


function FollowMoreTerms {: boolean};
  begin
  if Lex_token.kind = EolToken then FollowMoreTerms:= true
                               else FollowMoreTerms:= Lex_isMark(')')
                                                   or Lex_isMark(';');
  end {FollowMoreTerms};

function FollowMoreFactors {: boolean};
  begin
  if Lex_isMark('+') or Lex_isMark('-') then FollowMoreFactors:=true
                               else FollowMoreFactors:=FollowMoreTerms;
  end {FollowMoreFactors};


procedure Calculations;
  { A "starting symbol" to store the initial state. }
  label 66;
  var
    here: PositionType;
  begin
  Pos_Store(here);
  repeat
    Lex_Next;      if Lex_cancel then goto 66;
    Calculation;   if Lex_cancel then goto 66;
    Lex_exit;
  66: if Lex_cancel then Pos_Restore(here);
  until Lex_done;
  end {Calculations};


procedure CalcInit;
  begin
  Lex_Define('end',ord(EndSymbol));
  Lex_Define('bye',ord(EndSymbol));
  Lex_Ignore(EolCommentToken);
  Mem.Top:=0;        
  end {CalcInit};

begin { Calcul }

Seed_Init;

Ed_write('Calcul (Pascal) 1994.10.24A'); Ed_writeln;
Ed_write('Type in a formula, such as x = 2+2. Press Enter. Edit at will.');
                                                             Ed_writeln;
Ed_write('For help, type a question mark and press Enter.'); Ed_writeln;
Ed_write('To quit, type "end" or ctrl-X ctrl-C.'); Ed_writeln;
Ed_writeln;

CalcInit;
Calculations;
Ed_write('End calcul.'); Ed_writeln;

Seed_Fini;

end {Calcul}.
