{ lexline_f.p
  External declarations for lexline.p
}

var
  Lex_token: external TokenType;    { the current token }

procedure Lex_Store (var safe: Lex_PositionType { ; tag: vstring } );
  { Store a checkpoint of Lex state in the output parameter.
    The (currently omitted) tag is for tracing and debugging purposes. }
  external;

procedure Lex_Next;
  { Sets token := next token, changes position. }
  external;

procedure Lex_Error;
  { Signals an error at this point in the script. }
  external;

procedure Lex_exit;
  { Signals that a grain has completed parsing with success. }
  external;

function Lex_done: boolean;
  { Returns "true" if the current grain should be terminated 
    (with success or cancelled). Returns false if the current grain
    should be reprocessed. }
  external;

function Lex_cancel: boolean;
  { Whether the preceding Lex call "raised" the cancelled "exception". }
  external;

procedure Lex_Restore (safe: Lex_PositionType);
  { Tries to restore the state at "safe". If it succeeds,
    resets Lex.cancelled and Lex.exit to false ("exception handled").
    If it fails, sets them true ("exception reraised"). }
  external;

procedure Lex_Null;
  { Consume the current token. Lext_Get will fetch the next one. }
  external;

procedure Lex_Get;
  { If there is no current token, fetch the next one. }
  external;

function Lex_isMark (mark: char): boolean;
  { Is the token a given single-character special symbol? }
  external;

function Lex_isSymbol (num: integer): boolean;
  { Is the token an alphameric symbol with this number? }
  external;

procedure Lex_Int (var N: integer);
  { Get an integer number from the input.
    Return value in N, or raise cancelled exception.
    Lexical errors are handled internally. }
  external;

procedure Lex_Scan (var eol: boolean);
  { Skip blanks in input, stopping at end of line or first
    non-blank character. Which condition occurred is shown by
    the eol (= end of line) output parameter.
    The scan starts from the first position after the current
    (or just discarded) token.
    If there is a current token, it is discarded.
    If the scan stopped at a non-blank character, the next Lex_Get
    will look for a token starting with this character, while if
    Lex_Scan is called again, it does nothing.
    If the scan stopped at end of line, the next Lex_Get will
    get the first token on the next line, while if Lex_Scan is
    called again, it will scan the next input line.
    This routine is useful as a preliminary to Lex_Line. }
  external;

procedure Lex_Line (var line  : Lex_LineType;
                    var length: integer    );
  { Get rest of current input line, without lexical analysis.
    Returns text of line and its length. The current token (if
    any) is consumed. The next Lex_Get will fetch the first token
    on the next line. }
  external;


{ The following routines are intended for configuring the
  lexical analyzer. Their effects are permanent and are not
  part of the Store/Restore state. }

procedure Lex_Define (    word  : Lex_wordType;
                          symnum: integer);
  { Define an alphanumeric string to be a reserved symbol, and
    assign its symbol number. }
  external;

procedure Lex_Ignore (kind: TokenKind);
  { Specifies that tokens of the given kind are to be ignored.
    If the analyzer finds such a token, it is discarded and the
    analyzer continues to the next token. }
  external;

procedure Lex_FirstChar (kind: TokenKind; op: Lex_CsOpType; chars: vstring);
  { Define the characters that can start a token of the given kind.
    The 'op' is one of '=', '+', '-' and the character set is
    correspondingly set equal to the given 'chars' or the given
    chars are added to or removed from the set. (The type string
    is used, rather than 'set of char', to ease calls from C.)
    Some token classes are considered equivalent, and changing the
    starter set for a token class changes the starter sets for all
    equivalent classes. The equivalent classes are
      - NatNumToken and RealNumToken
      - SymbolToken and WordToken.
    For numeric tokens, the numeric value is computed only from the
    'usual' characters (digits 0..9, decimal point, exponent denotation)
    in the token; all other characters (such as '_' for Ada-style
    numeric literals) are discarded. The number base is always 10.
  }
  external;

procedure Lex_BodyChar (kind: TokenKind; op: Lex_CsOpType; chars: vstring);
  { Define the characters that can appear in the body of a token of
    the given kind. The parameters correspond to those of Lex_FirstChar,
    and token class equivalence holds here, also.
    For EolCommentToken, the 'body' means just the second character; if
    a blank space can appear in the body, a comment can be introduced by
    any one character from the allowed first character set.
    Thus, Ada-style comments are specified by first = '-', body  = '-',
    and C-style comments by first = '#', body = ' '. }
  external;

procedure Lex_QuotePair (left,right: char);
  { Define a pair of characters that can enclose quoted strings.
    A quoted string starts with one of the characters defined as
    'first' characters for the token kind - call it the open-quote.
    If Lex_QuotePair has been called with the open-quote as the 'left'
    parameter, then the 'right' parameter is the correponding close-
    quote character; otherwise the close-quote is the same as the
    open-quote. The string ends at the following close-quote character.
    Between the open and close quotes are allowed any characters defined
    as 'body' characters for the token kind, except the close-quote
    Thus, for Forth-style strings, use Lex_QuotePair ('(',')').
    Note, you must also specify the open-quote as a possible first-char
    for the StringToken class. }
  external;

procedure Lex_BlankChar (op: Lex_CsOpType; chars: vstring);
  { Define the set of blank or whitespace characters. The parameters
    are interpreted as in Lex_FirstChar, above. }
  external;

procedure Lex_Init;
  { Initialize the LexLine module. }
  external;
