/* seed_lexline.h
  Constants, types, and routines for lexline.p
*/


#if !defined(SEED_LEXLINE)
#define SEED_LEXLINE

/* line length limit (buffered mode) */
#define LEX_MAXLINE ED_MAXLINE
/* word (alphameric string) length limit */
#define LEX_MAXWORD 80

/* Opcodes for char-set functions */
typedef char lex_csoptype;

#define lex_set     '='
#define lex_include '+'
#define lex_exclude '-'

typedef ed_linetype lex_linetype;

typedef struct {
	int  length;
	char text[LEX_MAXWORD];
	} lex_wordtype;
typedef double lex_realtype;

typedef struct {
	scriptindex last;	/* last script position used by program */
	scriptindex start;	/* start of current token (if > last,
				   there is no current token) */
	int         length;	/* length of current token (if any) */
	char        character;	/* read-ahead character */
     /* vstring     tag; */	/* for tracing and debugging. */
	} lex_positiontype;

  /* Note that a Lex_PositionType.last is the last position used by the
     application program, not the last position read from the editor.
     The last position read is the eol on the line last read. */

typedef char tokenkind;
#define	nulltoken	 0
#define	cancelsignal	 1
#define	natnumtoken	 2
#define	realnumtoken	 3
#define	marktoken	 4
#define	symboltoken	 5
#define	wordtoken	 6
#define stringtoken      7
#define	eoltoken	 8
#define	eolcommenttoken	 9
#define	eoftoken	10

typedef struct {
	tokenkind kind;
	union {
	   int		natnumber;
	   lex_realtype	realnumber;
	   char		character;
	   int		symnum;
	   lex_wordtype	word;
	   struct {
	   	vstring quoted;
	   	char opener,closer;
	        }	string;
	   /* void */   /* nulltoken, cancelsignal,
	                   eoltoken, eolcommenttoken, eoftoken */
	} value;
     } tokentype;

/*
  External declarations for lexline.p
*/

extern tokentype lex_token;    /* the current token */

extern void lex_store (lex_positiontype *safe);
  /* Store a checkpoint of Lex state in the output parameter. */

extern void lex_next ();
  /* Sets token := next token, changes position. */

extern void lex_error ();
  /* Signals an error at this point in the script. */

extern void lex_exit ();
  /* Signals that a grain has completed parsing with success. */

extern seed_bool lex_done ();
  /* Returns "true" if the current grain should be terminated 
    (with success or cancelled). Returns false if the current grain
    should be reprocessed. */

extern seed_bool lex_cancel ();
  /* Whether the preceding Lex call "raised" the cancelled "exception". */

extern void lex_restore (lex_positiontype safe);
  /* Tries to restore the state at "safe". If it succeeds,
    resets Lex.cancelled and Lex.exit to false ("exception handled").
    If it fails, sets them true ("exception reraised"). */

extern void lex_null ();
  /* Consume the current token. lex_get will fetch the next one. */

extern void lex_get ();
  /* If there is no current token, fetch the next one. */

extern seed_bool lex_ismark (char mark);
  /* Is the token a given single-character special symbol? */

extern seed_bool lex_issymbol (int num);
  /* Is the token an alphameric symbol with this number? */

extern void lex_int (int *N);
  /* Get an integer number from the input.
    Return value in N, or raise cancelled exception.
    Lexical errors are handled internally. */

extern void lex_scan (seed_bool *eol);
  /* Skip blanks in input, stopping at end of line or first
    non-blank character. Which condition occurred is shown by
    the eol (= end of line) output parameter.
    The scan starts from the first position after the current
    (or just discarded) token.
    If there is a current token, it is discarded.
    If the scan stopped at a non-blank character, the next lex_get
    will look for a token starting with this character, while if
    lex_scan is called again, it does nothing.
    If the scan stopped at end of line, the next lex_get will
    get the first token on the next line, while if lex_scan is
    called again, it will scan the next input line.
    This routine is useful as a preliminary to lex_line. */

extern void lex_line (lex_linetype *line,
                      int          *length);
  /* Get rest of current input line, without lexical analysis.
    Returns text of line and its length. The current token (if
    any) is consumed. The next lex_get will fetch the first token
    on the next line. */


/* The following routines are intended for configuring the
  lexical analyzer. Their effects are permanent and are not
  part of the Store/Restore state. */

extern void lex_define (lex_wordtype word,
                        int          symnum);
  /* Define an alphanumeric string to be a reserved symbol, and
    assign its symbol number. */

extern void lex_ignore (tokenkind kind);
  /* Specifies that tokens of the given kind are to be ignored.
    If the analyzer finds such a token, it is discarded and the
    analyzer continues to the next token. */

extern void lex_firstchar (tokenkind    kind,
                           lex_csoptype op,
                           vstring      chars);
  /* Define the characters that can start a token of the given kind.
    The 'op' is one of '=', '+', '-' and the character set is
    correspondingly set equal to the given 'chars' or the given
    chars are added to or removed from the set. (The type string
    is used, rather than 'set of char', to ease calls from C.)
    Some token classes are considered equivalent, and changing the
    starter set for a token class changes the starter sets for all
    equivalent classes. The equivalent classes are
      - NatNumToken and RealNumToken
      - SymbolToken and WordToken.
    For numeric tokens, the numeric value is computed only from the
    'usual' characters (digits 0..9, decimal point, exponent denotation)
    in the token; all other characters (such as '_' for Ada-style
    numeric literals) are discarded. The number base is always 10.
  */

extern void lex_bodychar (tokenkind    kind,
                          lex_csoptype op,
                          vstring      chars);
  /* Define the characters that can appear in the body of a token of
    the given kind. The parameters correspond to those of Lex_FirstChar,
    and token class equivalence holds here, also.
    For EolCommentToken, the 'body' means just the second character; if
    a blank space can appear in the body, a comment can be introduced by
    any one character from the allowed first character set.
    Thus, Ada-style comments are specified by first = '-', body  = '-',
    and C-style comments by first = '#', body = ' '. */

extern void lex_quotepair (char left, char right);
  /* Define a pair of characters that can enclose quoted strings.
    A quoted string starts with one of the characters defined as
    'first' characters for the token kind - call it the open-quote.
    If Lex_QuotePair has been called with the open-quote as the 'left'
    parameter, then the 'right' parameter is the correponding close-
    quote character; otherwise the close-quote is the same as the
    open-quote. The string ends at the following close-quote character.
    Between the open and close quotes are allowed any characters defined
    as 'body' characters for the token kind, except the close-quote
    Thus, for Forth-style strings, use Lex_QuotePair ('(',')').
    Note, you must also specify the open-quote as a possible first-char
    for the StringToken class. */

extern void lex_blankchar (lex_csoptype op, vstring chars);
  /* Define the set of blank or whitespace characters. The parameters
    are interpreted as in Lex_FirstChar, above. */

extern void lex_init ();
  /* Initialize the LexLine module. */


/* The following are in cseed.c */

lex_wordtype *lex_word_c (char *str);
lex_wordtype *LW         (char *str);	/* Obsolete, = lex_word_c */

#endif
