 /*                                                                           *\

    TELA - Testing Environment for Learning Algorithms - version 1.995
    ==================================================================

    Module:       Alg_Database

    Contents:     TELA algorithm database

    Description:  offers run-time query facilities about the 
                  characteristics of the learning algorithms that 
                  are embedded in to TELA. 

    Author: Juho Rousu (1995,1996)

    Services:

    Database maintenance
    ====================
  
    ADB_init_db        - initialize the algorithm database (ADB) 
                         memory structures
    ADB_read_file      - read the ADB into memory
    ADB_write_alg_decl - write an algorithm description to file, unused
    ADB_write_file     - write the ADB to the disk, unused for the moment
    ADB_alg_decl_init - initialize an algorithm description
    ADB_add_algorithm - add an algorithm description into ADB

    ADB_write_alg_decl - write an algorithm description to file, unused
    ADB_write_file     - write the ADB to the disk, unused for the moment

    Query Capabilities
    ==================

    ADB_get_alg_ref       - get a reference to an algorithm by name
    ADB_alg_decl_by_ref   - fetch an algoritm description by reference
     ADB_alg_decl_by_name  - fetch an algorithm descriotion by name             

    ADB_get_exe_name      - get the name of the learning program executable
    ADB_get_alg_name      - get the name of an algorithm
     ADB_get_separator     - fetch the string corresponding to given separator
    ADB_is_equiv_sep      - check if two separators are equivalent

    ADB_get_filename_stem - output a non-clashing filename stem
    ADB_get_filename_ext  - fetch the filename extension for given filetype
    ADB_get_filename      - fetch proper filename for given filetype
    
    ADB_alg_file_support  - how shall the atts & exs be written

    ADB_alg_param_ok - check an algorithm input parameter
    ADB_alg_params_ok -  check an algorithm input parameter list
    ADB_alg_statistic_ok - check a statistics name 
  

    Internal Functions:

    find_free_alg_ref    - find the smallest unused algorithm reference
    check_alg_decl       - check the correctness of an algortihm description
    attr_decl_init       - alg. descriotion attribute section initialization
    ex_decl_init         - alg. description example section initialization
    interface_decl_init  - alg. description interface section initialzaiton
    cpy_quoted           - returns an unquoted copy of a string, very unsafe ! 
    alg_param_ok         - check the correctness of an algorithm input param.
\*                                                                            */

#include <stdio.h>
#include <unistd.h>
#include <string.h>

#include "Alg_Database.h"

t_ADB_database ADB_database;
int ADB_parser_rv = EXECUTION_OK;

extern char *TELA_call_path;


typedef enum {SC_CFTYPE,SC_NAME} t_sort_criterion;


/* ========================================================================== *\
                             INTERNAL FUNCTIONS     
\* ========================================================================== */


/* -------------------------------------------------------------------------- *\
                      Find an unused algorithm reference
\* -------------------------------------------------------------------------- */

t_algorithms find_free_alg_ref()
{
  t_alg_decl *p_alg_decl;

  if (list_get_last(&ADB_database.alg_decls,(void**)&p_alg_decl))
    return p_alg_decl->ref+1;
  else 
    return 0; 
}


/* -------------------------------------------------------------------------- *\
            Check the correctness of given algorithm description
\* -------------------------------------------------------------------------- */

int check_alg_decl(t_alg_decl *p_alg_decl)
{
  t_alg_decl *p_tmp_decl, *p_parent_decl;

  /* first, check that the algorithm name does not conflict with */
  /* existing names */
  if (ADB_alg_decl_by_name(p_alg_decl->alg_name,&p_tmp_decl)) 
  {
    printf("Duplicate description for algorithm %s.\n",p_alg_decl->alg_name);
    return 0;
  }

  /* fill empty fields with default values from the parent */
  /* TELA has no parent, and is not handled this way */
  if (strcmp(p_alg_decl->alg_name,"TELA"))
  {
    /* TELA is the default parent */
    if (!p_alg_decl->parent_name)
      alloc_string(&p_alg_decl->parent_name,"TELA");

    if (!ADB_alg_decl_by_name(p_alg_decl->parent_name,&p_parent_decl)) 
    {
      printf("Description for the parent %s missing.\n",
        p_alg_decl->parent_name);
      return 0;
    }

    /* fetch default values for empty fields */

		if (!p_alg_decl->cf_type)
			realloc_string(&p_alg_decl->cf_type,p_parent_decl->cf_type);

    if (p_alg_decl->attr_decl.class.b_first == -1)
      memcpy(&p_alg_decl->attr_decl.class,&p_parent_decl->attr_decl.class,
        sizeof(t_class));

    if (!p_alg_decl->attr_decl.att_sep)
      realloc_string(&p_alg_decl->attr_decl.att_sep,
        p_parent_decl->attr_decl.att_sep);
    if (!p_alg_decl->attr_decl.numatt_sep)
      realloc_string(&p_alg_decl->attr_decl.numatt_sep,
        p_parent_decl->attr_decl.numatt_sep);
    if (!p_alg_decl->attr_decl.nameval_sep)
      realloc_string(&p_alg_decl->attr_decl.nameval_sep,
        p_parent_decl->attr_decl.nameval_sep);
    if (!p_alg_decl->attr_decl.val_sep)
      realloc_string(&p_alg_decl->attr_decl.val_sep,
        p_parent_decl->attr_decl.val_sep);
    if (!p_alg_decl->attr_decl.typeval_sep)
      realloc_string(&p_alg_decl->attr_decl.typeval_sep,
        p_parent_decl->attr_decl.typeval_sep);

    if (!p_alg_decl->attr_decl.discrete_type)
      realloc_string(&p_alg_decl->attr_decl.discrete_type,
        p_parent_decl->attr_decl.discrete_type);
    if (!p_alg_decl->attr_decl.int_type)
      realloc_string(&p_alg_decl->attr_decl.int_type,
        p_parent_decl->attr_decl.int_type);
    if (!p_alg_decl->attr_decl.cont_type)
      realloc_string(&p_alg_decl->attr_decl.cont_type,
        p_parent_decl->attr_decl.cont_type);

    if (!p_alg_decl->attr_decl.val_unknown)
      realloc_string(&p_alg_decl->attr_decl.val_unknown,
        p_parent_decl->attr_decl.val_unknown);
    if (!p_alg_decl->attr_decl.val_dont_care)
      realloc_string(&p_alg_decl->attr_decl.val_dont_care,
        p_parent_decl->attr_decl.val_dont_care);

    if (p_alg_decl->ex_decl.b_class_first == -1)
      p_alg_decl->ex_decl.b_class_first =
        p_parent_decl->ex_decl.b_class_first;

    if (!p_alg_decl->ex_decl.ex_val_sep)
      realloc_string(&p_alg_decl->ex_decl.ex_val_sep,
        p_parent_decl->ex_decl.ex_val_sep);
    if (!p_alg_decl->ex_decl.example_sep)
      realloc_string(&p_alg_decl->ex_decl.example_sep,
        p_parent_decl->ex_decl.example_sep);

    if (!p_alg_decl->interface.atts_header)
      realloc_string(&p_alg_decl->interface.atts_header,
        p_parent_decl->interface.atts_header);
    if (!p_alg_decl->interface.exs_header)
      realloc_string(&p_alg_decl->interface.exs_header,
        p_parent_decl->interface.exs_header);
    if (!p_alg_decl->interface.rule_header)
      realloc_string(&p_alg_decl->interface.rule_header,
        p_parent_decl->interface.rule_header);
    if (!p_alg_decl->interface.both_header)
      realloc_string(&p_alg_decl->interface.both_header,
        p_parent_decl->interface.both_header);
    if (!p_alg_decl->interface.atts_footer)
      realloc_string(&p_alg_decl->interface.atts_footer,
        p_parent_decl->interface.atts_footer);
    if (!p_alg_decl->interface.exs_header)
      realloc_string(&p_alg_decl->interface.exs_footer,
        p_parent_decl->interface.exs_footer);
    if (!p_alg_decl->interface.rule_footer)
      realloc_string(&p_alg_decl->interface.rule_footer,
        p_parent_decl->interface.rule_footer);
    if (!p_alg_decl->interface.both_footer)
      realloc_string(&p_alg_decl->interface.both_footer,
        p_parent_decl->interface.both_footer);

    if (!p_alg_decl->interface.atts_exs_sep)
      realloc_string(&p_alg_decl->interface.atts_exs_sep,
        p_parent_decl->interface.atts_exs_sep);
    if (!p_alg_decl->interface.comment_sep)
      realloc_string(&p_alg_decl->interface.comment_sep,
        p_parent_decl->interface.comment_sep);

    if (!p_alg_decl->interface.atts_ext)
      realloc_string(&p_alg_decl->interface.atts_ext,
        p_parent_decl->interface.atts_ext);    
    if (!p_alg_decl->interface.exs_ext)
      realloc_string(&p_alg_decl->interface.exs_ext,
        p_parent_decl->interface.exs_ext);    
    if (!p_alg_decl->interface.both_ext)
      realloc_string(&p_alg_decl->interface.both_ext,
        p_parent_decl->interface.both_ext);    
    if (!p_alg_decl->interface.rule_ext)
      realloc_string(&p_alg_decl->interface.rule_ext,
        p_parent_decl->interface.rule_ext);    
    if (p_alg_decl->interface.file_support == UNDECLARED)
      p_alg_decl->interface.file_support =
        p_parent_decl->interface.file_support;    

    /* inherit also the name of the executable. Note that since TELA
      does not have an executable the parent may not either */
    if (!p_alg_decl->interface.exe_name && p_parent_decl->interface.exe_name)
      realloc_string(&p_alg_decl->interface.exe_name,
        p_parent_decl->interface.exe_name);    


  }
	else
	{
		if (!p_alg_decl->cf_type)
			alloc_string(&p_alg_decl->cf_type,"'Data Format'");
	}

  /* check that the final description is ok */
  if ((p_alg_decl->attr_decl.class.b_first == -1) ||
      !p_alg_decl->attr_decl.att_sep ||
      !p_alg_decl->attr_decl.numatt_sep ||
      !p_alg_decl->attr_decl.nameval_sep ||
      !p_alg_decl->attr_decl.val_sep ||
      !p_alg_decl->attr_decl.typeval_sep ||

      !p_alg_decl->attr_decl.discrete_type ||
      !p_alg_decl->attr_decl.int_type ||
      !p_alg_decl->attr_decl.cont_type ||

      !p_alg_decl->attr_decl.val_unknown ||
      !p_alg_decl->attr_decl.val_dont_care ||

      !p_alg_decl->ex_decl.ex_val_sep ||
      !p_alg_decl->ex_decl.example_sep ||
      !p_alg_decl->interface.atts_exs_sep || 
      !p_alg_decl->interface.comment_sep 
    )
  {
    printf("Nondeclared symbols in description %s.\n",p_alg_decl->alg_name);
    return 0;
  }  

  if (strcmp(p_alg_decl->alg_name,"TELA"))
  { 
    if (  !p_alg_decl->interface.atts_ext ||
          !p_alg_decl->interface.exs_ext  ||
          !p_alg_decl->interface.both_ext ||
          !p_alg_decl->interface.rule_ext )
     {
      printf("Nondeclared symbols in description %s.\n",p_alg_decl->alg_name);
      return 0;
    }  
  }


  /* either integer or float attributes must be supported */
  if (!p_alg_decl->attr_decl.int_type && !p_alg_decl->attr_decl.cont_type)
  {
    printf("INT and CONT fields cannot both be empty.\n");
    return 0;
  }

  if (p_alg_decl->attr_decl.att_sep)
  {
    if (p_alg_decl->attr_decl.val_sep)
    {
      if (!strcmp(p_alg_decl->attr_decl.att_sep, p_alg_decl->attr_decl.val_sep))
      {
        printf("ATTSEP and VALSEP may not be equal.\n");
        return 0;
      }
    }

    if (p_alg_decl->attr_decl.typeval_sep)
    {
      if (!strcmp(p_alg_decl->attr_decl.att_sep,
        p_alg_decl->attr_decl.typeval_sep))
      {
        printf("ATTSEP and TYPEVALSEP may not be equal.\n");
        return 0;
      }
    }

    if (p_alg_decl->attr_decl.nameval_sep)
    {
      if (!strcmp(p_alg_decl->attr_decl.att_sep,
        p_alg_decl->attr_decl.nameval_sep))
      {
        printf("ATTSEP and NAMEVALSEP may not be equal.\n");
        return 0;
      }
    }
  }

  if (p_alg_decl->attr_decl.numatt_sep)
  {
    if (p_alg_decl->attr_decl.val_sep)
    {
      if (!strcmp(p_alg_decl->attr_decl.numatt_sep,
         p_alg_decl->attr_decl.val_sep))
      {
        printf("NUMATTSEP and VALSEP may not be equal.\n");
        return 0;
      }
    }

    if (p_alg_decl->attr_decl.typeval_sep)
    {
      if (!strcmp(p_alg_decl->attr_decl.numatt_sep,
        p_alg_decl->attr_decl.typeval_sep))
      {
        printf("NUMATTSEP and TYPEVALSEP may not be equal.\n");
        return 0;
      }
    }

    if (p_alg_decl->attr_decl.nameval_sep)
    {
      if (!strcmp(p_alg_decl->attr_decl.numatt_sep,
        p_alg_decl->attr_decl.nameval_sep))
      {
        printf("NUMATTSEP and NAMEVALSEP may not be equal.\n");
        return 0;
      }
    }
  }


  /* ok, now let's build the absolute pathname to the algorithm executable */
  if (!p_alg_decl->interface.exe_name) return 1;

  if (p_alg_decl->interface.exe_name[1] != '\\')
  {
    char path_name[512];
    /* = TELA_call_path / exe_name */
    strcpy(path_name,"\'");
    strcat(path_name,TELA_call_path);
    strcat(path_name,"/");
    strcat(path_name,p_alg_decl->interface.exe_name+1);

    if (!realloc_string(&p_alg_decl->interface.exe_name,path_name))
    {
      printf("Out of memory\n");
      return 0;
    }
  }    
  return 1;
}


/* -------------------------------------------------------------------------- *\
            Initialize given attribute data format description
\* -------------------------------------------------------------------------- */

void attr_decl_init(t_attr_decl *p_attr_decl)
{
  memset(p_attr_decl,0,sizeof(t_attr_decl));
  p_attr_decl->class.b_unnamed = p_attr_decl->class.b_first = -1;
}


/* -------------------------------------------------------------------------- *\
            Initialize given example data format description
\* -------------------------------------------------------------------------- */

void ex_decl_init(t_ex_decl *p_ex_decl)
{
    memset(p_ex_decl,0,sizeof(t_ex_decl));
    p_ex_decl->b_class_first = -1;
}


/* -------------------------------------------------------------------------- *\
            Initialize given interface description
\* -------------------------------------------------------------------------- */

void interface_decl_init(t_interface *p_interface)
{
  memset(p_interface,0,sizeof(t_interface));
  list_init(&p_interface->options);
  list_init(&p_interface->runstats);
  list_init(&p_interface->teststats);
  p_interface->file_support = UNDECLARED;
}


/* -------------------------------------------------------------------------- *\
         Strip first and last character (hopefully quotes) of an string         

    Note: this is an extremely unelegant and unsafe function, use with caution!
\* -------------------------------------------------------------------------- */

void cpy_quoted(char * dst,char *src)
{  
  if ( (int)strlen(src) < 2 ) strcpy(dst,src); /* it's too short to be quoted */
  
  strcpy(dst,src+1);

  dst[strlen(dst)-1] = '\0';

}


/* -------------------------------------------------------------------------- *\
          Check the correctness of an algorithm input parameter
\* -------------------------------------------------------------------------- */

int alg_param_ok(t_run_node *p_run_node, t_alg_decl *p_alg_decl)
{
  int rv;
  char *p_option;

  rv = list_get_first(&p_alg_decl->interface.options,(void **)&p_option);

  while (rv)
  {
    /* only correctness of the parameter name is checked */
    if (!strcmp(p_option,p_run_node->selector)) return 1;
    rv = list_get_next(&p_alg_decl->interface.options,(void **)&p_option);
  }

  return 0;
}


/* -------------------------------------------------------------------------- *\
										Compare two algorithm descriptions
\* -------------------------------------------------------------------------- */

int cmp_field(t_alg_decl *p_1, t_alg_decl *p_2, t_sort_criterion field)
{
	switch (field)
	{
		case SC_CFTYPE:	return strcmp(p_1->cf_type,p_2->cf_type);
		case SC_NAME:   return strcmp(p_1->alg_name,p_2->alg_name);
	}
}


/* -------------------------------------------------------------------------- *\
										Sort index according to given field
\* -------------------------------------------------------------------------- */

void insertion_sort_index(t_alg_decl **index, int size,t_sort_criterion field)
{
		int i,j;
		t_alg_decl *tmp;

		for (i=1;i < size;i++)
		{			
			tmp = index[i];

			for (j=i-1;j >=0; j--)
			{
				if (cmp_field(tmp,index[j],field) >= 0) break;

				index[j+1] = index[j];		
			}

			index[j+1] = tmp; 
		}
}


/* -------------------------------------------------------------------------- *\
			Obtain a sorted index for the algorithm database
\* -------------------------------------------------------------------------- */

void make_sorted_index(t_alg_decl** index, t_sort_criterion field)
{
	int first_free=0,rv;
	t_alg_decl *p_alg_decl;

	rv = list_get_first(&ADB_database.alg_decls,(void **)&p_alg_decl);

  while (rv)
  {
		index[first_free++] = p_alg_decl;

    rv = list_get_next(&ADB_database.alg_decls,(void **)&p_alg_decl);
  }

	
	insertion_sort_index(index,first_free-1,field);
}



/* ========================================================================== *\
                             EXTERNAL SERVICES     
\* ========================================================================== */


/* -------------------------------------------------------------------------- *\
                      Initialize algorithm database
\* -------------------------------------------------------------------------- */

void ADB_init_db()
{
  ADB_database.alg_count = 0;
  list_init(&ADB_database.alg_decls);
}


/* -------------------------------------------------------------------------- *\
              Read the algorithm database from a external file.
\* -------------------------------------------------------------------------- */

int ADB_read_file(char *filename)
{
  FILE *p_dfd_file=0;
  int rv;
  t_alg_decl *p_alg_decl;
  ADB_parser_rv = EXECUTION_OK;

  if (!filename)
    p_dfd_file = freopen(".TELA_dfd_db","r",stdin);
  else
    p_dfd_file = freopen(filename,"r",stdin);

  if (p_dfd_file)
  {
    yyparse();
    fclose(p_dfd_file);    
  }
  else ADB_parser_rv = EXECUTION_FAILED;

  freopen("/dev/tty","r",stdin);

  return ADB_parser_rv == EXECUTION_OK;
}


/* -------------------------------------------------------------------------- *\
            Write an algorithm description to file
\* -------------------------------------------------------------------------- */

int ADB_write_alg_decl(FILE *dfd_file,t_alg_decl *p_alg_decl)
{


  t_attr_decl *p_attr_decl;
  t_ex_decl *p_ex_decl;
  t_interface *p_interface;
  char *option,*stat;

  /* header */
  fprintf(dfd_file,"BEGIN DESCRIPTION %s\n",p_alg_decl->alg_name);
  

  fprintf(dfd_file,"[ATTRIBUTES]\n");

  /* attribute declarations */
  p_attr_decl = &p_alg_decl->attr_decl;


  /* optional terminal symbol redeclarations */
  if (p_attr_decl->att_sep) 
    fprintf(dfd_file,"  ATTSEP= %s\n",p_attr_decl->att_sep);
  if (p_attr_decl->nameval_sep)
    fprintf(dfd_file,"  NAMEVALSEP= %s\n",p_attr_decl->nameval_sep);
  if (p_attr_decl->val_sep)
    fprintf(dfd_file,"  VALSEP= %s\n",p_attr_decl->val_sep);

  if (p_attr_decl->discrete_type)
    fprintf(dfd_file,"  DISCRETE = %s\n",p_attr_decl->discrete_type);
  if (p_attr_decl->int_type)
    fprintf(dfd_file,"  INT = %s\n",p_attr_decl->int_type);
  if (p_attr_decl->cont_type)
    fprintf(dfd_file,"  CONT = %s\n",p_attr_decl->cont_type);

  if (p_attr_decl->val_unknown)
    fprintf(dfd_file,"  UNKNOWN = %s\n",p_attr_decl->val_unknown);
  if (p_attr_decl->val_dont_care)
    fprintf(dfd_file,"  DONTCARE = %s\n",p_attr_decl->val_dont_care);

  /* example declarations */
  p_ex_decl = &p_alg_decl->ex_decl;

  if (p_ex_decl->ex_val_sep || p_ex_decl->example_sep)
  {
    fprintf(dfd_file,"[EXAMPLES]\n");

    if (p_ex_decl->ex_val_sep)
      fprintf(dfd_file,"  EXVALSEP = %s\n",p_ex_decl->ex_val_sep);
    if (p_ex_decl->example_sep)
      fprintf(dfd_file,"  EXAMPLESEP = %s\n",p_ex_decl->example_sep);
  }

  p_interface = &p_alg_decl->interface;

  /* calling interface declarations */
  fprintf(dfd_file,"[INTERFACE]\n");

  /* name of the executable */
  fprintf(dfd_file,"  EXENAME= %s\n",p_interface->exe_name);

  if (p_interface->file_support)
  {
    fprintf(dfd_file,"  ATTSEXSFILES= ");
    switch (p_interface->file_support)
    {
      case ONLY_COMB: fprintf(dfd_file,"COMBINED\n");break;    
      case ONLY_SEP: fprintf(dfd_file,"SEPARATE\n");break;
      case COMB_N_SEP: fprintf(dfd_file,"BOTH\n");break;
    }
  }

  if (p_interface->atts_ext)
    fprintf(dfd_file,"  ATTSEXT = %s\n",p_interface->atts_ext);
  if (p_interface->exs_ext)
    fprintf(dfd_file,"  EXSEXT = %s\n",p_interface->exs_ext);
  if (p_interface->both_ext)
    fprintf(dfd_file,"  BOTHEXT = %s\n",p_interface->both_ext);
  if (p_interface->rule_ext)
    fprintf(dfd_file,"  RULEEXT = %s\n",p_interface->rule_ext);

  /* algorithm options */  
  if (list_get_first(&p_interface->options,(void **)&option))
  {
    fprintf(dfd_file,"[OPTIONS]\n");
    do     {
      fprintf(dfd_file,"  %s\n",option);  
    }
    while (list_get_next(&p_interface->options,(void **)&option));
  }
  
  /* RUN-statistics */
  if (list_get_first(&p_interface->runstats,(void **)&stat))
  {
    fprintf(dfd_file,"[RUNSTATS]\n");
    
    do    {
      fprintf(dfd_file,"  %s\n",stat);  
    }
    while (list_get_next(&p_interface->runstats,(void **)&stat));    
  }

  /* TEST-statistics */
  if (list_get_first(&p_interface->teststats,(void **)&stat))
  {
    fprintf(dfd_file,"[TESTSTATS]\n");
    
    do    {
      fprintf(dfd_file,"  %s\n",stat);  
    }
    while (list_get_next(&p_interface->teststats,(void **)&stat));    
  }
  /* tailer */
  fprintf(dfd_file,"END DESCRIPTION \n");

  return 1;
}


/* -------------------------------------------------------------------------- *\
              Write the algorithm database to external storage
\* -------------------------------------------------------------------------- */

int ADB_write_file(char *filename)
{
  FILE *dfd_file;
  t_alg_decl *p_alg_decl;
  int rv;

  if (dfd_file = fopen(filename,"w"))
  {
    /* iterate trough ADB_database */
    rv = list_get_first(&ADB_database.alg_decls,(void**)&p_alg_decl);
    while (rv)
    {
      /* write alg.decl to file */
      if (!ADB_write_alg_decl(dfd_file,p_alg_decl))  return 0;

      rv = list_get_next(&ADB_database.alg_decls,(void**)&p_alg_decl);
    }
    fclose(dfd_file);
  }
  else return 0;
}


/* -------------------------------------------------------------------------- *\
        Add a new algorithm description to database (in memory)
\* -------------------------------------------------------------------------- */

int ADB_add_algorithm(t_alg_decl *p_alg_decl)
{
  t_alg_decl *p_new;

  /* check algorithm declaration validity first */
  if (!check_alg_decl(p_alg_decl)) return 0;

  p_new = (t_alg_decl *)malloc(sizeof(t_alg_decl));
  if (!p_new) return 0;
  memcpy(p_new,p_alg_decl,sizeof(t_alg_decl));
  p_new->ref = find_free_alg_ref();
  if (list_append(&ADB_database.alg_decls,(void *)p_new))
	{
		ADB_database.alg_count++;
		return 1;
	}
	else return 0;
}


/* -------------------------------------------------------------------------- *\
                    Initialize an algorithm description
\* -------------------------------------------------------------------------- */

void ADB_alg_decl_init(t_alg_decl *p_alg_decl)
{
  p_alg_decl->alg_name = 0;
  p_alg_decl->parent_name = 0;
	p_alg_decl->cf_type = 0;
  attr_decl_init(&p_alg_decl->attr_decl);
  ex_decl_init(&p_alg_decl->ex_decl);
  interface_decl_init(&p_alg_decl->interface);
}


/* -------------------------------------------------------------------------- *\
               Fetch an algorithm description by reference
\* -------------------------------------------------------------------------- */

int ADB_alg_decl_by_ref(t_algorithms alg, t_alg_decl **pp_alg_decl)
{
  if (list_get_first(&ADB_database.alg_decls,(void **)pp_alg_decl))
  {
    do
    {
      if ((*pp_alg_decl)->ref == alg) return 1;
    }
    while (list_get_next(&ADB_database.alg_decls,(void**)pp_alg_decl));
  }
  return 0;
}


/* -------------------------------------------------------------------------- *\
                  Fetch an algorithm description by name
\* -------------------------------------------------------------------------- */

int ADB_alg_decl_by_name(char *alg_name, t_alg_decl **pp_alg_decl)
{
  if (list_get_first(&ADB_database.alg_decls,(void **)pp_alg_decl))
  {
    do
    {
      if (!strcmp((*pp_alg_decl)->alg_name,alg_name)) return 1;
    }
    while (list_get_next(&ADB_database.alg_decls,(void**)pp_alg_decl));
  }
  return 0;
}


/* -------------------------------------------------------------------------- *\
              Get the name of the learning program executable
\* -------------------------------------------------------------------------- */

int ADB_get_exe_name(t_algorithms alg,char *buf)
{
  t_alg_decl *p_alg_decl;

  if (ADB_alg_decl_by_ref(alg,&p_alg_decl))
  {
    if (!p_alg_decl->interface.exe_name) 
    {
      buf[0] = '\0';
      return 0;
    }
    strcpy(buf,p_alg_decl->interface.exe_name+1);
    buf[strlen(buf)-1] = '\0';
    return 1;
  }
  else
  {
    errmsg("No such algorithm");
    return 0;
  }  
}


/* -------------------------------------------------------------------------- *\
                      Get algorithm reference by name
\* -------------------------------------------------------------------------- */

int ADB_get_alg_ref(char *name, t_algorithms *p_ref)
{
  t_alg_decl *p_alg_decl;
  if (!ADB_alg_decl_by_name(name,&p_alg_decl))
  {
    *p_ref = 0;
    return 0;
  }
  else
  {
    *p_ref = p_alg_decl->ref;
    return 1;
  }
}


/* -------------------------------------------------------------------------- *\
                              Get algorithm name
\* -------------------------------------------------------------------------- */

int ADB_get_alg_name(t_algorithms alg,char *buf)
{
  t_alg_decl *p_alg_decl;

  if (ADB_alg_decl_by_ref(alg,&p_alg_decl))
  {
    strcpy(buf,p_alg_decl->alg_name);
    return 1;
  }
  else
  {
    errmsg("No such algorithm");
    return 0;
  }  
}


/* -------------------------------------------------------------------------- *\
             Get a proper filename for a file of given type
\* -------------------------------------------------------------------------- */

int ADB_get_filename(t_algorithms alg, t_filetype f_type,char *filename)
{      
  char err[256];
  t_alg_decl *p_alg_decl;

  if (!ADB_alg_decl_by_ref(alg,&p_alg_decl)) 
  {
    sprintf(err,"Internal Error: No such algorithm: %d",alg);
    errmsg(err);
    return 0;
  }
  
  ADB_get_filename_stem(filename);
  ADB_get_filename_ext(alg,f_type,filename+strlen(filename));

}


/* -------------------------------------------------------------------------- *\
             Output an algorithm specific file name extension
\* -------------------------------------------------------------------------- */

int ADB_get_filename_ext(t_algorithms alg, t_filetype f_type,char *ext)
{      
  char err[256];
  t_alg_decl *p_alg_decl;

  if (!ADB_alg_decl_by_ref(alg,&p_alg_decl)) 
  {
    sprintf(err,"Internal Error: No such algorithm: %d",alg);
    errmsg(err);
    return 0;
  }
  
  switch (f_type)
  {
    case FT_ATTS: 
      if (strlen(p_alg_decl->interface.atts_ext)) 
        cpy_quoted(ext,p_alg_decl->interface.atts_ext);
      else
        strcpy(ext,".atts");
    break;
    case FT_EXS:
      if (strlen(p_alg_decl->interface.exs_ext))             
        cpy_quoted(ext,p_alg_decl->interface.exs_ext);
      else
        strcpy(ext,".exs");
    break;
    case FT_BOTH:
      if (strlen(p_alg_decl->interface.both_ext)) 
        cpy_quoted(ext,p_alg_decl->interface.both_ext);
      else
        strcpy(ext,".both");
    break;
    case FT_RULE:
    case FT_ATTSRULE:
      if (strlen(p_alg_decl->interface.rule_ext)) 
        cpy_quoted(ext,p_alg_decl->interface.rule_ext);
      else
        strcpy(ext,".rule");
    break;
    case FT_STATS:
      if (p_alg_decl->interface.stats_ext) 
        cpy_quoted(ext,p_alg_decl->interface.stats_ext);
      else
        strcpy(ext,".stats");
    break;
    default: 
    {
      sprintf(err,"No such file type: %d",f_type);
      errmsg(err);
      return 0;
    }
  }
  return 1;
}

/* -------------------------------------------------------------------------- *\
                     Get safe filename stem 
\* -------------------------------------------------------------------------- */

void ADB_get_filename_stem(char *stem)
{        
  /* a predefined prefix */
  strcpy(stem,TELA_FILENAME_PREFIX);
  
  /* make the filename unique by augmenting by process ID */
  sprintf(stem+strlen(stem),"_%ld",getpid());
}

/* -------------------------------------------------------------------------- *\
      Output the file support status for writing attributes and examples
\* -------------------------------------------------------------------------- */

t_file_support ADB_alg_file_support(t_algorithms alg)
{
  t_alg_decl *p_alg_decl;

  if (!ADB_alg_decl_by_ref(alg,&p_alg_decl)) return 0;

  return p_alg_decl->interface.file_support;
}


/* -------------------------------------------------------------------------- *\
      Check the correctness of algorithm input parameters
\* -------------------------------------------------------------------------- */

int ADB_alg_params_ok(t_run_node *p_run_node,t_algorithms alg)
{
  t_alg_decl *p_alg_decl;

  if (!ADB_alg_decl_by_ref(alg,&p_alg_decl)) return 0;   


  while (p_run_node)
  {
    if (!alg_param_ok(p_run_node,p_alg_decl)) return 0;
    p_run_node = p_run_node->next;
  }
  
  return 1;
}


/* -------------------------------------------------------------------------- *\
                 Check the properness of a statistics
\* -------------------------------------------------------------------------- */

int ADB_alg_statistic_ok(char *stat, t_algorithms alg, int mode)
{
  t_alg_decl *p_alg_decl;
  int rv;
  char *p_stat;

  if (!ADB_alg_decl_by_ref(alg,&p_alg_decl)) return 0;

  if (!mode)
  {
    rv = list_get_first(&p_alg_decl->interface.runstats,(void **)&p_stat);

    while (rv)
    {
      if (!strcmp(p_stat,stat)) return 1;
      rv = list_get_next(&p_alg_decl->interface.runstats,(void **)&p_stat);
    }
  }
  else
  {
    rv = list_get_first(&p_alg_decl->interface.teststats,(void **)&p_stat);

    while (rv)
    {
      if (!strcmp(p_stat,stat)) return 1;
      rv = list_get_next(&p_alg_decl->interface.teststats,(void **)&p_stat);
    }
  }


  return 0;  
}


/* -------------------------------------------------------------------------- *\
      Return the position of class attribute value in an example vector

    0 - last, 1 - first 
\* -------------------------------------------------------------------------- */


int ADB_get_class_place_in_ex(t_algorithms alg)
{
  t_alg_decl *p_alg_decl;
  ADB_alg_decl_by_ref(alg,&p_alg_decl);
  return p_alg_decl->ex_decl.b_class_first;
}

/* -------------------------------------------------------------------------- *\
      Return the position of class attribute declaration 
      in an attribute description

    0 - last, 1 - first 
\* -------------------------------------------------------------------------- */

int ADB_get_class_place_in_atts(t_algorithms alg)
{
  t_alg_decl *p_alg_decl;
  ADB_alg_decl_by_ref(alg,&p_alg_decl);
  return p_alg_decl->attr_decl.class.b_first;
}


/* -------------------------------------------------------------------------- *\
      Output a string corresponding a given separator in given data format
\* -------------------------------------------------------------------------- */

int ADB_get_separator(char *str, t_separator_type type, t_algorithms alg)
{
  t_alg_decl *p_alg_decl;
  char *source=0;
  if (!ADB_alg_decl_by_ref(alg,&p_alg_decl)) return 0;

  switch (type)
  {
    case ATTS_HEADER: source = p_alg_decl->interface.atts_header; break;
    case EXS_HEADER: source = p_alg_decl->interface.exs_header; break;
    case BOTH_HEADER: source = p_alg_decl->interface.both_header; break;
    case RULE_HEADER: source = p_alg_decl->interface.rule_header; break;
    case EXS_FOOTER: source = p_alg_decl->interface.exs_footer; break;
    case BOTH_FOOTER: source = p_alg_decl->interface.both_footer; break;
    case RULE_FOOTER: source = p_alg_decl->interface.rule_footer; break;
    case ATTS_FOOTER: source = p_alg_decl->interface.atts_footer; break;
    case ATTS_EXS_SEP: source = p_alg_decl->interface.atts_exs_sep; break;
    case COMMENT_SEP: source = p_alg_decl->interface.comment_sep; break;  

    case EXAMPLE_SEP: source = p_alg_decl->ex_decl.example_sep; break;
    case EXVAL_SEP: source = p_alg_decl->ex_decl.ex_val_sep; break;

    case ATT_SEP: source = p_alg_decl->attr_decl.att_sep; break;
    case NUMATT_SEP: source = p_alg_decl->attr_decl.numatt_sep; break;
    case NAME_VAL_SEP: source = p_alg_decl->attr_decl.nameval_sep; break;
    case VAL_SEP: source = p_alg_decl->attr_decl.val_sep; break;
    case TYPE_VAL_SEP: source = p_alg_decl->attr_decl.typeval_sep; break;  

    case INT_TYPE: source = p_alg_decl->attr_decl.int_type; break;  
    case CONT_TYPE: source = p_alg_decl->attr_decl.cont_type; break;  
    case NOM_TYPE: source = p_alg_decl->attr_decl.discrete_type; break;  

    case DONT_KNOW: source = p_alg_decl->attr_decl.val_unknown; break;
    case DONT_CARE: source = p_alg_decl->attr_decl.val_dont_care; break;

    default: return 0;
  }
  if (source)
  {
    if (source[0] == '\0')
    { 
      str[0] = '\0';
      return 0;    
    }

    if (source[1] == '\\' ? source[2] == 'n' : 0)
    {
      str[0] = '\n';str[1] = '\0';
    }
    else
    {
      strcpy(str,source+1);
      str[strlen(str)-1] = '\0';
    }
    return 1;
  }
  else 
  {
    str[0] = '\0';
    return 0;
  }
}


/* -------------------------------------------------------------------------- *\
        Check whether two separator are equivalent in given data format
\* -------------------------------------------------------------------------- */

int ADB_is_equiv_sep(t_separator_type sep1,t_separator_type sep2,
  t_algorithms alg)
{
  char buf1[64],buf2[64];
  ADB_get_separator(buf1,sep1,alg);
  ADB_get_separator(buf2,sep2,alg);

  return !strcmp(buf1,buf2);
}


/* -------------------------------------------------------------------------- *\
						List algorithms in order defined by their type
\* -------------------------------------------------------------------------- */

void ADB_list_algorithms()
{
	int i=0;
	char namebuf[256];

	t_alg_decl **index;

	index = (t_alg_decl**)malloc(ADB_database.alg_count*sizeof(t_alg_decl*));

	make_sorted_index(index,SC_CFTYPE);

	printf("Declared algorithms / data formats:\n\n");

	while (i < ADB_database.alg_count)
	{
		cpy_quoted(namebuf,index[i]->cf_type);
		printf("%s : ",namebuf);

		do	
		{
			printf("%s ", index[i]->alg_name);
			if (++i >= ADB_database.alg_count) break;
		}
		while (!strcmp(index[i-1]->cf_type,index[i]->cf_type));
		
		printf("\n");
	}
}
