#ifndef H_TELA_ALGORITHM_DB
#define H_TELA_ALGORITHM_DB

/*                                                                            *\

	  TELA - Testing Environment for Learning Algorithms - version 1.995
    ==================================================================

		Module:  			Alg_Database 

		Contents:    TELA algorithm database

		Description:	offers run-time query facilities about the 
									characteristics of the learning algorithms that 
									are embedded in to TELA. 

		Services:

		Database maintenance
		====================
	
	  ADB_init_db       - initialize the algorithm database (ADB) 
												memory structures
  	ADB_read_file     - read the ADB into memory
  	ADB_write_file    - write the ADB to the disk, unused for the moment
  	ADB_alg_decl_init - initialize an algorithm description
  	ADB_add_algorithm - add an algorithm description into ADB

	  Query Capabilities
	  ==================

	  ADB_get_alg_ref       - get a reference to an algorithm by name
	  ADB_alg_decl_by_ref   - fetch an algoritm description by reference
	 	ADB_alg_decl_by_name  - fetch an algorithm descriotion by name             

	  ADB_get_exe_name      - get the name of the learning program executable
	  ADB_get_alg_name      - get the name of an algorithm
 		ADB_get_separator     - fetch the string corresponding to given separator
		ADB_is_equiv_sep      - check if two separators are equivalent

		ADB_get_filename_stem - output a non-clashing filename stem
		ADB_get_filename_ext  - fetch the filename extension for given filetype
		ADB_get_filename      - fetch proper filename for given filetype

		ADB_alg_file_support  - how shall the atts & exs be written

		ADB_alg_param_ok - check an algorithm input parameter
		ADB_alg_params_ok -  check an algorithm input parameter list
		ADB_alg_statistic_ok - check a statistics name 
	
		Author: Juho Rousu (1995,1996)

\*                                                                            */

#include <stdio.h>

#include "tela.h"
#include "UI_parsers.h"

/* attribute declaration language */

/* how to find class attribute: these quirks are due the */
/* attribute declaration language of C4.5                */
typedef struct
{
	int b_unnamed; /* is class attribute unnamed */
	int b_first;   /* is it first or last */
}
t_class;

typedef struct
{
	/* optional redeclarations of terminal symbols */

	char *att_sep;					/* Attribute definition separator  */
	char *numatt_sep;				/* Numeric attribute decl. terminator */
	char *nameval_sep;			/* Attribute name terminator       */
	char *val_sep;					/* Attribute value separator       */
	char *typeval_sep;			/* Attribute type and value list separator */

	char *discrete_type;		/* Discrete, unordered attribute	 */
	char *int_type; 				/* Discrete, ordered attribute     */
	char *cont_type;				/* Continuous attribute            */

	char *val_unknown;  		/* Unknown value identifier        */
	char *val_dont_care;		/* Don't care identifier           */


	/* some additional fields: these are unused at the moment */
	char *firstcharset;			/* Valid identifier starters       */
	char *charset;					/* Valid identifier characters     */
	
	t_class class;  					/* How to find the class attribute  */

} t_attr_decl;

/* example vector format */
typedef struct
{
	int b_class_first;      /* is class attribute value first in line */
	char *ex_val_sep;				/* vector component separator      */
	char *example_sep;			/* example vector separator        */
} t_ex_decl;


typedef enum {ATTS_HEADER,EXS_HEADER,BOTH_HEADER,RULE_HEADER,TYPE_VAL_SEP,
							ATTS_FOOTER,EXS_FOOTER,BOTH_FOOTER,RULE_FOOTER, NO_SEPARATOR,
							ATTS_EXS_SEP, EXAMPLE_SEP, EXVAL_SEP, DONT_KNOW, DONT_CARE,
							ATT_SEP,NAME_VAL_SEP,VAL_SEP,COMMENT_SEP, NUMATT_SEP,
							INT_TYPE,CONT_TYPE,NOM_TYPE} t_separator_type;

/* DFD: algorithm interface declarations */

/* supported attribute and example file combinations */
typedef enum {UNDECLARED,
							ONLY_COMB, 		/* only combined file */ 
							ONLY_SEP,  		/* only separated atts and exs files */
							COMB_N_SEP		/* both modes supported */
}	t_file_support;

typedef struct
{
	char  		*exe_name;						/* name of the executable          		*/

	t_list	options;								/* algorithm option internal names		*/
	t_list 	runstats;								/* RUN-collectable statistics         */
	t_list	teststats;							/* TEST-collectable statistics				*/

	t_file_support file_support;		/* which a/e/b -file combinations 		*/

	char 			*atts_header;					/* headers, separators and footers		*/
	char 			*exs_header;
	char 			*both_header;
	char 			*rule_header;
	char 			*atts_footer;
	char			*exs_footer;
	char			*both_footer;
	char			*rule_footer;
	char 			*atts_exs_sep;
	char 			*comment_sep;

	char 			*atts_ext;						/* filename extensions    						*/
	char  		*exs_ext;						
	char			*both_ext;
	char 			*rule_ext;
	char			*stats_ext;

	/* some unused stuff */
	/* t_control control;							command passing mechanism, unused	*/
	/* char *input_cmd;								attribute and data input, unused		*/

} t_interface;


typedef struct s_alg_decl
{
	t_algorithms ref;								/* numeric identifier of the algorithm */
	char *alg_name;                 /* name of the algorithm               */
	char *parent_name;							/* fetch default values from parent    */
	char *cf_type;                  /* type of the produced classifier     */
	t_attr_decl	attr_decl;					/* attribute declaration format		     */
	t_ex_decl ex_decl;							/* example declaration format			     */
	t_interface interface;					/* interface components                */
	
	struct s_alg_decl *next;			 	/* next algorithm                      */
} t_alg_decl;


/* Algorithm DataBase */

typedef struct
{
	unsigned alg_count;
	t_list alg_decls;
} t_ADB_database;


/* ADB_database */
extern t_ADB_database ADB_database;

/* parser return value for Algorithm DataBase */
extern int ADB_parser_rv;


void ADB_init_db();
int ADB_read_file(char *filename);
int ADB_write_file(char *file_name);
int ADB_add_algorithm(t_alg_decl *p_alg_decl);
void ADB_alg_decl_init(t_alg_decl *p_alg_decl);

int ADB_alg_decl_by_ref(t_algorithms alg,t_alg_decl **p_alg_decl);
int ADB_alg_decl_by_name(char *alg_name,t_alg_decl **pp_alg_decl);
int ADB_get_alg_ref(char *name,t_algorithms *p_ref);

int ADB_get_exe_name(t_algorithms alg,char *buf);
int ADB_get_alg_name(t_algorithms alg,char *buf);
int ADB_get_separator(char* str, t_separator_type, t_algorithms alg);
int ADB_is_equiv_sep(t_separator_type sep1,t_separator_type sep2,
  t_algorithms alg);

void ADB_get_filename_stem(char *filename);
int ADB_get_filename_ext(t_algorithms alg,t_filetype f_type,char *ext);

t_file_support ADB_alg_file_support(t_algorithms alg);

int ADB_alg_param_ok(t_run_node *p_run_node,t_alg_decl *p_alg_decl);
int ADB_alg_params_ok(t_run_node *p_run_node,t_algorithms alg);
int ADB_alg_statistic_ok(char *stat,t_algorithms alg,int mode);


#endif
