/*                                                                            *\

	  TELA - Testing Environment for Learning Algorithms - version 1.995
    ==================================================================

		Module:      Alg_Exec_Machinery 


		Contents:    TELA Algorithm Execution Machinery

		Description: handles the execution of external learning programs, 
                 builds argument vectors and prepares input, runs the 
                 learning programs in a separate process and collects 
                 statistics.

    Author: Juho Rousu (1995,1996)

    Services:

    Input preparation
    =================

	  EXEC_write_atts_exs - write the attributes and examples 	
	  EXEC_build_RUN_instruction - build the argument vector for RUN-command
	  EXEC_build_TEST_instruction - build the argument vector for RUN-command

    Execution
    =========
		
	  EXEC_run_algorithm - execute the learning program in a separate process

    Output collection and cleanup
    =============================

	  EXEC_collect_stats - collect the recorded statistics
 	  EXEC_remove_temp_files - remove temporary attribute and example files


		Internal functions:

		print_cmd_args           - prints the contents of an argument vector
		get_run_lock             - grab an exclusive execution lock for an algorithm
		release_run_lock         - release the execution lock
		get_attr_exs_input_cmds  - print the commands indicating the names
															 of the atts/exs files

\*                                                                            */

#include <sys/times.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <unistd.h>
#include <libgen.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>

#include "tela.h"
#include "TESLA_statistics.h"
#include "Alg_Database.h"
#include "Alg_Exec_Machinery.h"

#define DEBUGGING_
#define DEBUG

extern char *TELA_call_path;

/* ========================================================================== *\
                             INTERNAL FUNCTIONS     
\* ========================================================================== */


/* -------------------------------------------------------------------------- *\
                   Output the contents of an argument vector
\* -------------------------------------------------------------------------- */

void print_cmd_args(char **table)
{	
	int i=0;
	printf("Argument list:\n");
	while (table[i]) printf("%s\n",table[i++]);
}


/* -------------------------------------------------------------------------- *\
           Get an exclusive execution lock for a learning program file
\* -------------------------------------------------------------------------- */

int get_run_lock(char *alg_name)
{
	char lock_file[256];
	int fd;

	sprintf(lock_file,"%s/.TELA_%s_lock",TELA_call_path,alg_name);

	if ((fd = open(lock_file, O_WRONLY|O_CREAT|O_EXCL,0644)) != -1)
	{
		close(fd);
		return 1;
	}
	else
	{
		char msg[256];
		sprintf(msg,"Could not set the exclusive run lock \n (%s).",lock_file);
		perror(msg);
		return 0;
	}
}


/* -------------------------------------------------------------------------- *\
                      Release an existing execution lock 
\* -------------------------------------------------------------------------- */

void release_run_lock(char *alg_name)
{
	char lock_file[256];

	sprintf(lock_file,"%s/.TELA_%s_lock",TELA_call_path,alg_name);

	remove(lock_file);

	return;
}


/* -------------------------------------------------------------------------- *\
       Print atts/exs input commands in the argument vector             
\* -------------------------------------------------------------------------- */

int get_attr_exs_input_cmds(t_algorithms alg,char *argv[],int *free_index)
{
	char buf[256];
	/* if separate attribute and example files are supported by this algorithm */
	/* we use separate files, otherwise put both in same file                  */

	if (ADB_alg_file_support(alg) == ONLY_SEP)
	{
		if (!alloc_string(argv+(*free_index)++,"ATTSFILE=")) return EXECUTION_FAILED;
		ADB_get_filename(alg,FT_ATTS,buf);
		if (!alloc_string(argv+(*free_index)++,buf)) return EXECUTION_FAILED;		
		if (!alloc_string(argv+(*free_index)++,"EXSFILE=")) return EXECUTION_FAILED;
		ADB_get_filename(alg,FT_EXS,buf);
		if (!alloc_string(argv+(*free_index)++,buf)) return EXECUTION_FAILED;		
	}
	else
	{
		if (!alloc_string(argv+(*free_index)++,"BOTHFILE=")) return EXECUTION_FAILED;
		ADB_get_filename(alg,FT_BOTH,buf);
		if (!alloc_string(argv+(*free_index)++,buf)) return EXECUTION_FAILED;
	}
	return EXECUTION_OK;
}



/* ============================================================================ *\
                              EXTERNAL SERVICES
\* ============================================================================ */


/* -------------------------------------------------------------------------- *\
                    Execute an algorithm in a separate process
\* -------------------------------------------------------------------------- */

int EXEC_run_algorithm(char **param_table,int *p_status,float *time)
{
	struct tms time1,time2;
	int fd,fd_lock;

	char tmp_output[256],file_lock[256];

	sprintf(tmp_output,".TELA_%ld_output",getpid());
	
	if (!get_run_lock(basename(param_table[0]))) return EXECUTION_FAILED;

	times(&time1);

	if (fork() == 0) 	/* child */
	{
		if ((fd = open(tmp_output,O_WRONLY | O_CREAT | O_TRUNC,0644)) == -1)
		{
			perror("Couldn't create temporary output file");
			_exit(EXECUTION_FAILED);
		}	
		if (dup2(fd,1) == -1) 
		{
			fprintf(stderr,"Error redirecting stdout.\n"); 
			_exit(EXECUTION_FAILED);
		}
		if (dup2(fd,2) == -1)
		{ 
			fprintf(stderr,"Error redirecting stderr.\n");
			_exit(EXECUTION_FAILED);	
		}	 
		execve(param_table[0],param_table,NULL);
			
		fprintf(stderr,"Operating system error. Couldn't run learning program");
		_exit(EXECUTION_FAILED);	
	}
	else 							/* parent */
	{
		wait(p_status);

		times(&time2);			
		release_run_lock(basename(param_table[0]));

		*time = ((float)time2.tms_cutime - (float)time1.tms_cutime)/100;
	}	
	
	if (*p_status != 0)
	{
		char msg[256];
		int i;
		sprintf(msg,"Algorithm %s \n execution unsuccesful, exit code:%d", 
			param_table[0],*p_status);

		errmsg(msg);
		return EXECUTION_FAILED;
	}

	return EXECUTION_OK;
}


/* -------------------------------------------------------------------------- *\
				 Collect recorded statistics from file to the statistics list
\* -------------------------------------------------------------------------- */

int EXEC_collect_stats(t_algorithms alg,t_stats_list *p_stats)
{
	FILE *stats_file;
	t_stats_node *p_stat_node;
	char buf[256],*stat_name,*token,*value;	
	int rv,no_match=1,lines=0;	

	ADB_get_filename(alg,FT_STATS,buf);
	if ((stats_file = fopen(buf,"r")) != 0) 
	{
		while (fgets(buf,256,stats_file))
		{			
			char tokbuf[256];
			char sep[3] = "= \n";
			float fvalue;

			lines++;

			/* stat_name or empty line */
			if (!(stat_name = strtok(buf,sep))) continue;

			/* value of the statistic */
			if (!(value = strtok((char*)0,sep)))
			{
				errmsg("Statistic value expected");
				return EXECUTION_FAILED;
			}
	

			if (!sscanf(value,"%f",&fvalue)) /* it should be a numeric value */
			{
				char msg[256];
			
#ifdef DEBUG
	printf("buf : %s \n name: %s \n value: %s\n",buf,stat_name,value);
#endif		

				sprintf(msg,
"Statistics must be numeric, %s is not a legal value",value);
				errmsg(msg);
				return EXECUTION_FAILED;
			}

			/* go trough the statistics list and search for a matching stats name */
			STATS_update_node(p_stats,stat_name,fvalue);
		}
		fclose(stats_file);

		return STATS_check_for_uncollected(p_stats);
	}
	else
	{
		/* was other statistics than time requested */
		if (p_stats->p_head != p_stats->p_tail)
			return EXECUTION_FAILED;
	}

	return EXECUTION_OK;

	cancel:
	errmsg("Syntax error");
	return EXECUTION_FAILED;		
}


/* -------------------------------------------------------------------------- *\
					Output the examples and attributes for execution of the LA
\* -------------------------------------------------------------------------- */


int EXEC_write_atts_exs(t_algorithms alg, t_ex_mask *p_mask)
{
	char filename[256];
	int rv,not_written = 0;
	
	if (ADB_alg_file_support(alg) != ONLY_SEP)
	{
		ADB_get_filename(alg,FT_BOTH,filename);
		rv = !IO_write_both_file(filename,alg,p_mask,1,&not_written);
	}
	else
	{ 
		ADB_get_filename(alg,FT_ATTS,filename);
		IO_write_atts_file(filename,alg,1);

		ADB_get_filename(alg,FT_EXS,filename);
		rv = !IO_write_exs_file(filename,alg,p_mask,1,&not_written);
	}

	if (not_written)
	{
		printf("%d examples were not forwarded to the algorithm.\n",not_written);				
		if (not_written == WS_get_ex_cnt())
		{
			errmsg("Algorithm execution failed");
			rv = EXECUTION_FAILED;
		}
	}
	
	return rv;
}


/* -------------------------------------------------------------------------- *\
		          Remove temporary files from filesystem
\* -------------------------------------------------------------------------- */

int EXEC_remove_temp_files(t_algorithms alg)
{
	char buf[256];

	if (ADB_alg_file_support(alg) == ONLY_SEP)
	{
		ADB_get_filename(alg,FT_ATTS,buf);
		remove(buf);
		ADB_get_filename(alg,FT_EXS,buf);
		remove(buf);
	}
	else
	{
		ADB_get_filename(alg,FT_BOTH,buf);
		remove(buf);
	}
	ADB_get_filename(alg,FT_STATS,buf);
	remove(buf);
}


/* -------------------------------------------------------------------------- *\
        Build the argument vector for the execution of a RUN-command
\* -------------------------------------------------------------------------- */

int EXEC_build_RUN_instruction(t_run_1 param, char ***params,int mode)
{
	t_stats_node *p_stat=0;
	int rv,next=0;
	t_run_node *p_param_list=0;
	char exe_name[256];
	char mode_buf[10];
	char buf[256];

	/* allocate room for 32 parameters or arguments */
	if (!(*params = (char **)calloc(32,sizeof(char *)) )) goto out_of_memory;

	/* name of the executable is in 'argv[0]' */
	if (!ADB_get_exe_name(param.alg,exe_name))
	{
		errmsg("Algorithm has no associated executable"); 
		goto out_of_memory;
	}
	if (!alloc_string(&(*params)[next++],exe_name)) goto out_of_memory;

	/* attribute and example input commands */
	if (get_attr_exs_input_cmds(param.alg,*params,&next) != EXECUTION_OK) 
		goto out_of_memory;

	/* now all the other parameters */
	p_param_list = param.params;
	while (p_param_list) 
	{
		strcpy(buf,p_param_list->selector);
		if (!alloc_string(*params+next++,buf)) goto out_of_memory;
		if (p_param_list->val)
		{ 
			strcat(buf,"=");
			if (!realloc_string(*params+next-1,buf))
				goto out_of_memory;
 			if (!alloc_string(*params+next++,p_param_list->val)) 
 				goto out_of_memory;
		}
		p_param_list = p_param_list->next;
	}

	/* statistics output file */
	if (!alloc_string(*params+next++,"STATSFILE=")) goto out_of_memory;
	ADB_get_filename(param.alg,FT_STATS,buf);
	if (!alloc_string(*params+next++,buf)) goto out_of_memory;


	/* all statistics collection affecting parameters */
	if (!alloc_string(*params+next++,"STAT=")) goto out_of_memory;

	rv = list_get_first(&param.stats,(void **)&p_stat);

	while (rv)	
	{
		/* if TELA times the execution, and not the algorithm */
		/* don't put "TIME"-statistics in as a instruction parameter */
		if (!strcmp(p_stat->name,"TIME") && param.b_time) 
		{
			rv = list_get_next(&param.stats,(void **)&p_stat);
			continue;
		}

		if (!alloc_string(*params+next++,p_stat->name)) goto out_of_memory;
		rv = list_get_next(&param.stats,(void **)&p_stat);	
	}


	/* classifier output file */
	if (!alloc_string(*params+next++,"RULEFILE=")) goto out_of_memory;
	ADB_get_filename(param.alg,FT_RULE,buf);
	if (!alloc_string(*params+next++,buf)) goto out_of_memory;


	/* run-mode (for learning programs that have integrated testing facilities) */
	
	if (!alloc_string(*params+next++,"MODE=")) goto out_of_memory;
	
	if (!alloc_string(*params+next++, "RUN")) goto out_of_memory;

	(*params)[next] = 0; /* argv must be null terminated */

#ifdef DEBUGGING
	print_cmd_args(*params);
#endif

	return EXECUTION_OK;

	out_of_memory:

 	/* do all the freeing of the memory here */
	return EXECUTION_FAILED;
		
}


/* -------------------------------------------------------------------------- *\
		Build the argument vector for execution of the TEST command
\* -------------------------------------------------------------------------- */

int EXEC_build_TEST_instruction(t_test_1 param, char ***params,int mode)
{
	t_stats_node *p_stat=0;
	int rv,next=0;
	char exe_name[256];
	char mode_buf[10];
	char buf[256];

	/* allocate room for 32 parameters or arguments */
	if (!(*params = (char **)calloc(32,sizeof(char *)) )) goto out_of_memory;

	/* name of the executable is in 'argv[0]' */
	if (!ADB_get_exe_name(param.alg,exe_name)) 
	{
		errmsg("Algorithm has no associated executable");
		goto out_of_memory;
	}
	if (!alloc_string(&(*params)[next++],exe_name)) goto out_of_memory;

	/* attribute and example input commands */
	if (get_attr_exs_input_cmds(param.alg,*params,&next) != EXECUTION_OK) 
		goto out_of_memory;

	/* statistics output file */
	if (!alloc_string(*params+next++,"STATSFILE=")) goto out_of_memory;
	ADB_get_filename(param.alg,FT_STATS,buf);
	if (!alloc_string(*params+next++,buf)) goto out_of_memory;


	/* all statistics collection affecting parameters */
	if (!alloc_string(*params+next++,"STAT=")) goto out_of_memory;

	rv = list_get_first(&param.stats,(void **)&p_stat);

	while (rv)	
	{
		/* if TELA times the execution, and not the algorithm */
		/* don't put "TESTTIME"-statistics in as a instruction parameter */
		if (!strcmp(p_stat->name,"TESTTIME") && param.b_testtime) 
		{
			rv = list_get_next(&param.stats,(void **)&p_stat);
			continue;
		}

		if (!alloc_string(*params+next++,p_stat->name)) goto out_of_memory;
		rv = list_get_next(&param.stats,(void **)&p_stat);	
	}


	/* classifier input file */
	if (!alloc_string(*params+next++,"RULEFILE=")) goto out_of_memory;
	ADB_get_filename(param.alg,FT_RULE,buf);
	if (!alloc_string(*params+next++,buf)) goto out_of_memory;


	if (!alloc_string(*params+next++,"MODE=")) goto out_of_memory;
	if (!alloc_string(*params+next++, "TEST")) goto out_of_memory;

	(*params)[next] = 0; /* argv must be null terminated */
	return EXECUTION_OK;

	out_of_memory:

 	/* do all the freeing of the memory here */
	return EXECUTION_FAILED;
		
}

