/*                                                                            *\

	  TELA - Testing Environment for Learning Algorithms - version 1.995
    ==================================================================

		Module:       DL_routines

		Contents:			Data logging routines of TELA. Generate an attribute
									description for a raw example set

		Services:

		DL_gen_attr_descr   - generate the attribute description from inference
													results
		DL_display_inference_stats - print inference statistics to stdout


		Internal functions:
		
		DL_get_attr_count 	- infer the length of the example vector
		DL_collect_values		- read the example set and log values
		DL_infer_types			- on the basis of collected log, infer attribute types 
		DL_set_class_attr   - find a suitable class attribute, and set it

		Author:       Juho Rousu (1995,1996)

\*                                                                            */ 


#include "DL_routines.h"
#include "WS_Database.h"
#include "Alg_Database.h"
#include "IO_scanner.h"

#include <stdio.h>
#include <string.h>
#include <limits.h>

#define HASH_COUNT 100


/* ========================================================================== *\
													INTERNAL FUNCTIONS
\* ========================================================================== */


/* -------------------------------------------------------------------------- *\
			Infer the length of the example vectors i.e. the number of attributes
\* -------------------------------------------------------------------------- */


int DL_get_attr_count(char *exs_file,char *exval_sep, char *ex_sep, 
	char *exs_header, int *count, int *is_header)
{
	FILE *fp;
	t_separator_type rv;
	char buf[512],ex_val[256];

	*count = 0;*is_header = 0;

	if (fp = fopen(exs_file,"r"))
	{
		/* if some header is declared b the algorithm, scan for it */
		if (strlen(exs_header))
		{
			fgets(buf,512,fp);
			if (strncmp(exs_header,buf,strlen(exs_header)))
			{
				/* no, it was no header, it's an example, start again */						
				*is_header = 0;

				fclose(fp);
				fp = fopen(exs_file,"r");
			}
			else *is_header = 1;
		}

		buf[0] = '\0';

		while ((rv = get_ex_token(fp,ex_val,exval_sep[0],ex_sep[0],""))
			 == EXVAL_SEP)
				(*count)++;

		if (*count==0 && rv == EXS_FOOTER)
		{
			printf("No examples in file %s.\n",exs_file);
			return 0;
		}

		/* note, that example vectors *must* be separated by the declared */
		/* example separator */
		if (rv == EXAMPLE_SEP) 
		{
			(*count)++;
			fclose(fp);
			return 1;
		}
		else 
		{
			printf("Example vector terminated incorrectly\n");
			fclose(fp);
			return 0;
		}
	}
	else
	{
		printf("Couldn't open the example file %s\n",exs_file);
		return 0;
	}
}


/* -------------------------------------------------------------------------- *\
								Read the example set and collect data into log
\* -------------------------------------------------------------------------- */

int DL_collect_values(char *exs_file,t_algorithms alg,
	t_dl_node *dl_nodes,int *is_nominals,int is_header,int *ex_cnt)
{
	FILE *fp;
	t_attr_ref a_ref;
	char buf[512];
	char exval_sep[8],example_sep[8], file_footer[256];
	char ex_val[256],a_name[8],dontcare[64],dontknow[64];		
	t_separator_type sep;

	ADB_get_separator(example_sep,EXAMPLE_SEP,alg);
	ADB_get_separator(exval_sep,EXVAL_SEP,alg);
	ADB_get_separator(file_footer,EXS_FOOTER,alg);
	ADB_get_separator(dontcare,DONT_CARE,alg);
	ADB_get_separator(dontknow,DONT_KNOW,alg);

	/* read the file a new */
	if (!(fp = fopen(exs_file,"r")))
	{
		printf("Couldn't open the example file");
		return 0;
	}		

	ex_val[0] = '\0';

	/* read the header */
	if (is_header) fgets(buf,512,fp);

	sep = get_ex_token(fp,ex_val,exval_sep[0],example_sep[0],file_footer);	
	if (sep == EXS_FOOTER || (sep == NO_SEPARATOR && !strlen(file_footer)))
		goto no_more;

	fprintf(stderr,"Collecting example data: ");
	
	while (!feof(fp))
	{
		float f_val;
		int i_val;
		int new_count,old_count;

		if (!(++(*ex_cnt) % HASH_COUNT))
		{
			fprintf(stderr,"[%d]",*ex_cnt);
		};

		WS_get_first_attr(&a_ref,M_JARJ);
		
		do
		{
			/* dontcare values don't affect the inference */
			if (strlen(dontcare))
				if (!strncmp(ex_val,dontcare,strlen(dontcare)))
					continue;
			/* neither do unknown values */
			if (strlen(dontknow))
				if (!strncmp(ex_val,dontknow,strlen(dontknow)))
					continue;

			/* regardless of the current inferred type of the attribute 				*/
			/* database stores the attribute as T_USER, so all different values */
			/* must be stored to the attribute description */
			if (WS_add_attr_val(a_ref,ex_val) != NO_ERROR) return EXECUTION_FAILED;
			old_count = dl_nodes[a_ref].val_count;
			WS_get_attr_domain_size(a_ref,&dl_nodes[a_ref].val_count);		
			new_count = dl_nodes[a_ref].val_count;

			if (!is_figure(ex_val))
			{
				if (old_count != new_count) dl_nodes[a_ref].d_val_cnt++;
			}
			else
			{
				sscanf(ex_val,"%f",&f_val);
				sscanf(ex_val,"%d",&i_val);

				if ((float)i_val == f_val && !strchr(ex_val,'.'))
				{
					if (old_count != new_count) dl_nodes[a_ref].i_val_cnt++;
				}
				else
				{ 
					if (old_count != new_count) dl_nodes[a_ref].f_val_cnt++;
				}									

				dl_nodes[a_ref].low = f_val < dl_nodes[a_ref].low ? f_val:
				 dl_nodes[a_ref].low;
				dl_nodes[a_ref].high = f_val > dl_nodes[a_ref].high ? f_val :
					dl_nodes[a_ref].high;
			}
	
			switch (dl_nodes[a_ref].a_type)
			{
				case T_USER : break;
				case T_FLOAT: 
					if (!is_figure(ex_val))
					{ 
						dl_nodes[a_ref].a_type = T_USER;
						*is_nominals = 1;
					}			
				break;
				case T_INT:		
					if (!is_figure(ex_val)) 
					{
						dl_nodes[a_ref].a_type = T_USER;
						*is_nominals = 1;
					}			
					else if (i_val != f_val || strchr(ex_val,'.')) 
						dl_nodes[a_ref].a_type = T_FLOAT;
				break;
			}
	
			if (get_ex_token(fp,ex_val,exval_sep[0],example_sep[0],file_footer) 
				== EXS_FOOTER) goto no_more;
		}			
		while (WS_get_next_attr(&a_ref,M_JARJ) == NO_ERROR);
	}	/* while !feof */


	no_more:

	fclose(fp);

	return 1;

}


/* -------------------------------------------------------------------------- *\
						Infer types for attributes on the basis of logged data
\* -------------------------------------------------------------------------- */

void DL_infer_types(t_dl_node *dl_nodes,int th,int is_nominals)
{
	t_attr_ref a_ref;
	int class_attribute_set = 0;

	WS_get_first_attr(&a_ref,M_JARJ);

	do
	{
		dl_nodes[a_ref].val_count = dl_nodes[a_ref].d_val_cnt+ 
			dl_nodes[a_ref].i_val_cnt+dl_nodes[a_ref].f_val_cnt;

		switch (dl_nodes[a_ref].a_type)
		{
			case T_USER:
			{
			 	dl_nodes[a_ref].high = (dl_nodes[a_ref].high == FLT_MIN) ? 
					0 : dl_nodes[a_ref].high;
				dl_nodes[a_ref].low =  (dl_nodes[a_ref].low == FLT_MAX) ? 
					0 : dl_nodes[a_ref].low;
			} 
			break;
			case T_INT:  /* change to T_INT, if enough values have been seen */
				if (dl_nodes[a_ref].i_val_cnt >= th)
				{
					t_attr_node *p_attr;
					char *a_name;

					WS_get_attr_name(a_ref,&a_name);
					WS_create_attr(a_name,T_INT,&p_attr);
					WS_replace_attr(a_ref,p_attr);
				}
				else dl_nodes[a_ref].a_type = T_USER;
			break;
			case T_FLOAT:
			{
				t_attr_node *p_attr;
				char *a_name;
				
				WS_get_attr_name(a_ref,&a_name);
				WS_create_attr(a_name,T_FLOAT,&p_attr);
				WS_replace_attr(a_ref,p_attr);
			}
			break;
		}
	}
	while (WS_get_next_attr(&a_ref,M_JARJ) == NO_ERROR);

	return;
}


/* -------------------------------------------------------------------------- *\
						Find a suitable class attribute and set it, if found
\* -------------------------------------------------------------------------- */

int DL_set_class_attr(t_algorithms alg)
{
	t_attr_types a_type;
	t_attr_ref a_ref;

	if (ADB_get_class_place_in_ex(alg))
	{
		/* class is first in examples */
		WS_get_first_attr(&a_ref,M_JARJ);	
		WS_get_attr_type(a_ref,&a_type);
		
		if (a_type == T_USER)
		{
			WS_set_class_attr(a_ref);
			return 1;
		}
	}	
	else
	{
		WS_get_last_attr(&a_ref,M_JARJ);
		WS_get_attr_type(a_ref,&a_type);

		if (a_type == T_USER)
		{
			WS_set_class_attr(a_ref);
			return 1;
		}
	}

	return 0;
}


/* ========================================================================== *\
													EXTERNAL FUNCTIONS
\* ========================================================================== */


/* -------------------------------------------------------------------------- *\
						Generate an attribute description for examples
\* -------------------------------------------------------------------------- */

int DL_gen_attr_descr(t_algorithms alg, char *exs_file, int th, t_dl_node **dl_data,int *ex_cnt)
{
	int attr_count,is_nominals=0,a_ref,is_header;
	FILE *fp;
	t_attr_node *p_attr_node=0;
	t_dl_node *dl_nodes;	
	char exval_sep[8],example_sep[8],a_name[64],exs_header[512];

	ADB_get_separator(example_sep,EXAMPLE_SEP,alg);
	ADB_get_separator(exval_sep,EXVAL_SEP,alg);
	ADB_get_separator(exs_header,EXS_HEADER,alg);

	WS_dispose_dbase();

	/* count values on the first example line */
	if (!DL_get_attr_count(exs_file,exval_sep,example_sep,exs_header,&attr_count,
		&is_header))
		return 0;

	printf("Inferred example vector length is %d.\n",attr_count);

	/* generate attribute nodes */
	for (a_ref=0;a_ref<attr_count;a_ref++)
	{
		sprintf(a_name,"unnamed%d",a_ref+1);
		if (WS_create_attr(a_name,T_USER,&p_attr_node) != NO_ERROR)
			return 0;
		if (WS_add_attr(p_attr_node) != NO_ERROR)
			return 0;
	}

	/* keep a temporary vector for inferred attribute types: they may 
		change along the way */
	dl_nodes = (t_dl_node*)calloc(attr_count,sizeof(t_dl_node));

	for (a_ref=0;a_ref< attr_count;a_ref++)
	{ 
		dl_nodes[a_ref].a_type = T_INT;
		dl_nodes[a_ref].low = FLT_MAX;
		dl_nodes[a_ref].high = FLT_MIN;
	}

	/* read the file and collect statistics of each example vector component */
	DL_collect_values(exs_file,alg,dl_nodes,&is_nominals,is_header,ex_cnt);
	
	/* based on the collected values, infer a type for each attribute */
	DL_infer_types(dl_nodes,th,is_nominals);

	*dl_data = dl_nodes;

	return 1;
}


/* -------------------------------------------------------------------------- *\
				Output statistics of the example file and inference process 
\* -------------------------------------------------------------------------- */

int DL_display_inference_stats(t_dl_node *p_dl_data, int ex_cnt)
{
	char *a_name;
	int i;
	t_attr_ref a_ref;
  t_command stat_cmd;

	printf("\nStatistics of inferred attribute declaration:\n");
	printf(  "---------------------------------------------\n");

	printf("Attribute    Inferred      Value counts by type     Boundary values\n");
	printf("  name        type          nom.  int.  cont.       low     high \n");
                   /*            123456123456123456                   */


	WS_get_first_attr(&a_ref,M_JARJ);

	do
	{
		WS_get_attr_name(a_ref,&a_name);
		 printf("%-13s",a_name);
		switch (p_dl_data[a_ref].a_type)
		{
			case T_USER: 
						  printf("nominal        "); break;
			case T_INT:
							printf("integer        "); break;
			case T_FLOAT:
							printf("continuous     "); break;
		}

		printf("%-6d",p_dl_data[a_ref].d_val_cnt);
		printf("%-6d",p_dl_data[a_ref].i_val_cnt);
		printf("%-12d",p_dl_data[a_ref].f_val_cnt);

		printf("%-10.2f",p_dl_data[a_ref].low);
		printf("%-6.2f",p_dl_data[a_ref].high);

		printf("\n");
	}
	while (WS_get_next_attr(&a_ref,M_JARJ) == NO_ERROR);
	
	printf("\nTotal example count: %d ",ex_cnt);

	if (WS_is_class_attr())
	{
		WS_get_class_attr(&a_ref);
		WS_get_attr_name(a_ref,&a_name);
		printf("Class attribute: %s\n",a_name);
	}
	else printf("No class attribute set.\n");

	printf("\n");

	return 1;
}
