/*                                                                            *\

	  TELA - Testing Environment for Learning Algorithms - version 1.995
    ==================================================================

		Module:  IO_server 

		Contents:    TELA I/O-server module

		Description:  handles the TELA I/0 operations, namely, the reading 
									and writing of attributes, examples and classifiers (rules) 
									in algorithm specific formats to/from filesystem. 
		
		Author: Tommi Timonen (1992), Juho Rousu (1995,1996)

		Services:

		General
		=======

		IO_initialize - initializes the I/0-server
		IO_set_round - tells the I/O-server which DO-round is in question (0 - none)
		IO_get_expanded_name - replaces '!' with current round index in the filename

	  Input operations
	  ================

	  IO_read_atts - read attribute description in given data format  
		IO_read_exs  - read examples in given data format
		IO_read_rule - read a rule

		IO_read_atts_file - read an attribute file
		IO_read_exs_file  - read an example file
  	IO_read_both_file - read a combined (atts&exs) file
  	IO_read_rule_file - read a rule file (with an attribute description)

		IO_read_alg_rule   - read the rule generated by an algorithm
		IO_write_alg_rule  - write the rule to algorithm

  	Output operations
  	=================

  	IO_write_atts - write attribute description in given data format  
  	IO_write_exs  - write examples in given data format
  	IO_write_rule - write a rule

  	IO_write_atts_file - write an attribute file
  	IO_write_exs_file   - write an example file
  	IO_write_both_file  - write a combined (atts&exs) file
  	IO_write_rule_file  - write a rule file (with an attribute description) 

		Internal functions:

		read_attr       - read a declaration for an attribute
    read_ex_val     - read an example vector component
		read_ex         - read an example vector

		write_attr_vals  - write an attribute domain specification
    write_ex_val     - write an example vector component 
		write_example    - write an example vector
		write_class_attr - write a class attribute declaration
		write_attr       - write an attribute declaration 

    attribute_struct_cmp  - compare two attribute descriptions

\*                                                                            */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>

#include "tela.h"
#include "IO_server.h"
#include "IO_scanner.h"
#include "Alg_Database.h"
#include "WS_Database.h"

static int round;   /* internal variable indicating the current DO-loop round */


/* ========================================================================== *\
                             INTERNAL FUNCTIONS     
\* ========================================================================== */


/* -------------------------------------------------------------------------- *\
					     Read an attribute declaration form a file 
\* -------------------------------------------------------------------------- */

int read_attr(FILE *atts_file,char *nv_s,char *tv_s, char *v_s, char *a_s, 
char *na_s, char *a_e_s,char *i_s,char *c_s,char *d_s, int *rv,t_algorithms alg)
{
	int len;
	t_separator_type sep;
	char buf[512],*rest,valbuf[64];
	char a_name[64];
	t_attr_ref a_ref;

	/* loop here until an non-empty line is read or eof */
	do
	{

		if (!fgets(buf,512,atts_file))
		{ 
			if (strlen(a_e_s)) /* should there have been an attribute separator */
				*rv = EXECUTION_FAILED; 
			else
				*rv = EXECUTION_OK;		
			return 0; 
		}

		rest = buf;

		sep = get_attr_token(rest,&len,nv_s,tv_s,v_s,a_s,na_s,a_e_s,c_s);
	
		if (sep == ATTS_EXS_SEP)
		{ 
			*rv = EXECUTION_OK; 
			return 0; 
		}
	}
	while (!len);

	/* attribute name ? */
	if (sep == NAME_VAL_SEP) /* named attribute */
	{
		strncpy(a_name,rest,len);
		a_name[len] = '\0';

		/* if a 'unnamed' attribute is encountered reassign it's name to
			avoid name clashes */
		if (!strncmp(a_name,"unnamed",7))
		{
			sprintf(a_name+7,"%d",WS_get_attr_cnt()+1);
		}

		rest = eatwhite(rest+len+strlen(nv_s)); /* move past separator */
			
		/* get fresh token */		
		sep = get_attr_token(rest,&len,nv_s,
			tv_s,v_s,a_s,na_s,a_e_s,c_s);
	}
	else
	{
		/* assign a generic name for the attribute*/
		sprintf(a_name,"unnamed%d",WS_get_attr_cnt()+1);
	}


	if (sep == NUMATT_SEP 
		|| sep == NO_SEPARATOR || sep == ATT_SEP)  /* integer or continuous */
	{
		/* rest points to a type specifier, which one ? */
		if (strlen(i_s) ? !strncmp(i_s,rest,strlen(i_s)):0)
		{
			WS_create_and_add_attr(a_name,T_INT,&a_ref);	 /* int */
			goto end;
		}
		else if (strlen(c_s) ? !strncmp(c_s,rest,strlen(c_s)):0)
		{
			WS_create_and_add_attr(a_name,T_FLOAT,&a_ref); /* float */
			goto end;
		}
	}	

	if (sep == TYPE_VAL_SEP ? 1: ADB_is_equiv_sep(sep,TYPE_VAL_SEP,alg))
	{
		/* attribute excplicitly declared as nominal */
		if (strncmp(d_s,rest,strlen(d_s)))
		{
			errmsg("Syntax Error: Invalid nominal type specifier");
			*rv = EXECUTION_FAILED;
			return 0;
		}

		rest = eatwhite(rest+len+strlen(tv_s));
		sep = get_attr_token(rest,&len,nv_s,tv_s,v_s,a_s,na_s,a_e_s,c_s);
	}
	
	WS_create_and_add_attr(a_name,T_USER,&a_ref);

	/* value list */
	while (sep == VAL_SEP)
	{
		strncpy(valbuf,rest,len);
		valbuf[len] = '\0';
		if (WS_add_attr_val(a_ref,valbuf) != NO_ERROR)
		{
			errmsg("Couldn't add value to the attribute");
			*rv = EXECUTION_FAILED;
			return 0;
		}
		rest = eatwhite(rest+len+strlen(v_s));
		sep = get_attr_token(rest,&len,nv_s,tv_s,v_s,a_s,na_s,a_e_s,c_s);
	}

	if (sep != ATT_SEP && sep != ATTS_EXS_SEP)
	{
		char msg[256];
		sprintf(msg,
			"Syntax Error: Value list of attribute %s terminated incorrectly",a_name);
		errmsg(msg);
		*rv = EXECUTION_FAILED;
		return 0;
	}
	else
	{
		strncpy(valbuf,rest,len);
		valbuf[len] = '\0';
		if (WS_add_attr_val(a_ref,valbuf) != NO_ERROR)
		{
			errmsg("Couldn't add value to the attribute");
			*rv = EXECUTION_FAILED;
			return 0;
		}
	}

	end:

	*rv = EXECUTION_OK;
	return 1;
}


/* -------------------------------------------------------------------------- *\
                  Reads an example vector component from a file
\* -------------------------------------------------------------------------- */

int read_ex_val(char *exval, t_attr_ref a_ref,char *dontcare,char* dontknow,
	int line)
{
	t_attr_types a_type;
	/* is it a dontcare symbol ? */
	if (!strcmp(exval,dontcare))
		WS_set_exval_kind(a_ref,line,T_DONT_CARE);	
	else if (!strcmp(exval,dontknow)) /* or an unknown value */
		WS_set_exval_kind(a_ref,line,T_DONT_KNOW);
	else /* ok, it is a normal value */
	{
		WS_set_exval_kind(a_ref,line,T_OK);
		WS_get_attr_type(a_ref,&a_type);
		switch (a_type)
		{
			case T_INT:
			{
				int intval;
				sscanf(exval,"%d",&intval);
				WS_set_integer_exval(a_ref,line,intval);
			}
			break;
			case T_FLOAT:
			{
				float fval;
				sscanf(exval,"%f",&fval);
				WS_set_float_exval(a_ref,line,fval);
			}
			break;
			case T_BINARY:
			case T_USER:
			{
				int index;
				if (WS_get_val_index(a_ref,exval,&index) == NO_ERROR)
					WS_set_nominal_exval(a_ref,line,index);
				else
				{
					char msg[256];
					sprintf(msg,"Syntax Error: Invalid attribute value: %s",exval);
					errmsg(msg);
					return 0;
				}
			}
		} /* switch a_type */
	}
	return 1;
}


/* -------------------------------------------------------------------------- *\
       Read an example vector from a file 
\* -------------------------------------------------------------------------- */

int read_ex(FILE *exs_file,char *ex_sep,char *exval_sep, char *file_footer, 
  char *dontcare, char *dontknow, char *comment_sep,  
  int *rv,int class_is_first)
{
	t_separator_type sep;
	char exval[256];
	int classified=0;
	t_attr_ref last,a_ref,cl_ref=-1;
	int len, e_count;

	if (feof(exs_file)) 
	{ 
		/* end of file encountered */

		if (strlen(file_footer))  /* if a file-footer is declared */
			*rv = EXECUTION_FAILED; /* return error code */
		else
			*rv = EXECUTION_OK;     /* else, all is ok */
		return 0;
	}	
			
	sep = get_ex_token(exs_file,exval,exval_sep[0],ex_sep[0],file_footer,
		comment_sep[0]);	
	if (sep == EXS_FOOTER || (sep == NO_SEPARATOR && !strlen(file_footer)))
	{ 
		*rv = EXECUTION_OK;
		return 0; 
	}

	e_count = WS_get_ex_cnt(); 
	WS_add_ex_row();

	if (classified = WS_is_class_attr()) WS_get_class_attr(&cl_ref);
	WS_get_last_attr(&last,M_JARJ);
	WS_get_first_attr(&a_ref,M_JARJ);

	/* is the place of the class attribute values other than last */
	if (classified && class_is_first)
	{		
		if (!read_ex_val(exval,cl_ref,dontcare,dontknow,e_count))
		{
			*rv = EXECUTION_FAILED;
			return 0;
		}

		if (sep != EXVAL_SEP)
		{
			errmsg("Syntax Error: Invalid example value separator");
			*rv = EXECUTION_OK;
			return 0;
		}
		sep = get_ex_token(exs_file,exval,exval_sep[0],ex_sep[0],file_footer
			, comment_sep[0]);
	}

	/* now, all the other attributes */
	do
	{
		if (classified && a_ref == cl_ref) 			/* skip over the class attribute */
		{
			if (WS_get_next_attr(&a_ref,M_JARJ) != NO_ERROR)
				break;
		}

		if (sep != EXVAL_SEP) 
		{
			/* the only case where we are allowed to see other separator than
				EXVAL_SEP is when we are handling the last attribute, and when the
				class attribute values are at the start of the example value vector */
			if (a_ref != last || (classified && class_is_first))
			{
				errmsg("Syntax Error: Example vector too short");
				*rv = EXECUTION_FAILED;
				return 0;
			}
		}

		if (!read_ex_val(exval,a_ref,dontcare,dontknow,e_count))
		{
			*rv = EXECUTION_FAILED;
			return 0;
		}
		
		next:

		if (WS_get_next_attr(&a_ref,M_JARJ) == NO_ERROR ? a_ref != cl_ref : 0)
			sep = get_ex_token(exs_file,exval,exval_sep[0],ex_sep[0],file_footer
				, comment_sep[0]);
		else break;
	}
	while (1);


	/* if the class attribute is unhandled, do handle it! */
	if (classified && !class_is_first)
	{			
		sep = get_ex_token(exs_file,exval,exval_sep[0],ex_sep[0],file_footer
			, comment_sep[0]);
		if (!read_ex_val(exval,cl_ref,dontcare,dontknow,e_count))
		{
			*rv = EXECUTION_FAILED;
			return 0;
		}

		if (sep != EXAMPLE_SEP && sep != EXVAL_SEP)
		{
			char msg[256];
			sprintf(msg,
				"Syntax Error: (Example %d) Invalid example separator",e_count+1);
			errmsg(msg);
			*rv = EXECUTION_FAILED;
			return 0;
		}
	}

	return 1;
}


/* -------------------------------------------------------------------------- *\
                      Write an example value to a file
\* -------------------------------------------------------------------------- */

int write_ex_val(FILE *exs_file,t_ex_node *p_ex, t_attr_ref a_ref,
	char *dont_care, char *dont_know, t_attr_types a_type,
	t_attr_types num_support)
{
	char *s_val;
	switch (p_ex->ok)
	{
		case T_DONT_CARE:
			if (strlen(dont_care))
				fprintf(exs_file,"%s",dont_care);
			else 
				return EXECUTION_FAILED; /* cannot print don't cares */
			break;
		case T_DONT_KNOW:					
			if (strlen(dont_know))
				fprintf(exs_file,"%s",dont_know);
			else 
				return EXECUTION_FAILED; /* cannot print don't knows */
			break;
		case T_OK:
			/* what kind of a value ? */
			switch (a_type)
			{				
				case T_FLOAT: 
					if (num_support != T_INT)
						fprintf(exs_file,"%f",p_ex->val.float_val); 
					else
						fprintf(exs_file,"%d",(int)p_ex->val.float_val);
					break;
				case T_INT: fprintf(exs_file,"%d",p_ex->val.int_val); break;
				case T_BINARY:
				case T_USER: 
				{
					WS_get_index_attr_val(a_ref,p_ex->val.int_val,&s_val);
					fprintf(exs_file,"%s",s_val);
				} break;
			}
		break;
	}
	return EXECUTION_OK;
}


/* -------------------------------------------------------------------------- *\
                     Write an example vector to file 
\* -------------------------------------------------------------------------- */

int write_example(FILE *exs_file,int ex_index, char *dn, char *dc, char *exval_sep, char *examp_sep, int cl_is_first, 
	int classified, t_attr_types num_support)
{
	t_attr_ref attr_ref,cl_attr_ref=-1;
	int i,a_count,b_first=1;

	int no_dn,no_dc;

	t_ex_ok kind;

	a_count = WS_get_attr_cnt();

	/* if target data format does not support don't care or don't know values */
	/* check if such exist at the place of this example                       */
	no_dn = !strlen(dn);
	no_dc = !strlen(dc);

	if (no_dn || no_dc)
	{
		if (WS_get_first_attr(&attr_ref,M_JARJ) == NO_ERROR)
		{
			do
			{
				WS_get_exval_kind(attr_ref,ex_index,&kind);

				switch (kind)
				{
					case T_OK:break;
					case T_DONT_KNOW: 
						if (no_dn) return EXECUTION_FAILED;
						break;
					case T_DONT_CARE: 
						if (no_dc) return EXECUTION_FAILED;
						break;
				}
			}
			while (WS_get_next_attr(&attr_ref,M_JARJ) == NO_ERROR);
		}
	}

	if (classified)
	{
		if (WS_get_class_attr(&cl_attr_ref) != NO_ERROR) return 0;
	
		/* is class attribute required to be the first */
		if (cl_is_first)
		{
			if (write_ex_val(exs_file,&WS_attr_table[cl_attr_ref].examples[ex_index],
				cl_attr_ref,dc,dn,WS_attr_table[cl_attr_ref].a_type, num_support) 
			!= EXECUTION_OK) 
				return EXECUTION_FAILED;		
			fprintf(exs_file,"%s",exval_sep); b_first=0;
		}
	}

	if (WS_get_first_attr(&attr_ref,M_JARJ) == NO_ERROR)
	{
		do
		{
			if (classified && attr_ref == cl_attr_ref) continue;

			if (!b_first) fprintf(exs_file,"%s",exval_sep);

			if (write_ex_val(exs_file,&WS_attr_table[attr_ref].examples[ex_index],
				attr_ref,dc,dn,WS_attr_table[attr_ref].a_type,num_support) 
			!= EXECUTION_OK) 
				return EXECUTION_FAILED;		

			b_first = 0;
		}
		while (WS_get_next_attr(&attr_ref,M_JARJ) == NO_ERROR);
	}

	if (classified && !cl_is_first)
	{
		fprintf(exs_file,"%s",exval_sep);

		if (write_ex_val(exs_file,&WS_attr_table[cl_attr_ref].examples[ex_index],
			cl_attr_ref,dc,dn,WS_attr_table[cl_attr_ref].a_type, num_support) 
		!= EXECUTION_OK) 
			return EXECUTION_FAILED;		
	}
	fprintf(exs_file,"%s%s",examp_sep,
		examp_sep[strlen(examp_sep)-1] == '\n' ? "" : "\n");

	return EXECUTION_OK;
}


/* -------------------------------------------------------------------------- *\
                Write a domain specification of and attribute
\* -------------------------------------------------------------------------- */

int write_attr_vals(FILE *atts_file, t_attr_ref a_ref, char *val_sep, 
	char *att_sep, char *numatt_sep, char *t_v_s, 
	char *int_decl, char *cont_decl, char *discr_decl)
{
	int val_count,i;
	char *val_name;

	switch (WS_attr_table[a_ref].a_type)
	{
		case T_INT: 
			if (strlen(int_decl))
			{
				fprintf(atts_file,"%s%s%s",int_decl,numatt_sep,
					numatt_sep[strlen(numatt_sep)-1] == '\n'? "": "\n" ); break;
			}
			/* fall trough if grammar doesn't support T_INT */
		case T_FLOAT: if (strlen(cont_decl)) 
				fprintf(atts_file,"%s%s%s",cont_decl,numatt_sep,		
					numatt_sep[strlen(numatt_sep)-1] == '\n'? "": "\n"); 
			else fprintf(atts_file,"%s%s%s",int_decl,numatt_sep,
				numatt_sep[strlen(numatt_sep)-1] == '\n'? "": "\n");
		break;
		case T_BINARY:
		case T_USER:
			if (strlen(discr_decl)) fprintf(atts_file,"%s ",discr_decl);
			if (strlen(t_v_s)) fprintf(atts_file,"%s",t_v_s);
	
			/* now all the discrete values */
			WS_get_attr_domain_size(a_ref,&val_count);

			for (i=0;i< val_count-1; i++)
			{
				WS_get_index_attr_val(a_ref,i,&val_name);
				fprintf(atts_file,"%s%s",val_name,val_sep);	
			}
			WS_get_index_attr_val(a_ref,val_count-1,&val_name);
			fprintf(atts_file,"%s",val_name);
			fprintf(atts_file,"%s%s",att_sep,att_sep[strlen(att_sep)-1] 
				== '\n' ? "" : "\n");
	
		break;
	}
	return EXECUTION_OK;
}


/* -------------------------------------------------------------------------- *\
            Write an attribute declaration to a file 
\* -------------------------------------------------------------------------- */

int write_attr(int place,FILE *atts_file, t_attr_ref a_ref, char *nameval_sep,
	char *att_sep, char *numatt_sep,  char *val_sep,char *t_v_s, char *int_decl, 
	char *cont_decl, char *discr_decl)
{
	char *name;

	WS_get_attr_name(a_ref,&name);

	if (!strncmp("unnamed",name,7))
		fprintf(atts_file,"unnamed%d%s",place,nameval_sep);
	else
		fprintf(atts_file,"%s%s",name,nameval_sep);

	return write_attr_vals(atts_file,a_ref,val_sep,att_sep,numatt_sep,t_v_s,
		int_decl,cont_decl,discr_decl);
}


/* -------------------------------------------------------------------------- *\
              Write a class attribute declaration to file
\* -------------------------------------------------------------------------- */

int write_class_attr(int place,FILE *atts_file,t_attr_ref a_ref, int unnamed,
	char *nameval_sep, char *att_sep, char *val_sep,char *t_v_s, 
	char *int_decl, char *cont_decl, char *discr_decl)
{
	if (!unnamed)
		return write_attr(place,atts_file,a_ref,nameval_sep,att_sep,"",
			val_sep,t_v_s, int_decl,cont_decl,discr_decl);
	else
		return write_attr_vals(atts_file,a_ref,val_sep,att_sep,"",t_v_s,
			int_decl,cont_decl, discr_decl);
}


/* -------------------------------------------------------------------------- *\
		      Compare the attribute description in memory to that in file
\* -------------------------------------------------------------------------- */

int attribute_struct_cmp (FILE *rule_file, t_algorithms alg) 
{
	if (WS_example_db_backup("cmp_attr_struct"))
	{
		errmsg("Couldn't back the example database up");
		return EXECUTION_FAILED;
	}
		
	WS_dispose_dbase();

	/* FT_BOTH lures the IO_read_atts to think he is reading attribute and
			example file instead of rule_file (and accept the ATTSEXSSEP) */
  if (!IO_read_atts(rule_file,alg,FT_BOTH))
	{
		errmsg("Couldn't read the attributes");
  	return 0;
  } 
	else if (!WS_attr_table_cmp ("cmp_attr_struct")) 
		return 0;

	return 1;

}



/* ========================================================================== *\
                             EXTERNAL SERVICES     
\* ========================================================================== */


/* -------------------------------------------------------------------------- *\
				                   Initialize I/O-server
\* -------------------------------------------------------------------------- */

void IO_initialize(void)
{
	round=0;
}


/* -------------------------------------------------------------------------- *\
              Tell I/0-server the number of the DO-round we are in, if any
						
		Note: round 0 (zero) means that we are not inside DO-loop 
\* -------------------------------------------------------------------------- */

void IO_set_round(int new_round)
{
	round=new_round;   
}


/* -------------------------------------------------------------------------- *\
						Replace '!' with DO-round counter in the given filename
\* -------------------------------------------------------------------------- */

char *IO_get_expanded_name(char *name)
{
	static char name2[MAXLENGTH+10];      
	int i=0,j=0;
	char *homedir;

	/* replace ~ with the user's home directory */

	if (name[0] == '~')
	{
		homedir = getenv("HOME");

		sprintf(name2,"%s",homedir);

		i=1;j=strlen(homedir);
	}


	for(;j<MAXLENGTH&&name[i]!='\0';i++)
	{
		if (name[i]=='!')
		{   			
			sprintf(&name2[j],"%i",round);
			j=j+strlen(&name2[j]);
		}
		else
			name2[j++]=name[i];
	}
	name2[j]='\0';
	return name2;
}


/* -------------------------------------------------------------------------- *\
					Write attributes to file in algorithm specific data format 
\* -------------------------------------------------------------------------- */

int IO_write_atts(FILE *atts_file,t_algorithms alg,int classified)
{
	t_alg_decl *p_alg_decl;
	t_attr_ref cl_attr_ref,attr_ref;
	char *a_name,nameval_sep[8],val_sep[8],att_sep[8],numatt_sep[8],t_v_s[8],
		int_decl[64],cont_decl[64],discr_decl[64];
	int i=1;

	ADB_get_separator(nameval_sep,NAME_VAL_SEP,alg);		
	ADB_get_separator(val_sep,VAL_SEP,alg);
	ADB_get_separator(numatt_sep,NUMATT_SEP,alg);		
	ADB_get_separator(att_sep,ATT_SEP,alg);
	ADB_get_separator(t_v_s,TYPE_VAL_SEP,alg);
	ADB_get_separator(int_decl,INT_TYPE,alg);
	ADB_get_separator(cont_decl,CONT_TYPE,alg);
	ADB_get_separator(discr_decl,NOM_TYPE,alg);

	if (!ADB_alg_decl_by_ref(alg,&p_alg_decl))
	{
		errmsg("No such algorithm");
		return 0;
	}

	if (!WS_is_class_attr())
	{
		int ex_place,at_place;

		ex_place = ADB_get_class_place_in_ex(alg);
		at_place = ADB_get_class_place_in_atts(alg);

		if (ex_place != at_place)
		{
			char msg[256];
			sprintf(msg,"Algorithm %s does not support writing unclassified data.",
			p_alg_decl->alg_name);
			errmsg(msg);
			return 0;
		}
	}

	if (classified)
	{

		if (WS_get_class_attr(&cl_attr_ref) != NO_ERROR)
		{
			errmsg("Class Attribute not defined.");
			return 0;
		}

		/* does the class attribute have to be first */
		if (p_alg_decl->attr_decl.class.b_first)
		{
			if (write_class_attr(i++,atts_file,cl_attr_ref,
				p_alg_decl->attr_decl.class.b_unnamed, nameval_sep, att_sep, val_sep,
				t_v_s, int_decl, cont_decl, discr_decl) != EXECUTION_OK)
			{
				errmsg("Couldn't write the class attribute");
				return 0;
			}
		}
	}

	/* other than the class attribute */	
	if (WS_get_first_attr(&attr_ref,M_JARJ) == NO_ERROR)
	{
		char *s_val;		
		do
		{
			if (classified && attr_ref == cl_attr_ref) continue;

			if (write_attr(i++,atts_file,attr_ref,nameval_sep,att_sep,
				numatt_sep,val_sep,t_v_s, int_decl,cont_decl,discr_decl) 
				!= EXECUTION_OK)
			{
				char msg[256],*name;
				WS_get_attr_name(attr_ref,&name);
				sprintf(msg,"Couldn't write the attribute %s",name);
				errmsg(msg);
				return 0;
			}
		}
		while (WS_get_next_attr(&attr_ref,M_JARJ) == NO_ERROR);
	}

	if (classified)
	{
		/* if not yet printed (as first) print class attribute last */
		if (!p_alg_decl->attr_decl.class.b_first)
		{
			if (write_class_attr(i,atts_file,cl_attr_ref,
				p_alg_decl->attr_decl.class.b_unnamed, nameval_sep, att_sep, val_sep,
				t_v_s, int_decl, cont_decl, discr_decl) != EXECUTION_OK)
			{
				errmsg("Couldn't write the class attribute");
				return 0;
			}
		}
	}

	return 1;
}


/* -------------------------------------------------------------------------- *\
					 Write examples to file in algorithm specific data format 
\* -------------------------------------------------------------------------- */

int IO_write_exs(FILE *exs_file,t_algorithms alg, 
	t_ex_mask *p_mask, int classified, int *not_written)
{	
	char dont_know[8],dont_care[8],exval_sep[8],examp_sep[8],
		int_type[64], cont_type[64];
	t_attr_types num_support;
	int first,i;

	ADB_get_separator(dont_know,DONT_KNOW,alg);
	ADB_get_separator(dont_care,DONT_CARE,alg);
	ADB_get_separator(exval_sep,EXVAL_SEP,alg);
	ADB_get_separator(examp_sep,EXAMPLE_SEP,alg);
	ADB_get_separator(int_type,INT_TYPE,alg);
	ADB_get_separator(cont_type,CONT_TYPE,alg);

	/* which numeric types are supported */
	num_support = strlen(int_type) ? T_ANY : T_FLOAT;
	num_support = strlen(cont_type) ? num_support : T_INT;

	first = ADB_get_class_place_in_ex(alg);
	*not_written = 0;

	for (i =0; i < WS_get_ex_cnt(); i++)
	{
		/* write only active examples */
		if (WS_is_active_ex(i,p_mask))
		{
			if (write_example(exs_file,i,dont_know,dont_care,
				exval_sep,examp_sep,first,classified,num_support) != EXECUTION_OK)
				(*not_written)++;
		}
	}
	return 1;
}


/* -------------------------------------------------------------------------- *\
                            Write rule to file
\* -------------------------------------------------------------------------- */

int IO_write_rule(FILE *rule_file,t_algorithms alg,char *comment)
{

	char ch;
	FILE *file;

	char internal_rulefile[512];

	WS_get_internal_rulename(internal_rulefile);

 	file = fopen(internal_rulefile,"r");

	if (!file) return 0;

	while ((ch = getc(file)) != EOF) putc(ch,rule_file);	

	fclose(file);

	return 1;
}


/* -------------------------------------------------------------------------- *\
                     Read rule from file
\* -------------------------------------------------------------------------- */

int IO_read_rule(FILE *rule_file,t_algorithms alg)
{
	char ch;
	FILE *file;

	char internal_rulefile[512];

	WS_get_internal_rulename(internal_rulefile);

 	file = fopen(internal_rulefile,"w");

	if (!file) return 0;

	while ((ch = getc(rule_file)) != EOF) putc(ch,file);	

	fclose(file);

	return 1;

}


/* -------------------------------------------------------------------------- *\
				 Read attributes from a file in algorithm specific data format
\* -------------------------------------------------------------------------- */

int IO_read_atts(FILE *atts_file,t_algorithms alg,t_filetype f_type)
{
	t_alg_decl *p_alg_decl;
	t_attr_ref cl_ref;
	t_attr_types a_type;
	int rv;
	char nameval_sep[8],val_sep[8],att_sep[8],typeval_sep[8],numatt_sep[8],
		end_marker[256],i_sep[64],c_sep[64],d_sep[64];

	ADB_get_separator(nameval_sep,NAME_VAL_SEP,alg);
	ADB_get_separator(typeval_sep,TYPE_VAL_SEP,alg);		
	ADB_get_separator(val_sep,VAL_SEP,alg);		
	ADB_get_separator(att_sep,ATT_SEP,alg);
	ADB_get_separator(numatt_sep,NUMATT_SEP,alg);
	ADB_get_separator(i_sep,INT_TYPE,alg);
	ADB_get_separator(c_sep,CONT_TYPE,alg);
	ADB_get_separator(d_sep,NOM_TYPE,alg);
	
	if (f_type == FT_BOTH)
		ADB_get_separator(end_marker,ATTS_EXS_SEP,alg);
	else
		ADB_get_separator(end_marker,ATTS_FOOTER,alg);

	/* read attribute description */
	while (read_attr(atts_file,nameval_sep,typeval_sep,val_sep,
		att_sep,numatt_sep,end_marker,i_sep,c_sep,d_sep,&rv,alg));

	if (rv == EXECUTION_FAILED) return 0;

	ADB_alg_decl_by_ref(alg,&p_alg_decl);

	if (WS_get_attr_cnt() < 2)
	{
		errmsg("At least two attributes must be declared");
		return 0;
	}

	if (p_alg_decl->attr_decl.class.b_first)
		WS_get_first_attr(&cl_ref,M_JARJ);	
	else
		WS_get_last_attr(&cl_ref,M_JARJ);

	WS_get_attr_type(cl_ref,&a_type);
	
	if (a_type == T_USER) WS_set_class_attr(cl_ref);

	return 1;
}


/* -------------------------------------------------------------------------- *\
			    Read examples from a file in algorithm specific data format
\* -------------------------------------------------------------------------- */

int IO_read_exs(FILE *exs_file,t_algorithms alg, t_filetype f_type)
{
	char comment_sep[8],ex_sep[8],exval_sep[8],dontcare[8],
		dontknow[8],footer[256];
	int rv,cl_f;

	ADB_get_separator(ex_sep,EXAMPLE_SEP,alg);
	ADB_get_separator(exval_sep,EXVAL_SEP,alg);
	ADB_get_separator(dontcare,DONT_CARE,alg);
	ADB_get_separator(dontknow,DONT_KNOW,alg);
	ADB_get_separator(comment_sep,COMMENT_SEP,alg);

	cl_f = ADB_get_class_place_in_ex(alg);

	/* get coorect filefooter for lookahead purposes */
	if (f_type == FT_EXS)
		ADB_get_separator(footer,EXS_FOOTER,alg);
	else
		ADB_get_separator(footer,BOTH_FOOTER,alg);

	while (read_ex(exs_file,ex_sep,exval_sep,footer,dontcare,
		dontknow,comment_sep,&rv,cl_f));

	if (rv == EXECUTION_FAILED) 
		return 0;
	else
		return 1;
}


/* -------------------------------------------------------------------------- *\
                   Write an attribute description file
\* -------------------------------------------------------------------------- */

int IO_write_atts_file(char *filename,t_algorithms alg, int classified)
{
	FILE *atts_file;
	t_alg_decl *p_alg_decl;
	char buf[256];

	if (!ADB_alg_decl_by_ref(alg,&p_alg_decl)) return 0;

	/* open file for writing */
	atts_file = fopen(filename,"w");
	if (!atts_file) return 0;

	/* write header */
	if (ADB_get_separator(buf,ATTS_HEADER,alg))
		fprintf(atts_file,"%s\n",buf);

	/* and the attributes */
	if (!IO_write_atts(atts_file,alg,classified)) return 0;

	if (ADB_get_separator(buf,ATTS_FOOTER,alg))
		fprintf(atts_file,"%s\n",buf);

	fclose(atts_file);

	return 1;
}


/* -------------------------------------------------------------------------- *\
                         Write an example vector file
\* -------------------------------------------------------------------------- */

int IO_write_exs_file(char *filename,
	t_algorithms alg, t_ex_mask *p_mask,int classified,int *not_written)
{
	FILE *exs_file;
	char buf[256];

	/* open file for writing */
	exs_file = fopen(filename,"w");
	if (!exs_file) return 0;

	/* write header */
	if (ADB_get_separator(buf,EXS_HEADER,alg))
		fprintf(exs_file,"%s\n",buf);

	/* and the examples */
	if (!IO_write_exs(exs_file,alg,p_mask,classified,not_written)) return 0;

	if (ADB_get_separator(buf,EXS_FOOTER,alg))
		fprintf(exs_file,"%s\n",buf);

	fclose(exs_file);

	return 1;

}


/* -------------------------------------------------------------------------- *\
                   Write a combined atts/exs file
\* -------------------------------------------------------------------------- */

int IO_write_both_file(char *filename, 
	t_algorithms alg, t_ex_mask *p_mask, int classified, int *not_written)
{
	FILE *both_file;
	char buf[256];

	/* open file for writing */
	both_file = fopen(filename,"w");
	if (!both_file) return 0;

	/* write header */	
	if (ADB_get_separator(buf,BOTH_HEADER,alg))
		fprintf(both_file,"%s\n",buf);

	/* attributes, */
	if (!IO_write_atts(both_file,alg,classified)) return 0;

	/* separator */
	if (ADB_get_separator(buf,ATTS_EXS_SEP,alg))
		fprintf(both_file,"%s\n",buf);

	/* examples */
	if (!IO_write_exs(both_file,alg,p_mask,classified,not_written)) return 0;

	/* footer */
	if (ADB_get_separator(buf,BOTH_FOOTER,alg))
		fprintf(both_file,"%s\n",buf);

	fclose(both_file);

	return 1;

}


/* -------------------------------------------------------------------------- *\
             Write a rule file 
\* -------------------------------------------------------------------------- */

int IO_write_rule_file(char *filename, t_algorithms alg,char *comment)
{
	FILE *rule_file;
	char buf[256];

	/* open file for writing */
	rule_file = fopen(filename,"w");
	if (!rule_file) return 0;

	if (ADB_get_separator(buf,RULE_HEADER,alg))
		fprintf(rule_file,"%s\n",buf);

	/* add a line containing the name of the generating algorithm */
	ADB_get_alg_name(alg,buf);
	fprintf(rule_file,"algorithm: %s\n",buf);

	/* attributes, */
	if (!IO_write_atts(rule_file,alg,1)) return 0;

	/* separator */
	if (ADB_get_separator(buf,ATTS_EXS_SEP,alg))
		fprintf(rule_file,"%s\n",buf);

	/* comment */
	if (comment)		
	{
		if (ADB_get_separator(buf,COMMENT_SEP,alg))
			fprintf(rule_file,"%s %s\n",buf,comment);
		else
		{
			printf("Comment separator undefined for algorithm.\n");
			printf("Comment not written in rule file.\n");
		}
	}

	/* rule */
	if (!IO_write_rule(rule_file,alg,comment)) return 0;

	/* footer */
	if (ADB_get_separator(buf,RULE_FOOTER,alg))
		fprintf(rule_file,"%s\n",buf);

	fclose(rule_file);

	return 1;

}


/* -------------------------------------------------------------------------- *\
			                Read an attribute description file 
\* -------------------------------------------------------------------------- */

int IO_read_atts_file(char *filename,t_algorithms alg)
{
	FILE *fp;
	char header[512],buf[512];

	WS_dispose_dbase(); /* get rid of the old example database */

	ADB_get_separator(header,ATTS_HEADER,alg);

	if (fp = fopen (filename, "r"))
	{
		if (strlen(header))
		{
			if (fgets(buf,512,fp) ? strncmp(header,buf,strlen(header)):1)
			{
				/* OK then, this wasn't a file header, presumably it an attribute. */
				/* start all over */
				fclose(fp);
				fp = fopen(filename,"r");
			}
		}

		if (!IO_read_atts(fp,alg,FT_ATTS))
		{
			errmsg("Couldn't read the attributes");
			return 0;
		}
		fclose(fp);
	}
	else 
	{
		errmsg("Couldn't open the attribute file for reading");
    return 0;
	}        

	return 1;          
}


/* -------------------------------------------------------------------------- *\
			                  Read an example vector file 
\* -------------------------------------------------------------------------- */

int IO_read_exs_file(char *filename, t_algorithms alg)
{
	
	FILE *fp;
	char header[512],buf[512];

	WS_remove_examples (); /* get rid of the old examples */

	ADB_get_separator(header,EXS_HEADER,alg);

	if (fp = fopen (filename, "r"))
	{
		if (strlen(header))
		{
			if (fgets(buf,512,fp) ? strncmp(header,buf,strlen(header)):1)
			{
				/* OK then, this wasn't a file header, presumably it an example. */
				/* start all over */
				fclose(fp);
				fp = fopen(filename,"r");
			}
		}

		if (!IO_read_exs(fp,alg,FT_EXS))
		{
			errmsg("Couldn't read the examples");
			return 0;
		}
		fclose(fp);
	}
	else 
	{
		errmsg("Couldn't open the example file for reading");
    return 0;
	}        

	return 1;          
}
/* -------------------------------------------------------------------------- *\
                    Read a combined atts/exs file
\* -------------------------------------------------------------------------- */

int IO_read_both_file(char *filename, t_algorithms alg)
{
	FILE *fp;
	char header[512],buf[512];

	WS_dispose_dbase(); /* get rid of the old example database */

	ADB_get_separator(header,BOTH_HEADER,alg);

	if (fp = fopen(filename,"r"))
	{
		if (strlen(header))
		{
			if (fgets(buf,512,fp) ? strncmp(header,buf,strlen(header)):1)
			{
				/* OK then, this wasn't a file header, presumably it an attribute. */
				/* start all over */
				fclose(fp);
				fp = fopen(filename,"r");
			}
		}

		if (!IO_read_atts(fp,alg,FT_BOTH))
		{
			errmsg("Couldn't read the attributes");
			return 0;
		}
		if (!IO_read_exs(fp,alg,FT_BOTH))
		{
			errmsg("Couldn't read the examples");
			return 0;
		}

		fclose (fp); 
	}
	else
	{
		errmsg("Couldn't open the attribute and example file for reading");
		return 0;
	}

  return 1;
}


/* -------------------------------------------------------------------------- *\
         Read a rule file woth an attribute description in front
\* -------------------------------------------------------------------------- */

int IO_read_rule_file(char *filename, t_algorithms alg)
{
	FILE *fp;
	char buf[512],namebuf[512],name[64];		
	char tmp_file[256];

	int not_written = 0;
		
	sprintf(tmp_file,".TELA_%ld_edb_rr",getpid());

	if (!(fp = fopen(filename,"r"))) 
	{
		errmsg("No such file");
		return 0;
	}

	ADB_get_separator(namebuf,RULE_HEADER,alg);
	if (strlen(namebuf))
	{
		if (!fgets(buf,512,fp))
		{
			errmsg("Empty rule file");
			return 0;
		}
		if (strncmp(buf,namebuf,strlen(namebuf))) 
		{
			errmsg("Incorrect file header found");
			return 0;
		}
	}

	if (!fgets(buf,512,fp))	
	{
		errmsg("Unexpected end-of-file");
		return 0;
	}

	sprintf(namebuf,"algorithm: ");
	ADB_get_alg_name(alg,name);
	strcat(namebuf,name);
	
	if (strncmp(buf,namebuf,strlen(namebuf)))
	{
		char msg[256];
		sprintf(msg,"Classifier was generated with algorithm %s",buf+ 
			strlen("algorithm: "));
		errmsg(msg);
		return 0;
	}

	/* now we have to a simple trick: since attribute_struct_cmp is            */
  /* extremely strict when comparing the attribute declarations, and         */
  /* doesn't allow the attributes be declared in varying orders, we have to  */
  /* write the example  database temporarily on disk in algorithm alg format,*/
  /*  and read it back right away. More efficient solution would be rewrite  */
  /* attribute_struct_cmp, but I don't think I bother...										 */									

	if (!WS_get_attr_cnt())
	{
		if (!IO_read_atts(fp,alg,FT_BOTH))
		{
			errmsg("Couldn't read the attributes");
  		return 0;
  	} 
	}
	else
	{
		if (WS_get_ex_cnt())
		{

			if (IO_write_both_file(tmp_file,alg,0,1,&not_written)?
				!IO_read_both_file(tmp_file,alg) : 1)
			{
				errmsg("Example data base is in inconsistent state");
				remove(tmp_file);
				return 0;
			}
			else if (not_written)
				printf("%d examples were discarded because of incompabilities",
					not_written);
		}
		else
		{
			if (IO_write_atts_file(tmp_file,alg,1)?
				!IO_read_atts_file(tmp_file,alg) :1);		
			{
				errmsg("Example data base is in inconsistent state");
				remove(tmp_file);
				return 0;
			}
		}

		remove(tmp_file);

		if (!attribute_struct_cmp(fp,alg))
		{
			errmsg("File doesn't match current attribute description");
			return 0;
		}
	}

	return IO_read_rule(fp,alg);
}

