/*                                                                            *\

	  TELA - Testing Environment for Learning Algorithms - version 1.995
    ==================================================================

		Module:  TESLA_exec_cmds (part of TESLA command server)

		Contents: Algorithm execution commands of TESLA language

		Services:

		RUN_command   - builds a classifier
		TEST_command  - tests the classifiers accuracy
		CV_command    - performs crossvalidation


		Author: Juho Rousu (1995,1996)

\*                                                                            */

#include <string.h>

#include "tela.h"
#include "WS_Database.h"
#include "TESLA_statistics.h"
#include "Alg_Database.h"
#include "Alg_Exec_Machinery.h"
#include "TESLA_cmds_EXEC.h"


/* -------------------------------------------------------------------------- *\
           Build a classifier using the given learning algorithm
\* -------------------------------------------------------------------------- */

int RUN_command(t_command cmd,int round)
{
	char **param_table;
	char rule_file[256],internal_rule[256];
	float cu_time;
	int status=0;

	t_algorithms alg;

	t_run_1 param = cmd.param.l_run_1;

	/* check if classified examples exist */

	if (!WS_get_ex_cnt())
	{
		errmsg("No examples in memory");
		return EXECUTION_FAILED;
	}
	else if (!WS_is_class_attr())
	{
		errmsg("Class attribute is not set");
		return EXECUTION_FAILED;
	}

	/* the correctness of algorithm parameters is checked here */
	if (!ADB_alg_params_ok(param.params,param.alg))
	{
		errmsg("Erroraneous parameters encountered");
		return EXECUTION_FAILED;	
	}

	/* build a learning program execution instruction */
	if (EXEC_build_RUN_instruction(param,&param_table,0) != EXECUTION_OK)
	{
		errmsg("Cannot build RUN-instruction");
		return EXECUTION_FAILED;
	}		

	/* write attributes and examples in a temporary file */
	if (EXEC_write_atts_exs(param.alg,param.a_mask) != EXECUTION_OK) goto cancel;

	/* if previously generated classifier by same algorithm exists */
	/* move classifer from internal file to temporary, this        */
	/* is done to support online algorithms                        */
	if (WS_get_active_classifier(&alg)? (alg == param.alg): 0)
	{
		ADB_get_filename(param.alg,FT_RULE,rule_file);
		WS_get_internal_rulename(internal_rule);
		if (rename(internal_rule,rule_file))
		{
			char msg[256];
			perror("Operating system error");
			sprintf(msg,
			"Moving classifier from internal storage (%s) to external (%s) failed",
				internal_rule,rule_file); 
			errmsg(msg);
			return EXECUTION_FAILED;
		}
	}

	/* invoke the learning program and statistics extraction*/
	if (EXEC_run_algorithm(param_table,&status,&cu_time) != EXECUTION_OK)
	{
		errmsg("Classifier generation unsuccesful.");
		goto cancel;		
	}

	/* collect and output statistics */
	if (EXEC_collect_stats(param.alg,&param.stats) != EXECUTION_OK)
	{
		errmsg("Couldn't collect statistics");	
		goto cancel;
	}
	
	/* if time was requested, but the learning algorithm didn't time it,
		we use the time clocked by TELA */
	if (param.b_time) STATS_update_node(&param.stats,"TIME",cu_time);

	if (STATS_write_file("RUN",1,param.alg,round,
		S_ONLY_IND,param.params, param.comment,param.stat_file_name,
		&param.stats) != EXECUTION_OK)
	{
		errmsg("Couldn't output statistics.");
		goto cancel;
	}
	
	/* copy generated classifier to an internal file (or, actually, rename file)*/
	ADB_get_filename(param.alg,FT_RULE,rule_file);

	WS_get_internal_rulename(internal_rule);
	if (rename(rule_file,internal_rule))
	{
		perror("Operating system error");
		errmsg("Couldn't store generated classifier");
		return EXECUTION_FAILED;
	}

	WS_set_active_classifier(param.alg);

	EXEC_remove_temp_files(param.alg);
	return EXECUTION_OK;

	cancel:
	/* free attribute table here */

	EXEC_remove_temp_files(param.alg);
	return EXECUTION_FAILED;

}


/* -------------------------------------------------------------------------- *\
    Test the accuracy of the active classifier against the active example set
\* -------------------------------------------------------------------------- */

int TEST_command(t_command cmd,int round)
{
	char **param_table;
	float cu_time;
	int status=0;	
	char rule_file[256],internal_rule[256];
	t_algorithms alg;

	t_test_1 param = cmd.param.l_test_1;

	/* check if classified examples exist */

	if (!WS_get_ex_cnt())
	{
		errmsg("No examples in memory");
		return EXECUTION_FAILED;
	}
	else if (!WS_is_class_attr())
	{
		errmsg("Class attribute is not set");
		return EXECUTION_FAILED;
	}

	if (!WS_get_active_classifier(&alg))
	{
		errmsg("No active classifier");
		return EXECUTION_FAILED;
	}
	else if (alg != param.alg)
	{
		errmsg("The classifier was built using another algorithm");
		return EXECUTION_FAILED;
	}

	if (EXEC_build_TEST_instruction(param,&param_table,1) != EXECUTION_OK)
	{		
		errmsg("Cannot build TEST-instruction");
		return EXECUTION_FAILED;
	}		

	/* write attributes and examples in a temporary file */
	if (EXEC_write_atts_exs(param.alg,param.a_mask) != EXECUTION_OK) goto cancel;

	/* copy classifer form internal file to temporary */
	ADB_get_filename(param.alg,FT_RULE,rule_file);

	WS_get_internal_rulename(internal_rule);
	if (rename(internal_rule,rule_file))
	{
		char msg[256];
		sprintf(msg,
			"Moving classifier from internal storage (%s) to external (%s) failed",
			internal_rule,rule_file); 
		errmsg(msg);
		return EXECUTION_FAILED;
	}

	/* invoke the testing facilities and statistics extraction*/
	if (EXEC_run_algorithm(param_table,&status,&cu_time) != EXECUTION_OK)
	{
		errmsg("Classifier testing unsuccessful.");
		goto cancel;		
	}

	/* collect and output statistics */
	if (EXEC_collect_stats(param.alg,&param.stats) != EXECUTION_OK)
	{
		errmsg("Couldn't collect statistics");
		goto cancel;
	}
		
	/* if time was requested, but the learning algorithm didn't time it,
		we use the time clocked by TELA */
	if (param.b_testtime) STATS_update_node(&param.stats,"TESTTIME",cu_time);

	if (STATS_write_file("TEST",1,param.alg, round,S_ONLY_IND,0,
		param.comment,param.stat_file_name,&param.stats) != EXECUTION_OK)
	{
		errmsg("Couldn't output statistics.");
		goto cancel;
	}	

	EXEC_remove_temp_files(param.alg);

	if (rename(rule_file,internal_rule))
	{		
		char msg[256];
		sprintf(msg,
			"Moving classifier from external storage (%s) to internal (%s) failed",   
        rule_file,internal_rule); 
		errmsg(msg);
		return EXECUTION_FAILED;
	}

	return EXECUTION_OK;

	cancel:

	/* do some cleaning up here */

	EXEC_remove_temp_files(param.alg);

	if (rename(rule_file,internal_rule))
	{		
		char msg[256];
		sprintf(msg,			
			"Moving classifier from external storage (%s) to internal (%s) failed",   
        rule_file,internal_rule); 
		errmsg(msg);
		return EXECUTION_FAILED;
	}

	return EXECUTION_FAILED;


}


/* -------------------------------------------------------------------------- *\
			      Perform a crossvalidation run for the active example set
\* -------------------------------------------------------------------------- */

int CV_command(t_command cmd, int round)
{
	t_cv *p_cv = &cmd.param.l_cv;
	t_command run,test;
	t_run_1 *runparam;
	t_test_1 *testparam;
	t_ex_mask *p_mask;
	int *p_tmp;
 	int prmode;

	int *freq;

	t_stats_list stats;
	float time,testtime,accuracy,sumtime=0,sumaccuracy=0;
	int dummy,i,rv,updated,fraction,index;

	list_init(&stats);

	freq = (int*)calloc(p_cv->count,sizeof(int));

	if (!WS_get_ex_cnt())
	{
		errmsg("No examples currently loaded");
		return EXECUTION_FAILED;
	}


	if (!(p_mask = (t_ex_mask *)malloc(sizeof(t_ex_mask) ) ) )
	{	
		errmsg("Out of memory");
		return EXECUTION_FAILED;
	}
	
	p_mask->count = WS_get_ex_cnt();

	if (!(p_mask->indexes = (int *)malloc(p_mask->count*sizeof(int)) ) ) 
	{
		free(p_mask);
		errmsg("Out of memory");
		return EXECUTION_FAILED;
	}	

	/* generate a partition, the size of each subset will be 
			floor(# examples / # cv rounds) ... ceil(# examples / # cv rounds) */
	
	fraction = p_mask->count/p_cv->count;

	if (!fraction) 
	{
		char buf[256];
		sprintf(buf,
			"Too few examples for a %d-fold crossvalidation",p_cv->count);
		errmsg(buf);
		free(p_mask);
		return EXECUTION_FAILED;
	}

	index = 0;
	for (i=1;i <= fraction ; i++)
	{
		updated = 0;
		
		while (updated < p_cv->count)
		{
			p_tmp = &(p_mask->indexes[index]);
			
			*p_tmp = RandomInt(0,p_cv->count-1);
			
			while (freq[*p_tmp] >= i)
			{
				(*p_tmp)++;
				if (*p_tmp > p_cv->count-1) *p_tmp = 0;
			}
			(freq[*p_tmp])++;
			updated++;
			index++;
		}
	}
	
	for (;index< p_mask->count;index++)
	{
		p_tmp = &(p_mask->indexes[index]);
			
		*p_tmp = RandomInt(0,p_cv->count-1);
			
		while (freq[*p_tmp] > fraction)
		{
			(*p_tmp)++;
			if (*p_tmp > p_cv->count-1) *p_tmp = 0;
		}
		(freq[*p_tmp])++;
	}

	prmode = p_cv->pr_mode == S_ONLY_AVG ? S_NO : S_BOTH; 

	/* now, generate a RUN- and a TEST-command */
	run.cmd_id = K_RUN;
	runparam = &run.param.l_run_1;
	runparam->b_time = !ADB_alg_statistic_ok("TIME",p_cv->alg,0); 
	runparam->alg = p_cv->alg;
	runparam->params = p_cv->params;
	runparam->stat_file_name = 0;
	runparam->a_mask = p_mask;
	list_init(&runparam->stats);
	runparam->pr_mode	= 
		STATS_copy_list(p_cv->alg,0,&runparam->stats,&p_cv->stats,prmode);
	
	test.cmd_id = K_TEST;
	testparam = &test.param.l_test_1;
	testparam->b_testtime = !ADB_alg_statistic_ok("TESTTIME",p_cv->alg,1); 
	testparam->alg = p_cv->alg;
	testparam->stat_file_name = 0;
	testparam->a_mask = p_mask;	
	list_init(&testparam->stats);
	testparam->pr_mode = 
		STATS_copy_list(p_cv->alg,1,&testparam->stats,&p_cv->stats,prmode);

	/* perform the cross-validation */
	for (i=0; i < p_cv->count ; i++)
	{
		if (p_cv->pr_mode >= 1)
			printf("Beginning round %d, %d training examples, %d testing examples...\n",i+1,p_mask->count-freq[i],freq[i]);
		p_mask->active_subset = i;
		/* RUN with all subsets except one */
		p_mask->all_except = 1;
		if (rv = RUN_command(run,0)) break;
		/* TEST with the remaining subset */
		p_mask->all_except = 0;
		if (rv = TEST_command(test,0)) break;
	}

	if (rv != EXECUTION_OK)
	{
		errmsg("Execution failed");
	}
	else 
	{
		if (p_cv->pr_mode >= 1)
		{
			printf("Average statistics over %d rounds:\n",p_cv->count); 
			fflush(stdout);
		}

		/* collect average values in statistic list */
/*		STATS_update_from_avg(p_cv->count,&p_cv->stats,&runparam->stats);
		STATS_update_from_avg(p_cv->count,&p_cv->stats,&testparam->stats);
*/
		list_catenate(&stats,&runparam->stats);
		list_catenate(&stats,&testparam->stats);

		STATS_update_from_avg(p_cv->count,&p_cv->stats,&stats);

		if (p_cv->pr_mode >= 1)
		if ((rv = STATS_write_file("CV",p_cv->count,p_cv->alg,round,
		   S_BOTH,p_cv->params,p_cv->comment,
				p_cv->stat_file_name,&stats)) != EXECUTION_OK)
			return EXECUTION_FAILED;
	}
		
	/* free some memory here */
	free(p_mask);

	return rv;

}
