/*                                                                            *\

	  TELA - Testing Environment for Learning Algorithms - version 1.995
    ==================================================================

		Module:  			TESLA_cmds_WS.c

		Contents:     Working set manipulation and general commands of TESLA

		Author: Ilkka Hyvrinen (1992,1993,1994) Juho Rousu (1995,1996)

		Services:

		Working Set Manipulation commands
		=================================

		CLASS_command		    - sets the class attribute
		DIVIDE_command      - partitions the active example set
		EXCLUDE_command		  - excludes specified attributes
		INCLUDE_command		  - includes (irrelevant) attributes
		MAP_command			    - maps and attributes domain to a new one
		NOISE_command		    - adds random noise
		RANDOMIZE_command	  - adds random noise
		SELECT_command      - chooses a subset of examples as the new example set 
		STAT_command		    - displays basic statistics of the dataset

		Supporting commands
		===================

		SET_command     - sets global variables, initializes the RND generator
		DELETE_command  - deletes a file
		INIT_command    - initializes the working set database
		BACKUP_command  - backups specified components of the working set
		RESTORE_command - restores a previously backed up working set portions
		QUIT_command   - exits TELA after cleanup


		Internal functions:

		randomize                - initialize the random number generator
		RandomFloat              - output a pseudorandom floating point figure
		Random                   - ditto, but form constrained interval
		RandomInt								 - ditto, but an integer
		flmax                    - get maximum of two floating numbers
		create_ex_mask           - create a mask for examples
		write_do_statistics_file - write statistics of a DO-loop to files
		ws_backup                - backup working set
		ws_restore               - restore working set
		ScrambleOneAttribute     - add random noise to an attribute's values
		GetAttrToHandle					 - get a handle to an attribute from a name
		OutputStatHeader         - output statistics header
		OutputNumericStats       - print statistics of a numeric attribute
		OutputDiscreteStats      - print statistics of a nominal attribute
		StatSingleAttr           - print statistics of a single attribute
\*                                                                            */ 
#include <stdio.h>
#include <stdlib.h>
#include <values.h>
#include <float.h>
#include <math.h>
#include <string.h>
#include <sys/time.h>

#include "tela.h"
#include "WS_Database.h"
#include "TESLA_cmds_WS.h"
#include "TESLA_cmds_EXEC.h"
#include "TESLA_cmds_IO.h"
#include "TESLA_statistics.h"
#include "IO_server.h"


#define private static
#define SQR(arg) ((double) arg * arg)

#define SELECT_IV_START 1
#define SELECT_IV_END 0

/* This hasn't been declared in the standard header files */
extern double drand48 (void);

/* Type for STAT-command to save the counts of values of discrete attributes */
typedef struct {
  int cnt;
  char *value;
} tDiscrVals;


/* Type definition for modes used to create different kinds of 
   noise to the examples. Used by RANDOMIZE and NOISE. */
typedef enum {DO_RANDOMIZE, DO_NOISE} tScrambleMode;


/* ========================================================================== *\
														INTERNAL FUNCTIONS
\* ========================================================================== */

void randomize (int seed);
float RandomFloat (void);
float Random (float lo, float hi); 
int RandomInt (int lo, int hi); 
float flmax(float a, float b);
t_ex_mask *create_ex_mask(int size);
int write_do_statistics_file(t_do_2 params);
int ws_backup(char *name,int *ae_backed_up,int *cf_backed_up,t_algorithms *cf);
int ws_restore(char *name, int ae_restore, int cf_restore, 
	t_algorithms cf, boolean cleanup);
int ScrambleOneAttribute (t_attr_ref aRef, float noise_p, 
	const tScrambleMode scrambleMode,int *changed_count);
int GetAttrToHandle (char *aName, t_attr_ref *aRef);
private void OutputStatHeader (char *statFName, char *comment); 
private void OutputNumericStats (char *statFName, char *aName,char *a_Type, 
	double mean, double stdDev, int okCnt, int unknownCnt, int dontCareCnt);
private void OutputDiscreteStats (char *statFName, char *aName, int valCnt,
	tDiscrVals dVals[], int okCnt, int unknownCnt, int dontCareCnt);
private int StatSingleAttr (char *statFName, char *comment, 
	char *aName, int cmdRow);



/* -------------------------------------------------------------------------- *\
				Initialize the random number generator with a given seed
\* -------------------------------------------------------------------------- */

void randomize (int seed) 
{
  srand48 (seed);
}


/* -------------------------------------------------------------------------- *\
								Generate a pseudo-random floating point number
\* -------------------------------------------------------------------------- */

float RandomFloat (void) 
{

  return (float) ((drand48 () * FLT_MAX / 2.0 - (FLT_MAX / 2.0)) * 2.0);
}


/* -------------------------------------------------------------------------- *\
			Generate a pseudo-random floating point number from a given interval
\* -------------------------------------------------------------------------- */

float Random (float lo, float hi) 
{
	
  return (float) ((drand48 () * (hi - lo)) + lo);
}  


/* -------------------------------------------------------------------------- *\
				Generate a pseudo-random integer number from a given interval
\* -------------------------------------------------------------------------- */

int RandomInt (int lo, int hi) 
{

	int rnd;

  if (hi < MAXINT)
    hi++;
  return ((int) (drand48 () * (unsigned) (hi - lo) + lo));
}


/* -------------------------------------------------------------------------- *\
						Return the maximum of two floating point figures
\* -------------------------------------------------------------------------- */

float flmax(float a, float b)
{
	return (a > b) ? a : b;
}


/* -------------------------------------------------------------------------- *\
												Create an example mask
\* -------------------------------------------------------------------------- */

t_ex_mask *create_ex_mask(int size)
{
	t_ex_mask *p_mask=0;

	if (!(p_mask = (t_ex_mask *)malloc(sizeof(t_ex_mask) ) ) )
	{	
		errmsg("Out of memory");
		return 0;
	}
	
	p_mask->count = size;

	if (!(p_mask->indexes = (int *)malloc(size*sizeof(int)) ) ) 
	{
		free(p_mask);
		errmsg("Out of memory");
		return 0;
	}
	
	return p_mask;
}


/* -------------------------------------------------------------------------- *\
	  Write the statistics from the execution of a DO-loop to appropriate files
\* -------------------------------------------------------------------------- */

int write_do_statistics_file(t_do_2 params)
{
	char s[256];
	t_command *cmd_ptr;
	FILE *stat_file=0;
	boolean stats_exist = 0;	 

	if (!params.stats.p_head) return EXECUTION_OK;
 
	cmd_ptr = params.commands;
	

	printf("Average statistics over %d rounds:\n",params.do_cnt);
	if (params.stat_file_name)
	{
		if (stat_file = fopen(params.stat_file_name,"a"))
		{
			STATS_print_parameters("DO",params.do_cnt,0,0,stat_file,params.comment);
			fprintf(stat_file,"\n");
			fclose(stat_file);
		}
		else
		{
			errmsg("Couldn't open statistics file");
			return EXECUTION_FAILED;
		}
	}

	/* Collect all the averages of recorded statistics */
	while (cmd_ptr != NULL) 
	{
		char s1[256];
    char s2[256];
	
		switch (cmd_ptr->cmd_id)
		{
			case K_RUN:
			{				
				t_run_2 *p_run = &cmd_ptr->param.l_run_2;

				if (!p_run->stats.p_head) break;
				STATS_write_file("RUN",params.do_cnt,p_run->alg,1, 
					S_ONLY_AVG, p_run->params, p_run->comment,params.stat_file_name,
					&p_run->stats);				
				STATS_update_from_avg(params.do_cnt,&params.stats,&p_run->stats);
    	} 
			break;
			case K_TEST:
			{				
				t_test_2 *p_test = &cmd_ptr->param.l_test_2;
	
				if (!p_test->stats.p_head) break;
				STATS_write_file("TEST",params.do_cnt,p_test->alg,1,
					S_ONLY_AVG,0, p_test->comment,params.stat_file_name,
					&p_test->stats);				
				STATS_update_from_avg(params.do_cnt,&params.stats,&p_test->stats);
    	}
			break;
			case K_CV:
			{
				t_cv *p_cv = &cmd_ptr->param.l_cv;				
				
				if (!p_cv->stats.p_head) break;
				STATS_write_file("CV",params.do_cnt,p_cv->alg,1,S_ONLY_AVG,0,
					p_cv->comment,params.stat_file_name,&p_cv->stats);				
				STATS_update_from_avg(params.do_cnt,&params.stats,&p_cv->stats);
			}
			break;
		} /* switch */
    cmd_ptr = cmd_ptr->next;	
	} /* while */	

	return EXECUTION_OK;

}


/* -------------------------------------------------------------------------- *\
											Back up the working set
\* -------------------------------------------------------------------------- */

int ws_backup(char *name,int *ae_backed_up,int *cf_backed_up,t_algorithms *cf)
{
	int rv = EXECUTION_OK;
	char buf[256];

	*ae_backed_up = 0;*cf_backed_up = 0;

	if (WS_get_attr_cnt()) /* backup example database */	
	{
		sprintf(buf,"%s_eb_%ld",name,getpid());
		rv = WS_example_db_backup(buf);
		*ae_backed_up = !rv;
	}
	
	if (!rv && WS_get_active_classifier(cf)) /* if classifier exits back it up */
	{		
		sprintf(buf,"%s_cf_%ld",name,getpid());
		rv = WS_rule_backup(*cf,buf);
		*cf_backed_up = !rv; 
	}

	return rv;
}


/* -------------------------------------------------------------------------- *\
												Restore the working set 
\* -------------------------------------------------------------------------- */

int ws_restore(char *name, int ae_restore, int cf_restore, 
	t_algorithms cf, boolean cleanup)
{
	int rv = EXECUTION_OK;
	char buf[256];

	if (ae_restore)	
	{
		sprintf(buf,"%s_eb_%ld",name,getpid());
		if (cleanup)			
			rv = WS_example_db_restore(buf); /* delete backup when restoring */
		else
			rv = WS_example_db_get(buf);
	}
	else
		WS_dispose_dbase();

	if (!rv && cf_restore) /* restore classifier if it exists */
	{
		sprintf(buf,"%s_cf_%ld",name,getpid());
		rv = WS_rule_restore(cf,buf);
		if (cleanup)
			remove(buf);								/* delete backup */
	}
	else
		WS_remove_active_classifier(); /* no active classifier */

	return rv;
}


/* -------------------------------------------------------------------------- *\
										Add noise to an attribute's values
\* -------------------------------------------------------------------------- */

int ScrambleOneAttribute (t_attr_ref aRef, float noise_p, 
			  const tScrambleMode scrambleMode,int *changed_count) 
{

t_attr_types 	aTyyppi;
int 		eCnt, valCnt;
t_ex_ok 	eLaji;
int 		oldIndex, newIndex;
int 		oldInt, newInt;
float 		oldFloat, newFloat;
int 		retVal;
int 		row;

	*changed_count = 0;

  if (retVal = WS_get_attr_type (aRef, &aTyyppi))
    return (retVal);
  if (aTyyppi == T_USER || aTyyppi == T_BINARY) {
    if (retVal = WS_get_attr_domain_size (aRef, &valCnt))
      return (retVal);
    if (valCnt <= 1)
      return (NO_ERROR);
  };
    
  eCnt = WS_get_ex_cnt ();
  for (row = 0; row < eCnt; row++) {

    /* Change example if random number ok.
       NB. unknown or not cared of examples are ignored. */
    if (retVal = WS_get_exval_kind (aRef, row, &eLaji))
      return (retVal);
    if ((Random (0, 1) <= noise_p) && (eLaji == T_OK)) 
		{
      switch (aTyyppi) {
        case T_USER : 
          WS_get_nominal_exval (aRef, row, &oldIndex);
          if (scrambleMode == DO_NOISE) {
            if ((newIndex = RandomInt (0, valCnt - 2)) >= oldIndex)
              newIndex++;
          } else
            newIndex = RandomInt (0, valCnt - 1);
          WS_set_nominal_exval (aRef, row, newIndex);
					if (oldIndex != newIndex) (*changed_count)++;
          break;
        case T_BINARY :
          WS_get_nominal_exval (aRef, row, &oldIndex);
          if (scrambleMode == DO_NOISE) {
            if (oldIndex == 0)
              WS_set_nominal_exval (aRef, row, 1);
            else
              WS_set_nominal_exval (aRef, row, 0);
						(*changed_count)++;
          } 
					else
					{
						newIndex = RandomInt(0,1);
            WS_set_nominal_exval (aRef, row, newIndex);
						if (newIndex != oldIndex) (*changed_count)++;
					}
          break;
        case T_INT : /* what follows is a totally useless noise model */

          WS_get_integer_exval (aRef, row, &oldInt);
					newInt = oldInt;
          while ((newInt += RandomInt (0,abs(oldInt)/2)-oldInt/2) == oldInt) ;
          WS_set_integer_exval (aRef, row, newInt);
					(*changed_count)++;
          break;
        case T_FLOAT : /* and so is this */				
				{
					int i=0;
          WS_get_float_exval (aRef, row, &oldFloat);
					newFloat = oldFloat;
          while (i++ <= 10)
					{ 
						if ((newFloat += Random(0,abs(oldFloat))-oldFloat/2) != oldFloat)
						{
 							(*changed_count)++;         
							WS_set_float_exval (aRef, row, newFloat);
							break;
						}
					}
	        break;
				}
      }  /* switch */
    }  /* if */
  }  /* for */
  return (NO_ERROR);
}  /* routine */



/* -------------------------------------------------------------------------- *\
						Make a reference from an attribute name
\* -------------------------------------------------------------------------- */ 

int GetAttrToHandle (char *aName, t_attr_ref *aRef) 
{
	
	int	retVal;
	
  if (!strcmp (aName, "") || !strcmp(aName, "class"))
    retVal = WS_get_class_attr (aRef);
  else
    retVal = WS_get_attr_ref (aName, aRef);
    
  return retVal;
}


/* -------------------------------------------------------------------------- *\
								Print header text for the STAT-command
\* -------------------------------------------------------------------------- */ 
private void OutputStatHeader (char *statFName, char *comment) 
{
	
	char		wstr[200];

  /* sprintf (wstr, "\n     #%s", comment); */
  sprintf (wstr, "Attribute         Type  Values   Average   Std deviation   Unknown  Don't Care");
  fprintf (stdout, "%s\n", wstr);
  if (statFName != NULL)
    STATS_write_line (statFName, wstr);

  sprintf (wstr, "name (* = class)        (Nom.)     (Numeric only)          count      count");
  printf ("%s\n", wstr);
  if (statFName != NULL)
    STATS_write_line (statFName, wstr);
}


/* -------------------------------------------------------------------------- *\
							Print statistics concerning a numeric attribute
\* -------------------------------------------------------------------------- */ 
private void OutputNumericStats (char *statFName, char *aName,char *a_Type, 
				 double mean, double stdDev, 
				 int okCnt, int unknownCnt, 
				 int dontCareCnt) 
{
				 
char wstr[501];  /* Max string len 500 */

  sprintf (wstr, "%-17s %s%s%18.2f%16.2f%10d%12d", aName,
 		((int)strlen(aName)<17)?"":"\n                  ", a_Type, 
		mean, stdDev, unknownCnt, dontCareCnt);
  printf ("%s\n", wstr);
  if (statFName != NULL)
    STATS_write_line (statFName, wstr);
}


/* -------------------------------------------------------------------------- *\
							Print statistics concernng a nominal attribute
\* -------------------------------------------------------------------------- */ 
private void OutputDiscreteStats (char *statFName, char *aName, int valCnt,
				  tDiscrVals dVals[], int okCnt, 
				  int unknownCnt, int dontCareCnt) 
{

int	i;
t_attr_ref a_ref;
char	wstr[501],cl[2] = "",*cl_name,new_name[64];

	strcpy(new_name,aName);

	if (WS_is_class_attr())
	{
		WS_get_class_attr(&a_ref);
		WS_get_attr_name(a_ref,&cl_name);
		if (!strcmp(aName,cl_name)) 
			strcat(new_name,"*");
	}



  sprintf (wstr, "%-17s%s%s DISCR%43d%12d", new_name, 
		((int)strlen(new_name)<17)?"":"\n                 ",cl, 
		unknownCnt, dontCareCnt);
  printf ("%s\n", wstr);
  if (statFName != NULL)
    STATS_write_line (statFName, wstr);
  
  for (i = 0; i < valCnt; i++) {
    sprintf (wstr, "%17s%13d", dVals[i].value, dVals[i].cnt);
    printf ("%s\n", wstr);
    if (statFName != NULL)
      STATS_write_line (statFName, wstr);
  }
}


/* -------------------------------------------------------------------------- *\
			Output statistics for a single attribute
\* -------------------------------------------------------------------------- */ 
private int StatSingleAttr (char *statFName, char *comment, 
			    char *aName, int cmdRow) 
{
	
t_attr_ref 	aRef;
t_attr_types 	aType;
tDiscrVals	*dVals=0;
int		exCnt, eRow;
t_ex_ok 	exStatus;
int		i;
int		okCnt, unKnownCnt, dontCareCnt, wInt;
int		retVal;
double		sum, sqrSum, mean, stdDev;
float		wFloat;
int		valCnt;

  okCnt = unKnownCnt = dontCareCnt = sum = sqrSum = mean = stdDev = 0;
  
  if (retVal = GetAttrToHandle (aName, &aRef))
    return retVal;
  if (retVal = WS_get_attr_type (aRef, &aType))
    return retVal;
   
  /* Malloc storage to hold counts && values of discrete attributes
     per possible value. Also init the string pointers */
  if (aType == T_USER || aType == T_BINARY) {
    if (retVal = WS_get_attr_domain_size (aRef, &valCnt))
      return (retVal);
    if ((dVals = (tDiscrVals *) malloc (valCnt * sizeof (tDiscrVals))) == NULL) {
      fprintf (stdout, "FATAL ERROR in routine StatSingleAttr: ");
      fprintf (stdout, "Memory allocation failed!\n");
      return (FATAL_ERROR);
    }
      
    /* String pointer init */
    for (i = 0; i < valCnt; i++)
      if (retVal = WS_get_index_attr_val (aRef, i, &(dVals[i].value))) 
			{
      	if (dVals) free (dVals);
        return retVal;
      } else {
      	dVals[i].cnt = 0;
      }
  }
   
  /* Loop through all examples */ 
  exCnt = WS_get_ex_cnt ();
  for (eRow = 0; eRow < exCnt; eRow++) 
	{
    if (retVal = WS_get_exval_kind (aRef, eRow, &exStatus))
      return (retVal);
    switch (exStatus) {
      case T_OK :
        switch (aType) {
          case (T_INT) :
            if (retVal = WS_get_integer_exval (aRef, eRow, &wInt)) {
              if (dVals) free (dVals);
              return (retVal);
            }
            sum += wInt;
            sqrSum += SQR (wInt);
            break;
          case T_FLOAT :
            if (retVal = WS_get_float_exval (aRef, eRow, &wFloat)) {
              if (dVals) free (dVals);
              return (retVal);
            }
            sum += wFloat;
            sqrSum += SQR (wFloat);
            break;
          default :
            /* Update discrete value table */
            if (retVal = WS_get_nominal_exval (aRef, eRow, &wInt)) {
              if (dVals) free (dVals);
              return (retVal);
            }
            dVals[wInt].cnt++;
            break;
        }  /* switch (aType) */
        okCnt++;
        break;
      case T_DONT_KNOW :
        unKnownCnt++;
        break;
      case T_DONT_CARE :
        dontCareCnt++;
        break;
    }  /* switch (exStatus) */
  }  /* for */

  if (aType == T_FLOAT || aType == T_INT) {
    if (okCnt) {
      mean = sum / okCnt;
      stdDev = sqrt ((sqrSum - okCnt * SQR (mean)) / (okCnt - 1));
    }
    OutputNumericStats(statFName, aName, aType == T_FLOAT? "CONT": "INT ",
			mean, stdDev, okCnt, unKnownCnt, dontCareCnt);
  } else {
    OutputDiscreteStats (statFName, aName, valCnt, dVals,
    			 okCnt, unKnownCnt, dontCareCnt);
  }
  if (dVals) free (dVals);
  return (NO_ERROR);
}  /* StatSingleAttr */



/* ========================================================================== *\
														EXTERNAL SERVICES
\* ========================================================================== */


/* -------------------------------------------------------------------------- *\
				                Execute a TESLA command
\* -------------------------------------------------------------------------- */

int TESLA_cmd_execute(t_command *cmd)
{
	int rv; 

	switch (cmd->cmd_id)
	{
		case K_CLASS: 
			rv = CLASS_command(*cmd);break;
		case K_EXCLUDE: 
		case K_EXCLUDE_ALL_EXCEPT: 
			rv = EXCLUDE_command(*cmd);break;
		case K_DO: 
			rv = DO_command(*cmd);break; 
		case K_DIVIDE: 
			rv = DIVIDE_command(*cmd);break;
		case K_INCLUDE: 
			rv = INCLUDE_command(*cmd);break;
		case K_MAP: 
			rv = MAP_command(*cmd);break;
		case K_NOISE: 
			rv = NOISE_command(*cmd);break;
		case K_RANDOMIZE: 
			rv = RANDOMIZE_command(*cmd);break;
		case K_WRITE_ATTS:
		case K_WRITE_EXS:
		case K_WRITE_ATTS_AND_EXS:
		case K_WRITE_BOTH:
		case K_WRITE_RULE:  
			rv = WRITE_command(*cmd);break;
		case K_STAT:
		case K_STAT_ALL: 
			rv = STAT_command(*cmd);break;
		case K_READ_ATTS:
		case K_READ_EXS:
		case K_READ_ATTS_AND_EXS:
		case K_READ_BOTH:
		case K_READ_RULE:  
			rv = READ_command(*cmd);break;
		case K_RUN: 
			rv = RUN_command (*cmd,0);break;
		case K_TEST:
			rv = TEST_command(*cmd,0);break;
		case K_CV: 
			rv = CV_command(*cmd,0);break; 
		case K_SELECT:
			rv = SELECT_command(*cmd); break;
		case K_SET_RANDSEED:
		case K_SET_RANDCLOCK:
		case K_SET_ALG:
			rv = SET_command(*cmd); break;
		case K_DELETE:
			rv = DELETE_command(*cmd);break;
		case K_BACKUP_ATTS_EXS:
		case K_BACKUP_RULE:
			rv = BACKUP_command(*cmd);break;
		case K_RESTORE_ATTS_EXS:
		case K_RESTORE_RULE:
			rv = RESTORE_command(*cmd);break;
		case K_INIT:
			rv = INIT_command(*cmd);break;
		case K_QUIT:
			rv = QUIT_command(*cmd);break;
		default: rv = EXECUTION_FAILED;break;
	}	

	return rv;
}


/* -------------------------------------------------------------------------- *\
       								Set the CLASS attribute
\* -------------------------------------------------------------------------- */

int CLASS_command (t_command p) 
{

t_attr_ref 	aRef;
int 		retVal;
	
	if (p.param.l_class_1.attr_spec == A_NONE)
	{
		WS_set_class_attr(-1);
		printf("The examples are now unclassified.\n");
		return EXECUTION_OK;
  }

  if (retVal = GetAttrToHandle(p.param.l_class_1.attr_name, &aRef))
    return retVal;
    
  if (retVal = WS_set_class_attr (aRef))
    return retVal;

	printf("Class attribute set to %s.\n",p.param.l_class_1.attr_name);
    
  return EXECUTION_OK;
}


/* -------------------------------------------------------------------------- *\
						Partition the example set into disjoint subsets
\* -------------------------------------------------------------------------- */

int DIVIDE_command (t_command cmd)
{
	t_div_node *p_div_node= cmd.param.l_divide_1.first;
	t_algorithms alg;
	float val,sum=0;
	int i,index=0;
	int not_written=0;

	t_ex_mask *p_mask = create_ex_mask(WS_get_ex_cnt());
	if (!p_mask)
	{
		errmsg("Partition mask allocation failed.\n"); 
		return EXECUTION_FAILED;
	}

	/* write mode is : active subset gets written, others don't */
	p_mask->all_except = 0;

	for (i=0; i < WS_get_ex_cnt(); i++)
	{
		val = Random(0,1);

		index=0;sum=0;p_div_node = cmd.param.l_divide_2.first;

		/* search a node whose cumulative probability is bigger than val */
		while (sum + p_div_node->part < val)
		{
			index++;		
			sum += p_div_node->part;
			p_div_node = p_div_node->next;
		}

		/* assign mask digit of current example to index */
		p_mask->indexes[i] = index;
	}

	/* what is the active algorithm, use it's grammar */
	if (!WS_get_active_classifier(&alg)) alg = 0;

	p_div_node = cmd.param.l_divide_2.first;
	index=0;

	/* write each subset in turn to a file */
	while (p_div_node)
	{
		/* only index'th subset is written*/
		p_mask->active_subset = index;
		if (!IO_write_both_file(IO_get_expanded_name(p_div_node->f_name),
			alg,p_mask,WS_is_class_attr(),&not_written))
		{
			errmsg("Couldn't write exampleset.");
			goto cancel;
		}
		else if (not_written)
			printf("%d examples were discarded because of incompabilities.\n");

		index++;
		p_div_node = p_div_node->next;
	}

	/* read first subset back in memory */
	p_div_node = cmd.param.l_divide_2.first;
	if (!IO_read_both_file(IO_get_expanded_name(p_div_node->f_name),alg))
	{
		errmsg("Couldn't read training examples");
		goto cancel;
	}
	
	printf("Divide complete. %d examples left in working set.\n",WS_get_ex_cnt());

	return EXECUTION_OK;

	cancel:
	free(p_mask);
	return EXECUTION_FAILED;
}	


/* -------------------------------------------------------------------------- *\
												Execute a DO-loop
\* -------------------------------------------------------------------------- */

DO_command(t_command cmd)
{
	int rv = EXECUTION_OK;
	t_do_1 *p_do = &cmd.param.l_do_1;
	t_command *p_cmd_start = p_do->commands,*p_cmd_iter;
	int i,ae,cf;
	t_algorithms alg;

	/* cumulating statistics */
	float 	avg_time = 0;
	float 	avg_size1 = 0;
	float 	avg_size2 = 0;
	int run_count = 0;
	float 	avg_rank = 0;
	float 	avg_testtime = 0;
	int test_count = 0;

	/* back up the working set */	
	if (rv = ws_backup(".TELA_do",&ae,&cf,&alg))
	{	
		errmsg("Couldn't back up the working set");
		return rv;
	}

	/* the loop */
	for (i=1;i<=p_do->do_cnt;i++)
	{
		/* notify TELA's file server */
		IO_set_round(i);
		
		fprintf (stdout,"Beginning round %i ...\n", i);

		/* execute the commands */
		p_cmd_iter = p_cmd_start;
		while (p_cmd_iter && rv == EXECUTION_OK)
		{
			switch (p_cmd_iter->cmd_id)
			{
				case K_END: p_cmd_iter = 0; continue; /* get out of while-loop */
				case K_RUN: rv = RUN_command(*p_cmd_iter,i); break;
				case K_TEST:rv = TEST_command(*p_cmd_iter,i); break;
				case K_CV:rv = CV_command(*p_cmd_iter,i); break;
				default:
					rv = TESLA_cmd_execute(p_cmd_iter);
			}
			if (rv)
			{
				errmsg("Command execution failed");
				goto restore_ws; 
			}
			p_cmd_iter = p_cmd_iter->next;
		}
		
		/* for the next iteration, we need to restore the working set */
		if (rv = ws_restore(".TELA_do",ae,cf,alg,0))
		{
			errmsg("Couldn't restore the working set");
			return rv;
		}
	} 
	IO_set_round(0);

	/* output appropriate statistics */
	write_do_statistics_file(*p_do);

	restore_ws:

	if (ws_restore(".TELA_do",ae,cf,alg,1))
	{
		errmsg("Fatal error: Couldn't restore the working set");
		return EXECUTION_FAILED;
	}

	return rv;

}


/* -------------------------------------------------------------------------- *\
								Compare two values using a given relation
\* -------------------------------------------------------------------------- */

boolean cmp_select(float val_1,float val_2,boolean eq_allow,boolean cmp_to_start)
{
	if (eq_allow)
		if (cmp_to_start)
			return (val_1 >= val_2);
		else
			return (val_1 <= val_2);
	else
		if (cmp_to_start)
			return (val_1 > val_2);
		else
			return (val_1 < val_2);
}


/* -------------------------------------------------------------------------- *\
		Evaluate an examples attr. value against a subset of an attributes domain
\* -------------------------------------------------------------------------- */

boolean evaluate(int ex_index,t_select_node *p_node)
{
	t_attr_ref a_ref;
	t_val_node *p_val_node = p_node->val_list;
	t_val_type val_type;
	t_val val;
	if (GetAttrToHandle(p_node->attr_name,&a_ref) == NO_ERROR)
		if (!WS_get_attr_val(ex_index,a_ref,&val_type,&val))
		{
			switch (val_type)
			{
				case l_dont_care:
				case l_dont_know: return 0;
				case l_float:
				{ 
					if (p_node->start_encl != -1)
					{
						if (!cmp_select(val.f,p_val_node->val.f,p_node->start_encl,
							SELECT_IV_START)) 
							return p_node->neg;
						p_val_node = p_val_node->p_next; 
					}	
					if (p_node->end_encl != -1)
					{
						if (!cmp_select(val.f,p_val_node->val.f,p_node->end_encl,
							SELECT_IV_END))
							return p_node->neg;
					}
					return !p_node->neg;
				}
				case l_int:				
				{ 
					if (p_node->start_encl != -1)
					{
						if (!cmp_select(val.i,p_val_node->val.i,p_node->start_encl,
							SELECT_IV_START)) 
							return p_node->neg;
						p_val_node = p_val_node->p_next; 
					}	
					if (p_node->end_encl != -1)
					{
						if (!cmp_select(val.i,p_val_node->val.i,p_node->end_encl,
							SELECT_IV_END))
							return p_node->neg;
					}
					return !p_node->neg;
				}

				case l_discrete:
					while (p_val_node)
					{
						if (!eq_q_str(p_val_node->val.discrete,val.discrete)) 
						{
							free(val.discrete);
							return !p_node->neg;
						}
						p_val_node = p_val_node->p_next;
					}
					return p_node->neg;
			}
		}

	return 0;
}


/* -------------------------------------------------------------------------- *\
					Evaluate an example against a conjunction of tests
\* -------------------------------------------------------------------------- */

boolean ex_passed(int ex_index,t_select *p_select)
{
	t_select_node *p_param = p_select->p_first;
	while (p_param)
	{
		if (!evaluate(ex_index,p_param)) return 0;
		p_param = p_param->p_next;
	}
	return 1;
}


/* -------------------------------------------------------------------------- *\
			Select a subset of examples on the basis of a conjunction of tests
\* -------------------------------------------------------------------------- */

int SELECT_command(t_command cmd)
{
	int i,rv,select_counter=0,no_of_examples = WS_get_ex_cnt();
	t_select *p_select = &cmd.param.l_select;	
	
	for (i=0; i < no_of_examples; i++)
	{
		if (ex_passed(i,p_select))	
		{
			/* copy example to the place indicated by select_counter */
			WS_move_ex_row(i,select_counter);			
			select_counter++;
		}	
	}
	/* compress examples database,room for only 'select_counter' examples */ 
	rv = WS_exs_remove_to_end(select_counter);
	
	if (rv == EXECUTION_OK)
	{
		printf("%d example(s) selected.\n",select_counter);
	}
	else
	{
		errmsg("Selection failed.\n");
	}
	return rv;
}


/* -------------------------------------------------------------------------- *\
									Exclude attributes form the example set
\* -------------------------------------------------------------------------- */

int EXCLUDE_command (t_command p) 
{

	char 			*aName;	
	t_attr_ref 		aRef;
	int 			found;
	struct t_exclude_node 	*fPtr;
	int 			retVal;
	struct t_exclude_node 	*wPtr;
	char buf[256];

	fPtr = wPtr = p.param.l_exclude_1.first;

	/* replace the generic "class" with the actual attribute name */
	while (wPtr)
	{
		if (!strcmp(wPtr->attr_name,"class"))
		{
			char *tmp_name;
			t_attr_ref cl_ref;
			if (WS_get_class_attr(&cl_ref))
			{	
				errmsg("The class attribute is not set");
				return EXECUTION_FAILED;
			}
			WS_get_attr_name(cl_ref,&tmp_name);
			realloc_string(&wPtr->attr_name,tmp_name);
		}
		else if (retVal = GetAttrToHandle(wPtr->attr_name, &aRef))
      return retVal;

		wPtr = wPtr->next;
	}

	wPtr = fPtr;

  if (p.cmd_id == K_EXCLUDE_ALL_EXCEPT) {
    /* This isn't nice. I depend on the fact that attributes are stored
       in an array and the attribute handles that are given to DBServer
       routines are actually indices to that array. 
       
       This is done because WS_get_next_attr doesn't survive when removing
       attribute node(s) between successive calls. 

       (Removing a set of attributes should actually be a low level operation?
        -ilkka 25.5.94) */
    aRef = 0;
    while (aRef < WS_get_attr_cnt ()) {
      wPtr = fPtr;
      if (retVal = WS_get_attr_name (aRef, &aName))
        return retVal;
        
      /* Go through param list. This makes this O (x^2), too bad. */
      while (!(found = (!eq_q_str(aName, wPtr->attr_name))) &&
      	     ((wPtr = wPtr->next) != NULL));

      /* remove if current attribute not in the list */
      if (!found) { /* This attribute gets crunched! */
        if (retVal = WS_remove_attr (aRef))  
          return retVal;
      } else 
        aRef++;  /* Goto next one */
    }  /* while aRef < WS_get_attr_cnt ... */
  } else  /* if exclude all except */

    /* Much easier than the previous one. Just go through the list and whamm! */
    for (wPtr = fPtr; wPtr != NULL; wPtr = wPtr->next) {
      if (retVal = GetAttrToHandle(wPtr->attr_name, &aRef))
        return retVal;
      if (retVal = WS_remove_attr (aRef))
        return retVal;
    }
  printf ("Specified attributes excluded.\n");
  return (NO_ERROR);
}


/* -------------------------------------------------------------------------- *\
			Transform the include command parameters into a more suitable form
\* -------------------------------------------------------------------------- */

t_attr_node *optimize_include (t_include_1 *p_include) 
{

  struct t_incl_val_node *ptr;
  t_attr_node		 *a_node;

  if (WS_create_attr (p_include->attr_name, T_USER, &a_node) != NO_ERROR) 
		return 0;
 
  for (ptr = &(p_include->val_list_head); ptr; ptr = ptr->next)
    if (WS_join_val_low (a_node, ptr->attr_val) != NO_ERROR) return 0;
	
	return a_node;

}


/* -------------------------------------------------------------------------- *\
		Include new (irrelevant) attributes into the working set
\* -------------------------------------------------------------------------- */

int INCLUDE_command (t_command p) 
{

t_attr_types 	aTyyppi;
int 		aRef,attr_cnt = 0;
t_attr_node	*p_attr_node;
int 		retVal, eCnt, cloneCnt=1, row, i, j;
t_ex_node 	*s;
t_include_1 *p_tmp,*p_include = &p.param.l_include_1;


  if ((eCnt = WS_get_ex_cnt ()) != 0) 
	{  
  	/* First we'll clone the existing example columns */
		/* the amount of new examples generated is determined by the product */
		/* of sizes of included attributes domain's */
		p_tmp = p_include;
		while (p_tmp)
		{
			attr_cnt++; /* count the number of attributes. Stupid, ha? */
			cloneCnt *= p_tmp->val_cnt;			

			/* check for duplicate names */
			if (WS_is_attr_name_used(p_tmp->attr_name))
			{
				errmsg("Attribute already exists");
				return EXECUTION_FAILED;
			}

			p_tmp = p_tmp->next;

		}
	
		WS_get_first_attr(&aRef,A_JARJ);

		do
		{      
			WS_get_attr_type (aRef, &aTyyppi);

			if (WS_create_ex_column (aTyyppi, cloneCnt * eCnt, &s) != EXECUTION_OK)
				return EXECUTION_FAILED;
	
			if (WS_clone_ex_column (aRef, s, cloneCnt) != EXECUTION_OK) 
				return EXECUTION_FAILED;

			if (WS_replace_ex_column (aRef, s) != EXECUTION_OK) 
			{
				WS_dispose_dbase();
   			return EXECUTION_FAILED;
			}

		}
		while (WS_get_next_attr(&aRef,A_JARJ) == NO_ERROR );
	}

	WS_set_ex_cnt (cloneCnt * eCnt);

	/* now, there's room for everyone */  	p_tmp = p_include;
	while (p_include)
	{
		/* create an attribute node from t_include_1 */
		if (!(p_attr_node = optimize_include(p_include)))
			return EXECUTION_FAILED;
		
		if (WS_add_attr(p_attr_node) != EXECUTION_OK)
			return EXECUTION_FAILED;
		else
			WS_get_attr_ref (p_include->attr_name, &aRef);

		if (eCnt > 0) 
		{
    	if (WS_create_ex_column (T_USER, cloneCnt * eCnt, &s) != EXECUTION_OK)
				return EXECUTION_FAILED;
			if (WS_replace_ex_column (aRef, s) != EXECUTION_OK)
				return EXECUTION_FAILED;
      
    	/* Init new column */
			row = 0;
			while (row < cloneCnt*eCnt) 
			{
    		for (i = 0; i < p_include->val_cnt; i++)
				{
	    	  WS_set_exval_kind (aRef, row, T_OK);
					WS_set_nominal_exval (aRef, row, i);
					row++;
				}
			}
 		}
		p_include = p_include->next;
	}

	printf("Working set contains now %d attributes and %d examples.\n",
		WS_get_attr_cnt(),WS_get_ex_cnt());

  return (NO_ERROR);
}


/* -------------------------------------------------------------------------- *\
								Map an attributes domain to a new one
\* -------------------------------------------------------------------------- */

int MAP_command (t_command p) 
{
	t_attr_ref	aRef;
	int		exCnt, i, lAttr;
	t_attr_node	*newAttr;
	t_ex_node	*newColPtr;
	int		onLuokkaAttr, retVal;

  if (!(exCnt = WS_get_ex_cnt ()))
    return (NO_ERROR);
  if (retVal = GetAttrToHandle (p.param.l_map_2.attr_name, &aRef))
    return retVal;
  if (retVal = WS_create_ex_column (T_USER, exCnt, &newColPtr))
    return retVal;
    
  /* Create local copy of the received attribute node */
  if (retVal = WS_copy_attr (p.param.l_map_2.attribute_table, &newAttr))
    return retVal;
    
  /* Loop where we map the examples. WS_map_domain () is a DBServer routine
     for modularity and efficiency (we use straight access to the
     data structures). */
  for (i = 0; i < exCnt; i++)
    if (retVal = WS_map_domain (aRef, i, newColPtr, newAttr,
    			 p.param.l_map_2.first, p.param.l_map_2.else_val, 
    	    		 p.param.l_map_2.dont_know_val, 
    	    		 p.param.l_map_2.dont_care_val))
      return retVal;
      	    
  /* Delete old attribute node and add new one + new examples.
     See, if we're mapping the class attribute and take care of it. */
  if (WS_is_class_attr () && ((retVal = WS_get_class_attr (&lAttr)) > 0)) {
    WS_free_attr (newAttr);
    return retVal;
  } else
    onLuokkaAttr = (lAttr == aRef);
  if (retVal = WS_remove_attr (aRef)) {
    WS_free_attr (newAttr);
    return retVal;
  }
  if (retVal = WS_add_attr (newAttr)) {
    WS_free_attr (newAttr);
    return retVal;
  }
  if (retVal = GetAttrToHandle (p.param.l_map_2.attr_name, &aRef)) {
    WS_free_attr (newAttr);
    return retVal;
  }
  if (retVal = WS_replace_ex_column (aRef, newColPtr)) {
    WS_free_attr (newAttr);
    return retVal;
  }
  if (onLuokkaAttr)
    WS_set_class_attr (aRef);
  
	printf("Mapping complete.\n");

  /* We're done */
  return (NO_ERROR);
}


/* -------------------------------------------------------------------------- *\
						Add random noise to an attribute's values
		Note: for discrete attributes, the new value is quaranteed to differ 
				from the original one
\* -------------------------------------------------------------------------- */ 
int NOISE_command (t_command p) 
{
	
char 			*aName;
int ch_cnt;
t_attr_ref 		aRef, class;
int 			found, retVal;
struct t_noise_node	*wPtr;
	
  for (wPtr = p.param.l_noise_1.first; wPtr != NULL; wPtr = wPtr->next) {
     if (retVal = GetAttrToHandle (wPtr->attr_name, &aRef))
       return retVal;
     ScrambleOneAttribute (aRef, wPtr->noise_p, DO_NOISE,&ch_cnt);
		 printf("%s: %d out of %d values changed.\n",
			wPtr->attr_name,ch_cnt,WS_get_ex_cnt());
  };
  
  /* others-probability */
  if (p.param.l_noise_1.others > 0.001) {
    retVal = WS_get_first_attr (&aRef, A_JARJ);
    if (WS_is_class_attr ())
      WS_get_class_attr (&class);
    else
      class = -1;
    while (!retVal) {
      if (aRef != class) {
        WS_get_attr_name (aRef, &aName);
        wPtr = p.param.l_noise_1.first;
        while (wPtr != NULL && !(found = !strcmp (wPtr->attr_name, aName)))
          wPtr = wPtr->next;
        if (!found)
				{
          ScrambleOneAttribute 
						(aRef, p.param.l_noise_1.others, DO_NOISE,&ch_cnt);		 
					printf("%s: %d out of %d values changed.\n",
						aName,ch_cnt,WS_get_ex_cnt());
				}
      }
      retVal = WS_get_next_attr (&aRef, A_JARJ);
    }
  }
  return NO_ERROR;
}


/* -------------------------------------------------------------------------- *\
							Add random noise to an attribute's values
		Note: is NOT quaranteed to change value, new and original value may be 
			the same
\* -------------------------------------------------------------------------- */ 
int RANDOMIZE_command (t_command p) 
{
	
char 			*aName;
int ch_cnt;
t_attr_ref 		aRef, class;
int 			found, retVal;
struct t_noise_node 	*wPtr;
	
  for (wPtr = p.param.l_noise_1.first; wPtr != NULL; wPtr = wPtr->next) {
     if (retVal = GetAttrToHandle (wPtr->attr_name, &aRef))
       return retVal;
     ScrambleOneAttribute (aRef, wPtr->noise_p, DO_RANDOMIZE,&ch_cnt);
     printf("%s: %d out of %d values changed.\n",
				wPtr->attr_name,ch_cnt,WS_get_ex_cnt());
 
 };
  
  /* others-probability */
  if (p.param.l_noise_1.others > 0.001) {
    retVal = WS_get_first_attr (&aRef, A_JARJ);
    if (WS_is_class_attr ())
      WS_get_class_attr (&class);
    else
      class = -1;
    while (!retVal) {
      if (aRef != class) {
        WS_get_attr_name (aRef, &aName);
        wPtr = p.param.l_noise_1.first;
        while (wPtr != NULL && !(found = !strcmp (wPtr->attr_name, aName)))
          wPtr = wPtr->next;
        if (!found)
				{
          ScrambleOneAttribute (aRef, p.param.l_noise_1.others, 
						DO_RANDOMIZE,&ch_cnt);
					printf("%s: %d out of %d values changed.\n",
						aName,ch_cnt,WS_get_ex_cnt());
				}
      }
      retVal = WS_get_next_attr (&aRef, A_JARJ);
    }
  }
  return NO_ERROR;
}


/* -------------------------------------------------------------------------- *\
			Output basic statistics of the active working set
\* -------------------------------------------------------------------------- */ 
int STAT_command (t_command p) 
{
	
	char		*aName;
	t_attr_ref	aRef;
	t_snode		*curr_attr;
	int 		ret_val, stop;

  OutputStatHeader (p.param.l_stat_1.stat_file_name,
  		    p.param.l_stat_1.comment);
  if (p.cmd_id == K_STAT) {
    curr_attr = p.param.l_stat_1.first;
    while (curr_attr != NULL) 
		{
      ret_val = StatSingleAttr (p.param.l_stat_1.stat_file_name,
    				p.param.l_stat_1.comment,
    				curr_attr->attr_name, p.row);
      if (ret_val)
        return ret_val;
      curr_attr = curr_attr->next;
    }
  } else {  /* wir mussen loop hier */
    if (stop = WS_get_first_attr (&aRef, A_JARJ))
     return stop;
    while (!stop) {
      if (stop = WS_get_attr_name (aRef, &aName))
       return stop;
      if (stop = StatSingleAttr (p.param.l_stat_1.stat_file_name,
      			     	 p.param.l_stat_1.comment,
      			     	 aName,
      			     	 p.row))
       return stop;
      if (!stop)
        stop = WS_get_next_attr (&aRef, A_JARJ);
    }  /* while */
  }
  printf ("\nTotal example count: %d\n", WS_get_ex_cnt ());
	if (p.param.l_stat_1.stat_file_name)
	{
		printf("Statistics written to file %s.\n",p.param.l_stat_1.stat_file_name);
	}
  return (NO_ERROR);
}  /* STAt_command */


/* -------------------------------------------------------------------------- *\
									DELETE files from the file system

		Note: this command is not yet enabled for the user, it could be
			someday, though
\* -------------------------------------------------------------------------- */ 
int DELETE_command(t_command cmd)
{
	int i,rv = 0;
	char *name;
	t_delete_node *p_node = cmd.param.l_delete_1.first;
	while (p_node)
	{
		for (i = 0; i <= cmd.param.l_delete_1.top_exclam_index; i++)
		{
			IO_set_round(i);
			name = IO_get_expanded_name(p_node->f_name);		
			remove(name);
		}
		IO_set_round(0);
		p_node = p_node->next;
	}
	return rv; 
}


/* -------------------------------------------------------------------------- *\
									Initialize the working set database

		Note: this command is not enabled for the user
\* -------------------------------------------------------------------------- */ 
int INIT_command(t_command cmd)
{
	WS_dispose_dbase();
	WS_remove_active_classifier();
	return 0;
}


/* -------------------------------------------------------------------------- *\
						BACKUP given portions of the working set
\* -------------------------------------------------------------------------- */ 
int BACKUP_command(t_command cmd)
{
	int rv= EXECUTION_OK;

	switch (cmd.cmd_id)
	{
		case K_BACKUP_ATTS_EXS:
			rv =WS_example_db_backup(cmd.param.l_backup.ae_name);
		break;
		case K_BACKUP_RULE:			
		{
			t_algorithms alg = cmd.param.l_backup.alg;
			char *cf_name = cmd.param.l_backup.cf_name;
			if (alg)
				rv = WS_rule_backup(alg,cf_name);
		}
		break;
		default: rv = EXECUTION_FAILED; 
		break;
	}
	return rv;
}


/* -------------------------------------------------------------------------- *\
									Restore a portion of the working set
\* -------------------------------------------------------------------------- */ 
int RESTORE_command(t_command cmd)
{
	int rv = EXECUTION_OK;

	switch (cmd.cmd_id)
	{
		case K_RESTORE_ATTS_EXS:
			rv =WS_example_db_restore(cmd.param.l_restore.ae_name);
		break;
		case K_RESTORE_RULE:
		{
			t_algorithms alg = cmd.param.l_restore.alg;
			char *cf_name = cmd.param.l_restore.cf_name;
			if (alg)
			{
				rv =WS_rule_restore(alg,cf_name);
				WS_set_active_classifier(alg);
			}
			else
				WS_remove_active_classifier();		
		}
		break;
		default: rv = EXECUTION_FAILED; 
		break;
	}
	return rv;
}


/* -------------------------------------------------------------------------- *\
													SET global options 
\* -------------------------------------------------------------------------- */ 
int SET_command(t_command cmd)
{
	switch (cmd.cmd_id)
	{
		case K_SET_RANDSEED:
		{
			randomize(cmd.param.l_set.seed);
			printf("Random number generator initialized with seed %d.\n",
				cmd.param.l_set.seed);
		}
		break;
		case K_SET_RANDCLOCK:
		{
			struct timeval	t;
			struct timezone	tz;

  		gettimeofday (&t, &tz);

			randomize(t.tv_sec);			
			printf("Random number generator initialized with seed %d.\n",
				t.tv_sec);
		}
		break;
		case K_SET_ALG:
		default:
		{
			errmsg("Command not implemented");
			return EXECUTION_FAILED;
		}	
	}

	return EXECUTION_OK;
}


/* -------------------------------------------------------------------------- *\
															QUIT from TELA
\* -------------------------------------------------------------------------- */ 
int QUIT_command(t_command cmd)
{
	return USER_QUIT;
}



