/*                                                                            *\

	  TELA - Testing Environment for Learning Algorithms - version 1.995
    ==================================================================

		Module:       TESLA_statistics.h

		Contents:    Statistics handling and output routines.	
								 
		Author: Juho Rousu (1995,1996)
	
		Services:

		STATS_check_for_uncollected   - check whether some stats are not collected
		STATS_copy_list								- make a copy of a statistics list
		STATS_create_node 						- create a statistics node
		STATS_merge_lists							- merge to statistics lists
		STATS_open_file								- open statistics file for writing
		STATS_write_file							- write parameters and statistics to file
		STATS_STATS_print_parameters				- print parametres to file
		STATS_update_node							- update a statistics node
		STATS_update_from_avg					- update stats from a list of avg. stats
		STATS_delete_list							- delete the statistics list
		STATS_write_line							- write a line to the statsitics file
		STATS_find_node								- find a statistics node by name

		Internal Functions:

		stat_is_relevant							- is given statistic relevant to given task
		print_statistics							- print statistics to file
		avg_stats_val									- calculate an average value of the statistic
		std_dev_val										- calculate the std. deviation of the stat
		copy_stats_list								- make a copy of stats list
\*  																																					*/

#include <math.h>
#include <time.h>

#include "TESLA_statistics.h"
#include "IO_server.h"


/* ========================================================================== *\
														INTERNAL FUNCTIONS
\* ========================================================================== */

int stat_is_relevant(char *name,t_algorithms alg, int mode);
void print_statistics(t_stats_list *p_list, int rounds, FILE* s_file);
float avg_stats_val(t_stats_node *p_stats_node);
float std_dev_val(t_stats_node *p_stats_node);
void delete_statistics_list(t_stats_list *p_stats);
void print_timestamp(FILE *stats_file);


/* -------------------------------------------------------------------------- *\
					check if given statistic is relevant to task mode
\* -------------------------------------------------------------------------- */

int stat_is_relevant(char *name,t_algorithms alg, int mode)
{
	if (mode == -1) /* both RUN and TEST come to question */
	{
		if (ADB_alg_statistic_ok(name,alg,0) || 
			ADB_alg_statistic_ok(name,alg,1)) return 1;
	}
	else
		if (ADB_alg_statistic_ok(name,alg,mode)) return 1;

	/* if algorithm isn't able to handle timings, that's all right */
	if (mode != 1 && !strcmp(name,"TIME")) return 1;
	if (mode != 0 && !strcmp(name,"TESTTIME")) return 1;

	return 0;
}


/* -------------------------------------------------------------------------- *\
 												print statistics to a file 
\* -------------------------------------------------------------------------- */

void print_statistics(t_stats_list *p_list, int pr_mode, FILE* s_file)
{
	t_stats_node *p_s;
	int first=1;

	if (list_get_first(p_list,(void**)&p_s))
	{
		do
		{
			/* if we are printing averages and avg.value is printable */
			if ((pr_mode == S_BOTH || pr_mode == S_ONLY_AVG) 
					&& p_s->printing != S_ONLY_IND && p_s->printing != S_NO)
			{
				if (first) fprintf(s_file,"");
				fprintf(s_file," %s =%4.3f (+/-%1.3f)",p_s->name,
					avg_stats_val(p_s),std_dev_val(p_s));
				first=0;
			}
			else if (pr_mode == S_ONLY_IND  && 
				p_s->printing != S_ONLY_AVG && p_s->printing != S_NO)
			{
				if (first) fprintf(s_file,"");
				fprintf(s_file," %s =%4.3f",p_s->name,p_s->value);
				first=0;
			}
		}
		while (list_get_next(p_list,(void**)&p_s));

		if (!first) fprintf(s_file,"\n");
	}
}


/* -------------------------------------------------------------------------- *\
								Calculate the average value of the statistic
\* -------------------------------------------------------------------------- */

float avg_stats_val(t_stats_node *p_stats_node)
{
	return p_stats_node->sum_val/(float)p_stats_node->terms;
}


/* -------------------------------------------------------------------------- *\
							Calculate the std. deviation of the statistic
\* -------------------------------------------------------------------------- */

float std_dev_val(t_stats_node *p_stats_node)
{
	float avg = avg_stats_val(p_stats_node);
	return (float)sqrt(p_stats_node->sqrsum_val/(float)p_stats_node->terms 
		- avg*avg);
}


/* -------------------------------------------------------------------------- *\
											Print a timestamp to a file
\* -------------------------------------------------------------------------- */

void print_timestamp(FILE *stats_file)
{
	char date[256];
   
  time_t t;
  
  t = time(0);
 
  ctime_r(&t,date,256);
   
  fprintf(stats_file,"%s",date);      
}


/* ========================================================================== *\
													 			PUBLIC				
\* ========================================================================== */


/* -------------------------------------------------------------------------- *\
					find a statistics node and return a handle to it 
\* -------------------------------------------------------------------------- */

int STATS_find_node(t_stats_list *p_stats,
	char *stat_name, t_stats_node **pp_node)
{
	t_stats_node *p_stats_node;
	int rv,no_match=1;
	void *p;
	
	rv = list_get_first(p_stats,&p);

	while (rv && no_match)
	{
		p_stats_node = (t_stats_node*)p;

		if (!strcmp(p_stats_node->name,stat_name))
		{
			no_match = 0;
			*pp_node = p_stats_node;
			break;	
		}
		rv = list_get_next(p_stats,&p);
	}

	return no_match ? EXECUTION_FAILED : EXECUTION_OK;

}


/* -------------------------------------------------------------------------- *\
												create a new statistics node
\* -------------------------------------------------------------------------- */

t_stats_node *STATS_create_node(char *name)
{
	t_stats_node *p_stats_node = (t_stats_node*)malloc(sizeof(t_stats_node));
	alloc_string(&p_stats_node->name,name);
	p_stats_node->value=0;
	p_stats_node->sum_val=0;
	p_stats_node->sqrsum_val=0;
	p_stats_node->printing=1;
	p_stats_node->terms=0;
	return p_stats_node;
}


/* -------------------------------------------------------------------------- *\
 							find statistic from list and update its value
\* -------------------------------------------------------------------------- */

int STATS_update_node(t_stats_list *p_stats,char *stat_name,float value)
{
	t_stats_node *p_stats_node;
	/* go trough the statistics list and search for a matching stats name */
	int rv,no_match = 1;
	
	rv = list_get_first(p_stats,(void **)&p_stats_node);

	while (rv && no_match)
	{
		if (!strcmp(p_stats_node->name,stat_name))
		{
			no_match = 0;
			p_stats_node->value = value;
			p_stats_node->sum_val += value;
			p_stats_node->sqrsum_val += value*value;
			p_stats_node->terms++;
			break;	
		}
		rv = list_get_next(p_stats,(void**)&p_stats_node);
	}

	return no_match ? EXECUTION_FAILED : EXECUTION_OK;
}


/* -------------------------------------------------------------------------- *\
		check whether in the list exists statistics that have not been updated
\* -------------------------------------------------------------------------- */

int STATS_check_for_uncollected(t_stats_list *p_stats)
{	
	t_stats_node *p_stats_node;
	
	int rv;
	
	rv = list_get_first(p_stats,(void **)&p_stats_node);

	while (rv)
	{
		if (!p_stats_node->terms)
		{
			if (strcmp(p_stats_node->name,"TIME") && 
				strcmp(p_stats_node->name,"TESTTIME"))
			{			
				char msg[256];
				sprintf(msg,"Statistic \"%s\" was not measured",p_stats_node->name);
				errmsg(msg);
				return EXECUTION_FAILED;
			}
		}

		rv = list_get_next(p_stats,(void**)&p_stats_node);
	}

	return EXECUTION_OK;

}


/* -------------------------------------------------------------------------- *\
		 									merge two statistics lists

		Note: this merge is asymmetric in the sense that the first list is
			treted as 'indenpendently printable' and the second as 
			being of 'print only averages' type.
\* -------------------------------------------------------------------------- */

int STATS_merge_lists(t_algorithms alg,int mode, t_stats_list *p_list1,
	t_stats_list *p_list2)
{
	t_stats_node *p_stats_node, *p_dummy;
	int rv;

	if (!p_list1->p_head) 
		rv = S_ONLY_AVG;
	else rv = S_ONLY_IND;

	/* go trough the second statistics list */
	if (list_get_first(p_list2,(void**)&p_stats_node))
	{
		do
		{ 
			/* if relevant statistic is encountered*/
			if (stat_is_relevant(p_stats_node->name,alg,mode))
			{
				/* if it doesn't exist already in first list */
				if (STATS_find_node(p_list1,p_stats_node->name,&p_dummy) 
					!= EXECUTION_OK)
				{
					/* add it to first list with 'printing' set to 0 */
					t_stats_node *p_new = STATS_create_node(p_stats_node->name);
					p_new->printing = S_ONLY_AVG; /* only average printing */
					list_append(p_list1,p_new);
					if (rv == S_ONLY_IND) rv = S_BOTH;
				}
				else rv = p_dummy->printing = S_BOTH; /* both modes of printing */
			}
		} 		
		while (list_get_next(p_list2,(void**)&p_stats_node));
	}
	return rv;
}


/* -------------------------------------------------------------------------- *\
		updates stats in first list from average values in second list
\* -------------------------------------------------------------------------- */

void STATS_update_from_avg(float rounds, t_stats_list *p_list1, 
	t_stats_list *p_list2)
{
	t_stats_node *p_stats_node;

	if (list_get_first(p_list2,(void**)&p_stats_node))
		do 
			STATS_update_node
				(p_list1,p_stats_node->name,avg_stats_val(p_stats_node));
		while (list_get_next(p_list2,(void**)&p_stats_node));
}


/* -------------------------------------------------------------------------- *\
		  		Print algorithm execution parameters to file
\* -------------------------------------------------------------------------- */

void STATS_print_parameters(char *cmd_string,int rounds,t_algorithms alg,
	t_run_node *param_list, FILE* s_file,char *comment)
{
	char alg_name[256];

	if (comment) fprintf(s_file,"# %s\n",comment);

	fprintf(s_file,"%s ",cmd_string);

	/* 'TELA' as an algorithm name is not written to file */	
	if (alg ? ADB_get_alg_name(alg,alg_name):0)
	{
		fprintf(s_file,"%s ",alg_name);	
	}

	if (rounds > 1)
	{
		fprintf(s_file,"rounds: %d ", rounds);
	}

	if (param_list)
	{
		do
		{
			fprintf(s_file,"%s",param_list->selector);
			if (param_list->val)
				fprintf(s_file,"=%s ",param_list->val);
		} 
		while (param_list = param_list->next);
	}

	fprintf(s_file,"\n");
}


/* -------------------------------------------------------------------------- *\
									Print parameters and statistics to a file 
\* -------------------------------------------------------------------------- */

int STATS_write_file(char *cmd_str,int rounds, 
	t_algorithms alg,int in_do,int pr_mode, t_run_node *p_params, char *comment,
	char *file_name,t_stats_list *p_list)
{
	if (!p_list) return EXECUTION_OK; /* no statistics recorded, no printing */

	if (pr_mode == -1 ) return EXECUTION_OK;


	if (file_name)
	{
		FILE *stat_file;
		if (stat_file = STATS_open_file(IO_get_expanded_name(file_name)))
		{
			print_timestamp(stat_file);
			STATS_print_parameters(cmd_str,rounds,alg,p_params,stat_file,comment);
			print_statistics(p_list,pr_mode,stat_file);
			fclose(stat_file);
		}		
		else
		{
			errmsg("Couldn't open statistics file");
			return EXECUTION_FAILED;
		}
	}

	/* if printing from within do-loop print the parameters in front  */
	/* to help reading statistics on screen. */	
	if (in_do)
	{
		STATS_print_parameters(cmd_str,rounds,alg,0,stdout,comment);
	}

	print_statistics(p_list,pr_mode,stdout);

	return EXECUTION_OK;	
}

/* -------------------------------------------------------------------------- *\
										Write a line to statistics file
\* -------------------------------------------------------------------------- */

void STATS_write_line(char *file_name,char *line)
{
	FILE *fp;
		
	if (!file_name) return;

	if (fp=STATS_open_file(file_name))
	{
		fprintf(fp,"%s\n",line);
		fclose(fp);
	}
}


/* -------------------------------------------------------------------------- *\
							open a statistics file for apped
\* -------------------------------------------------------------------------- */

FILE *STATS_open_file(char *stat_file_name)
{
	FILE *fp;
	char buf[512];

	/* check if it exists already */
	if (fp = fopen(stat_file_name,"r"))
	{
		/* is is a statistics file */
		if (fgets(buf,512,fp))
		{
			if (strncmp(buf,"# TELA statistics file",22))
			{
				sprintf(buf,"The file %s is not a TELA statistics file",stat_file_name);
				errmsg(buf);
				return 0;
			}
		}
		fclose(fp);
	}
	else
	{
		/* write the header to file */
		if (fp = fopen(stat_file_name,"w"))
		{
			fprintf(fp,"# TELA statistics file\n");
			fclose(fp);
		}
		else
		{
			sprintf(buf,"Could not open the statistics file %s for writing",
				stat_file_name);
			errmsg(buf);
			return 0;
		}
	}

	return fopen(stat_file_name,"a");

}


/* -------------------------------------------------------------------------- *\
											Delete the statistics list
\* -------------------------------------------------------------------------- */

void STATS_delete_list(t_stats_list *p_stats)
{
	t_stats_node *p_node;

	while (list_remove_first(p_stats,(void**)&p_node))
	{
		free(p_node);
	}
}


/* -------------------------------------------------------------------------- *\
										make a copy of a statistics list
\* -------------------------------------------------------------------------- */

int STATS_copy_list(t_algorithms alg,int mode,
	t_stats_list *p_list1, t_stats_list *p_list2,int printing)
{
	t_stats_node *p_stats_node, *p_dummy;
	int rv;

	rv = printing;

	list_init(p_list1);

	/* go trough the second statistics list */
	if (list_get_first(p_list2,(void**)&p_stats_node))
	{
		do
		{ 
			/* if relevant statistic is encountered*/
			if (stat_is_relevant(p_stats_node->name,alg,mode))
			{
				t_stats_node *p_new = STATS_create_node(p_stats_node->name);
				p_new->printing = printing; 
				list_append(p_list1,p_new);
			}
		} 		
		while (list_get_next(p_list2,(void**)&p_stats_node));
	}
	return rv;
}


