/*                                                                            *\

	  TELA - Testing Environment for Learning Algorithms - version 1.995
    ==================================================================

		Module:       UI_basic

		Contents:     Parsers for some common language elements in TESLA 

		Author: Juho Rousu (1995,1996)
	
		Services:

		TESLA_word             - parse an unquoted string
		TESLA_string           - parse a quoted string, strip quotes
		TESLA_qstring		       - parse a quotes string, keep the quotes
		TESLA_nat_val          - parse s natural number
		TESLA_cont_val         - parse a floating point number
		TESLA_int_val          - parse an integer value
		TESLA_nominal_val      - parse a nominal value: character, word or string
		TESLA_attribute_value  - parse a value that is legal for an attribute
		TESLA_percent_val      - parse an percentage
		TESLA_relation         - parse an arithmetic relation
		TESLA_attribute_name   - parse an attribute name
		TESLA_algorithm        - parse an algorithm name
		TESLA_filename         - parse a filename
		TESLA_param_value      - parse a parameter value
		TESLA_symbol           - parse a reserved word
		TESLA_datatype         - parse a data (file) type
		TESLA_wordtoken_cmp	   - compare a string to current (word) token
		TESLA_is_separator     - is current token given separator character
		TESLA_is_reserved_mark - is given character a reserved separator
		TESLA_is_reserved_word - is current token given reserved word 
\*       
                                                                     */
#include <string.h>
#include <stdio.h>

#include "UI_basic.h"
#include "WS_Database.h"


/* -------------------------------------------------------------------------- *\
												Parses an unquoted string
\* -------------------------------------------------------------------------- */

int TESLA_word(char **wd)
{
	char *tmp_str;

	if (lex_token.kind == wordtoken)
	{
		tmp_str = lex_token.value.word.text;
    tmp_str[lex_token.value.word.length] = '\0'; /* pad null to end */
		*wd = (char *)malloc((lex_token.value.word.length+1)*sizeof(char));
		strcpy(*wd,tmp_str);	
	}
	else return PARSE_FAILED;
	
	return PARSE_OK;
}


/* -------------------------------------------------------------------------- *\
                             Parses a string
\* -------------------------------------------------------------------------- */

int TESLA_string(char **str)
{
	char *tmp_str;

	if (lex_token.kind == stringtoken)
	{
		tmp_str = lex_token.value.string.quoted.text;
    tmp_str[lex_token.value.string.quoted.length] = '\0'; /* pad zero to end */
		*str = (char*)malloc((lex_token.value.string.quoted.length+1)*sizeof(char));
		strcpy(*str,tmp_str);	
	}
	else return PARSE_FAILED;

	return PARSE_OK;
}


/* -------------------------------------------------------------------------- *\
							Parses a string and put's double quotes around
\* -------------------------------------------------------------------------- */

int TESLA_qstring(char **str)
{
	char *tmp_str;

	if (lex_token.kind == stringtoken)
	{
		tmp_str = lex_token.value.string.quoted.text;
    tmp_str[lex_token.value.string.quoted.length] = '\0'; /* pad zero to end */
		*str = (char*)malloc((lex_token.value.string.quoted.length+3)*sizeof(char));
		(*str)[0] = '\"';
		strcpy(*str+1,tmp_str);		
		(*str)[strlen(*str)+1] = '\0';	
		(*str)[strlen(*str)] = '\"';	
	}
	else return PARSE_FAILED;

	return PARSE_OK;
}


/* -------------------------------------------------------------------------- *\
													Parses a natural number
\* -------------------------------------------------------------------------- */

int TESLA_nat_val(unsigned *p_n)
{
	if (lex_token.kind == natnumtoken)
	{
		*p_n = (unsigned)lex_token.value.natnumber;
		return PARSE_OK;
	}
	return PARSE_FAILED;
}	


/* -------------------------------------------------------------------------- *\
										Parse a floating point figure
\* -------------------------------------------------------------------------- */

int TESLA_cont_val(float *p_f)
{
	unsigned i;
	int sign = 1;	

	lex_positiontype here;

	lex_store(&here);		

	if (lex_ismark('-'))
	{
		sign = -1;
		lex_next(); if (lex_cancel()) goto cancel;
	}
	else if (lex_ismark('+'))
	{
		sign = 1;
		lex_next(); if (lex_cancel()) goto cancel;
	}

	if (lex_token.kind == realnumtoken)
	{
 		*p_f = lex_token.value.realnumber*sign;
	}
	else if (!TESLA_nat_val(&i)) *p_f = i*sign;
	else return PARSE_FAILED;

	return PARSE_OK;

	cancel: 

	lex_restore(here);
	return PARSE_CANCELLED;

} 


/* -------------------------------------------------------------------------- *\
														Parse an integer
\* -------------------------------------------------------------------------- */

int TESLA_int_val(int *p_n)
{
	float tmp_f;
	int sign = 1;
	unsigned u_tmp;

	lex_positiontype here;

	lex_store(&here);		

	if (lex_ismark('-'))
	{
		sign = -1;
		lex_next(); if (lex_cancel()) goto cancel;
	}
	else if (lex_ismark('+'))
	{
		sign = 1;
		lex_next(); if (lex_cancel()) goto cancel;
	}

	if (TESLA_nat_val(&u_tmp))
	{
		lex_restore(here);
		if (TESLA_cont_val(&tmp_f)) goto cancel;

		*p_n = (int)tmp_f*sign;							/* truncate to int */

		/* was there fractional digits */
		if (*p_n != tmp_f) goto cancel;
	}
	else
		*p_n = (int)u_tmp*sign;

	return PARSE_OK;	

cancel:

	lex_restore(here);
	return PARSE_CANCELLED;

}

/* -------------------------------------------------------------------------- *\
													Parse a nominal value
\* -------------------------------------------------------------------------- */

int TESLA_nominal_val(char **p_val)
{
	/* is it a unquoted string ? */
	if (!TESLA_word(p_val)) return PARSE_OK;

	if (!TESLA_qstring(p_val))
	{
		if ((int)strlen(*p_val) >= 3)	   /* "" does not qualify for a value */
			return PARSE_OK;
		else 
			return PARSE_FAILED;
	}

	/* if not, then, is it a single character? */
	if (lex_token.kind == marktoken)
	{
		*p_val = (char *)malloc(2*sizeof(char));
		if (!TESLA_is_reserved_mark(lex_token.value.character))
		{
			(*p_val)[0] = lex_token.value.character;
			(*p_val)[1] = '\0';
			return PARSE_OK;
		}
	}
	return PARSE_FAILED;
}


/* -------------------------------------------------------------------------- *\
												Parse an attribute value
\* -------------------------------------------------------------------------- */

int TESLA_attribute_value(char **pp_val)
{
	int i;
	float f;
	char buf[256];	

	if (!TESLA_int_val(&i))
	{
		sprintf(buf,"%d",i);
		if (!alloc_string(pp_val,buf)) 
		{
			errmsg("Out of memory");
			return PARSE_CANCELLED;
		} 
		return PARSE_OK;
	}

	if (!TESLA_cont_val(&f)){
		sprintf(buf,"%f",f);
		if (!alloc_string(pp_val,buf)) 
		{
			errmsg("Out of memory");
			return PARSE_CANCELLED;
		} 
	 	return PARSE_OK;
	}

	if (!TESLA_nominal_val(pp_val)) return PARSE_OK;


	return PARSE_FAILED;
}


/* -------------------------------------------------------------------------- *\
											Parse a precentage figure 
\* -------------------------------------------------------------------------- */

int TESLA_percent_val(float *p_n)
{
	unsigned tmp_n;
	if (TESLA_nat_val(&tmp_n))
	{
		if (TESLA_cont_val(p_n))
			return PARSE_FAILED;
	}
	else
		*p_n = tmp_n;	

	if (0 <= *p_n )
		if (*p_n <= 100)
			return PARSE_OK;
		
	return PARSE_FAILED;
}


/* -------------------------------------------------------------------------- *\
											Parse an arithmetic relation
\* -------------------------------------------------------------------------- */

int TESLA_relation(t_relation *p_rel)
{
	lex_positiontype here;

	lex_store(&here);

	if (lex_token.kind != marktoken)
		return PARSE_FAILED;	

	if (lex_ismark('<'))
	{
		lex_next(); if (lex_cancel()) return PARSE_CANCELLED;
		if (lex_token.kind == marktoken) 
		{
			if (lex_ismark('='))
				*p_rel = LE;
			else if (lex_ismark('>'))
				*p_rel = NE;
			else
				return PARSE_FAILED;
		}
		else
		{
			lex_restore(here);
			*p_rel = LT;
		}
	}
	else if (lex_ismark('>'))
	{
		lex_next(); if (lex_cancel()) return PARSE_CANCELLED;
		if (lex_token.kind == marktoken) 
		{
			if (lex_ismark('='))
				*p_rel = GE;
			else
				return PARSE_FAILED;
		}
		else
		{
			lex_restore(here);
			*p_rel = GT;
		}
	}
	else if (lex_ismark('='))
		*p_rel = EQ;
	else
	  return PARSE_FAILED;


	return PARSE_OK;
}


/* -------------------------------------------------------------------------- *\
												Parse an attribute name
\* -------------------------------------------------------------------------- */

int TESLA_attribute_name(char **name)
{
	t_attr_ref a_ref;
	char *tmp;

	if (TESLA_is_reserved_word(class_sym))
	{
		if (alloc_string(name,"class")) /* represents class-attr */
			return PARSE_OK;
		else
		{
			errmsg("Out of memory");
			return PARSE_FAILED;
		}
	}
	
	if (!TESLA_qstring(name)) return PARSE_OK;

	if (!TESLA_word(name)) return PARSE_OK;

	return PARSE_FAILED;
}


/* -------------------------------------------------------------------------- *\
												Parse an algorithm name (optional)
\* -------------------------------------------------------------------------- */

int TESLA_algorithm(t_algorithms *p_alg_ref)
{
	char *name;


	if (!TESLA_word(&name))	/* is it ? */
	{
		if (!ADB_get_alg_ref(name,p_alg_ref)) /* does it exist? */
		{
			free(name);
			return PARSE_FAILED;
		}
		free(name);
	}
	else
	{
		return PARSE_FAILED;
	}

	return PARSE_OK;
}


/* -------------------------------------------------------------------------- *\
															Parse a filename
\* -------------------------------------------------------------------------- */

int TESLA_filename(t_algorithms alg, t_filetype f_type, char **fn)
{
	char tmp_str[256];

	/* from our point of view, a filename is either a quoted */
	if (!TESLA_string(fn))
		return PARSE_OK;
	else if (!TESLA_word(fn)) /* or an unquoted string */
	{  
 		ADB_get_filename_ext(alg,f_type,tmp_str);

		/* avoid generating <stem>.atts.atts filenames*/
		if (strstr(*fn,tmp_str))
			return PARSE_OK;

		*fn = (char*)realloc(*fn,(strlen(*fn)+strlen(tmp_str)+1)*sizeof(char));
		strcat(*fn,tmp_str);
		return PARSE_OK;
	}

	return PARSE_FAILED;
}


/* -------------------------------------------------------------------------- *\
					 							Parse a parameter value
\* -------------------------------------------------------------------------- */

int TESLA_param_value(char **pp_val)
{
	char tmp[256];
	float f;
	int d;
	if (!TESLA_word(pp_val)) return PARSE_OK;
	if (!TESLA_string(pp_val)) return PARSE_OK;	
	if (!TESLA_int_val(&d)) 
	{
		sprintf(tmp,"%d",d);
	  *pp_val = (char*)malloc((strlen(tmp)+1)*sizeof(char));
		strcpy(*pp_val,tmp);
		return PARSE_OK;
	}
	if (!TESLA_cont_val(&f)) 
	{
		sprintf(tmp,"%f",f);
	  *pp_val = (char*)malloc((strlen(tmp)+1)*sizeof(char));
		strcpy(*pp_val,tmp);
		return PARSE_OK;
	}		
	return PARSE_FAILED;
}


/* -------------------------------------------------------------------------- *\
												Parse a reserved word
\* -------------------------------------------------------------------------- */

int TESLA_symbol(int *id)
{
	if (lex_token.kind == symboltoken)
	{
		*id = lex_token.value.symnum;
		return PARSE_OK;
	}
	
	return PARSE_FAILED;
}


/* -------------------------------------------------------------------------- *\
								Parse a data (file) type
\* -------------------------------------------------------------------------- */

int TESLA_datatype(t_filetype *p_id)
{
	int id;
	if (!TESLA_symbol(&id))
	{
		switch (id)
		{
			case atts_sym: *p_id = FT_ATTS; break;
			case exs_sym:*p_id = FT_EXS; break;
			case both_sym:*p_id = FT_BOTH; break;
			case rule_sym:*p_id = FT_ATTSRULE; break;
			default:
				return PARSE_FAILED;
		}
		return PARSE_OK;
	}
	else return PARSE_FAILED;
}


/* -------------------------------------------------------------------------- *\
			Check if current token is  the same word as given string 
\* -------------------------------------------------------------------------- */

int TESLA_wordtoken_cmp(char* str)
{
	if (lex_token.kind == wordtoken)
		return strncmp(lex_token.value.word.text, str, 
       	    	  lex_token.value.word.length);
	else
		return 1;
}


/* -------------------------------------------------------------------------- *\
								Is current token given separator character
\* -------------------------------------------------------------------------- */

boolean TESLA_is_separator(char c)
{
	if (lex_ismark(c)) 
		return 1;
	else
		return 0;
}


/* -------------------------------------------------------------------------- *\
							Is given character reserved in TESLA language
\* -------------------------------------------------------------------------- */

boolean TESLA_is_reserved_mark(char c)
{
	switch (c)
	{
		case ';':
		case ':':
		case ',':
		case '<':
		case '>':
		case '=':
			return 1;
		default: 
			return 0;
	}
}


/* -------------------------------------------------------------------------- *\
									Is current token a given reserved word
\* -------------------------------------------------------------------------- */

boolean TESLA_is_reserved_word(int id)
{
	if (lex_token.kind == symboltoken)
	{
		int tokval = lex_token.value.symnum;
			if (tokval == id)
			return 1;
	}

	return 0;
}


