/*                                                                            *\

	  TELA - Testing Environment for Learning Algorithms - version 1.995
    ==================================================================

		Module:  			UI_interpreter 

		Contents:     The command interpreter of TELA

		Description:  Interprets and executes TESLA-commands that user writes into 
									the SEED session editor. Is responsible (jointly with SEED) 
									of the consistency between the SEED transcript and the 
									internal state of TELA.

		Author: Juho Rousu (1995,1996)

		Services:

		TESLA_interpreter - interpret and execute TESLA commands

		Internal functions;

		Parsing and execution
		=====================

		parse_cmd	   - Parse a TESLA command
		command      - parse and execute a TESLA command
		confirm      - execute a confirmation dialog

		Backup / Restore data
		==========================================

		gen_ATTS_EXS_backup - gen. a command for backing up the 
													attributes and examples
		gen_RULE_backup     - gen. a command for backing up the rule
		gen_backup_cmds 		- generate backup commands

		gen_INIT_cmd      	- generate an init command		
		gen_BACKUP_restore  - gen. a command for fetching the backed up information
		gen_CLASS_restore   - gen. a command for restoring the class attribute
		gen_SET_restore     - gen. a command for restoring the effects of a
 													SET-command
		gen_DIVIDE_restore  - gen. a command for restoring the effects of a DIVIDE
		gen_WRITE_restore   - gen. a command for restoring  the effects of a WRITE
		gen_DO_restore      - gen. commands for restoring the effects of 
											  	commands in a DO-loop
		gen_restore_cmds  	- gen. commands for restoring the effects of a given 
													command

		Checkpoints and Command reversal
		================================

		update_chk_pnt  - update a checkpoint and generate a new one, if needed
		restore_safe    - reverse until a safe transcript position in found

		Initilization and cleanup
		=========================

		init_keymap            - ensure that there exists a keymap file for SEED
		init_user_interface		 - prepare SEED for interperation of TESLA commands
		remove_backup_files    - remove temporary files
		cleanup_user_interface - perform cleanup duties

\*  
                                                                          */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <unistd.h>
#include <libgen.h>

#include "cseed.h"
#include "seed_lexline.h"

#include "TESLA_cmds_WS.h"
#include "TESLA_cmds_IO.h"
#include "TESLA_cmds_EXEC.h"
#include "Alg_Database.h"

#include "UI_basic.h"
#include "UI_chk_pnt.h"
#include "UI_parsers.h"
#include "UI_interpreter.h"

#define private static

#define DEBUG_
#ifdef DEBUG
	FILE* df;
#endif

char *TELA_call_path;


/* ========================================================================== *\
                             INTERNAL FUNCTIONS     
\* ========================================================================== */

int command(t_chk_pnt **pp_chk_pnt);
int TESLA_cmd_execute(t_command *cmd);
int parse_cmd(t_command **cmd, int in_do, lex_positiontype *p_lex_pos);
void restore_safe(t_chk_pnt **pp_chk_pnt);
t_command *gen_backup_cmds(t_chk_pnt *p_chk_pnt, t_command *p_cmd);	
t_command *gen_restore_cmds(t_command *p_cmd, int exclam_limit);
t_command *ATTS_EXS_backup();
t_command *RULE_backup();
t_command *BACKUP_restore(t_command *p_cmd);
t_command *CLASS_restore(t_command *p_cmd );
t_command *DIVIDE_restore(t_command *p_cmd);
void init_user_interface(int argc, char *argv[]);
t_command *WRITE_restore(t_command *p_cmd, int exclam_limit);
t_command *SET_restore(t_command *p_cmd);
t_command *gen_INIT_cmd();
t_command *DO_restore(t_command *p_cmd);
void cleanup_user_interface();
void remove_backup_files();

/* -------------------------------------------------------------------------- *\
					                 Parse a TESLA command
\* -------------------------------------------------------------------------- */

int parse_cmd(t_command **pp_cmd, int in_do, lex_positiontype *p_lex_pos)
{
	int symbol_id, rv = PARSE_FAILED;
	lex_positiontype here;	

	*pp_cmd = 0;

	lex_store(&here);if (p_lex_pos) lex_store(p_lex_pos);
	lex_next(); 
	if (lex_cancel())
	{ 
		rv = PARSE_CANCELLED;                 /* user edited */
		goto end;
	}

	/* scan past all the comments */
	while (lex_token.kind == eolcommenttoken)
		lex_next();

	if (!TESLA_symbol(&symbol_id))  					/* is it a command? */
	{
		/* parse a command */
		switch (symbol_id)
		{
			case class_sym:	rv = parse_class(pp_cmd);break;
			case div_sym: rv = parse_divide(pp_cmd); break;
			case do_sym: 
				if (!in_do) 
					rv = parse_do(pp_cmd); 
				else { 
					rv = PARSE_FAILED; 
					errmsg("Nested do-commands not allowed");
				}
			break;
			case end_sym: 
				if (in_do)
					rv = parse_end(pp_cmd);
				else {
					rv = PARSE_FAILED;
					errmsg("An END without DO\n");
				}
			break; 
			case excl_sym: rv = parse_exclude(pp_cmd);break;
			case incl_sym: rv = parse_include(pp_cmd);break;
			case map_sym: rv = parse_map(pp_cmd);break;
			case noise_sym: rv = parse_noise(pp_cmd);break;
			case quit_sym: rv = parse_quit(pp_cmd);break;
			case rand_sym: if (!(rv = parse_noise(pp_cmd))) 
				(*pp_cmd)->cmd_id = K_RANDOMIZE;
			break;
			case read_sym: rv = parse_read(pp_cmd);break;
			case stat_sym: rv = parse_stat(pp_cmd);break;
			case write_sym: rv = parse_write(pp_cmd);break;
			case run_sym: rv = parse_run(pp_cmd);break;
			case test_sym: rv = parse_test(pp_cmd);break;
			case cv_sym: rv = parse_cv(pp_cmd);break;  
			case select_sym: rv = parse_select(pp_cmd);break;
			case set_sym: rv = parse_set(pp_cmd);break;
			default: errmsg("Command not implemented");break;
		}	
	}
	else   /* error: this is not a command word */ 
	{
		errmsg("No such command");
		rv = PARSE_FAILED;
	}	

	end:

	if (rv != PARSE_OK) lex_restore(here);

	return rv;
}


/* -------------------------------------------------------------------------- *\
					          Parse and Execute a TESLA command
\* -------------------------------------------------------------------------- */

int command(t_chk_pnt **pp_chk_pnt)
{
	t_command *p_tmp_cmd=0; /* temporary command structures */
	int rv = EXECUTION_FAILED,quit =0;
	lex_positiontype lex_pos;

	do
	{
		if (parse_cmd(&p_tmp_cmd,0,&lex_pos) == PARSE_OK)
		{
			/* update, or possibly create a new checkpoint */
			update_chk_pnt(pp_chk_pnt,p_tmp_cmd,lex_pos);
			
			if ((rv = TESLA_cmd_execute(p_tmp_cmd))!=EXECUTION_FAILED)
				lex_exit();
			else
				errmsg("Command execution failed");			
		}
		
	} while (!lex_done()); /* until executed or user invalidated the checkpoint */  
	return rv;
}


/* -------------------------------------------------------------------------- *\
											Execute a confirmation dialog
\* -------------------------------------------------------------------------- */

int confirm(char *msg)
{
	int rv;
	lex_positiontype here;
	
	printf(msg);

	lex_store(&here);
	do
	{
		lex_restore(here);
		lex_next(); if (lex_cancel()) return 0;
		if (!TESLA_wordtoken_cmp("yes"))
		{		
			rv = 1; 
			lex_null();
			lex_exit();
		}
		else if (!TESLA_wordtoken_cmp("no"))
		{
			rv=0;
			lex_null();
			lex_exit();
		}
		else 
			errmsg("Answer \"yes\" or \"no\".");
	}
	while (!lex_done());

	return rv;
}


/* -------------------------------------------------------------------------- *\
	  Generate a command for backing up the active attribute description and 
															example set
\* -------------------------------------------------------------------------- */

t_command *gen_ATTS_EXS_backup(t_chk_pnt *p_chk_pnt)
{
	t_command *p_cmd=0;
	char buf[16];
	p_cmd = alloc_cmd();
	if (!p_cmd)
	{
		errmsg("Out of memory");
		return p_cmd;
	}
	p_cmd->cmd_id = K_BACKUP_ATTS_EXS;
	sprintf(buf,"%d",p_chk_pnt->lex_pos.last);
	if (!alloc_string(&p_cmd->param.l_backup.ae_name,buf))
	{
		free_cmd(p_cmd,0);
		return 0;
	} 
	p_cmd->param.l_backup.cf_name = 0;
	p_cmd->next = 0;
	p_chk_pnt->ae_backup = 1;
	return p_cmd;	
}


/* -------------------------------------------------------------------------- *\
		        Genarate a command for backing up the active classifier
\* -------------------------------------------------------------------------- */

t_command *gen_RULE_backup(t_chk_pnt *p_chk_pnt)
{	
	t_command *p_cmd=0;
	char buf[256];
	t_algorithms alg;

	WS_get_active_classifier(&alg);

	p_cmd = alloc_cmd();
	p_cmd->cmd_id = K_BACKUP_RULE;

	/* if algorithm is different from TELA, make up a name bor the backup */
	if (alg)
	{
		sprintf(buf,".TELA_%ld_rb_%d",getpid(),p_chk_pnt->lex_pos.last);
		if (!alloc_string(&p_cmd->param.l_backup.cf_name,buf))
		{
			free_cmd(p_cmd,0); return 0;
		}
	}
	p_cmd->param.l_backup.alg = alg;
	p_cmd->param.l_backup.ae_name = 0;
	p_cmd->next = 0;
	p_chk_pnt->cf_backup = 1;
	return p_cmd;	
}


/* -------------------------------------------------------------------------- *\
			  Generate backup commands that are required by given command
\* -------------------------------------------------------------------------- */

t_command *gen_backup_cmds(t_chk_pnt *p_chk_pnt, t_command *p_cmd)
{	
	t_algorithms alg;

	switch (p_cmd->cmd_id)
	{		
		case K_DO:
		case K_DIVIDE: 
		case K_EXCLUDE: 
		case K_EXCLUDE_ALL_EXCEPT:
		case K_SELECT:
		case K_MAP:
		case K_READ_ATTS:
		case K_READ_EXS:
		case K_READ_ATTS_AND_EXS:
		case K_READ_BOTH:
		case K_NOISE:
		case K_RANDOMIZE:
		case K_INCLUDE: 
			/* backup ws only if not already backed up */
			if (!p_chk_pnt->ae_backup && p_chk_pnt->p_prev)
				return gen_ATTS_EXS_backup(p_chk_pnt);
			else return 0;
		break;			
		case K_CV:
		{
			t_command *p_backup=0;
			if (!p_chk_pnt->cf_backup)
				p_backup = gen_RULE_backup(p_chk_pnt);
			return p_backup;
		}
		break;
		case K_RUN: 	
		case K_READ_RULE: 
			if (!p_chk_pnt->cf_backup /* && WS_get_active_classifier(&alg) */)
				return gen_RULE_backup(p_chk_pnt);
		default: return 0;
	}
}


/* -------------------------------------------------------------------------- *\
		                    Generate an init command
\* -------------------------------------------------------------------------- */

t_command *gen_INIT_cmd()
{
	t_command *p_tmp=0;

	if (!(p_tmp = alloc_cmd()))
	{
		errmsg("Out of memory");
		return p_tmp;
	}
	p_tmp->cmd_id = K_INIT;
	p_tmp->next = p_tmp;
	return p_tmp;
}


/* -------------------------------------------------------------------------- *\
		    Generate a command for fetching the backed up information
\* -------------------------------------------------------------------------- */

t_command *gen_BACKUP_restore(t_command *p_cmd)
{
	t_command *p_new=0;

	if (!(p_new = alloc_cmd()))
	{
		errmsg("Out of memory"); return 0;	
	};
	switch (p_cmd->cmd_id)
	{
		case K_BACKUP_ATTS_EXS : 	
			p_new->cmd_id = K_RESTORE_ATTS_EXS;
			if (alloc_string(&p_new->param.l_restore.ae_name,
				p_cmd->param.l_backup.ae_name))
				p_new->next = p_new;
			else goto cancel;
		break;
		case K_BACKUP_RULE :
		{
			t_delete_1 *p_delete; 
			/* restore rule */
			p_new->cmd_id = K_RESTORE_RULE; 
			p_new->param.l_restore.alg = p_cmd->param.l_backup.alg;
			if (p_cmd->param.l_backup.alg)
			{
				if (!alloc_string(&p_new->param.l_restore.cf_name,
					p_cmd->param.l_backup.cf_name)) goto cancel;
				/* and delete backup file */
				if (p_new->next = alloc_cmd())
				{
					p_new->next->cmd_id = K_DELETE;
					p_new->next->next = p_new;
					p_delete = &p_new->next->param.l_delete_1;
					p_delete->top_exclam_index =0;
					if (!(p_delete->first =
					 (t_delete_node*)malloc(sizeof(t_delete_node))))
					{
						errmsg("Out of memory");
						goto cancel;
					}	
					else
					{
						if (!alloc_string(&p_delete->first->f_name,
							p_new->param.l_restore.cf_name))
						{
							errmsg("Out of memory");
							goto cancel;
						}
						p_delete->first->next = 0;
					}
				}
				else 
				{	
					errmsg("Out of memory");	
					goto cancel;
				}
				p_new = p_new->next;
			}
			else p_new->next = p_new;
		}
		break;
		default: 
			goto cancel;
	}
	return p_new;

	cancel:
		free_cmd(p_new,0); return 0;
}


/* -------------------------------------------------------------------------- *\
		    Generate a command for restoring the class attribute
\* -------------------------------------------------------------------------- */

t_command *gen_CLASS_restore(t_command *p_cmd )
{
	char buf[256];
	t_command *p_tmp=0;		
	t_class_2 param = p_cmd->param.l_class_1;
		
	p_tmp = alloc_cmd();
	p_tmp->param.l_class_1.attr_spec = A_NORMAL;
	p_tmp->cmd_id = K_CLASS;
	p_tmp->next = p_tmp;
	
	if (WS_get_class_attr_name(buf)) 
	{
	  /* no class attribute, restoring requires to 
			get rid of the class attribute */
		p_tmp->param.l_class_1.attr_spec = A_NONE;
		return p_tmp;
	}

	if (param.attr_spec != A_NONE)
	{
		if (!strcmp(buf,param.attr_name))
			return 0; /* the commands does nothing, no restoring */
	}	

	if (!alloc_string(&p_tmp->param.l_class_1.attr_name, buf))
	{		
		errmsg("Out of memory");
		return 0;
	}

	return p_tmp;
}


/* -------------------------------------------------------------------------- *\
		     Generate a command for restoring the effects of a SET-command
\* -------------------------------------------------------------------------- */

t_command *gen_SET_restore(t_command *p_cmd )
{
	/* in general, it is not possible to restore the status of a random 
		number generator, no other SET variants are currently implemented */

	return 0;
}


/* -------------------------------------------------------------------------- *\
				  Generate a command for restoring a DIVIDE-command
\* -------------------------------------------------------------------------- */

t_command *gen_DIVIDE_restore(t_command *p_cmd)
{
	t_command *p_restore=0;
	t_delete_node *p_delete=0;
	t_div_node *p_div = p_cmd->param.l_divide_1.first;
	
	p_restore = alloc_cmd();
	p_restore->cmd_id = K_DELETE;
	p_restore->param.l_delete_1.top_exclam_index=0;
	p_restore->next = p_restore; /* make it a ring */

	/* get first file to delete */
	if ((p_delete = (t_delete_node *)malloc(sizeof(t_delete_node))) ?
		!alloc_string(&p_delete->f_name,p_div->f_name) : 1)
	{
		errmsg("Out of memory"); return 0;	
	}
	p_div = p_div->next;
	p_restore->param.l_delete_1.first = p_delete;

	/* get rest of the files to delete */
	while (p_div != 0)
	{
		if ((p_delete->next = (t_delete_node *)malloc(sizeof(t_delete_node))) ? 
			!alloc_string(&p_delete->next->f_name,p_div->f_name) : 1)
		{
			errmsg("Out of memory"); return 0;	
		}

		p_div = p_div->next;
		p_delete = p_delete->next;
	}
	p_delete->next = 0;

	return p_restore;

}


/* -------------------------------------------------------------------------- *\
			Generate a command for restoring  the effects of a WRITE-command
\* -------------------------------------------------------------------------- */

t_command *gen_WRITE_restore(t_command *p_cmd,int exclam_limit)
{	
	t_command *p_tmp=0;

	t_write_1 *p_wr_params = &p_cmd->param.l_write_1; 			
	t_delete_node *p_delete = 0,**pp_first;

	p_tmp = alloc_cmd();

	pp_first = &p_tmp->param.l_delete_1.first;
		
	p_tmp->cmd_id = K_DELETE;
	p_tmp->param.l_delete_1.top_exclam_index= exclam_limit;
	p_tmp->next = p_tmp;

	if (p_wr_params->a_filename)
	{
		if ((*pp_first = (t_delete_node*)malloc(sizeof(t_delete_node))) ?
			!alloc_string(&(*pp_first)->f_name,p_wr_params->a_filename) : 1)	
		{
			errmsg("Out of memory"); return 0;	
		}
		pp_first = &(*pp_first)->next;
	}
	if (p_wr_params->e_filename)
	{
		if ((*pp_first = (t_delete_node*)malloc(sizeof(t_delete_node))) ?
			!alloc_string(&(*pp_first)->f_name, p_wr_params->e_filename) : 1)	
		{
			errmsg("Out of memory"); return 0;	
		}

		pp_first = &(*pp_first)->next;
	}
	if (p_wr_params->r_filename)
	{
		if ((*pp_first = (t_delete_node*)malloc(sizeof(t_delete_node))) ?
			!alloc_string(&(*pp_first)->f_name, p_wr_params->r_filename) : 1)
		{
			errmsg("Out of memory"); return 0;	
		}

		pp_first = &(*pp_first)->next;
	}
	*pp_first = 0;

	return p_tmp;
}


/* -------------------------------------------------------------------------- *\
			Generate commands for restoring the effects caused by commands in the 
																DO-loop
\* -------------------------------------------------------------------------- */

t_command *gen_DO_restore(t_command *p_cmd)
{
	t_command *p_rev_cmds=0,*p_cmds = p_cmd->param.l_do_1.commands;
	int exclam_limit = p_cmd->param.l_do_1.do_cnt;
	while (p_cmds->cmd_id != K_END)
	{
		insert_to_cmd_list(&p_rev_cmds,gen_restore_cmds(p_cmds,exclam_limit));
		p_cmds = p_cmds->next;
	}
	return p_rev_cmds;
}


/* -------------------------------------------------------------------------- *\
		     Generate commands for restoring the effects of a given command
\* -------------------------------------------------------------------------- */

t_command *gen_restore_cmds(t_command *p_cmd, int exclam_limit)
{
	switch (p_cmd->cmd_id)
	{
		case K_BACKUP_ATTS_EXS:
		case K_BACKUP_RULE: if (!exclam_limit) return gen_BACKUP_restore(p_cmd);break;
		case K_CLASS: return gen_CLASS_restore(p_cmd); break;
		case K_DIVIDE: return gen_DIVIDE_restore(p_cmd); break;
		case K_WRITE_ATTS_AND_EXS: 
		case K_WRITE_BOTH: 
		case K_WRITE_ATTS:
 		case K_WRITE_EXS: 
		case K_WRITE_RULE: return gen_WRITE_restore(p_cmd,exclam_limit); break;
		case K_DO: return gen_DO_restore(p_cmd); break;
		case K_SET_RANDSEED:
		case K_SET_RANDCLOCK:
		case K_SET_ALG: return gen_SET_restore(p_cmd); break;
		default: return 0;
	}
}


/* -------------------------------------------------------------------------- *\
		Update a checkpoint as required by a command that is to be executed by
 		generating and executing backup commands and storing instructions for
																	 restoration
\* -------------------------------------------------------------------------- */

update_chk_pnt(t_chk_pnt **pp_chk_pnt, t_command *p_cmd,lex_positiontype lex_pos)
{
	t_command *p_backup=0;
	/* the only case when we don't have to generate a new checkpoint */
	/* is when we have a new C_WSMANIP-command to add to an existing */
	/* C_WSMANIP-block  */
	if ((cmd_type(p_cmd) == C_PROCESSING) 
		|| ((*pp_chk_pnt == 0)? 1 : (**pp_chk_pnt).cp_type == C_PROCESSING))
	{
		new_chk_pnt(pp_chk_pnt,cmd_type(p_cmd),lex_pos);
	}

	if ((**pp_chk_pnt).p_prev == 0)
	{
		if ((**pp_chk_pnt).p_rev_cmds == 0)	
			insert_to_cmd_list(&(**pp_chk_pnt).p_rev_cmds,gen_INIT_cmd());
		return;
	}

	/* generate commands that are needed to back up any alterations */
	/* the current command  might do to the system, if backup commands */
	/* are not needed, then restoration commands aren't either */ 
	if (p_backup = gen_backup_cmds(*pp_chk_pnt,p_cmd))
	{
		t_command *p_tmp=0;
		/* execute them right away, error makes this checkpoint unsafe: */
		while (p_backup)
		{
			if (TESLA_cmd_execute(p_backup))
			{ 
				(**pp_chk_pnt).unsafe = 1;
				break; /* it makes no sense to do any further backing up */
			}
			else /* generate the corresponding restore command */
				insert_to_cmd_list(&(**pp_chk_pnt).p_rev_cmds,
					gen_restore_cmds(p_backup,0));

			p_tmp = p_backup;
			p_backup = p_backup->next;
			free_cmd(p_tmp,1);
		}
	}	
	/* and finally, commands that restore the actual commands effects */
	insert_to_cmd_list(&(**pp_chk_pnt).p_rev_cmds,gen_restore_cmds(p_cmd,0));
}


/* -------------------------------------------------------------------------- *\
			        Reverse until a safe transcript position in found
\* -------------------------------------------------------------------------- */

void restore_safe(t_chk_pnt **pp_chk_pnt)
{
	t_command *p_cmd=0;
	t_chk_pnt *p_tmp_chk=0;

	while (*pp_chk_pnt ? (lex_cancel() || (**pp_chk_pnt).unsafe) : 0)
	{
		/* reverse to the checkpoint in editor*/
		lex_restore((**pp_chk_pnt).lex_pos);
#ifdef DEBUG
		df = fopen("debug_log","a");
		fprintf(df,"rs : %d\n",(**pp_chk_pnt).lex_pos.last);
		fclose(df);
#endif

		/* execute all reverse commands */
		if (p_cmd = (**pp_chk_pnt).p_rev_cmds)
		{
			do
			{			
				p_cmd = p_cmd->next;
				if (TESLA_cmd_execute(p_cmd))
				{ 
					(**pp_chk_pnt).unsafe = 1;
#ifdef DEBUG	
		df = fopen("debug_log","a");
		fprintf(df,"%s, unsafe: %d\n", cmd_name(p_cmd),
			(**pp_chk_pnt).lex_pos.last);		
		fclose(df);
#endif
				}
			} 
			while (p_cmd != (**pp_chk_pnt).p_rev_cmds);
		}
		/* find next checkpoint */
		p_tmp_chk = *pp_chk_pnt;
		*pp_chk_pnt = (*pp_chk_pnt)->p_prev;
		free_chk_pnt(&p_tmp_chk);						 /* free this checkpoint     */ 
	}	
}


/* -------------------------------------------------------------------------- *\
							Ensure that there exists a keymap file for SEED
\* -------------------------------------------------------------------------- */

void init_keymap()
{
	FILE *fp_src,*fp_dst;
	int ch;
	char *home_dir = getenv("HOME");

	if (fp_src = fopen(".seed-keymap","r"))
	{
		fclose(fp_src);
	}
	else
	{
		char map_file[512];

		if (!home_dir)
		{
			printf("Could not find the keymap file .seed-keymap\n");
			_exit(1);
		}

		sprintf(map_file,"%s/.seed-keymap",home_dir);

		if (fp_src = fopen(map_file,"r"))
		{
			fclose(fp_src);
		}
		else
		{
			/* copy the keymap file from TELA's directory */
			if (fp_dst = fopen(map_file,"w"))
			{
				char keymap_src[512];
				sprintf(keymap_src,"%s/.seed-keymap",TELA_call_path);

				if (fp_src = fopen(keymap_src,"r"))
				{
					while ((ch = fgetc(fp_src)) != EOF) fputc(ch,fp_dst);
					fclose(fp_src);
				}	
				else
				{
					printf("Could not open the keymap file %s for reading.\n",keymap_src);
					_exit(1);
				}
				fclose(fp_dst);
			}
			else
			{
				printf("Could not write the keymap file to directory %s\n",home_dir);
				_exit(1);
			}
		}
	}
}


/* -------------------------------------------------------------------------- *\
						Prepare SEED for interperation of TESLA commands
\* -------------------------------------------------------------------------- */

void init_user_interface(int argc, char *argv[])
{
	init_keymap();

  seed_init (argc, argv); lex_init ();
  if (seed_stdout () == -1) exit (1);

	/* define command words for SEED*/
  lex_define (*lex_word_c ("CLASS"    ), class_sym);
  lex_define (*lex_word_c ("DIVIDE"   ), div_sym);
  lex_define (*lex_word_c ("DO"       ), do_sym);
  lex_define (*lex_word_c ("DUMP"     ), dump_sym);
  lex_define (*lex_word_c ("END"      ), end_sym);
  lex_define (*lex_word_c ("EXCLUDE"  ), excl_sym);
  lex_define (*lex_word_c ("INCLUDE"  ), incl_sym);
  lex_define (*lex_word_c ("EXCLUDE"  ), excl_sym);
  lex_define (*lex_word_c ("MAP"      ), map_sym);
  lex_define (*lex_word_c ("NOISE"    ), noise_sym);
  lex_define (*lex_word_c ("QUIT"     ), quit_sym);
  lex_define (*lex_word_c ("RANDOMIZE"), rand_sym);
  lex_define (*lex_word_c ("READ"     ), read_sym);
  lex_define (*lex_word_c ("STAT"     ), stat_sym);
  lex_define (*lex_word_c ("WRITE"    ), write_sym);
	lex_define (*lex_word_c ("RUN"			), run_sym);
	lex_define (*lex_word_c ("TEST"			), test_sym);
	lex_define (*lex_word_c ("CV"				), cv_sym);
	lex_define (*lex_word_c ("SELECT" 	), select_sym);
  lex_define (*lex_word_c ("SET"      ), set_sym);


	lex_define (*lex_word_c ("ATTS"				), atts_sym);
  lex_define (*lex_word_c ("EXS"				), exs_sym);
  lex_define (*lex_word_c ("BOTH"				), both_sym);
  lex_define (*lex_word_c ("RULE"				), rule_sym);
	lex_define (*lex_word_c ("RANDSEED"   ), randseed_sym);
	lex_define (*lex_word_c ("ALGORITHM"  ), algorithm_sym);
	lex_define (*lex_word_c ("CLOCK"      ), clock_sym);
  
	lex_define (*lex_word_c ("OTHERS"   	), others_sym);
	lex_define (*lex_word_c ("EXCEPT"	), except_sym);
	lex_define (*lex_word_c ("AS"  				), as_sym);
	lex_define (*lex_word_c ("ELSE"    		), else_sym); 
  lex_define (*lex_word_c ("ALL"   			), all_sym);
	lex_define (*lex_word_c ("NOT"   			), not_sym);
	lex_define (*lex_word_c ("IN"					), in_sym);
	lex_define (*lex_word_c ("NONE"       ), none_sym);  

  lex_define (*lex_word_c ("INTO"       ), into_sym);
  lex_define (*lex_word_c ("COMMENT"		), comment_sym);
  lex_define (*lex_word_c ("RECORD"     ), record_sym);
  lex_define (*lex_word_c ("TIME"  			), time_sym);
  lex_define (*lex_word_c ("TESTTIME"   ), testtime_sym);

  lex_ignore (eoltoken);

	lex_firstchar(eolcommenttoken,'=',*vstring_c("#"));
  lex_bodychar (eolcommenttoken,'=',*vstring_c(" "));
	

	atexit(&remove_backup_files);
}


/* -------------------------------------------------------------------------- *\
		Remove any temporary files that arer augmented by PID of the TELA process
\* -------------------------------------------------------------------------- */

void remove_backup_files()
{	
	DIR *def_dir;
	struct dirent *p_entry;			
	char spid[32];

	/* search for process ID to recognize a temporary file */
	sprintf(spid,"_%ld",getpid());

	if (def_dir = opendir("."))
	{
		while (p_entry = readdir(def_dir))
		{
			if (strstr(p_entry->d_name,spid))
				remove(p_entry->d_name);
		}
		closedir(def_dir);
	}
}


/* -------------------------------------------------------------------------- *\
								   Perform cleanup duties before exit
\* -------------------------------------------------------------------------- */

void cleanup_user_interface()
{
	remove_backup_files();
	seed_fini(0);
}


/* ========================================================================== *\
                             EXTERNAL SERVICES     
\* ========================================================================== */


/* -------------------------------------------------------------------------- *\
			            Interpret and execute TESLA commands
\* -------------------------------------------------------------------------- */

void TESLA_interpreter(int argc, char *argv[])
{

	t_chk_pnt *p_chk_pnt = 0;
	int rv = EXECUTION_OK;	
	char alg_db_file[512];	
	
	/* first, figure out, where we are called from */
	TELA_call_path = dirname(argv[0]);
	sprintf(alg_db_file,"%s/.TELA_dfd_db",TELA_call_path);

	/* start SEED */
	init_user_interface(argc,argv);	

	/* read in algorithm descriptions */
	if (!ADB_read_file(alg_db_file))	
	{
		printf("Algorithm database initialization did not succeed.\n");
		printf("Press ^X^C to exit TELA.\n");
		lex_next();
		return;
	}


	printf ("\nWelcome to TELA version %s !\n\n",TELA_VERSION);
	ADB_list_algorithms();

  randomize ();
  IO_initialize(); 

	/* main command interpreting loop: executes commands until user quits */
	while (rv != USER_QUIT)
	{		
		if (lex_cancel()) 
			restore_safe(&p_chk_pnt);
		rv = command(&p_chk_pnt); 	
	}	
	cleanup_user_interface();
}
