/*                                                                            *\

	  TELA - Testing Environment for Learning Algorithms - version 1.995
    ==================================================================

		Module:       tela

		Contents:     Common routines for all TELA modules 

		Services:

		alloc_string      - allocate space and make copy of a string
		realloc_string    - ditto, but re-allocate space
		eq_q_str          - compare two strings, ignore quotes
	
		list_init         - initialize an empty list
		list_append       - add new item at the end of the list
		list_insert       - insert a new item before the first tiem in the list
		list_remove_first - remove the first item in the list
		list_get_first    - get the first item in the list
		list_get_next     - get the next item in the list
		list_get_last     - get the last item in the list
		list_catenate     - concatenate two lists

		Author: Juho Rousu (1995,1996)
\*  
                                                                          */
#include <string.h>
#include "tela.h"


/* -------------------------------------------------------------------------- *\
						      Allocate space and copy a string
\* -------------------------------------------------------------------------- */

boolean realloc_string(char **dest,char *source) 
{
  *dest = (char *) realloc (*dest,strlen (source) + 1); 
  if (!*dest) return 0;
  strcpy (*dest, source);
	return 1;
}


/* -------------------------------------------------------------------------- *\
								  Reallocate space and copy a string
\* -------------------------------------------------------------------------- */

boolean alloc_string(char **dest,char *source) 
{
  *dest = (char *) malloc (strlen (source) + 1); 
  if (!*dest) return 0;
  strcpy (*dest, source);
	return 1;
}


/*--------------------------------------------------------------------------*\
								Compare two strings, ignore quotes

Note: for any two strings s1 and s2
	eq_q_str('"'+s1+'"',s2) == strcmp(s1,s2) and
	eq_q_str(s1,'"'+s2+'"') == strcmp(s1,s2)
\*--------------------------------------------------------------------------*/

int eq_q_str(const char* str1,const char* str2)
{
	char *p1,*p2;
	int len1,len2;
	int s1_quoted = 0,s2_quoted =0;
	
	p1 = (char *)str1;
	p2 = (char *)str2;


	/* check for starting quotes */

	if (*p1 == '"') 
	{
		s1_quoted = 1;
		p1++;
	}
	if (*p2 == '"') 
	{
		s2_quoted = 1;
		p2++;
	}

	/* check char by char; without length of a string or a sentinel */
	/* we have to check for ending NULL-char */
	while (*p1==*p2)
	{
		if (*p1 == '\0')  								/* end of both strings ? 	 */
			return s1_quoted != s2_quoted;  /* are quotes in balance ? */
		p1++;p2++;	
	}

	/* now we have longest (possibly empty) common prefix of both strings behind 
	  and have *p1 != *p2, let's check for missing ending quote */ 

	if (s1_quoted == s2_quoted) 
  	return strcmp(p1,p2);     /* postfixes define the actual ordering */  

	/* only two special cases are left, where the strings migth be equal,*/
	/* namely, when the other is unquoted and the other quoted */

	if (s1_quoted && *p1 == '"' && *(p1+1) == '\0' && *p2 == '\0') 
	{
		return 0;
	}

	if (s2_quoted && *p2 == '"' && *(p2+1) == '\0' && *p1 == '\0')
	{
		return 0;
	}

	/* let strcmp find out the actual ordering */
		
	return strcmp(p1,p2);	
}


/* -------------------------------------------------------------------------- *\
													Initialize an empty list
\* -------------------------------------------------------------------------- */

void list_init(t_list *p_list)
{
	p_list->p_head = 0;
	p_list->p_tail = 0;
}


/* -------------------------------------------------------------------------- *\
									Add a new item at the end of the list
\* -------------------------------------------------------------------------- */

int list_append(t_list *p_list,void* item_to_add)
{
	t_list_item *p_new;

	if (!(p_new = (t_list_item*)malloc(sizeof(t_list_item)) ) )
		return 0;
	p_new->item = item_to_add;
	p_new->p_next = 0;

	if (p_list->p_tail)
	{
		p_list->p_tail->p_next = p_new;
		p_list->p_tail = p_new; 
	}
	else p_list->p_tail = p_list->p_head = p_new;	

	return 1;
}


/* -------------------------------------------------------------------------- *\
					Insert a new item before the first item in the list
\* -------------------------------------------------------------------------- */

int list_insert(t_list *p_list,void* item_to_insert)
{
	t_list_item *p_new;

	if (!(p_new = (t_list_item*)malloc(sizeof(t_list_item)) ) )
		return 0;
	p_new->item = item_to_insert;
	p_new->p_next = 0;

	if (p_list->p_head)
	{
		p_new->p_next = p_list->p_head;
		p_list->p_head = p_new;
	}
	else p_list->p_head = p_list->p_tail = p_new;	

	return 1;
}


/* -------------------------------------------------------------------------- *\
									Remove the first item in the list
\* -------------------------------------------------------------------------- */

int list_remove_first(t_list *p_list,void **pp_item)
{
	if (p_list->p_head)
	{
		*pp_item = p_list->p_head->item;
		free(p_list->p_head);
		p_list->p_head = p_list->p_head->p_next;
		if (p_list->p_head) p_list->p_tail = 0;
		return 1;	
	}
	else return 0;
}


/* -------------------------------------------------------------------------- *\
								Get the first item in the list
\* -------------------------------------------------------------------------- */

int list_get_first(t_list *p_list,void **pp_item)
{
	if (p_list->p_head)
	{
		*pp_item = p_list->p_head->item;
		p_list->p_curr = p_list->p_head->p_next;
		return 1;
	}
	else return 0;
}


/* -------------------------------------------------------------------------- *\
										Get the next item in the list
\* -------------------------------------------------------------------------- */

int list_get_next(t_list *p_list,void **pp_item)
{
	if (p_list->p_curr)
	{
		*pp_item = p_list->p_curr->item;
		p_list->p_curr = p_list->p_curr->p_next;
		return 1;
	}
	else return 0;
}


/* -------------------------------------------------------------------------- *\
											Get the last item in the list
\* -------------------------------------------------------------------------- */

int list_get_last(t_list *p_list,void **pp_item)
{
	if (p_list->p_tail)
	{
		*pp_item = p_list->p_tail->item;
		return 1;
	}
	else return 0;
}


/* -------------------------------------------------------------------------- *\
												Concatenate two lists
\* -------------------------------------------------------------------------- */

void list_catenate(t_list *p_list1,t_list *p_list2)
{
	t_list_item *p_tmp;

	if (!p_list2->p_tail) return;

	p_tmp = p_list1->p_tail;
	p_list1->p_tail  = p_list2->p_tail;
	if (p_tmp)
		p_tmp->p_next = p_list2->p_head;
	else p_list1->p_head = p_list2->p_head;
}



