#ifndef H_TELA
#define H_TELA

/*                                                                            *\

	  TELA - Testing Environment for Learning Algorithms - version 1.995
    ==================================================================

		Module:       tela.h

		Contents:     Common type definitions and routines for all TELA modules.									 
		Services:

		alloc_string    - allocate space and make copy of a string
		realloc_string  - ditto, but re-allocate space
		eq_q_str        - compare two strings, ignore quotes

		list_init         - initialize an empty list
		list_append       - add new item at the end of the list
		list_insert       - insert a new item before the first tiem in the list
		list_remove_first - remove the first item in the list
		list_get_first    - get the first item in the list
		list_get_next     - get the next item in the list
		list_get_last     - get the last item in the list
		list_catenate     - concatenate two lists

		Author: Tommi Timonen (1992), Ilkka Hyvarinen (1993,1994),
						Juho Rousu (1995,1996)

\*  																																					*/
   
#include <stdio.h>

       
#define private static

#define TELA_VERSION "1.995"
#define TELA_FILENAME_PREFIX "SAMPLE"
#define MAXLENGTH 200  /* maximum length of file names (IO_server)*/
                                                               
typedef int boolean;
typedef int t_algorithms;
typedef int t_attr_ref;


/* Error codes */
#define NO_ERROR	0
#define FATAL_ERROR     1
#define NON_FATAL_ERROR	2
#define PARSE_OK 0
#define EXECUTION_OK 0
#define PARSE_FAILED 1
#define PARSE_CANCELLED 2
#define EXECUTION_FAILED 3
#define USER_QUIT 4


/* supported filetypes for TELA */
typedef enum 
{
	FT_ATTS,			/* attribute file */
	FT_EXS,				/* examples file */
	FT_BOTH,			/* example and attribute file */
	FT_RULE,			/* classifier file */
	FT_ATTSRULE,	/* attribute and rule file */
	FT_STATS,			/* statistics file */
} t_filetype;


/* a generic list type */
typedef struct s_list_item
{
	void *item;
	struct s_list_item *p_next;
} t_list_item;

typedef struct s_list
{
	t_list_item *p_head;
	t_list_item *p_tail;
	t_list_item *p_curr;
} t_list;


/* -------------------------------------------------------------------------- *\
		  					Working set data structure definitions	

		Note: it is somewhat unfortunate that these are declared here, modularity
			suffers, but this is unavoidable for historical reasons.
\* -------------------------------------------------------------------------- */


/* Is attribute value of an example normal (OK), unknown or not cared about */
typedef enum 
{
  T_OK, 
	T_DONT_KNOW, 
	T_DONT_CARE
} t_ex_ok;

/* Relations needed for MAP and some other parsers */
typedef enum {LT, LE, EQ, GE, GT, IN, NE} t_relation;

/* Single attribute value of an example */
typedef struct 
{
  t_ex_ok ok;  
  union	
	{
    int   int_val;
    float float_val;
  } val;           
} t_ex_node;

/* Supported attribute types */
typedef enum {
	T_USER,
  T_INT, 
	T_FLOAT,
	T_BINARY,
	T_ANY
} t_attr_types;

/* struct holding one value of a discrete symbolic attribute */
typedef struct discrete_val 
{
  char *val;      /* The value itself */
  int  next_val;  /* The values are kept in a list in input order */
} t_discrete_val;

/* Attribute node */
typedef struct attr_node 
{
  char     	 *name;
  t_attr_types	 a_type;

  int  		 val_cnt;  	  						
  t_discrete_val *disc_val_table;		/* nominal attribute domain */

  int	first_read;   						/* link to first attribute in input order */	 
  int next_attr;	  						/* link to next attribute in input order */
  t_ex_node *examples; 	  			/* example values */
} t_attr_node;

/* example activation mask */
typedef struct {
	int *indexes;       /* partition of examples */
	int count;          /* length of vector, should always be count of exs also */
	boolean all_except; /* inactivate or activate following subset */
	int active_subset;  /* subset to activate */
} t_ex_mask;


/* -------------------------------------------------------------------------- *\
													TESLA data structures
\* -------------------------------------------------------------------------- */

#define t_read_2 t_read_1
#define l_read_2 l_read_1

#define t_write_2 t_write_1
#define l_write_2 l_write_1

#define t_divide_2 t_divide_1
#define l_divide_2 l_divide_1


#define t_noise_2 t_noise_1
#define l_noise_2 l_noise_1

#define t_exclude_2 t_exclude_1
#define l_exclude_2 l_exclude_1

#define t_class_2 t_class_1
#define l_class_2 l_class_1


#define t_stat_2 t_stat_1
#define l_stat_2 l_stat_1

#define t_stat_all_2 t_stat_1
#define l_stat_all_2 l_stat_all_1

#define t_run_2 t_run_1
#define l_run_2 l_run_1

#define t_test_2 t_test_1
#define l_test_2 l_test_1

#define t_do_2 t_do_1
#define l_do_2 l_do_1

typedef enum {
	K_NULL,
	K_QUIT,
	K_INIT,
	K_BACKUP_ATTS_EXS,K_BACKUP_RULE,
	K_RESTORE_ATTS_EXS,K_RESTORE_RULE,
  K_BEGIN,
  K_READ_ATTS, K_READ_EXS, K_READ_ATTS_AND_EXS, K_READ_BOTH, K_READ_RULE,
  K_WRITE_ATTS, K_WRITE_EXS, K_WRITE_ATTS_AND_EXS, K_WRITE_BOTH, K_WRITE_RULE,
  K_DIVIDE,
  K_NOISE,
  K_RANDOMIZE,
  K_EXCLUDE, K_EXCLUDE_ALL_EXCEPT,
  K_INCLUDE,
  K_CLASS,
  K_MAP,
  K_STAT, K_STAT_ALL,
  K_DO, K_END,
  K_RUN,
  K_TEST,
	K_CV,
	K_SELECT,
	K_SET_ALG,K_SET_RANDSEED,K_SET_RANDCLOCK,
	K_DELETE
} t_cmd_id;

typedef enum {
  l_float, l_int, l_discrete, l_dont_care, l_dont_know
} t_val_type;

typedef union {
  float f;
  int   i;
  char  *discrete;
} t_val;



/* READ */

typedef struct {
  t_algorithms alg;
  char         *a_filename, *e_filename, *r_filename;
} t_read_1;

/* WRITE */

typedef struct {
  t_algorithms alg;
  char         *a_filename;
  char	       *e_filename, *r_filename;
  char         *comment;
} t_write_1;

/* DIVIDE */

typedef struct s_div_node {
  char              *f_name;
  float             part;
  struct s_div_node *next;
  FILE              *handle;	/* for run-time use */
} t_div_node;

typedef struct {
	int count;
  t_div_node *first;     /* begin list of division nodes */
} t_divide_1;

/* DELETE */

typedef struct s_delete_node 
{
	char  *f_name;
	struct s_delete_node *next;
} t_delete_node;

typedef struct 
{	
	unsigned top_exclam_index; /* highest value of exclamation point */  
	t_delete_node *first;
}	t_delete_1;


/* NOISE, RANDOMIZE */

struct t_noise_node {
  char                *attr_name;
  float               noise_p;    /* noise probability (or percentage) */
  struct t_noise_node *next;
};

typedef struct {
  struct t_noise_node *first;
  float                others;
} t_noise_1;

/* EXCLUDE */

struct t_exclude_node {
  char                  *attr_name;
  struct t_exclude_node *next;
};

typedef struct {
	struct t_exclude_node *first;
} t_exclude_1;

/* INCLUDE */

struct t_incl_val_node {
  char                   *attr_val;
  struct t_incl_val_node *next;
};

typedef struct incl_node {
  char                   *attr_name;
	int 	val_cnt;
  struct t_incl_val_node val_list_head;
  struct incl_node	 *next;
} t_include_1;

/* CLASS */

typedef enum {A_NORMAL,A_CLASS,A_NONE} t_attr_spec;

typedef struct {
	t_attr_spec attr_spec;
  char *attr_name;
} t_class_1;

/* MAP */

typedef struct map_node {
  t_relation      relation;
  t_val_type      old_type;
  t_val           old_val;
  char            *new_val;
  struct map_node *next;
} t_map_node;

typedef struct map_node_oth {
  char   *attr_val;
  struct map_node_oth *next;
} t_map_node_oth;

typedef struct {
  char           *attr_name;
  t_map_node     *first;
  char           *else_val;
  t_map_node_oth *others;
} t_map_1;

typedef struct {
  char        *attr_name;
  t_map_node  *first;
  char        *else_val;
  char        *dont_know_val;
  char        *dont_care_val;
  t_attr_node *attribute_table;
} t_map_2;

/* STAT */

typedef struct s_snode {
  char *attr_name;
  struct s_snode *next;
} t_snode;

typedef struct {
  t_snode *first;
  char    *stat_file_name;
  char    *comment;
} t_stat_1;

typedef struct {    /* to be removed? */
  char *stat_file_name;
  char *comment;
} t_stat_all_1;

/* statistics collection node and list */
typedef struct      
{
	char *name;
	float value;
	float sum_val;
	float sqrsum_val;
	int terms;
	int printing;
} t_stats_node;

typedef t_list t_stats_list;


/* DO */

typedef struct {
  int       	   do_cnt;
  char      	   *stat_file_name;

	t_stats_list stats;	

  char      	   *comment;
  struct s_command *commands;
} t_do_1;

/* RUN */

typedef struct s_run_node {
  char              *selector;
  char 	            *val; 		/* NULL if no value */
  struct s_run_node *next;
} t_run_node;

typedef struct {
  t_algorithms alg;     /* TELA-type not allowed */

	t_ex_mask *a_mask;    /* activation mask for the examples */

  t_run_node   *params;
  boolean      b_time;

	t_stats_list	 stats; /* list of names of the statistics to be collected */
	int 				pr_mode;
  char         *stat_file_name;
  char         *comment;

} t_run_1;

/* TEST */

typedef struct {	

	t_ex_mask *a_mask;    /* activation mask for the examples */

  boolean      b_testtime;

	t_stats_list stats; /* statistics list (see RUN) */
  char         *stat_file_name;
	int pr_mode;

  char         *comment;

  t_algorithms alg; 	         /* TELA not allowed */
  /* Information about the algorithm is determined at run time
     (specified by the last RUN or READ RULE command) */

} t_test_1;

/* BACKUP */

typedef struct
{
	t_algorithms alg;
	char *ae_name; /* name of the attribute and example backup */
	char *cf_name; /* name of the classifier backup 					 */	
}	t_backup;


/* RESTORE */

typedef struct
{
	t_algorithms alg;
	char *ae_name;	/* name of the attribute and example backup */
	char *cf_name;	/* name of the classifier backup 					 */	
} t_restore;


/* SELECT */

typedef struct s_val_node
{
	t_val val;
	struct s_val_node *p_next;
} t_val_node;

typedef struct s_select_node
{
	char *attr_name;
	t_relation rel;
	int start_encl; /* -1 = -oo, 0 = ], 1 = [ */
	int end_encl;   /* -1 = oo, 0 = [, 1 = ] */
	int neg;
	t_val_node *val_list;
	struct s_select_node *p_next;
} t_select_node;


typedef struct
{
	t_select_node *p_first;	
} t_select; 

/* CV */

typedef struct 
{
	t_algorithms alg;

  t_run_node   *params;
	
	t_stats_list stats;

	char *comment;
	int pr_mode; 

	unsigned count;

	char *stat_file_name;

} t_cv;

/* SET */

typedef struct
{
  int seed;
	t_algorithms alg;
} t_set;


/* Command tree */

typedef struct s_command {
  int              row;
  t_cmd_id         cmd_id;
  struct s_command *next, *prev;
  union {
    /* Members ending _1 are used during parse and passed to optimizer.
       Members ending _2 are valid after optimization.
       Command server gets its params in _2 ending members.
    */
    t_read_1     l_read_1;     /* For all read variants */
    t_write_1    l_write_1;    /* For all write variants */
    t_divide_1   l_divide_1;
    t_noise_1    l_noise_1;    /* Used also for randomize */
    t_exclude_1  l_exclude_1;  /* used also for all except */
    t_include_1  l_include_1;
    t_class_1    l_class_1;
    t_map_1      l_map_1;
    t_stat_1     l_stat_1;
    t_stat_all_1 l_stat_all_1;
    t_do_1       l_do_1;
    t_run_1      l_run_1;
    t_test_1     l_test_1;
		t_delete_1	 l_delete_1;
		t_restore		 l_restore;
		t_backup		 l_backup;
		t_select	   l_select;
		t_set	       l_set;
		t_cv				 l_cv;

 /*   t_include_2  l_include_2; */
    t_map_2      l_map_2;
  } param;
} t_command;


extern char *TELA_call_path;

boolean realloc_string(char **dest,char *source);
boolean alloc_string(char **dest,char *source);
int eq_q_str(const char* str1,const char* str2);

void list_init(t_list *p_list);
int list_append(t_list *p_list,void* item_to_add);
int list_insert(t_list *p_list,void* item_to_insert);
int list_remove_first(t_list *p_list,void **pp_item);
int list_get_first(t_list *p_list,void **pp_item);
int list_get_next(t_list *p_list,void **pp_item);
int list_get_last(t_list *p_list,void **pp_item);
void list_catenate(t_list *p_list1,t_list *p_list2);

/* this function is actually not implemented in this module 
			but in user interface */
void errmsg(char *msg);

#endif
